/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Contains code derived from D.L. Carroll's FORTRAN Genetic Algorithm 
Driver v1.7.0 (v1.7a Copyright 2001 David L. Carroll). 

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/* ga.c - John Hemmings (08 May 2013) */

/* Contains optimizer code based on a microgenetic algorithm (Krishnakumar, 
   1989) following an implementation by David L. Carroll (Carroll, 1996)

   This code was written with reference to Carroll's FORTRAN GA driver code 
   v1.7.0 which is freely available for non-commercial use from
   http://www.cuaerospace.com/Technology/GeneticAlgorithm

   References:
   Krishnakumar, K., "Micro-Genetic Algorithms for Stationary and 
   Non-Stationary Function Optimization," SPIE: Intelligent Control and 
   Adaptive Systems, Vol. 1196, Philadelphia, PA, 1989. doi:10.1117/12.969927.
   Carroll, D. L., "Genetic Algorithms and Optimizing Chemical Oxygen-Iodine
   Lasers," Developments in Theoretical and Applied Mechanics, Vol. XVIII,
   eds. H.B. Wilson, R.C. Batra, C.W. Bert, A.M.J. Davis, R.A. Schapery, D.S.
   Stewart, and F.F. Swinson, School of Engineering, The University of Alabama,
   1996, pp.411-424. 
*/

#include "optimizer.h"
#include "ranwell.h"

#define RAN_ECHO 0   /* Set to 1 to echo random variates to stdout */

static void ga_evalout (struct gadatarec *, OBJF_FLOAT_TYPE *, double [], 
                        int *);
static int ga_writegenreport (FILE *, struct gadatarec *, OBJF_FLOAT_TYPE *, 
                              double [], int *);
static void ga_generation (FILE *, struct gadatarec *, double [], int *);
static OBJF_FLOAT_TYPE *ga_decode (int , struct gadatarec *, OBJF_FLOAT_TYPE *);
static int *ga_code (int, OBJF_FLOAT_TYPE *, struct gadatarec *);
static void selectn (struct gadatarec *, int *, int *[], int *[], int [], 
                     double []);
static int *selectbest (int, int, int *, double *, int *);
static int shuffle (int, int, int *, double *);
static void crosovr (struct gadatarec *, int [], int [], int [], int *);
static void newgen (FILE *, struct gadatarec *, int *);
static void gamicro (FILE *, struct gadatarec *, int *);


int minimize_ga (struct gadatarec *g, OBJF_FLOAT_TYPE *parmval, 
                 double fitness[], int *psetbest)

{ int i,j,k,status; 
  struct gactrlrec *ctrl;

  ctrl=g->ctrl;

  randominit(ctrl->seed);
  if (ctrl->psetini==1)
    for (i=0;i<ctrl->npopsiz;i++) ga_code(i,parmval,g); 
  else
    { for (i=0;i<ctrl->npopsiz*g->nchrome;i++)
        g->iparent[i]=(int)(randomu(RAN_ECHO)+0.5);
    }

  g->gennum=1;
  g->popconverged=0;
  for (i=0; i<ctrl->maxgen && !g->popconverged ;i++)
    { if (i==0) *psetbest=0;
      else ga_generation(g->logfile,g,fitness,psetbest);
      ga_evalout(g,parmval,fitness,psetbest); 
      if (g->logfile)
        { status=ga_writegenreport(g->logfile,g,parmval,fitness,psetbest);
          if (status!=OK) return status; 
        }
    }

  return OK; 
}


static void ga_generation (FILE *outfile, struct gadatarec *g, 
                           double fitness[], int *psetbest)

{ int ncross,ipick,j;
  int *genomemate1,*genomemate2,*genomechild;
  struct gactrlrec *ctrl;

  ctrl=g->ctrl;

  ncross=0;
  ipick=ctrl->npopsiz-1;
  for (j=0;j<ctrl->npopsiz;j+=ctrl->nchild)
    { genomechild=&g->ichild[j*g->nchrome];
      selectn(g,&ipick,&genomemate1,&genomemate2,genomechild,fitness);
      crosovr(g,genomemate1,genomemate2,genomechild,&ncross);
    }

  if (outfile) fprintf(outfile, "Number of crossovers = %d\n",ncross);

  newgen(outfile,g,psetbest); 
  g->gennum++;
  gamicro(outfile,g,psetbest);
}


OBJF_FLOAT_TYPE *ga_decode (int i, struct gadatarec *g, 
                            OBJF_FLOAT_TYPE *parmval)

{ int j,k,l,mfirst,mlast,iparam;
  int *genome;
  OBJF_FLOAT_TYPE *parmvec;
  struct parmctrlrec *parmctrl;

  parmctrl=g->parmctrl;

  genome=&g->iparent[i*g->nchrome];
  parmvec=&parmval[i*parmctrl->nparm];

  l=0;
  for (k=0;k<parmctrl->nparm;k++)
    { iparam=0;
      mfirst=l;
      mlast=mfirst+parmctrl->nbits[k]-1;
      for (j=mfirst;j<=mlast;j++)
        { l++;
          iparam+=genome[j]*(int)(pow(2.0,(double)(mlast-j))+0.5);
        }
      if (parmctrl->logtrans[k])
        parmvec[k]=
          pow(10,parmctrl->logmin[k]+parmctrl->step[k]*(0.5+iparam));
      else
        parmvec[k]=parmctrl->min[k]+parmctrl->step[k]*(0.5+iparam);
    }

  return parmvec;
}


static int *ga_code (int i, OBJF_FLOAT_TYPE *parmval, struct gadatarec *g) 

{ int j,k,l,mfirst,mlast,iparam,inew;
  int *genome;
  OBJF_FLOAT_TYPE *parmvec;
  struct parmctrlrec *parmctrl;

  parmctrl=g->parmctrl;

  genome=&g->iparent[i*g->nchrome];
  parmvec=&parmval[i*parmctrl->nparm];

  l=0;
  for (k=0;k<parmctrl->nparm;k++)
    { mfirst=l;
      mlast=mfirst+parmctrl->nbits[k]-1;
      if (g->parmctrl->logtrans[k])
        iparam=(int)((log10(parmvec[k])-parmctrl->logmin[k])/parmctrl->step[k]);
      else
        iparam=(int)((parmvec[k]-parmctrl->min[k])/parmctrl->step[k]);
      for (j=mlast;j>=mfirst;j--)
        { inew=iparam/2;
          genome[j]=iparam-2*inew;
          iparam=inew;
        }
      l=mlast+1;
    }

  return genome;
}


static void ga_evalout (struct gadatarec *g, OBJF_FLOAT_TYPE *parmval,
                        double fitness[], int *psetbest) 

{ OBJF_FLOAT_TYPE *parmvec;
  int *genomebest;
  double func,best;
  int j,k,jbest,duplicate;

  for (j=0;j<g->ctrl->npopsiz;j++)

    { /* Check whether individual 'j' is identical to the previous one */
 
      if (j==0) duplicate=0;
      else
        { duplicate=1;
          for (k=0;k<g->nchrome && duplicate; k++)
            { if (g->iparent[j*g->nchrome+k]!=g->iparent[(j-1)*g->nchrome+k])
                duplicate=0;
            }
        }

      /* Get parameter values for individual 'j' */

      parmvec=ga_decode(j,g,parmval); 

      /* Get fitness if already known and call function driver. 
         Driver evaluates objective function if needed or returns 
         the known fitness. */
      
      g->psetnum=j+1;
      if (duplicate) g->knownfitness=&fitness[j-1];
      else if (g->psetnum==*psetbest) g->knownfitness=&g->fitnessbest;
      else g->knownfitness=NULL;

      fitness[j]=(*g->func)(g,parmvec);

      /* Check whether fitness of individual 'j' is best */

      if (j==0 || fitness[j]>best)
        { best=fitness[j];
          jbest=j;
        }
    }

  /* Update record of best individual */

  *psetbest=jbest+1;
  genomebest=&g->iparent[jbest*g->nchrome];
  for (j=0;j<g->nchrome;j++) g->ibest[j]=genomebest[j];
  g->fitnessbest=best;
}


static int ga_writegenreport (FILE *outfile, struct gadatarec *g, 
                              OBJF_FLOAT_TYPE *parmval, double fitness[],
                              int *psetbest)

{ char *varname[N_FREEPARM_MAX+2];
  int sigdig[NFIELD_MAX];
  int fw[N_FREEPARM_MAX+2];
  int status,nvar,npage,pageno,i,j,k,ipage;
  OBJF_FLOAT_TYPE sum,pval,pmin,pmax;
  time_t systime;
  char *s;
  struct parmctrlrec *parmctrl;
  int npopsiz,nparm;

  if (!outfile) return OK;

  npopsiz=g->ctrl->npopsiz;
  parmctrl=g->parmctrl;
  nparm=parmctrl->nparm;

  /* Convert genes to string format */

  s=g->genome_fieldlist;
  for (i=0;i<npopsiz;i++)
    { g->genome[i]=s;
      for (j=0;j<g->nchrome;j++)
        { *s='0'+g->iparent[i*g->nchrome+j];
          s++;
        }
      *s='\0';
      s++;
    }

  fprintf(outfile,"\n#################  Generation %d  #################\n\n",
          g->gennum);

  time(&systime);
  s=ctime(&systime);
  fprintf(outfile,"%s\n",s);

  /* Set variable names
     Remove string fields in page buffer for numerical data to force
     use of values in number field */

  nvar=0;
  varname[nvar]="individual";
  for (i=0;i<NPAGEREC_MAX;i++) dpage[i].field[nvar]=NULL;
  nvar++;
  varname[nvar]="genome";
  nvar++;
  for (j=0;j<nparm;j++)
    { varname[nvar]=g->parmname[j];
      for (i=0;i<NPAGEREC_MAX;i++) dpage[i].field[nvar]=NULL;
      nvar++;
    }
  varname[nvar]="func";
  for (i=0;i<NPAGEREC_MAX;i++) dpage[i].field[nvar]=NULL;
  nvar++;

  npage=1+(int)((npopsiz-1)/NPAGEREC_MAX);
  pageno=1;
  i=0;
  while (i<npopsiz)

    { /* Set up table data on current logical page */

      for (ipage=0; ipage<NPAGEREC_MAX && i<npopsiz ;ipage++)

        { k=0;
          dpage[ipage].x[k++]=(double)(i+1);
          dpage[ipage].field[k++]=g->genome[i];
          for (j=0;j<nparm;j++) dpage[ipage].x[k++]=(double)parmval[i*nparm+j];
          dpage[ipage].x[k]=(double)(-fitness[i]);

          i++;
        }

      /* Write logical page */

      sigdig[0]=DEFAULT;
      k=2;
      for (j=0;j<nparm;j++) sigdig[k++]=OBJF_SIGDIG_DEF;
      sigdig[k]=OBJF_SIGDIG_DEF;

      setcolwidths(nvar,varname,ipage,dpage,sigdig,CWIDTH_MAX,fw);

      if (i>NPAGEREC_MAX) fprintf(outfile,"\n");
      if (npage>1) fprintf(outfile,"[Page %d of %d]\n",pageno++,npage);
      status=writelogicalpage(outfile,nvar,1,varname,fw,sigdig,DEFAULT,
                              COLSEP_LOG,PAGEWIDTH,ipage,dpage);
      if (status!=OK) return status;
    }

  fprintf(outfile,"\n");

  fprintf(outfile,"----- Summary Statistics -----\n\n");

  /* Set variable names
     Remove string fields in page buffer for numerical data to force
     use of values in number field */

  nvar=0;
  varname[nvar]="parmname";
  nvar++;
  varname[nvar]="mean";
  for (i=0;i<NPAGEREC_MAX;i++) dpage[i].field[nvar]=NULL;
  nvar++;
  varname[nvar]="range";
  for (i=0;i<NPAGEREC_MAX;i++) dpage[i].field[nvar]=NULL;
  nvar++;
  varname[nvar]="min";
  for (i=0;i<NPAGEREC_MAX;i++) dpage[i].field[nvar]=NULL;
  nvar++;
  varname[nvar]="max";
  for (i=0;i<NPAGEREC_MAX;i++) dpage[i].field[nvar]=NULL;
  nvar++;

  npage=1+(int)((nparm-1)/NPAGEREC_MAX);
  pageno=1;
  i=0;
  while (i<nparm)

    { /* Set up table data on current logical page */

     for (ipage=0; ipage<NPAGEREC_MAX && i<nparm ;ipage++)

        { sum=0;
          for (j=0;j<npopsiz;j++)
            { pval=parmval[j*nparm+i];
              if (j==0 || pval<pmin) pmin=pval;
              if (j==0 || pval>pmax) pmax=pval;
              sum+=pval;
            }

          k=0;
          dpage[ipage].field[k++]=g->parmname[i];
          dpage[ipage].x[k++]=(double)sum/npopsiz;
          dpage[ipage].x[k++]=(double)(pmax-pmin);
          dpage[ipage].x[k++]=(double)pmin;
          dpage[ipage].x[k++]=(double)pmax;

          i++;
        }

      /* Write logical page */

      sigdig[1]=OBJF_SIGDIG_DEF;
      sigdig[2]=OBJF_SIGDIG_DEF;
      sigdig[3]=OBJF_SIGDIG_DEF;
      sigdig[4]=OBJF_SIGDIG_DEF;

      setcolwidths(nvar,varname,ipage,dpage,sigdig,CWIDTH_MAX,fw);

      if (i>NPAGEREC_MAX) fprintf(outfile,"\n");
      if (npage>1) fprintf(outfile,"[Page %d of %d]\n",pageno++,npage);
      status=writelogicalpage(outfile,nvar,1,varname,fw,sigdig,DEFAULT,
                              COLSEP_LOG,PAGEWIDTH,ipage,dpage);
      if (status!=OK) return status;
    }

  fprintf(outfile,"\n");

  sum=0;
  for (j=0;j<npopsiz;j++) sum+=-fitness[j];
  fprintf(outfile,"Mean function value     = %.*g\n",
          OBJF_SIGDIG_DEF,sum/npopsiz);
  fprintf(outfile,"Minimum function value  = %.*g (individual = %d)\n",
          OBJF_SIGDIG_DEF,-fitness[*psetbest-1],*psetbest);

  fprintf(outfile,"\n####################################################\n\n");
  fflush(outfile);

  return OK;
}


static void selectn (struct gadatarec *g, int *ipick, int *genomemate1[],
                     int *genomemate2[], int genomechild1[], double fitness[])

{ int n;
  int *genomechild2;

  *genomemate1=
    selectbest(g->ctrl->npopsiz,g->nchrome,g->iparent,fitness,ipick);
  *genomemate2=
    selectbest(g->ctrl->npopsiz,g->nchrome,g->iparent,fitness,ipick);

  for (n=0;n<g->nchrome;n++) genomechild1[n]=(*genomemate1)[n];
  if (g->ctrl->nchild==2)
    { genomechild2=&genomechild1[g->nchrome];
      for (n=0;n<g->nchrome;n++) genomechild2[n]=(*genomemate2)[n];
    }
}


static int *selectbest (int npopsiz, int nchrome, int iparent[],
                        double fitness[], int *ipick)

{ int ifirst,isecond;

  /* Select best of two possible parents for mating */

  if ((*ipick)+1>=npopsiz)
    *ipick=shuffle(npopsiz,nchrome,iparent,fitness);
  ifirst=(*ipick);
  isecond=(*ipick)+1;
  (*ipick)+=2;

  if (fitness[ifirst]>fitness[isecond]) return &iparent[ifirst*nchrome];
  else return &iparent[isecond*nchrome];
}


static int shuffle (int npopsiz, int nchrome, int iparent[], double fitness[])

{ int j,iother,n,itemp;
  int *genomej,*genomeiother;
  double temp;

  for (j=0;j<npopsiz-1;j++)
    { iother=j+1+(int)((npopsiz-(j+1))*randomu(RAN_ECHO));
      genomej=&iparent[j*nchrome];
      genomeiother=&iparent[iother*nchrome];
      for (n=0;n<nchrome;n++)
        { itemp=genomeiother[n];
          genomeiother[n]=genomej[n];
          genomej[n]=itemp;
        }
      temp=fitness[iother];
      fitness[iother]=fitness[j];
      fitness[j]=temp;
    }

  return 0;
}


static void crosovr (struct gadatarec *g, int genomemate1[], int genomemate2[],
                     int genomechild1[], int *ncross)

{ int n,icross;
  int *genomechild2;
  struct gactrlrec *ctrl;

  ctrl=g->ctrl;

  genomechild2=&genomechild1[g->nchrome];

  if (ctrl->iunifrm)

    { /* Uniform crossover between the randomly selected pair */

      for (n=0;n<g->nchrome;n++)
        { if (randomu(RAN_ECHO)<=ctrl->pcross)
            { (*ncross)++;
              genomechild1[n]=genomemate2[n];
              if (ctrl->nchild==2) genomechild2[n]=genomemate1[n];
            }
        }
    }

  else

    { /* Single-point crossover at a random chromosome point */

      if (randomu(RAN_ECHO)>ctrl->pcross) return;

      (*ncross)++;
      icross=1+(int)((g->nchrome-1)*randomu(RAN_ECHO));
      for (n=icross;n<g->nchrome;n++) genomechild1[n]=genomemate2[n];
      if (ctrl->nchild==2)
        for (n=icross;n<g->nchrome;n++) genomechild2[n]=genomemate1[n];
    }
}


static void newgen (FILE *outfile, struct gadatarec *g, int *psetbest)

{ int j,jelite,kelite,n,irand;
  int *genomechild,*genomeparent;
  struct gactrlrec *ctrl;

  ctrl=g->ctrl;

  /* Write child back into parent array for new generation and
     check to see if the best parent was replicated (record new position) */

  *psetbest=0;
  kelite=0;
  for (j=0;j<ctrl->npopsiz;j++)
    { genomechild=&g->ichild[j*g->nchrome];
      genomeparent=&g->iparent[j*g->nchrome];
      jelite=0;
      for (n=0;n<g->nchrome;n++)
        { genomeparent[n]=genomechild[n];
          if (genomeparent[n]==g->ibest[n]) jelite++;
        }
      if (!kelite && jelite==g->nchrome)
        { kelite=1;
          *psetbest=j+1;
        }
    }

  /* If elitism is enabled and best parent was not replicated then
     reproduce best parent into a random slot (record new position) */

  if (ctrl->ielite && !kelite)
    { irand=(int)(ctrl->npopsiz*randomu(RAN_ECHO));
      genomeparent=&g->iparent[irand*g->nchrome];
      for (n=0;n<g->nchrome;n++) genomeparent[n]=g->ibest[n];
      if (outfile)
        fprintf(outfile,"Elitist reproduction to new individual %d\n",irand+1);
      *psetbest=irand+1;
    }
}


static void gamicro (FILE *outfile, struct gadatarec *g, int *psetbest)

{ int icount,j,n;
  int *genome;
  double diffrac;
  struct gactrlrec *ctrl;

  ctrl=g->ctrl;

  /* Count number of bits different from best parent */

  icount=0;
  for (j=0;j<ctrl->npopsiz;j++)
    { genome=&g->iparent[j*g->nchrome];
      for (n=0;n<g->nchrome;n++)
        if (genome[n]!=g->ibest[n]) icount++;
    }

  /* If less than 5% of bits differ then consider population to be converged
     - restart with best parent and random others
       (record new position of best parent) */

  diffrac=(double)icount/(double)((ctrl->npopsiz-1)*g->nchrome);
  if (outfile) fprintf(outfile,
"New generation: %.3g%% of bits in gene pool differ from best parent\n",
                       diffrac*100);

  if (diffrac<0.05)

    { g->popconverged=1;
      if (outfile) fprintf(outfile,"Population converged\n");

      if (g->gennum<ctrl->mingen)

        { for (n=0;n<g->nchrome;n++) g->iparent[n]=g->ibest[n];
          *psetbest=1;
          for (n=g->nchrome; n<ctrl->npopsiz*g->nchrome; n++)
            g->iparent[n]=(int)(randomu(RAN_ECHO)+0.5);

          if (outfile) fprintf(outfile,
 "%%%%%%%%%%%%%%  Restart micro-population at generation %d  %%%%%%%%%%%%%%\n",
                              g->gennum);
          g->popconverged=0;
        }
    }
}

