!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!=======================================================================
!  MODULE trmix
!  John Hemmings - 29 Nov 2012
!
!  MarMOT layer mixing subroutines:
!
!   - mixinglevels - determine boundary layer mixing parameters
!   - mixing - mix tracers over upper boundary layer
!      - mix_tracer - do mixing for single tracer
!   - mixsubset - mix selected tracer-related forcing variables over
!                 upper boundary layer 
!
!=======================================================================

  module trmix

  use modconfig, only: kind_r_ext,kind_r

  implicit none

  contains

!=======================================================================

  subroutine mixinglevels(km,zbot,f_mld,kmixed,dzpart)

  implicit none

! Input arguments

  integer km                     ! Number of levels
  real(kind_r) zbot(:)           ! Depth at bottom of level (m)
  real(kind_r_ext) f_mld         ! Upper mixed layer depth (m)

! Output arguments

  integer kmixed                 ! No of levels fully mixed
  real(kind_r) dzpart            ! Penetration of mixing into
                                 !   partially mixed level (m)
!-----------------------------------------------------------------------

  integer i,j,k

  if (f_mld<=0 .or. km==1) then

    kmixed=0
    dzpart=0.0

  else if (f_mld>=zbot(km)) then

    kmixed=km
    dzpart=0.0

  else

    k=1
    do while (zbot(k)<=f_mld)
      k=k+1
    enddo

    kmixed=k-1
    dzpart=f_mld-zbot(kmixed)

  endif

  end subroutine

!=======================================================================

  subroutine mixing(scheme,kmixed,dz,dzpart,tracer)

  use trlink, only: ntracer,itracer

  implicit none

! Input arguments

  integer scheme                 ! Mixing scheme
  integer kmixed                 ! No of levels fully mixed
  real(kind_r) dz(:)             ! Level height (m)
  real(kind_r) dzpart            ! Penetration of mixing into
                                 !   partially mixed level (m)

! In/out argument

  real(kind_r_ext) tracer(:,:)   ! Tracer values for update

!-----------------------------------------------------------------------

  real(kind_r),parameter :: dzpart0=0.
  integer i,j

  do i=1,ntracer
    j=itracer(i)
    if (j/=0) then
      select case(scheme)

      case(1)  ! No partial mixing
        call mix_tracer(kmixed,dz,dzpart0,tracer(:,j))

      case(2)  ! Partially mix layer at mixed layer depth
        call mix_tracer(kmixed,dz,dzpart,tracer(:,j))

      end select
    endif
  enddo

  end subroutine

!=======================================================================

  subroutine mixsubset(scheme,kmixed,dz,dzpart,np,ip,updflag,prof)

  implicit none

! Input arguments

  integer scheme                 ! Mixing scheme
  integer kmixed                 ! No of levels fully mixed
  real(kind_r) dz(:)             ! Level height (m)
  real(kind_r) dzpart            ! Penetration of mixing into
                                 !   partially mixed level (m)
  integer np                     ! Number of tracer-related profiles
  integer ip(:)                  ! Tracer indices
  logical updflag(:)             ! Flag to enable profile mixing  

! In/out argument

  real(kind_r_ext) prof(:,:)     ! Tracer-related profiles

!-----------------------------------------------------------------------

  real(kind_r),parameter :: dzpart0=0.
  integer i,j

  do i=1,np
    j=ip(i)
    if (j/=0 .and. updflag(j)) then
      select case(scheme)
      case(1)
        call mix_tracer(kmixed,dz,dzpart0,prof(:,j))
      case(2)
        call mix_tracer(kmixed,dz,dzpart,prof(:,j))
      end select
    endif
  enddo

  end subroutine

!=======================================================================

  subroutine mix_tracer(kmixed,dz,dzpart,tracer)

  implicit none

! Input arguments

  integer kmixed                 ! Number of levels fully mixed 
  real(kind_r) dz(:)             ! Level height (m)
  real(kind_r) dzpart            ! Penetration into part mixed level (m)

! In/out argument

  real(kind_r_ext) tracer(:)     ! Profile to mix

!-----------------------------------------------------------------------

  real(kind_r) sum,sumdz,extra,mean
  integer k

  if (kmixed/=0) then

    sum=0.0
    sumdz=0.0
    do k=1,kmixed
      sum=sum+tracer(k)*dz(k)
      sumdz=sumdz+dz(k)
    enddo
    if (dzpart>0.0) then
      sum=sum+dzpart*tracer(kmixed+1)
    endif
    mean=sum/(sumdz+dzpart)

    do k=1,kmixed
      tracer(k)=mean
    enddo
    if (dzpart>0.0) then
      extra=dzpart*(tracer(kmixed+1)-mean)
      tracer(kmixed+1)=tracer(kmixed+1)-extra/dz(kmixed+1)
    endif

  endif

  end subroutine

!=======================================================================

  end module trmix
       
