!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!=======================================================================
!  MODULE trperturb 
!  John Hemmings - 04 May 2013
!
!  MarMOT tracer perturbation subroutines:
!
!   - pertsetup - set up routine for tracer perturbation 
!   - perturbation - perturb tracers
!      - perturb_tracer - do perturbation for single tracer
!   - rlxconstraint - determine tendency limitation factor for 
!                     constrained tracer relaxation
!
!  Peturbations take the form of an autoregressive random process with
!  a prescribed mean and standard deviation.  
!=======================================================================

  module trperturb

  use modconfig, only: kind_r_ext,kind_r
  use trlink, only: ntpri,ntracer,ntracer_in,itracer
  use random, only: randwkspace

  implicit none

! Perturbation scheme configuration

  logical pertflag(ntpri)      ! Perturbation switch for each tracer
  logical pertflag_mean(ntpri) ! Prescribed perturbation switch for each tracer 
  logical pertflag_ran(ntpri)  ! Random perturbation switch for each tracer
  integer perttransform(ntpri) ! Transformations for activated tracers
  logical pertactive           ! Switch to show if perturbation is needed
  logical pertactive_ran       ! Switch to show if random perturbation is needed
  integer pertmixscheme        ! Upper boundary layer mixing scheme for 
                               !   modifying perturbations 
  real(kind_r) pertautocorr    ! Random perturbation autocorrelation after 24 h

! Perturbation state

  real(kind_r) pertstate(ntpri)        ! Autoregressive random perturbation 
                                       !   for activated tracers
  integer,allocatable :: pertseed(:,:) ! Pseudo-random number generator state 
                                       !   for activated tracers
                                       !   (n integer array for each tracer;
                                       !    n is system-dependent) 
  type(randwkspace) pertwk(ntpri)      ! Work space for tracer perturbation 

  contains

!=======================================================================

  subroutine pertsetup(pmeanflag,pranflag,transform,pautocorr,seed,mixopt)

  use random, only: randnorminit

  implicit none

! Input arguments

  integer pmeanflag(:)         ! Tracer perturbation flag
                               !   0: no prescribed perturbation
                               !   1: presecribed perturbation active
  integer pranflag(:)          ! Tracer perturbation flag
                               !   0: no random perturbation
                               !   1: random perturbation active
  integer transform(:)         ! Tracer perturbation transformation
                               !   0: no variable transformation
                               !   1: log10 transformation
                               !   2: sqrt transformation
  real(kind_r_ext) pautocorr   ! Perturbation autocorrelation after 24 h
  integer seed(:)              ! Tracer perturbation seed for random
                               !   number generation
  integer mixopt               ! Mixing option for modifying perturbation (0-2)

!-----------------------------------------------------------------------

  integer i,j,n,status

  pertactive=.FALSE.
  pertactive_ran=.FALSE.
  n=0
  do i=1,ntracer_in
    j=itracer(i)
    if (j/=0) then
      if (j>ntpri) stop &
"*** Non-primary tracer reference in 'pertsetup' - bad 'modconfig' module? ***" 
      pertflag_mean(j)=(pmeanflag(j)/=0)
      pertflag_ran(j)=(pranflag(j)/=0)
      pertflag(j)=(pertflag_mean(j).or.pertflag_ran(j))
      if (pertflag(j)) then
        pertactive=.TRUE.
        perttransform(j)=transform(j)
        if (pertflag_ran(j)) then
          pertactive_ran=.TRUE.
          pertstate(j)=0.0
          if (n==0) then
            if (allocated(pertseed)) deallocate(pertseed)
            call random_seed(size=n)
            allocate(pertseed(n,ntracer_in),stat=status)
            if (status/=0) stop &
              "*** Memory allocation failed for 'pertseed' in 'pertsetup' ***"
          endif
          pertseed(1:n,j)=seed(j)
          call randnorminit(pertwk(j))
        endif
      endif
    endif
  enddo

  if (pertactive) then
    pertmixscheme=mixopt
    pertautocorr=max(0.0,min(1.0_kind_r,pautocorr))
  else
    pertmixscheme=0
  endif

  end subroutine pertsetup

!=======================================================================

  subroutine perturbation(km,dt,dtpert,f_pert,f_pertsd,tracer,ddt,ddtp)

  use trlink, only: linkedtrend,ntracer_in

  implicit none

! Input arguments

  integer km                      ! Number of levels 
  real(kind_r) dt                 ! Time step (d)
  real(kind_r) dtpert             ! Elapsed time since last call for 
                                  !   autoregressive process (d)
  real(kind_r_ext) f_pert(:,:)    ! Perturb rates (trans. units/d)
  real(kind_r_ext) f_pertsd(:,:)  ! Perturb s.d.s (trans. units/d)
  real(kind_r_ext) tracer(:,:)    ! Tracer values 

! In/out argument

  real(kind_r_ext) ddt(:,:)       ! Tracer tendencies for update 

! Output argument 

  real(kind_r) ddtp(:,:)          ! Tracer tendencies due to perturbation 

!-----------------------------------------------------------------------

  real(kind_r) rana,ranscale
  integer i,j,nstepday

! Get random perturbation parameters appropriate to the elapsed time 'dtpert' 
! (negative 'dtpert' initializes a new process).
! 'rana' is the autocorrelation 
! 'ranscale' is the scaling factor required to give unit expected s.d. 
  
  if (pertactive_ran) then
    if (dtpert<0.0) then
      rana=0.0
      ranscale=1.0_kind_r
    else
      nstepday=1.0_kind_r/dtpert
      rana=pertautocorr**dtpert
      if (rana==1.0_kind_r) then
        ranscale=0.0
      else
        ranscale=sqrt((1.0_kind_r-rana**2)/(1.0_kind_r-rana**(2*nstepday)))
      endif
    endif
  endif

! Initialize tendencies for all tracers used 

  do i=1,ntracer
    j=itracer(i)
    if (j/=0) ddtp(1:km,j)=0.0
  enddo

! Get perturbation tendencies for selected tracers

  do i=1,ntracer_in
    j=itracer(i)
    if (j/=0 .and. pertflag(j)) then
      call perturb_tracer(1,km,dt,pertflag_mean(j),pertflag_ran(j), &
                          rana,ranscale,perttransform(j),f_pert(:,j), &
                          f_pertsd(:,j),tracer(:,j),pertstate(j), &
                          pertseed(:,j),pertwk(j),ddtp(:,j))
    endif
  enddo

! Get tendencies for linked tracers 

  call linkedtrend(1,km,tracer,ddtp)

! Update tendencies for all tracers used 
  
  do i=1,ntracer
    j=itracer(i)
    if (j/=0) ddt(1:km,j)=ddt(1:km,j)+ddtp(1:km,j)
  enddo

  end subroutine

!=======================================================================

  subroutine perturb_tracer(kmin,kmax,dt,mflag,ranflag,rana,ranscale, &
                            transform,ratemean,ratesd,tracer, &
                            ranpert,seed,wk,ddt)

  use random, only: randnorm,randwkspace

  implicit none

! Input arguments

  integer kmin                   ! Minimum level
  integer kmax                   ! Maximum level
  real(kind_r) dt                ! Time step (d)
  logical mflag                  ! Prescribed perturbation switch
  logical ranflag                ! Random perturbation switch
  real(kind_r) rana              ! Autocorrelation between time steps
  real(kind_r) ranscale          ! Scaling factor for random perturbation 
  integer transform              ! Variable transformation option
  real(kind_r_ext) ratemean(:)   ! Prescribed perturbation rate
                                 !   (transformed units/d)
  real(kind_r_ext) ratesd(:)     ! Random perturbation s.d.
                                 !   (transformed units/d)
  real(kind_r_ext) tracer(:)     ! Tracer profile (tracer units)

! In/out argument

  real(kind_r) ranpert           ! Autoregressive random perturbation
  integer seed(:)                ! Seed for random number generator
  type(randwkspace) wk           ! Workspace for random number generator

! Output argument

  real(kind_r) ddt(:)            ! Perturbation tendency (tracer units /d)

!-----------------------------------------------------------------------

  integer k
  real(kind_r) dxdt,x

! Get next random perturbation value 

  if (ranflag) ranpert=rana*ranpert+ranscale*randnorm(seed,wk)

  do k=kmin,kmax

!   Determine perturbation tendency to apply

    if (mflag) then
      dxdt=ratemean(k)
    else
      dxdt=0.0
    endif

    if (ranflag .and. ratesd(k)>0.0) dxdt=dxdt+ratesd(k)*ranpert

!   Determine equivalent perturbation tendency in untransformed space.

    select case(transform)

    case(0)  ! no transformation 
             ! (flux limitation to avoid growth of -ve concentrations)

      ddt(k)=max(dxdt,-max(0.0,tracer(k))/dt) 

    case(1)  ! log transformation

      ddt(k)=max(0.0,tracer(k))*log(10.0_kind_r)*dxdt

    case(2)  ! sqrt transformation

      ddt(k)=2.0_kind_r*sqrt(max(0.0,tracer(k)))*dxdt

    end select

  enddo

  end subroutine

!=======================================================================

  real(kind_r) function rlxconstraint(dt,kmin,kmax,ctrlfact,relaxflag, &
                                      f_pertsd,f_ref,tracer)

  implicit none

! Input arguments

  real(kind_r) dt                   ! Length of time step (d)
  integer kmin                      ! Minimum relaxation level
  integer kmax                      ! Maximum relaxation level
  real(kind_r) ctrlfact             ! Relaxation contol factor 
  logical relaxflag(:)              ! Relaxation switch for each tracer
  real(kind_r_ext) f_pertsd(:,:)    ! Random perturbation s.d.
                                    !   (transformed units/d)
  real(kind_r_ext) f_ref(:,:)       ! Reference tracer profiles
  real(kind_r_ext) tracer(:,:)      ! Tracer profiles

!-----------------------------------------------------------------------
! Returns a relaxation rate scaling factor that limits the magnitude of
! a relaxation tendency in transformed variable space to a maximum of 
! 'ctrlfact'*'f_pertsd', provided that the input relaxation rate is not 
! more than the reciprocal of the time step (this being the maximum 
! relaxation rate without overshoot in the absence of other tendencies).
!
! A uniform scaling factor is determined that ensures the maximum rate
! is not exceeded anywhere for any relaxed tracer when the input rate
! is the reciprocal of the time step. Structural relationships between
! tracers and depth levels are thus preserved in the case of a uniform
! input relaxation rate: the final multivariate state (comprising
! all relaxed tracers in the unmasked depth range) will be a weighted
! average of the model state and the reference state.
!
! Note 1: Log transformations in any variable will lead to very strong 
! constraints at low concentrations of that variable - USE WITH CARE!
!
! Note 2: No distinction is made between "relaxed" tracers and tracers
! that are only ever overridden (see 'trrelax' module). The latter can
! therefore influence the constraint even though they are unaffected by 
! it. The tendency limitation scheme is not generally intended for use 
! in conjunction with such tracers.
!-----------------------------------------------------------------------

  real(kind_r) rlxfrac              ! Fraction of max. relaxation
                                    ! rate permitted
  integer i,j,k
  real(kind_r) dxdt,ddtpot,ddtmax

  rlxfrac=1.0_kind_r

  do i=1,ntracer
    j=itracer(i)
    if (j/=0 .and. relaxflag(j)) then
      do k=kmin,kmax

!       Determine size of tendency for relaxation rate 1/dt 

        ddtpot=abs(f_ref(k,j)-tracer(k,j))/dt

        if (ddtpot>0.0) then

!         Determine max. permitted size of tendency 

          dxdt=ctrlfact*f_pertsd(k,j)

          select case(perttransform(j))

          case(0)  ! no transformation

            ddtmax=dxdt

          case(1)  ! log transformation

            ddtmax=max(0.0,tracer(k,j))*log(10.0_kind_r)*dxdt

          case(2)  ! sqrt transformation

            ddtmax=2.0_kind_r*sqrt(max(0.0,tracer(k,j)))*dxdt

          end select

!         Update fraction permitted

          rlxfrac=min(rlxfrac,ddtmax/ddtpot)

        endif

      enddo
    endif
  enddo

  rlxconstraint=rlxfrac

  end function

!=======================================================================

  end module trperturb
