!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!=======================================================================
!  MODULE trrelax 
!  John Hemmings - 04 Apr 2013
!
!  MarMOT tracer relaxation subroutines:
!
!   - relaxsetup - setup routine for tracer relaxation 
!   - relaxation - relax tracers towards reference profiles
!   - override - override tracers with reference profile values
!
!=======================================================================

  module trrelax 

  use modconfig, only: kind_r_ext,kind_r
  use trlink, only: ntpri,ntracer,ntracer_in,itracer,ntracer_linked, &
                    linkedtrend

  implicit none

  real(kind_r),parameter :: relaxupperlimit=1000.0
                             ! Maximum relaxation rate not causing override (/d)
  real(kind_r_ext) :: modelrelaxulim=relaxupperlimit
                             ! Copy visible to marmot.c

  logical relaxflag(ntpri)   ! Relaxation switches for activated tracers
  real(kind_r) relaxctrlfact ! Multiple of perturbation s.d. for limiting
                             ! relaxation rate (zero suppresses all relaxation)
  integer relaxmaskflag      ! Tracer relaxation mask flag
  integer relaxmaskopt       ! Tracer relaxation mask option
  integer relaxmixscheme     ! Upper boundary layer mixing after relaxation
  logical relaxactive        ! Relaxation flag set if relaxation step needed

  contains

!=======================================================================

  subroutine relaxsetup(rlxflag,ctrlfact,maskflag,maskopt,mixopt)

  implicit none

! Input arguments

  integer rlxflag(ntpri)       ! Tracer relaxation flag
                               !   0: no relaxation
                               !   1: relaxation active
  real(kind_r_ext) ctrlfact    ! Multiple of perturbation s.d. for
                               ! limiting relaxation rate
  integer maskflag             ! Tracer relaxation mask flag
                               !  -1: mask levels at z or above
                               !   0: no mask
                               !   1: mask levels at z or below
  integer maskopt              ! Tracer relaxation mask option
                               !   1: z = upper mixed layer depth
                               !   2: z = euphotic zone depth
                               !   3: z = max. of (1) and (2)
  integer mixopt               ! Mixing option after relaxation (0-2)

!-----------------------------------------------------------------------

  integer i,j

  relaxactive=.FALSE.
  if (ctrlfact/=0.0) then
    do i=1,ntracer_in
      j=itracer(i)
      if (j/=0) then
        if (j>ntpri) stop &
"*** Non-primary tracer reference in 'relaxsetup' - bad 'modconfig' module? ***"
        relaxflag(j)=(rlxflag(j)/=0)
        if (relaxflag(j)) relaxactive=.TRUE.
      endif
    enddo
  endif

  if (relaxactive) then
    relaxctrlfact=ctrlfact
    relaxmaskflag=maskflag
    relaxmaskopt=maskopt
    relaxmixscheme=mixopt
  else
    relaxmixscheme=0
  endif

  end subroutine relaxsetup

!=======================================================================

  subroutine relaxation(km,zbot,dt,f_mld,f_rlx,f_ref,f_pertsd,zeuphotic, & 
                        tracer,ddt,ddtr,kmin,kmax,ratefrac)

  use trperturb, only: rlxconstraint

  implicit none

! Input arguments
 
  integer km                        ! Number of levels
  real(kind_r) zbot(:)              ! Depth at bottom of level (m) 
  real(kind_r) dt                   ! Length of time step (d)
  real(kind_r_ext) f_mld            ! Upper mixed layer depth (m)
  real(kind_r_ext) f_rlx(:,:)       ! Relaxation rates (/d)
  real(kind_r_ext) f_ref(:,:)       ! Reference tracer profiles
  real(kind_r_ext) f_pertsd(:,:)    ! Random perturbation s.d.
                                    !   (trans. units/d)
  real(kind_r) zeuphotic            ! Euphotic zone depth (m)
  real(kind_r_ext) tracer(:,:)      ! Tracer values 

! In/out argument

  real(kind_r_ext) ddt(:,:)         ! Tracer tendencies for update

! Output argument

  real(kind_r) ddtr(:,:)            ! Tracer tendencies due to relaxation 
  integer kmin                      ! Shallowest level for relaxation 
  integer kmax                      ! Deepest level for relaxation 
  real(kind_r) ratefrac             ! Relaxation rate limitation factor 

!-----------------------------------------------------------------------

  real(kind_r) z
  integer i,j,k

! If masking above or below reference depth has been set up then all
! layers down/up to reference depth are masked including layer spanning
! reference depth

  if (relaxmaskflag==0) then

    kmin=1
    kmax=km

  else

    select case(relaxmaskopt)
    case(1)
      z=f_mld
    case(2)
      z=zeuphotic
    case(3)
      z=max(f_mld,zeuphotic)
    end select

    if (z<=0) then

      if (relaxmaskflag<0) then  ! No masking
        kmin=1
        kmax=km
      else  ! Full masking
        kmin=0
        kmax=0
      endif

    else if (z>=zbot(km)) then

      if (relaxmaskflag<0) then  ! Full masking
        kmin=0
        kmax=0
      else  ! No masking
        kmin=1
        kmax=km
      endif

    else

      k=1
      do while (zbot(k)<z)
        k=k+1
      enddo

      if (relaxmaskflag<0) then  ! Mask k and above
        kmin=k+1
        kmax=km
      else 
        if (zbot(k)>z) then  ! Mask k and below
          kmin=1
          kmax=k-1
        else  ! Mask k+1 and below
          kmin=1
          kmax=k
        endif
      endif

    endif

  endif

! Initialize tendencies for all tracers used

  do i=1,ntracer
    j=itracer(i)
    if (j/=0) ddtr(1:km,j)=0.0
  enddo

! Input relaxation rates exceeding 'relaxupperlimit' indicate that the 
! reference state should fully override the model state at the end of the 
! time step. The corresponding updates are deferred - done in 'override'.
! If a valid relaxation control factor is present (i.e. 'relaxctrlfact'>=0)
! a tendency limitation scheme applies to the remaining relaxation rates. 
! A relaxation control factor of zero de-activates all relaxation apart from
! overrides.

  if (relaxctrlfact/=0.0) then

    if (relaxctrlfact<0.0) then
      ratefrac=1.0
    else 
      ratefrac=rlxconstraint(dt,kmin,kmax,relaxctrlfact, &
                             relaxflag,f_pertsd,f_ref,tracer)
    endif

!   Get relaxation tendencies for selected tracers where relaxation rate
!   does not indicate an override request

    do i=1,ntracer_in  
      j=itracer(i)
      if (j/=0 .and. relaxflag(j)) then
        do k=kmin,kmax
          if (f_rlx(k,j)>0.0 .and. f_rlx(k,j) <= relaxupperlimit) &
            ddtr(k,j)=ratefrac*f_rlx(k,j)*(f_ref(k,j)-tracer(k,j))
        enddo
      endif
    enddo

!   Get tendencies for linked tracers

    call linkedtrend(kmin,kmax,tracer,ddtr)

!   Update tendencies for selected tracers

    do i=1,ntracer_in
      j=itracer(i)
      if (j/=0 .and. relaxflag(j)) & 
        ddt(kmin:kmax,j)=ddt(kmin:kmax,j)+ddtr(kmin:kmax,j)
    enddo

!   Update tendencies for linked tracers

    do i=1,ntracer_linked
      j=itracer(ntracer_in+i)
      if (j/=0) ddt(kmin:kmax,j)=ddt(kmin:kmax,j)+ddtr(kmin:kmax,j)
    enddo

  endif

  end subroutine

!=======================================================================

  subroutine override(km,kmin,kmax,dt,f_rlx,f_ref,tracer,ddtr)

  use modconfig, only: nt,nk

  implicit none

! Input arguments

  integer km                        ! Number of levels
  integer kmin                      ! Shallowest level for relaxation 
  integer kmax                      ! Deepest level for relaxation 
  real(kind_r) dt                   ! Length of time step (d)
  real(kind_r_ext) f_rlx(:,:)       ! Relaxation rates (/d)
  real(kind_r_ext) f_ref(:,:)       ! Reference tracer profiles

! In/out argument

  real(kind_r_ext) tracer(:,:)      ! Tracer values
  real(kind_r) ddtr(:,:)            ! Tracer tendencies due to relaxation

!-----------------------------------------------------------------------

  real(kind_r) ddtr_tmp(nk,nt)      ! Override tendencies for activated tracers
                                    ! (array shape matching 'ddtr')
  integer i,j,k

  if (size(ddtr,1)/=nk .or. size(ddtr,2)/=nt) &
    stop "*** Unexpected array dimensions for 'ddtr' argument to 'override' ***"
 
! Initialize tendencies for all tracers used

  do i=1,ntracer
    j=itracer(i)
    if (j/=0) ddtr_tmp(kmin:kmax,j)=0.0
  enddo

! Get override tendencies for selected tracers where relaxation rate 
! indicates an override request 

  do i=1,ntracer_in
    j=itracer(i)
    if (j/=0 .and. relaxflag(j)) then
      do k=kmin,kmax
        if (f_rlx(k,j) > relaxupperlimit) &
          ddtr_tmp(k,j)=(f_ref(k,j)-tracer(k,j))/dt
      enddo
    endif
  enddo

! Get tendencies for linked tracers

  call linkedtrend(kmin,kmax,tracer,ddtr_tmp)

! Update selected tracers where relaxation rate indicates an override request

  do i=1,ntracer_in
    j=itracer(i)
    if (j/=0 .and. relaxflag(j)) then
      do k=kmin,kmax
        if (f_rlx(k,j) > relaxupperlimit) tracer(k,j)=f_ref(k,j)
      enddo
    endif
  enddo

! Update linked tracers 

  do i=1,ntracer_linked
    j=itracer(ntracer_in+i)
    if (j/=0) tracer(kmin:kmax,j)=tracer(kmin:kmax,j)+ddtr_tmp(kmin:kmax,j)*dt
  enddo

! Add relaxation + override tendencies for diagnostics

  do i=1,ntracer
    j=itracer(i)
    if (j/=0) ddtr(kmin:kmax,j)=ddtr(kmin:kmax,j)+ddtr_tmp(kmin:kmax,j)
  enddo  

  end subroutine 

!=======================================================================

  end module trrelax 


