!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!===============================================================================
! MODULE MEDUSA 
! John Hemmings - 24 Jul 2013 
!
! Contains MarMOT interface routines for the MEDUSA model
! 
! MarMOT interface routines: 
!
!  - codever - provide version notes to MarMOT 
!  - defparm - define parameter set for MarMOT 
!  - getinfo - provide parameter set specific info to MarMOT
!  - setup - set-up fixed model environment 
!  - setparm - set model parameters
!  - initstate - initialize MarMOT state: derived tracers & ratios
!  - getbiotrans - output tracer-specific velocities for MarMOT advection 
!  - getphotin - output data to MarMOT for photosynthesis calculation
!  - biotrend - add source-minus-sink to MarMOT transport tendencies 
!  - getdiag - output model diagnostics to MarMOT
!
!===============================================================================

  module medusa 

  use modconfig, only: kind_r_ext,kind_r,nk,atomc
  use medusa_globdata, only: wp,jpi,jpj,jpim1,jpjm1,jpk,trn,nt

  implicit none

  integer,parameter :: ji=jpim1        ! Active array element in i dimension
  integer,parameter :: jj=jpjm1        ! Active array element in j dimension
  integer, parameter :: nparm=77       ! Number of parameters 

  integer kt                           ! Time step counter
  integer jsink                        ! Detrital sinking option switch

  real(wp) tratrans(jpi,jpj,nk,nt)     ! Tendencies due to transport (units /s)

! MarMOT interface mapping

  integer,parameter :: nr=9

  integer itracer(nt)     ! Indices of tracers in MarMOT array
  integer iratio(nr)      ! Indices of ratios in MarMOT array
  integer idiagprof(25)  ! Indices of profile diagnostics in MarMOT array
  integer idiagscal(2)  ! Indices of scalar diagnostics in MarMOT array


  contains

!===============================================================================

  subroutine codever(version,description)

  use modconfig, only: vertextlen 

  implicit none

! Arguments (output)

  character(len=vertextlen+1) version      ! Code version details 
  character(len=vertextlen+1) description  ! Model description document 

!-------------------------------------------------------------------------------

  version='JCH_2013JUL24; adapted original code'
  description='Yool et al 2011, Geosci. Model Dev., 4, 381-417'

  end subroutine

!===============================================================================

  subroutine defparm(np,pname,pdesc,pmin,pmax)

  use modconfig, only: pnamelen,pdesclen

  implicit none

! Arguments (output)

  integer np                                        ! No. of parameters
  character(len=pnamelen+1), dimension(:) :: pname  ! Parameter names
  character(len=pdesclen+1), dimension(:) :: pdesc  ! Descriptions
  real(kind_r_ext) pmin(:)                          ! Minimum values 
  real(kind_r_ext) pmax(:)                          ! Maximum values

!-------------------------------------------------------------------------------

  integer i 

  if (size(pname)<nparm) then
    write(*,*) "Setup error in MEDUSA model - defining too many parameters"
    stop
  endif

  np=nparm

! Stochiometric ratio limits

  i=1
  pname(i)='rcchlminndia'
  pdesc(i)='minimum C to Chl ratio for non-diatoms (gC/gChl) [=1/xthetam]'
  pmin(i)=0.

  i=2
  pname(i)='frcchlmindiat'
  pdesc(i)='minimum C to Chl ratio for diatoms relative to non-diatoms [=xthetam/xthetamd]'
  pmin(i)=0.

  i=3
  pname(i)='rsinpdiatmin'
  pdesc(i)='minimum diatom Si:N ratio [=xsin0]'
  pmin(i)=0.

  i=4
  pname(i)='rsinpdiatmax'
  pdesc(i)='maximum diatom Si:N ratio [=1/xnsi0]'
  pmin(i)=0.

! Stochiometric ratios

  i=5
  pname(i)='rcnphy'
  pdesc(i)='phytoplankton C:N ratio (molC/molN) [=xthetap]'
  pmin(i)=0.

  i=6
  pname(i)='rcnzmi'
  pdesc(i)='microzoo. C:N ratio (molC/molN) [=xthetazmi]'
  pmin(i)=0.

  i=7
  pname(i)='rcnzme'
  pdesc(i)='mesozoo. C:N ratio (molC/molN) [=xthetazme]'
  pmin(i)=0.

  i=8
  pname(i)='rcndet'
  pdesc(i)='detritus C:N ratio (molC/molN) [=xthetad]'
  pmin(i)=0.

  i=9
  pname(i)='rfen'
  pdesc(i)='phytoplankton Fe:N ratio (molFe/molN) [=xrfn]'
  pmin(i)=0.

! Primary production and chl related quantities

  i=10
  pname(i)='alphachlndia'
  pdesc(i)='chl. specific initial slope of P-I curve for non-diatoms (gC/gChl/(W/m^2)/d) [=xaln]' 
  pmin(i)=0.

  i=11
  pname(i)='falphachldiat'
  pdesc(i)='chl. specific initial slope of P-I curve for diatoms relative to non-diatoms [=xald/xaln]'
  pmin(i)=0.

  i=12
  pname(i)='photmax0ndia'
  pdesc(i)='maximum growth rate for non-diatoms at 0 degC (/d) [=xvpn]'      
  pmin(i)=0.

  i=13
  pname(i)='fphotmax0diat'
  pdesc(i)='maximum growth rate for diatoms relative to non-diatoms at 0 degC [=xvpd/xvpn]'      
  pmin(i)=0.

  i=14
  pname(i)='ursininf'
  pdesc(i)='hypothetical growth ratio at infinite Si:N ratio [=xuif]'      
  pmin(i)=0.

! Nutrient limitation

  i=15
  pname(i)='kdinndia'
  pdesc(i)='half-sat. conc. for DIN uptake by non-diatoms (mmolN/m^3) [=xnln]' 
  pmin(i)=0.

  i=16
  pname(i)='fkdindiat'
  pdesc(i)='half-sat. conc. for DIN uptake by diatoms relative to non-diatoms [=xnld/xnln]'
  pmin(i)=0.
 
  i=17
  pname(i)='ksidiat'
  pdesc(i)='half-sat. conc. for Si uptake by diatoms (mmolSi/m^3) [=xsld]'
  pmin(i)=0.

  i=18
  pname(i)='kfendia'
  pdesc(i)='half-sat. conc. for Fe uptake by non-diatoms (mmolFe/m^3) [=xfln]'
  pmin(i)=0.

  i=19
  pname(i)='fkfediat'
  pdesc(i)='half-sat. conc. for Fe uptake by diatoms relative to non-diatoms [=xfld/xfln]'
  pmin(i)=0.

! Grazing

  i=20
  pname(i)='gmaxzmi'
  pdesc(i)='microzoo. maximum grazing rate (/d) [=xgmi]'      
  pmin(i)=0.

  i=21
  pname(i)='fgmaxzme'
  pdesc(i)='mesozoo. maximum grazing rate relative to microzoo. [=xgme/xgmi]'      
  pmin(i)=0.

  i=22
  pname(i)='kgzmi'
  pdesc(i)='microzoo. grazing half-sat. conc. (mmolN/m^3) [=xkmi]'      
  pmin(i)=0.

  i=23
  pname(i)='fkgzme'
  pdesc(i)='mesozoo. grazing half-sat. conc. relative to microzoo. [=xkme/xkmi]'      
  pmin(i)=0.

  i=24
  pname(i)='fgmessy'
  pdesc(i)='micro/mesozoo. grazing inefficiency [=xphi]'      
  pmin(i)=0.
  pmax(i)=1.

  i=25
  pname(i)='betan'
  pdesc(i)='micro/mesozoo. N assimilation efficiency [=xbetan]'    
  pmin(i)=0.
  pmax(i)=1.

  i=26
  pname(i)='offbetac'
  pdesc(i)='C assim. efficiency offset as frac. of max. [=(xbetac-xbetan)/(1-xbetan) OR (xbetac-xbetan)/xbetan]'    
  pmin(i)=0.
  pmax(i)=1.

  i=27
  pname(i)='kc'
  pdesc(i)='micro/mesozoo. net C growth efficiency [=xkc]'       
  pmin(i)=0.
  pmax(i)=1.

  i=28
  pname(i)='prfzmilive'
  pdesc(i)='grazing preference of microzoo. for live food (=non-diatoms) [=xpmipn]'    
  pmin(i)=0.
  pmax(i)=1.

  i=29
  pname(i)='prfzmelive'
  pdesc(i)='grazing preference of mesozoo. for live food [=xpmepn+xpmepd+xpmezmi]'    
  pmin(i)=0.
  pmax(i)=1.

  i=30
  pname(i)='cprfzmep'
  pdesc(i)='grazing preference of mesozoo. for phyto. given live food [=(xpmepn+xzmepd)/(xpmepn+xpmepd+xpmezmi)]'    
  pmin(i)=0.
  pmax(i)=1.

  i=31
  pname(i)='cprfzmepndia'
  pdesc(i)='grazing preference of mesozoo. for non-diatoms given phytoplankton [=xpmepn/(xpmepn+xpmepd)]'   
  pmin(i)=0.
  pmax(i)=1.

! Metabolic losses

  i=32
  pname(i)='metapndia'
  pdesc(i)='non-diatom metabolic loss rate (/d) [=xmetapn]'   
  pmin(i)=0.

  i=33
  pname(i)='fmetapdiat'
  pdesc(i)='diatom metabolic loss rate relative to non-diatoms [=xmetapd/xmetapn]'   
  pmin(i)=0.

  i=34
  pname(i)='fmetazmi'
  pdesc(i)='microzoo. metabolic loss rate relative to non-diatom phyto. [=xmetazmi/xmetapn]'  
  pmin(i)=0.

  i=35
  pname(i)='fmetazme'
  pdesc(i)='mesozoo. metabolic loss rate relative to microzoo. [=xmetazme/xmetazmi]'  
  pmin(i)=0.

! Mortality losses

  i=36
  pname(i)='mortmaxpndia'
  pdesc(i)='non-diatom maximum mortality rate (/d) [=xmpn]'      
  pmin(i)=0.

  i=37
  pname(i)='kmortpndia'
  pdesc(i)='non-diatom mortality half-sat. conc. (mmolN/m^3) [=xkphn]'     
  pmin(i)=0.

  i=38
  pname(i)='fmortmaxpdiat'
  pdesc(i)='diatom maximum mortality rate relative to non-diatoms [=xmpd/xmpn]'      
  pmin(i)=0.

  i=39
  pname(i)='fkmortpdiat'
  pdesc(i)='diatom mortality half-sat. conc. relative to non-diatoms [=xkphd/xkphn]'     
  pmin(i)=0.

  i=40
  pname(i)='fmortmaxzmi'
  pdesc(i)='microzoo. maximum mortality rate relative to non-diatom phyto. [=xmzmi/xmpn]'     
  pmin(i)=0.

  i=41
  pname(i)='fkmortzmi'
  pdesc(i)='microzoo. mortality half-sat. conc. relative to non-diatom phyto. [=xkzmi/xkphn]'     
  pmin(i)=0.

  i=42
  pname(i)='fmortmaxzme'
  pdesc(i)='mesozoo. maximum mortality rate relative to microzoo. [=xmzme/xmzmi]'     
  pmin(i)=0.

  i=43
  pname(i)='fkmortzme'
  pdesc(i)='mesozoo. mortality half-sat. conc. relative to microzoo. [=xkzme/xkzmi]'     
  pmin(i)=0.

! Remineralisation

  i=44
  pname(i)='remin0'
  pdesc(i)='detrital nitrogen remineralisation rate at 0 degC (/d) [=xmd]'
  pmin(i)=0.

! Iron parameters

  i=45
  pname(i)='rfedust'
  pdesc(i)='soluble iron fraction of aeolian dust deposition [=xfe_sol]'   
  pmin(i)=0.
  pmax(i)=1.

  i=46
  pname(i)='sedfe'
  pdesc(i)='iron input from sediment (mmolFe/m2/d) [=xfe_sed]'   
  pmin(i)=0.

  i=47
  pname(i)='ligand'
  pdesc(i)='total ligand concentration (mmol/m^3) [=xLgT*1e-3]'      
  pmin(i)=0.

  i=48
  pname(i)='kfelig'
  pdesc(i)='dissociation constant for (Fe + ligand) (/(mmolFe/m^3)) [=xk_FeL*1e3]'      
  pmin(i)=0.

  i=49
  pname(i)='scavfe'
  pdesc(i)='scavenging rate of "free" iron (/d) [=xk_sc_Fe]'  
  pmin(i)=0

! Fast-sinking detritus parameters

  i=50
  pname(i)='ffastmortdiat'
  pdesc(i)='fast-sinking fraction of diatom mortality losses [=xfdfrac1]'  
  pmin(i)=0.
  pmax(i)=1.

  i=51
  pname(i)='ffastmortzme'
  pdesc(i)='fast-sinking fraction of mesozoo. mortality losses [=xfdfrac2]'  
  pmin(i)=0.
  pmax(i)=1.

  i=52
  pname(i)='rco3fast90'
  pdesc(i)='polar CaCO3 to organic C ratio [=xcaco3a]'   
  pmin(i)=0.

  i=53
  pname(i)='rco3fast0'
  pdesc(i)='equatorial CaCO3 to organic C ratio [=xcaco3b]'   
  pmin(i)=0.

  i=54
  pname(i)='rprotca'
  pdesc(i)='calcium carbonate protection ratio (gC/gC) [=xprotca]'   
  pmin(i)=0.
  pmax(i)=1.

  i=55
  pname(i)='rprotsi'
  pdesc(i)='biogenic silicon protection ratio (gC/gSi) [=xprotsi]'   
  pmin(i)=0.
  pmax(i)=1.

  i=56
  pname(i)='lfastc'
  pdesc(i)='excess organic C dissolution length scale (m) [=xfastc]'    
  pmin(i)=0.

  i=57
  pname(i)='lfastca'
  pdesc(i)='calcium carbonate dissolution length scale (m) [=xfastca]'   
  pmin(i)=0.

  i=58
  pname(i)='lfastsi'
  pdesc(i)='biogenic silicon dissolution length scale (m) [=xfastsi]'   
  pmin(i)=0.

! Miscellaneous

  i=59
  pname(i)='sidiss'
  pdesc(i)='diatom frustule dissolution rate (/d) [=xsdiss]'    
  pmin(i)=0.

! Gravitational sinking

  i=60
  pname(i)='sinkdet'
  pdesc(i)='detritus gravitational sinking rate (m/d) [=vsed*86400]'      
  pmin(i)=0.

! Optical model parameters

  i=61
  pname(i)='attenwaterg'
  pdesc(i)='downwelling attenuation of blue-green band due to water (/m) [=xkg0]' 
  pmin(i)=0.

  i=62
  pname(i)='attenwaterr'
  pdesc(i)='downwelling attenuation of red band due to water (/m) [=xkr0]' 
  pmin(i)=0.

  i=63
  pname(i)='attenupigg'
  pdesc(i)='downwelling attenuation of blue-green band due to 1 mg/m^3 of pigment (/m) [=xkgp]' 
  pmin(i)=0.

  i=64
  pname(i)='attenupigr'
  pdesc(i)='downwelling attenuation of red band due to 1 mg/m^3 of pigment (/m) [=xkrp]' 
  pmin(i)=0.

  i=65
  pname(i)='attenexpigg'
  pdesc(i)='exposant for pigment (mg/m^3) absorption in blue-green band [=xlg]'
  pmin(i)=0.

  i=66
  pname(i)='attenexpigr'
  pdesc(i)='exposant for pigment (mg/m^3) absorption in red band [=xlr]'  
  pmin(i)=0.

  i=67
  pname(i)='rchlpig'
  pdesc(i)='chl. to total pigment ratio [=rpig]' 
  pmin(i)=tiny(pmin(i))
  pmax(i)=1.

! Option switches 

  i=68
  pname(i)='photmaxtdep'
  pdesc(i)='temperature-dependent phytoplankton max. growth (0: off, 1: on) [=jphy]'
  pmin(i)=0.
  pmax(i)=1.

  i=69
  pname(i)='mortfnpndia'
  pdesc(i)='non-diatom mortality functional form (1: linear, 2: quadratic, 3: hyperbolic, 4: sigmoid) [=jmpn]'
  pmin(i)=1.
  pmax(i)=4.

  i=70
  pname(i)='mortfnpdiat'
  pdesc(i)='diatom mortality functional form (1: linear, 2: quadratic, 3: hyperbolic, 4: sigmoid) [=jmpd]'
  pmin(i)=1.
  pmax(i)=4.

  i=71
  pname(i)='mortfnzmi'
  pdesc(i)='microzoo mortality functional form (1: linear, 2: quadratic, 3: hyperbolic, 4: sigmoid) [=jmzmi]'
  pmin(i)=1.
  pmax(i)=4.

  i=72
  pname(i)='mortfnzme'
  pdesc(i)='mesozoo mortality functional form (1: linear, 2: quadratic, 3: hyperbolic, 4: sigmoid) [=jmzme]'
  pmin(i)=1.
  pmax(i)=4.

  i=73
  pname(i)='remintdep'
  pdesc(i)='temperature-dependent detrital remineralisation (0: off, 1: on) [=jmd]'
  pmin(i)=0.
  pmax(i)=1.

  i=74
  pname(i)='photopt'
  pdesc(i)='implementation of photosynthesis (0: external, 1: internal) [=jphot]'
  pmin(i)=0.
  pmax(i)=1.

  i=75
  pname(i)='dsinkopt'
  pdesc(i)='implementation of detrital sinking (0: external, 1: internal) [=jsink]'
  pmin(i)=0.
  pmax(i)=1.

  i=76
  pname(i)='dustopt'
  pdesc(i)='conversion of dust to soluble iron (0: external, 1: internal) [=jdust]'
  pmin(i)=0.
  pmax(i)=1.

  i=77
  pname(i)='logopt'
  pdesc(i)='log output to stdout (0: off, 1: on) [=lwp]'
  pmin(i)=0.
  pmax(i)=1.

  end subroutine

!===============================================================================

  subroutine getinfo(parm,photflag,latflag,advectflag,errmsg)

  use modconfig, only: msgtextlen
  use medusa_globdata, only: jphot,xpmepn,xpmepd,xpmezmi 

  implicit none

! Input argument

  real(kind_r_ext) parm(:)             ! Model-specific parameters

! Output arguments

  integer photflag                     ! Photosynthesis model flag
                                       ! -1: external phot model prohibited
                                       ! 1: external phot model required
                                       ! 2: external phot model with 
                                       !    phot:(alpha*par) ratio required 
  integer latflag                      ! Latitude flag (1 if required)
  integer advectflag                   ! Advection scheme flag (1 if required)
  character(len=msgtextlen+1) errmsg   ! Error message

!-------------------------------------------------------------------------------

  if (size(parm)==0) then
    jphot=1
    jsink=1
  else if (size(parm)>=nparm) then
    xpmepn=parm(31)*parm(30)*parm(29)
    xpmepd=(1.-parm(31))*parm(30)*parm(29)
    xpmezmi=(1.-parm(30))*parm(29)
    if (xpmepn+xpmepd+xpmezmi>1.) then
      errmsg='mesozooplankton feeding preferences sum to more than 1'
    endif
    jphot=nint(parm(74))
    jsink=nint(parm(75))
  else
    write(*,*) "Setup error in MEDUSA model - too few parameters"
    stop
  endif

  latflag=1

  if (jphot==1) then
    photflag=-1
  else
    photflag=2
  endif

  if (jsink==0) advectflag=1

  end subroutine

!===============================================================================

  subroutine setup(kmax,zbot,lat,parm, &
                   ntinit,ntlinked,itbase,it,nrinit,ir, &
                   nphyext,reqsol,reqdust,reqfedust,reqtemp,dsinkext)

  use modconfig, only: i_pndia,i_pdiat,i_zmi,i_zme,i_din,i_dsi,i_dfe,i_det, &
    i_chlndia,i_chldiat,i_pdiatsi,i_rcchlndia,i_rcchldiat,i_rsinpdiat, &
    i_rcnpndia,i_rcnpdiat,i_rcnzmi,i_rcnzme,i_rcndet,i_rfen, &
    i_kdpar,i_parfracztop,i_parztop, &
    i_photmaxndia,i_pmaxndia,i_alphandia,i_pkndia,i_limdinndia,i_limfendia, &
    i_photndia,i_photmaxdiat,i_pmaxdiat,i_alphadiat,i_pkdiat,i_limdindiat, &
    i_limfediat,i_limsidiat,i_photdiat,i_wdet,i_insinkn,i_insinksi, &
    i_insinkfe,i_insinkc,i_insinkca,i_fescav,i_feinsurf,i_feinbed, &
    atomc,atomo,atomsi,atomca,atomfe 

  use medusa_globdata, only: fsdepw,fse3t,mbathy,hmld,gphit, &
    xfe_mass,xmassc,xmassca,xmasssi,jphy,jmpn,jmpd,jmzmi,jmzme,jmd,jphot, &
    jdust,lwp,jpphn,jpphd,jpzmi,jpzme,jpdin,jpsil,jpfer,jpdet,jpchn,jpchd,jppds

  implicit none

  integer,parameter :: ntlnk=3

! Input arguments

  integer kmax                    ! Number of levels
  real(kind_r_ext) zbot(:)        ! Depth at bottom of level (m)
  real(kind_r_ext) lat            ! Latitude (deg N)
  real(kind_r_ext) parm(:)        ! Model-specific parameters

! Output arguments

  integer ntinit                  ! Number of tracers to be initialized
  integer ntlinked                ! Number of linked tracers
  integer itbase(:)               ! Local indices of tracers that are linked to
  integer it(:)                   ! Indices of tracers in MarMOT array
  integer nrinit                  ! Number of ratios to be initialized
  integer ir(:)                   ! Indices of ratios in MarMOT array
  integer nphyext                 ! Number of phytoplankton types for
                                  !   external photosynthesis model
  integer reqsol                  ! Flag requesting solar radiation
  integer reqdust                 ! Flag requesting dust deposition 
  integer reqfedust               ! Flag requesting aeolian iron input
  integer reqtemp                 ! Flag requesting temperature 
  integer dsinkext                ! Flag indicating external handling of
                                  !   detrital sinking
!-------------------------------------------------------------------------------

  integer jk 

! Set up vertical grid information

  jpk=kmax
  if (jpk > nk-1) jpk=nk-1
  mbathy(ji,jj)=jpk+1

  fsdepw(ji,jj,1)=0.
  do jk=2,jpk+1 
    fsdepw(ji,jj,jk)=zbot(jk-1)
  enddo

  fse3t(ji,jj,1)=zbot(1)
  do jk=2,jpk 
    fse3t(ji,jj,jk)=zbot(jk)-zbot(jk-1)
  enddo

  hmld(ji,jj)=0.  ! Set a dummy MLD for unused internal mixed layer diagnostics
 
! Initialize time step counter

  kt=0

! Set latitude

  gphit(ji,jj)=lat 

! Set fixed parameters

  xfe_mass  = atomfe 
  xmassc    = atomc 
  xmassca   = atomca+atomc+3*atomo
  xmasssi   = atomsi+2*atomo 

! Set model options

  if (size(parm)==0) then

    jphy=1
    jmpn=3
    jmpd=3
    jmzmi=3
    jmzme=3
    jmd=1
    jphot=1
    jsink=1
    jdust=0
    lwp=.FALSE.
   
  else if (size(parm)>=nparm) then

    jphy=nint(parm(68))
    jmpn=nint(parm(69))
    jmpd=nint(parm(70))
    jmzmi=nint(parm(71))
    jmzme=nint(parm(72))
    jmd=nint(parm(73))
    jphot=nint(parm(74))
    jsink=nint(parm(75))
    jdust=nint(parm(76))
   
    if (parm(77)>0.5) then
      lwp=.TRUE.
    else
      lwp=.FALSE.
    endif

  else

    write(*,*) "Setup error in MEDUSA model - too few parameters"
    stop

  endif

! Set indices for model access to MarMOT arrays 

  itracer(jpphn)=i_pndia
  itracer(jpphd)=i_pdiat
  itracer(jpzmi)=i_zmi
  itracer(jpzme)=i_zme
  itracer(jpdin)=i_din
  itracer(jpsil)=i_dsi
  itracer(jpfer)=i_dfe
  itracer(jpdet)=i_det
  itracer(jpchn)=i_chlndia
  itracer(jpchd)=i_chldiat
  itracer(jppds)=i_pdiatsi

  iratio(1)=i_rcchlndia
  iratio(2)=i_rcchldiat
  iratio(3)=i_rsinpdiat
  iratio(4)=i_rcnpndia
  iratio(5)=i_rcnpdiat
  iratio(6)=i_rcnzmi
  iratio(7)=i_rcnzme
  iratio(8)=i_rcndet
  iratio(9)=i_rfen

  idiagprof(1)=i_kdpar
  idiagprof(2)=i_parfracztop
  idiagprof(3)=i_parztop
  idiagprof(4)=i_photmaxndia
  idiagprof(5)=i_pmaxndia
  idiagprof(6)=i_alphandia
  idiagprof(7)=i_pkndia
  idiagprof(8)=i_limdinndia
  idiagprof(9)=i_limfendia
  idiagprof(10)=i_photndia
  idiagprof(11)=i_photmaxdiat
  idiagprof(12)=i_pmaxdiat
  idiagprof(13)=i_alphadiat
  idiagprof(14)=i_pkdiat
  idiagprof(15)=i_limdindiat
  idiagprof(16)=i_limfediat
  idiagprof(17)=i_limsidiat
  idiagprof(18)=i_photdiat
  idiagprof(19)=i_wdet
  idiagprof(20)=i_insinkn
  idiagprof(21)=i_insinksi
  idiagprof(22)=i_insinkfe
  idiagprof(23)=i_insinkc
  idiagprof(24)=i_insinkca
  idiagprof(25)=i_fescav

  idiagscal(1)=i_feinsurf
  idiagscal(2)=i_feinbed

! Set indices for socket layer access to MarMOT tracer & ratio arrays

  if (size(it)<nt .or. size(itbase)<ntlnk) then
    write(*,*) "Setup error in MEDUSA model - required tracers not accessible"
    stop
  endif

  ntinit=nt-ntlnk
  ntlinked=ntlnk
  itbase(1)=jpphn
  itbase(2)=jpphd
  itbase(3)=jpphd
  it(1:nt)=itracer(1:nt)

  if (size(ir)<nr) then
    write(*,*) "Setup error in MEDUSA model - required ratios not accessible"
    stop
  endif

  nrinit=3
  ir(1:nrinit)=iratio(1:nrinit)

! Set forcing data request flags, external photosynthesis configuration &
! external detrital sinking indicator

  if (jphot==0) then
    nphyext=2
  else
    reqsol=1
    nphyext=0
  endif

  if (jdust==1) then
    reqdust=1
  else
    reqfedust=1
  endif

  if (jphy/=0 .or. jmd/=0) reqtemp=1

  if (jsink==0) dsinkext=1

  end subroutine

!===============================================================================

  subroutine setparm(parm,ratio_default,linkratio_def,linkratio_min, &
                     linkratio_max)

  use medusa_globdata

  implicit none

! Input arguments

  real(kind_r_ext) parm(:)       ! Model parameters

! Output arguments

  real(kind_r) ratio_default(:)  ! Default values for initial ratios
  real(kind_r) linkratio_def(:)  ! Default tendency ratio for linked tracers
  real(kind_r) linkratio_min(:)  ! Min. tendency ratio for linked tracers
  real(kind_r) linkratio_max(:)  ! Max. tendency ratio for linked tracers

!-------------------------------------------------------------------------------

  if (size(parm)==0) then

!   Use values from Experiment no. 276. This is the standard experiment
!   presented in Yool et al 2011.

!   Primary production and chl related quantities

    xaln      = 15.0     
    xald      = 11.25  
    xvpn      = 0.530   
    xvpd      = 0.500  
    xthetam   = 0.05  
    xthetamd  = 0.05  

!   Diatom silicon parameters

    xsin0     = 0.2  
    xnsi0     = 0.2 
    xuif      = 1.5

!   Nutrient limitation

    xnln      = 0.5  
    xnld      = 0.75 
    xsld      = 0.75 
    xfln      = 0.00033 
    xfld      = 0.00067 

!   Grazing

    xgmi      = 2.0 
    xgme      = 0.5 
    xkmi      = 0.8 
    xkme      = 0.3 
    xphi      = 0.20
    xbetan    = 0.69 
    xbetac    = 0.69 
    xkc       = 0.80 
    xpmipn    = 0.75
    xpmepn    = 0.15
    xpmepd    = 0.35
    xpmezmi   = 0.35

!   Metabolic losses

    xmetapn   = 0.02
    xmetapd   = 0.02
    xmetazmi  = 0.02
    xmetazme  = 0.02

!   Mortality losses

    xmpn      = 0.1 
    xkphn     = 0.5 
    xmpd      = 0.1 
    xkphd     = 0.5  
    xmzmi     = 0.1  
    xkzmi     = 0.5 
    xmzme     = 0.2 
    xkzme     = 0.75

!   Remineralisation

    xmd       = 0.016  

!   Stochiometric ratios

    xthetap   = 6.625     
    xthetazmi = 6.625     
    xthetazme = 6.625     
    xthetad   = 6.625     
    xrfn      = 30.0e-6   

!   Iron parameters

    xfe_sol   = 0.014 
    xfe_sed   = 0.0
    xLgT      = 1.0     
    xk_FeL	= 100.0 
    xk_sc_Fe  = 1.e-3   

!   Fast-sinking detritus parameters

    xfdfrac1  = 0.75   
    xfdfrac2  = 1.00  
    xcaco3a   = 0.02  
    xcaco3b   = 0.10  
    xprotca   = 0.070  
    xprotsi   = 0.026  
    xfastc    = 188.0  
    xfastca   = 3500.0 
    xfastsi   = 2000.0 

!   Miscellaneous

    xsdiss    = 0.006    

!   Gravitational sinking

    vsed      = 3.472e-5  

!   Optical model parameters

    xkg0 = 0.0232
    xkr0 = 0.225
    xkgp = 0.074
    xkrp = 0.037
    xlg  = 0.629
    xlr  = 0.674
    rpig = 0.7

  else if (size(parm)>=nparm) then

!   Primary production and chl related quantities

    xaln      = parm(10) 
    xald      = parm(10)*parm(11) 
    xvpn      = parm(12)
    xvpd      = parm(12)*parm(13) 
    xthetam   = 1./parm(1)
    xthetamd  = 1./(parm(1)*parm(2))

!   Diatom silicon parameters

    xsin0     = parm(3)
    xnsi0     = 1./parm(4)
    xuif      = parm(14)

!   Nutrient limitation

    xnln      = parm(15)
    xnld      = parm(15)*parm(16)
    xsld      = parm(17)
    xfln      = parm(18)
    xfld      = parm(18)*parm(19)

!   Grazing

    xgmi      = parm(20)
    xgme      = parm(20)*parm(21)
    xkmi      = parm(22) 
    xkme      = parm(22)*parm(23)
    xphi      = parm(24)
    xbetan    = parm(25)

    if (parm(26) > 0.) then
      xbetac = parm(25)+parm(26)*(1.-parm(25))
    else
      xbetac = parm(25)+parm(26)*parm(25)
    endif

    xkc       = parm(27)
    xpmipn    = parm(28)
    xpmepn    = parm(31)*parm(30)*parm(29)
    xpmepd    = (1.-parm(31))*parm(30)*parm(29)
    xpmezmi   = (1.-parm(30))*parm(29)

!   Metabolic losses

    xmetapn   = parm(32)
    xmetapd   = parm(32)*parm(33)
    xmetazmi  = parm(32)*parm(34)
    xmetazme  = parm(32)*parm(34)*parm(35)

!   Mortality losses

    xmpn      = parm(36)
    xkphn     = parm(37)
    xmpd      = parm(36)*parm(38)
    xkphd     = parm(37)*parm(39)
    xmzmi     = parm(36)*parm(40)
    xkzmi     = parm(37)*parm(41)
    xmzme     = parm(36)*parm(40)*parm(42)
    xkzme     = parm(37)*parm(41)*parm(43)

!   Remineralisation

    xmd       = parm(44)

!   Stochiometric ratios

    xthetap   = parm(5)
    xthetazmi = parm(6)
    xthetazme = parm(7)
    xthetad   = parm(8)
    xrfn      = parm(9)

!   Iron parameters

    xfe_sol   = parm(45) 
    xfe_sed   = parm(46) 
    xLgT      = parm(47)*1e3
    xk_FeL    = parm(48)*1e-3
    xk_sc_Fe  = parm(49)

!   Fast-sinking detritus parameters

    xfdfrac1  = parm(50)
    xfdfrac2  = parm(51)
    xcaco3a   = parm(52)
    xcaco3b   = parm(53)
    xprotca   = parm(54)
    xprotsi   = parm(55)
    xfastc    = parm(56)
    xfastca   = parm(57)
    xfastsi   = parm(58)

!   Miscellaneous

    xsdiss    = parm(59)

!   Gravitational sinking

    vsed      = parm(60)/86400.

!   Optical model parameters

    xkg0 = parm(61)
    xkr0 = parm(62)
    xkgp = parm(63)
    xkrp = parm(64)
    xlg  = parm(65)
    xlr  = parm(66)
    rpig = parm(67)

  else

    write(*,*) "Setup error in MEDUSA model - too few parameters"
    stop 

  endif

! Derived parameters

  xxi=1./(xthetap*atomc)
  xthetapn=xthetap
  xthetapd=xthetap

  xpmid=1.-xpmipn
  xpmed=1.-(xpmepn+xpmepd+xpmezmi)

  if (xpmed<0) then 
    write(*,*) &
    "Setup error in MEDUSA model - bad mesozoo. feeding preference combination"
    stop 
  endif

! Default ratios (in MarMOT ratio units)

  ratio_default(1)=1./xthetam 
  ratio_default(2)=1./xthetamd 
  ratio_default(3)=xsin0

! Ratio bounds & default for tracer linking (in raw tracer ratio units) 

  linkratio_def(1)=xthetam/xxi
  linkratio_min(1)=0.
  linkratio_max(1)=xthetam/xxi
  linkratio_def(2)=xthetamd/xxi
  linkratio_min(2)=0.
  linkratio_max(2)=xthetamd/xxi 

  linkratio_def(3)=xsin0
  linkratio_min(3)=xsin0
  linkratio_max(3)=1./xnsi0

  end subroutine

!===============================================================================

  subroutine initstate(tracer,ratio)

  use medusa_globdata, only: xthetam,xthetamd,xsin0,xnsi0,xthetapn,xthetapd, &
                             xthetazmi,xthetazme,xthetad,xrfn, &
                             jpchn,jpchd,jpphn,jpphd,jppds

  implicit none

! Arguments (in/out)

  real(kind_r_ext) tracer(:,:)      ! MarMOT tracer array
  real(kind_r_ext) ratio(:,:)       ! MarMOT composition ratio array

!-------------------------------------------------------------------------------

  real(wp) conv
  integer jk

! Update C:Chl, Si:N if out of range 

  do jk=1,jpk
    ratio(jk,iratio(1))=max(ratio(jk,iratio(1)),1./xthetam)
    ratio(jk,iratio(2))=max(ratio(jk,iratio(2)),1./xthetamd)
    ratio(jk,iratio(3))=min(max(ratio(jk,iratio(3)),xsin0),1./xnsi0)
  enddo

! Calculate derived tracers 

  conv=xthetapn*atomc
  do jk=1,jpk
    tracer(jk,itracer(jpchn))=tracer(jk,itracer(jpphn))*conv/ratio(jk,iratio(1))
  enddo

  conv=xthetapd*atomc
  do jk=1,jpk
    tracer(jk,itracer(jpchd))=tracer(jk,itracer(jpphd))*conv/ratio(jk,iratio(2))
  enddo

  do jk=1,jpk
    tracer(jk,itracer(jppds))=tracer(jk,itracer(jpphd))*ratio(jk,iratio(3))
  enddo
  
! Output fixed composition ratios

  do jk=1,jpk
    ratio(jk,iratio(4))=xthetapn    !  non-diatom C:N ratio
    ratio(jk,iratio(5))=xthetapd    !  diatom C:N ratio
    ratio(jk,iratio(6))=xthetazmi   !  microzoo C:N ratio
    ratio(jk,iratio(7))=xthetazme   !  mesozoo C:N ratio
    ratio(jk,iratio(8))=xthetad     !  detritus C:N ratio
    ratio(jk,iratio(9))=xrfn        !  phytoplankton Fe:N ratio

  enddo

  end subroutine

!===============================================================================

  subroutine getbiotrans(wdet)

  use medusa_globdata, only: vsed 

  implicit none

! Output argument

  real(kind_r) wdet         ! Detritus velocity due to sinking (m/d, up)

! Set active tracer velocity for advection scheme if sinking is to be handled
! externally

  if (jsink==0) then
    wdet=-vsed*86400.
  else
    wdet=0.
  endif

  end subroutine

!===============================================================================

  subroutine getphotin(f_sol,f_temp,tracer,par0,pigment,alpha,chlflag,cchl,vp)

  use modconfig, only: EpdtoW
  use medusa_globdata, only: xald,xaln,jpphn,jpphd,jpchd,jpchn,jphy,rpig, &
                             xxi,xvpn,xvpd

  implicit none

! Input arguments

  real(kind_r_ext) f_sol        ! Surface irradiance (W/m^2)
  real(kind_r_ext) f_temp(:)    ! Temperature (degC)
  real(kind_r_ext) tracer(:,:)  ! MarMOT tracer array 

! Output arguments

  real(kind_r) par0             ! Surface PAR (E/m^2/d)
  real(kind_r) pigment(:)       ! Pigment conc. (mg/m^3) 
  real(kind_r) alpha(:)         ! Initial slope of P-E curve for scalar
                                !   PAR at depth 0+ (gC/gChl/(E/m^2))
  logical chlflag(:,:)          ! Switch for presence of chlorophyll 
  real(kind_r) cchl(:,:)        ! C:Chl (gC/gChl)
  real(kind_r) vp(:,:)          ! Phytoplankton max. growth rate (/d)

!-------------------------------------------------------------------------------

  real(wp) zphn
  real(wp) zphd
  real(wp) zchn
  real(wp) zchd
  real(wp) fun_T 
  integer jk

  par0=0.43*f_sol/EpdtoW
  alpha(1)=xaln*EpdtoW
  alpha(2)=xald*EpdtoW

  do jk=1,jpk

    zchn=max(0.,tracer(jk,itracer(jpchn)))
    zchd=max(0.,tracer(jk,itracer(jpchd)))
    zphn=max(0.,tracer(jk,itracer(jpphn)))
    zphd=max(0.,tracer(jk,itracer(jpphd)))

    if (zchn.eq.0.) zphn=0.
    if (zchd.eq.0.) zphd=0.
    if (zphn.eq.0.) zchn=0.
    if (zphd.eq.0.) zchd=0.

    pigment(jk)=(zchn+zchd)/rpig
    chlflag(jk,1)=(zchn/=0.0)
    chlflag(jk,2)=(zchd/=0.0)
    cchl(jk,1)=1./max(tiny(zchn),(zchn*xxi)/(zphn+tiny(zphn)))
    cchl(jk,2)=1./max(tiny(zchd),(zchd*xxi)/(zphd+tiny(zphd)))

    if (jphy.eq.1) then
      fun_T=1.066**(f_temp(jk))
      vp(jk,1)=xvpn*fun_T
      vp(jk,2)=xvpd*fun_T
    else
      vp(jk,1)=xvpn
      vp(jk,2)=xvpd
    endif

  enddo

  end subroutine

!===============================================================================

  subroutine biotrend(f_sol,f_temp,f_dust,f_fedust,phot,rphotai,tracer,ddt, &
                      zeuphotic) 

  use medusa_globdata, only: tra,qsr,dust,tn,xfe_mass,jphot,jdust,xze, &
                             fjln_in,fjld_in,fchn_in,fchd_in

  use medusa_opt, only: trc_opt_medusa
  use medusa_bio, only: trc_bio_medusa
  use medusa_sed, only: trc_sed_medusa

  implicit none

! Input arguments 

  real(kind_r_ext) f_sol       ! Surface irradiance (W/m^2)
  real(kind_r_ext) f_temp(:)   ! Temperature (degC)
  real(kind_r_ext) f_dust      ! Aeolian dust deposition (mg/m^2/d)
  real(kind_r_ext) f_fedust    ! Aeolian input of soluble iron (mmolFe/m^2/d)
  real(kind_r) phot(:,:)       ! Photosythetic rate (/d)  
  real(kind_r) rphotai(:,:)    ! Photosythesis:(alpha*par) ratio 
  real(kind_r_ext) tracer(:,:) ! MarMOT tracer array 

! In/out argument

  real(kind_r_ext) ddt(:,:)    ! MarMOT tendency array (units/d)

! Output argument

  real(kind_r) zeuphotic       ! Euphotic zone depth (m)

!-------------------------------------------------------------------------------

  integer i,j,jk

! Copy state and transport tendencies from MarMOT

  do i=1,nt
    j=itracer(i)
    do jk=1,jpk
      trn(ji,jj,jk,i)=tracer(jk,j)
      tra(ji,jj,jk,i)=ddt(jk,j)/86400.
    enddo
  enddo

! Set forcing data

  if (jphot==1) qsr(ji,jj)=f_sol          
  if (jdust==1) then
    dust(ji,jj)=f_dust
  else
    dust(ji,jj)=f_fedust
  endif 
  do jk=1,jpk
    tn(ji,jj,jk)=f_temp(jk)
  enddo

! MEDUSA model
 
  kt=kt+1

  if (jphot==1) then
    call trc_opt_medusa(kt)   ! optical model 
    zeuphotic=xze(ji,jj)
  else
    fjln_in(:)=phot(:,1)
    fjld_in(:)=phot(:,2)
    fchn_in(:)=rphotai(:,1)
    fchd_in(:)=rphotai(:,2)
  endif

  call trc_bio_medusa(kt)     ! biological model

  if (jsink==1) call trc_sed_medusa(kt)   ! sedimentation model

! Copy updated tendencies to MarMOT

  do i=1,nt
    j=itracer(i)
    do jk=1,jpk
      ddt(jk,j)=tra(ji,jj,jk,i)*86400.
    enddo
  enddo

  end subroutine 

!===============================================================================

  subroutine getdiag(diagprof,diagscal)

  use modconfig, only: EpdtoW

  use medusa_globdata, only: jphot,fse3t,zparr_top_diag,zparg_top_diag, &
    xvpnT_diag,faln_diag,fjln_diag,fnln_diag,ffln_diag,fprn_diag, &
    xvpdT_diag,fald_diag,fjld_diag,fnld_diag,ffld_diag,fsld2_diag,fprd_diag, &
    vsed,freminn_diag,freminsi_diag,freminfe_diag,freminc_diag,freminca_diag, &
    ftempn_diag,ftempsi_diag,ftempfe_diag,ftempc_diag,ftempca_diag, &
    ffescav_diag,ffetopflx_diag,ffebotflx_diag 
        
  implicit none
        
  real(kind_r_ext) diagprof(:,:)   ! MarMOT profile diagnostics array
  real(kind_r_ext) diagscal(:)     ! MarMOT scalar diagnostics array
!-------------------------------------------------------------------------------

  real(wp) zpar_top,zpar_bot,zpar_surf
  
  integer jk

  if (jphot==1) then 
    zpar_top=zparr_top_diag(1)+zparg_top_diag(1)
    zpar_surf=zpar_top
  endif

  do jk=1,jpk

    ! Light atten. coeff. (/m)

    if (jphot==1) then
      zpar_bot=zparr_top_diag(jk+1)+zparg_top_diag(jk+1)
      diagprof(jk,idiagprof(1))=-(1/fse3t(ji,jj,jk))*log(zpar_bot/zpar_top) 
      diagprof(jk,idiagprof(2))=zpar_top/zpar_surf
      diagprof(jk,idiagprof(3))=zpar_top
      zpar_top=zpar_bot
    endif

    ! Non-diatom photosynthesis

    if (jphot==1) then
      diagprof(jk,idiagprof(4))=xvpnT_diag(jk)  ! Max. photos. (/d) 
      diagprof(jk,idiagprof(5))=xvpnT_diag(jk)  ! Max. photos. for P-I curve
      diagprof(jk,idiagprof(6))=faln_diag(jk)*EpdtoW  
                                                ! Initial slope (/(E/m^2))  
      diagprof(jk,idiagprof(7))=fjln_diag(jk)   ! Potential photos. (/d) 
    endif
    diagprof(jk,idiagprof(8))=fnln_diag(jk)     ! Nitrogen limitation 
    diagprof(jk,idiagprof(9))=ffln_diag(jk)     ! Iron limitation 
    diagprof(jk,idiagprof(10))=fprn_diag(jk)    ! Realized photos. (/d) 

    ! Diatom photosynthesis

    if (jphot==1) then
      diagprof(jk,idiagprof(11))=xvpdT_diag(jk) ! Max. photos. (/d)        
      diagprof(jk,idiagprof(12))=xvpdT_diag(jk) ! Max. photos. for P-I curve
      diagprof(jk,idiagprof(13))=fald_diag(jk)*EpdtoW 
                                                ! Initial slope (/(E/m^2))
      diagprof(jk,idiagprof(14))=fjld_diag(jk)  ! Potential photos. (/d)
    endif
    diagprof(jk,idiagprof(15))=fnld_diag(jk)    ! Nitrogen limitation 
    diagprof(jk,idiagprof(16))=ffld_diag(jk)    ! Iron limitation 
    diagprof(jk,idiagprof(17))=fsld2_diag(jk)   ! Silicon limitation 
    diagprof(jk,idiagprof(18))=fprd_diag(jk)    ! Realized photos. (/d)

    ! Detrital sinking (m/d), net input of material from fast sinking 
    ! detritus (mmol/m^3/d), net carbonate dissolution (mmol/m^3/d)

    if (jsink==1) diagprof(jk,idiagprof(19))=-vsed*86400.     
    diagprof(jk,idiagprof(20))=freminn_diag(jk)-ftempn_diag(jk)   
    diagprof(jk,idiagprof(21))=freminsi_diag(jk)-ftempsi_diag(jk)   
    diagprof(jk,idiagprof(22))=freminfe_diag(jk)-ftempfe_diag(jk)   
    diagprof(jk,idiagprof(23))=freminc_diag(jk)+freminca_diag(jk) &
                               -ftempc_diag(jk)-ftempca_diag(jk)   
    diagprof(jk,idiagprof(24))=freminca_diag(jk)-ftempca_diag(jk)   

    ! Iron scavenging 

    diagprof(jk,idiagprof(25))=ffescav_diag(jk)
 
  enddo

  if (jphot==1) then
    diagprof(jpk+1,idiagprof(2))=zpar_top/zpar_surf
    diagprof(jpk+1,idiagprof(3))=zpar_top
  endif

  ! Atmospheric & benthic iron input

  diagscal(idiagscal(1))=ffetopflx_diag
  diagscal(idiagscal(2))=ffebotflx_diag

  end subroutine

!===============================================================================

  end module medusa 

