!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright NEMO Consortium                                                !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! Code imported from trcsed_medusa.F90 was written by other authors.       !
! See imported header for details.                                         !   
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!===============================================================================
! MODULE medusa_sed
! John Hemmings - 18 Jun 2012 
!
! Contains a version of the NEMO subroutine 'trc_sed_medusa' of Yool et al
! (2011), adapted for use with MarMOT.
!
! Pre-processing keys key_trc_diaadd & key_trc_diabio are assumed
! to be undefined.
!
! The original code is modified to include only the detrital sinking.
! N.B. Level 'jpk' may now be part of water column.
!
! Reference:
! Yool, A., Popova, E.E., and Anderson, T.R., 2011. Medusa-1.0: a new 
! intermediate complexity plankton ecosystem model for the global domain, 
! Geosci. Model Dev., 4, 381–417, doi:10.5194/gmd-4-381-2011.
!===============================================================================

!******************* Imported code [trcsed_medusa.F90] *************************
   !!======================================================================
   !!                         ***  MODULE trcsed_medusa  ***
   !! TOP :   MEDUSA Compute loss of organic matter in the sediments
   !!======================================================================
   !! History :    -   !  1995-06 (M. Levy)  original code
   !!              -   !  2000-12 (E. Kestenare)  clean up
   !!             2.0  !  2007-12  (C. Deltel, G. Madec)  F90 + simplifications
   !!              -   !  2008-08  (K. Popova) adaptation for MEDUSA
   !!              -   !  2008-11  (A. Yool) continuing adaptation for MEDUSA
   !!              -   !  2010-03  (A. Yool) updated for branch inclusion
   !!----------------------------------------------------------------------
   !!----------------------------------------------------------------------
   !!   'key_medusa'                                      MEDUSA bio-model
   !!----------------------------------------------------------------------
   !!   trc_sed_medusa        :  Compute loss of organic matter in the sediments
   !!----------------------------------------------------------------------
!************************** End of imported code *******************************

  MODULE medusa_sed

  USE medusa_globdata

  IMPLICIT NONE
  PRIVATE

  PUBLIC trc_sed_medusa

!******************* Imported code [trcsed_medusa.F90] *************************
   !!----------------------------------------------------------------------
   !! NEMO/TOP 2.0 , LOCEAN-IPSL (2007) 
   !! $Id: trcsed_medusa.F90 1146 2008-06-25 11:42:56Z rblod $ 
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE trc_sed_medusa( kt )
      !!---------------------------------------------------------------------
      !!                     ***  ROUTINE trc_sed_medusa  ***
      !!
      !! ** Purpose :   compute the now trend due to the vertical sedimentation of
      !!              detritus and add it to the general trend of detritus equations
      !!
      !! ** Method  :   this ROUTINE compute not exactly the advection but the
      !!              transport term, i.e.  dz(wt) and dz(ws)., dz(wtr)
      !!              using an upstream scheme
      !!              the now vertical advection of tracers is given by:
      !!                      dz(trn wn) = 1/bt dk+1( e1t e2t vsed (trn) )
      !!              add this trend now to the general trend of tracer (ta,sa,tra):
      !!                             tra = tra + dz(trn wn)
      !!        
      !!              IF 'key_trc_diabio' is defined, the now vertical advection
      !!              trend of passive tracers is saved for futher diagnostics.
      !!---------------------------------------------------------------------
      INTEGER, INTENT( in ) ::   kt      ! ocean time-step index      
!************************** End of imported code *******************************
      INTEGER  ::   ji, jj, jk
      REAL(wp) ::   ztra
      REAL(wp), DIMENSION(jpi,jpj,jpk+1) ::   zwork

!******************* Imported code [trcsed_medusa.F90] *************************

      IF( kt == nit000 ) THEN
         IF(lwp) WRITE(numout,*)
         IF(lwp) WRITE(numout,*) ' trc_sed_medusa: MEDUSA sedimentation'
         IF(lwp) WRITE(numout,*) ' ~~~~~~~'
      ENDIF

!************************** End of imported code *******************************

      ! sedimentation of detritus  : upstream scheme
      ! --------------------------------------------

      ! for detritus sedimentation only - jpdet

      zwork(:,:,1  ) = 0.e0      ! surface value set to zero
      zwork(:,:,jpk+1) = 0.e0    ! bottom value  set to zero  

      ! tracer flux at w-point: we use -vsed (downward flux)  with simplification : no e1*e2

      DO jk = 2, jpk+1  
         zwork(:,:,jk) = -vsed * trn(:,:,jk-1,jpdet)
      END DO

      ! tracer flux divergence at t-point added to the general trend

      DO jk = 1, jpk   
!******************* Imported code [trcsed_medusa.F90] *************************
         DO jj = 1, jpj
            DO ji = 1,jpi
               ztra  = - ( zwork(ji,jj,jk) - zwork(ji,jj,jk+1) ) / fse3t(ji,jj,jk)
               tra(ji,jj,jk,jpdet) = tra(ji,jj,jk,jpdet) + ztra
# if defined key_trc_diabio
               trbio(ji,jj,jk,8) = ztra
# endif
# if defined key_trc_diaadd 
               trc2d(ji,jj,8) = trc2d(ji,jj,8) + ztra * fse3t(ji,jj,jk) * 86400.
# endif
            END DO
         END DO
      END DO

!************************** End of imported code *******************************

  END SUBROUTINE trc_sed_medusa

  END MODULE medusa_sed
