!**************************************************************************!
!                                                                          ! 
! SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013                     ! 
!                                                                          !
! Copyright Natural Environment Research Council, UK                       !
!                                                                          !
! Author: John Hemmings, National Oceanography Centre                      !
! e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013)                         !
! Alternative e-mail for author: marmot@wessexenv.mail1.co.uk              !
!                                                                          !
! This software is governed by the CeCILL  license under French law and    !
! abiding by the rules of distribution of free software.  You can  use,    ! 
! modify and/ or redistribute the software under the terms of the CeCILL   !
! license as circulated by CEA, CNRS and INRIA at the following URL        !
! "http://www.cecill.info".                                                ! 
!                                                                          !
! As a counterpart to the access to the source code and  rights to copy,   !
! modify and redistribute granted by the license, users are provided only  !
! with a limited warranty  and the software's author,  the holder of the   !
! economic rights,  and the successive licensors  have only  limited       !
! liability.                                                               ! 
!                                                                          !
! In this respect, the user's attention is drawn to the risks associated   !
! with loading,  using,  modifying and/or developing or reproducing the    !
! software by the user in light of its specific status of free software,   !
! that may mean  that it is complicated to manipulate,  and  that  also    !
! therefore means  that it is reserved for developers  and  experienced    !
! professionals having in-depth computer knowledge. Users are therefore    !
! encouraged to load and test the software's suitability as regards their  !
! requirements in conditions enabling the security of their systems and/or ! 
! data to be ensured and,  more generally, to use and operate it in the    ! 
! same conditions as regards security.                                     ! 
!                                                                          ! 
! The fact that you are presently reading this means that you have had     !
! knowledge of the CeCILL license and that you accept its terms.           !
!                                                                          !
!**************************************************************************!
!=======================================================================
! MODULE og99npzd 
! John Hemmings - 13 November 2013 
!
! Contains a MarMOT-compatible implementation of the Oschlies & Garcon 
! 1999 NPZD model.
! MarMOT interface routines: 
!
!  - codever - provide version notes to MarMOT 
!  - defparm - define parameter set for MarMOT 
!  - getinfo - provide parameter set specific info to MarMOT
!  - setup - set-up fixed model environment 
!  - setparm - set model parameters
!  - getbiotrans - output tracer-specific velocities for MarMOT advection
!  - getphotin - output data to MarMOT for photosynthesis calculation
!  - biotrend - add source-minus-sink to MarMOT transport tendencies
!  - getdiag - output model diagnostics to MarMOT
!
! Reference:
! Oschlies A. & Garcon V., 1999. An eddy-permitting coupled physical-
! biological model of the North Atlantic. 1. Sensitivity to advection 
! numerics and mixed layer physics. Global Biogeochemical Cycles 13, 
! 135-160.
!=======================================================================

  module og99npzd 

  use modconfig, only: kind_r_ext,kind_r,nk,EpdtoW, &
                       i_din,i_phy,i_zoo,i_det, &
                       i_photmax,i_limdin,i_phot

  implicit none

  integer km                    ! Number of levels

! Model parameters 

  integer,parameter :: nparm=15

  real(kind_r) alpha,parfrac,aphotmax,bphotmax,cphotmax,kdin,muphy, &
    gamma1,gmax,epsilon,muzoo,gamma2,mudet,wsink,rphypig

! Diagnostics/ working variables 

  real(kind_r) vp(nk)           ! Phytoplankton max. growth rate (/d)
  real(kind_r) nlim_diag(nk)    ! DIN limitation factor 
  real(kind_r) pgrowth_diag(nk) ! Photosynthesis diagnostic (/d)

  contains

!=======================================================================

  subroutine codever(version,description)

  use modconfig, only: vertextlen 

  implicit none

! Arguments (output)

  character(len=vertextlen+1) version      ! Code version details 
  character(len=vertextlen+1) description  ! Model description document 

!-----------------------------------------------------------------------

  version='JCH_2013NOV13;' 
  description='Oschlies & Garcon 1999, GBC 13, 135-160'

  end subroutine

!=======================================================================

  subroutine defparm(np,pname,pdesc,pmin,pmax)

  use modconfig, only: pnamelen,pdesclen

  implicit none

! Arguments (output)

  integer np                            ! Number of parameters
  character(len=pnamelen+1), dimension(:) :: pname  ! Parameter names
  character(len=pdesclen+1), dimension(:) :: pdesc  ! Descriptions
  real(kind_r_ext) pmin(:)              ! Minimum values 
  real(kind_r_ext) pmax(:)              ! Maximum values

!-----------------------------------------------------------------------

  if (size(pname)<nparm) then
    write(*,*) "Setup error in OG99 NPZD model - defining too many parameters"
    stop
  endif

  np=nparm

  pname(1)='alpha'
  pdesc(1)='initial slope of photosynthesis-PAR curve (/(E/m^2)) [=alpha]'
  pmin(1)=0.

  pname(2)='rparsol'
  pdesc(2)='ratio of PAR to total downwelling solar irradiance' // &
           ' at sea surface [=parfrac]'
  pmin(2)=0.
  pmax(2)=1.

  pname(3)='aphotmax'
  pdesc(3)='max. photosynthetic rate at 0 deg C (/d) [=aphotmax]'
  pmin(3)=1e-10

  pname(4)='bphotmax'
  pdesc(4)='max. photosynthesis - base for temperature variation [=bphotmax]'
  pmin(4)=1e-10

  pname(5)='cphotmax'
  pdesc(5)='max. photosynthesis - temperature sensitivity' // &
           ' of exponent (/degC) [=cphotmax]'

  pname(6)='kdin'
  pdesc(6)='half-saturation conc. for DIN uptake (mmol/m^3) [=kdin]'
  pmin(6)=1e-10

  pname(7)='pmort'
  pdesc(7)='phytoplankton mortality rate (/d) [=muphy]' 
  pmin(7)=0.

  pname(8)='betap'
  pdesc(8)='zooplankton food assimilation efficiency [=gamma1]' 
  pmin(8)=0.
  pmax(8)=1.

  pname(9)='gmax'
  pdesc(9)='maximum grazing rate (/d) [=gmax]' 
  pmin(9)=0.

  pname(10)='epsfood'
  pdesc(10)='prey capture rate (/d/(mmolN/m^3)^2) [=epsilon]' 
  pmin(10)=0.

  pname(11)='zmortdd'
  pdesc(11)='density dependent zooplankton mortality' // &
            ' (/d/(mmolN/m^3)) [=muzoo]'
  pmin(11)=0.

  pname(12)='zexcr'
  pdesc(12)='zooplankton excretion rate (/d) [=gamma2]' 
  pmin(12)=0.

  pname(13)='remin'
  pdesc(13)='detrital remineralization rate (/d) [=mudet]'
  pmin(13)=0.

  pname(14)='dsink'
  pdesc(14)='detrital sinking rate (m/d) [=wsink]'
  pmin(14)=0.

  pname(15)='rphypig'
  pdesc(15)='ratio of phytoplankton nitrogen to total pigment' // &
           ' (molN/g) [=rphypig]'
  pmin(15)=1e-10

  end subroutine

!===============================================================================

  subroutine getinfo(parm,advectflag)

  implicit none

! Input argument

  real(kind_r_ext) parm(:)        ! Model-specific parameters

! Output argument

  integer advectflag              ! Advection scheme flag (1 if required)

!-------------------------------------------------------------------------------

! Advection scheme is needed to handle detrital sinking 
! ('parm' argument not currently used but included to aid possible extension) 

  advectflag=1

  end subroutine

!=======================================================================

  subroutine setup(kmax,ntinit,it,nphyext,reqtemp,dsinkext)

  implicit none

  integer,parameter :: nt=4

! Input argument

  integer kmax                 ! Number of depth levels

! Output arguments

  integer ntinit               ! Number of tracers to be initialized
  integer it(:)                ! Indices of tracers in MarMOT array
  integer nphyext              ! Number of phytoplankton types for 
                               !   external photosynthesis model
  integer reqtemp              ! Temperature request flag
  integer dsinkext             ! Flag to indicate external handling of 
                               !   detrital sinking 
!-----------------------------------------------------------------------

  integer i,k

  km=kmax
  if (km > nk) km=nk

! Set indices for socket layer access to MarMOT tracer arrays 

  if (size(it)<nt) then
    write(*,*) &
       "Setup error in OG99 NPZD model - required tracers not accessible"
    stop
  endif

  ntinit=nt

  it(1)=i_din
  it(2)=i_phy
  it(3)=i_zoo
  it(4)=i_det

! Set number of phytoplankton types for external photosynthesis model

  nphyext=1

! Set flags

  reqtemp=1
  dsinkext=1

  end subroutine

!=======================================================================

  subroutine setparm(parm)

  implicit none

  real(kind_r_ext) parm(:)    ! Input parameters

!-----------------------------------------------------------------------

  if (size(parm)==0) then

    alpha    = 0.063  ! initial slope of P-E curve (/(E/m^2)) 
                      !  = 0.025/d/(W/m^2) using 1E/d=2.52W
    parfrac  = 0.43   ! PAR fraction of surface solar radiation
    aphotmax = 0.6    ! max P growth rate at 0 degC (/d)
    bphotmax = 1.066  ! Eppley factor
    cphotmax = 1.0    ! Eppley factor (/degC)
    kdin     = 0.5    ! half-saturation constant for N uptake (mmol/m^3)
    muphy    = 0.03   ! phytoplankton mortality rate (/d)  
    gamma1   = 0.75   ! assimilation efficiency 
    gmax     = 2.     ! maximum grazing rate (/d)
    epsilon  = 1.     ! prey capture rate (/(mmolN/m^3)^2/d)
    muzoo    = 0.2    ! zooplankton mortality (/(mmolN/m^3)/d)
    gamma2   = 0.03   ! zooplankton excretion (/d) 
    mudet    = 0.05   ! detrital remineralisation rate (/d)
    wsink    = 5.0    ! detrital sinking velocity (m/d)
    rphypig  = 0.5    ! ratio of phytoplankton to pigment (molN/g)

  else if (size(parm)>=nparm) then

    alpha    = parm(1) 
    parfrac  = parm(2)
    aphotmax = parm(3) 
    bphotmax = parm(4) 
    cphotmax = parm(5) 
    kdin     = parm(6) 
    muphy    = parm(7) 
    gamma1   = parm(8)  
    gmax     = parm(9)

    if (gmax <=0.) then
      gmax = 2.
      epsilon = 0.
    else
      epsilon = parm(10) 
    endif

    muzoo     = parm(11) 
    gamma2    = parm(12) 
    mudet     = parm(13) 
    wsink     = parm(14) 
    rphypig   = parm(15) 

  else

    write(*,*) "Setup error in OG99 NPZD model - too few parameters"
    stop 

  endif

  end subroutine

!===============================================================================

  subroutine getbiotrans(wdet)

  implicit none

! Output argument

  real(kind_r) wdet         ! Detritus velocity due to sinking (m/d, up)

  wdet=-wsink

  end subroutine

!=======================================================================

  subroutine getphotin(f_sol,f_temp,tracer,par0,pigment,alpha_phot, &
                       phyflag,vp_phot)

  implicit none

! Input arguments

  real(kind_r_ext) f_sol       ! Surface irradiance (W/m^2)
  real(kind_r_ext) f_temp(:)   ! Temperature (degC)
  real(kind_r_ext) tracer(:,:) ! MarMOT tracer array 

! Output arguments

  real(kind_r) par0            ! Surface PAR (E/m^2/d)
  real(kind_r) pigment(:)      ! Pigment conc. (mg/m^3) 
  real(kind_r) alpha_phot      ! Initial slope of P-I curve (/(E/m^2))
  logical phyflag(:)           ! Switch for presence of phytoplankton
  real(kind_r) vp_phot(:)      ! Phytoplankton max. growth rate (/d)

!-----------------------------------------------------------------------

  integer k

  par0=parfrac*f_sol/EpdtoW
  alpha_phot=alpha

  do k=1,km
    pigment(k)=max(tracer(k,i_phy),0.)/rphypig
    phyflag=(tracer(k,i_phy)/=0.0)
    vp(k)=aphotmax*bphotmax**(cphotmax*f_temp(k))
    vp_phot(k)=vp(k)
  enddo

  end subroutine

!=======================================================================

  subroutine biotrend(f_temp,avphot,tracer,ddt) 

  implicit none

! Arguments (input)

  real(kind_r_ext) f_temp(:)    ! Temperature (degC)
  real(kind_r) avphot(:)        ! Level mean photosynthesis (/d)
  real(kind_r_ext) tracer(:,:)  ! MarMOT tracer array 
  real(kind_r_ext) ddt(:,:)     ! MarMOT transport tendency array 

!-----------------------------------------------------------------------

  real(kind_r) bion,biop,bioz,biod
  real(kind_r) pgrowth,grazep
  integer k

  do k=1,km

    bion = max(0.,tracer(k,i_din)) 
    biop = max(0.,tracer(k,i_phy))
    bioz = max(0.,tracer(k,i_zoo))
    biod = max(0.,tracer(k,i_det))

    vp(k)=aphotmax*bphotmax**(cphotmax*f_temp(k))
    pgrowth = min(avphot(k),vp(k)*bion/(kdin+bion))

    nlim_diag(k) = bion/(kdin+bion)
    pgrowth_diag(k) = pgrowth 

    grazep = gmax*epsilon*biop*biop/(gmax+epsilon*biop*biop)

    ddt(k,i_phy) = ddt(k,i_phy) + pgrowth*biop - grazep*bioz - muphy*biop 

    ddt(k,i_zoo) = ddt(k,i_zoo) + & 
      gamma1*grazep*bioz - gamma2*bioz - muzoo*bioz*bioz 

    ddt(k,i_det) = ddt(k,i_det) + &
      (1.0_kind_r-gamma1)*grazep*bioz + muphy*biop + muzoo*bioz*bioz &
      - mudet*biod

    ddt(k,i_din) = ddt(k,i_din) + mudet*biod + gamma2*bioz - pgrowth*biop

  enddo

  end subroutine 

!=======================================================================

  subroutine getdiag(diagprof)
        
  implicit none
        
  real(kind_r_ext) diagprof(:,:)   ! MarMOT profile diagnostics array

!-----------------------------------------------------------------------

  integer k

  do k=1,km
    diagprof(k,i_photmax)=vp(k)         ! Phytop. max. growth rate (/d)
    diagprof(k,i_limdin)=nlim_diag(k)   ! DIN limitation factor
    diagprof(k,i_phot)=pgrowth_diag(k)  ! Realized photosynthesis (/d)
  enddo

  end subroutine

!=======================================================================

  end module og99npzd
