/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/

/* tabcalc.c - John Hemmings (Oct 2000) */

/* v1.01 - parsing changed to allow blank series, meaning NODATA,
           on RHS of = and != conditions
   v1.02 - added functions 'dtr' and 'rtd' and constant 'pi'
   v1.03 - Nov 2000 - fixed bug in parsing a leading minus sign after certain
           delimiters by proper initialisation of 'dlim' in each function +
           error messages changed to output full expressions
   v1.04 - Jan 2001 - 'isvnamech' function used for identifying variable names
           + new variable names validated
   v1.05 - Jan 2001 - added functions 'ran1' and 'ran1norm' (corresponding to
           Numerical Recipes -2nd ed- functions 'ran1' and 'gasdev')
   v1.06 - Jun 2001 - added functions 'gammaln' and 'betainc' (corresponding to
           Numerical Recipes -2nd ed- functions 'gammln' and 'betai'
           NB betainc(x,a,b) is equivalent to betai(a,b,x) and is
           not defined here for a or b <=0)
   v1.07 - Aug 2001 - fixed system dependent bug where constant was
           overwritten in getfactor 
   v1.08 - Nov 2001 - added output only option + changed parsing to allow 
           constants beginning with '.' 
   v1.09 - Dec 2001 - fixed bug parsing exponents (allowed delimiters now 
           include ')' and ',') 
   v1.10 - Nov 2004 - added 'tdifd' function 
   v1.11 - May 2005 - added 'norm' and 'erf' functions (erf is approximate
           with maximum error of 3E-7, ref: C. Hastings, Jr., Approximations 
           for digital computers. Princeton Univ. Press, Princeton, N.J., 
           1955)
   v1.12 - Aug 2005 - fixed to avoid passing NULL to fclose 
   v1.13 - May 2006 - added '.ttb' output file name recognition and option to 
           suppress header for text output to stdout ('outfile' = '.' or can be 
           omitted; 'infile' can also be omitted if no input data)
   v1.14 - Nov 2006 - added 'tyears', 'tofyd', 'mofy', 'dofm', 'dofy' and 
           'isleap' functions + changed 'tdifd' to use astronomical year and 
           double f.p. arithmetic + consolidated pre-screening of function 
           arguments for nodata value
   v1.15 - Feb 2010 - removed Numerical Recipes routines (these or 
           equivalent functions are now made available via tabcalc.h)
           + retrieved PI from atan instead of defining directly
*/

#include "toolbox.h"
#include "tabcalc.h"
#include <limits.h>

#define INFILE 1
#define OUTFILE 2
#define PI (4*atan(1.0)) 

enum {BADEXPR_AND=-99,BADEXPR_COND1,BADEXPR_COND2,BADEXPR_SER,BADEXPR_TERM,
      BADEXPR_FACT1,BADEXPR_FACT2,BADEXPR_FACT3,BADEXPR_FACT4,BADEXPR_FACT5,
      BADEXPR_FACT6,BADEXPR_FACT7,BADEXPR_FACT8,BADEXPR_FACT9,BADEXPR_FACT10,
      BADEXPR_FACT11,BADEXPR_FACT12,BADEXPR_FACT13};

enum {CONST,ID,MINUS,DIV,SUM,PROD,AND,OR,LN,LOG10,EXP,POW,SQR,SQRT,SIN,COS,TAN,
      ASIN,ACOS,ATAN,SINH,COSH,TANH,CEIL,FLOOR,INT,FRAC,ABS,MIN,MAX,
      IF,EQ,NE,LT,GT,LE,GE,NOT,DTR,RTD,NORM,RAN1,RAN1NORM,ERF,GAMMALN,BETAINC,
      TDIFD,TYEARS,TOFYD,MOFY,DOFM,DOFY,ISLEAP};

struct node {
  int op;
  int index;
  double value;
#if defined(EXTFUNC_RANDOM)
  struct ranwkrec ranwk;
#endif
  struct node *arg;
  struct node *arg2;
  struct node *arg3;
  struct node *arg4;
  struct node *next;
};

int parse (char **, char *, int, char * [], struct node **);
int getandexpr (char **, char *, int, char * [], struct node **);
int getcond (char **, char *, int, char * [], struct node **);
int getseries (char **, char *, int, char * [], struct node **);
int getterm (char **, char *, int, char * [], struct node **);
int getfactor (char **, char *, int, char * [], struct node **);
double evalnode (struct node *, double [], double);
double evalsum (struct node *, double [], double);
double evalprod (struct node *, double [], double);
double evaland (struct node *, double [], double);
double evalor (struct node *, double [], double);

/* Time processing functions */
/* N.B. Fractional part of input year, month and day are ignored 
   but time in days (tdays) is treated as a fraction */

double tdifd (double, double, double, double); /* {y2,tdays2} - {y1,tdays1} */
double tyears (double, double, double, double *); /* {y,tdays} - {y,0} in yrs and days */
double mofy (double, double, double); /* month of year from {y,tdays} */
double dofm (double, double, double); /* day of month from {y,tdays} */
double dofy (double, int, int, double); /* day of year from {year,month,day} */
int isleap (double);


int main (int argc, char *argv[])

{ FILE *infile=NULL, *outfile=NULL;
  char outfbuff[FILENAME_MAX+1];
  char *outfnamereq;
  char *outfname;
  char *fname;
  char varlist[LINE_MAX+1];
  char *invarname[NFIELD_MAX],*newvarname[NFIELD_MAX],*outvarname[NFIELD_MAX];
  static double x[NFIELD_MAX];
  static char fieldlist[LINE_MAX+1];
  static char *field[NFIELD_MAX];
  struct datarec d={x,fieldlist,field,NULL};
  struct node *expr[NFIELD_MAX];
  char currexpr[LINE_MAX+1];
  char *s,dlim;
  static double value;
  struct datarec dnew={&value,NULL,NULL,NULL};
  int ninvar,nnewvar,noutvar,status,i,ttb,ttbout,lineno,dummyflag,
      voffset=3,noinfile=0,nooutfilehead=0;

#if defined(EXTFUNC)
  sprintf(currexpr,"1.15 (external functions %s)",EXTFUNC);
  s=currexpr;
#else
  s="1.15"; 
#endif

  version(s,VHDR,argc,argv);

  if (argc<2)
    { fprintf (stderr,"Usage: %s infile outfile [var1=expr1 ...]\n",argv[0]);
      fprintf (stderr,"   or: %s [infile] =expr1 [...]\n",argv[0]);
      exit(EXIT_FAILURE);
    }

  for (i=1;i<argc;i++)
    if (strpbrk(argv[i],WHITESPACE)!=NULL) exiterror(BADARG,argv[0],argv[i],0);

  if (*argv[INFILE]=='=') 
    { noinfile=1;
      nooutfilehead=1;
      voffset=1;
    }
  else if (argc<3 || *argv[OUTFILE]=='=')
    { nooutfilehead=1;
      voffset=2;
    }

  if (!noinfile && strcmp(argv[INFILE],".")==0) noinfile=1;
  if (!nooutfilehead && strcmp(argv[OUTFILE],".")==0) nooutfilehead=1;

  if (noinfile)

    { ninvar=0;
      ttb=0;
    }

  else

    { if (strcmp(argv[INFILE],"-")==0) infile=stdin;
      else if ((infile=fopen(argv[INFILE],"r"))==NULL)
        exiterror(OPENINERR,argv[0],argv[INFILE],0);

      status=readvar(infile,varlist,invarname,&ninvar,&ttb,&lineno);
      if (status!=OK) exiterror(status,argv[0],argv[INFILE],lineno);
    }
 
  nnewvar=argc-voffset;
  noutvar=ninvar+nnewvar;
  if (noutvar>NFIELD_MAX) exiterror(OUTVARERR,argv[0],NULL,0);
 
  currexpr[LINE_MAX]='\0';
  for (i=0;i<nnewvar;i++)
    { strncpy(currexpr,argv[i+voffset],LINE_MAX);
      newvarname[i]=argv[i+voffset];
      s=strchr(newvarname[i],'=');
      if (s==NULL) s="";
      else *(s++)='\0';
      if (!nooutfilehead && !validvname(newvarname[i]))
        exiterror(BADEXPR,argv[0],newvarname[i],0);
      status=parse(&s,&dlim,ninvar,invarname,&expr[i]);
      if (status==VARERR) exiterror(VARERR,argv[0],s,0);
      if (status==BADFUNC) exiterror(BADFUNC,argv[0],s,0);
      if (status<0) exiterror(BADEXPR,argv[0],currexpr,0);
      if (status!=OK) exiterror(status,argv[0],currexpr,0);
      if (dlim!='\0') exiterror(BADEXPR,argv[0],currexpr,0);
    }

  if (nooutfilehead) outfnamereq=".";
  else outfnamereq=argv[OUTFILE];
  ttbout=ttb;
  if ((outfile=outfopen(outfnamereq,outfbuff,&outfname,&ttbout))==NULL)
    exiterror(OPENOUTERR,argv[0],outfname,0);

  if (!nooutfilehead)
    { for (i=0;i<ninvar;i++) outvarname[i]=invarname[i];
      for (i=0;i<nnewvar;i++) outvarname[ninvar+i]=newvarname[i];
      status=writevar(outfile,ttbout,noutvar,outvarname,NULL,DEFAULT);
      if (status!=OK) exiterror(status,argv[0],outfname,0);
    }

  while ((!noinfile && 
          (status=readdata(infile,ttb,ninvar,0,NULL,0,NULL,&d,&lineno))==OK) ||
         (noinfile && status==OK))

    { if (!nooutfilehead) 
        { status=writedata(outfile,ttbout,&d,ninvar,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,nnewvar==0?1:0);
          if (status!=OK) exiterror(status,argv[0],outfname,0);
        }

      if (!ttb) for (i=0;i<ninvar;i++) d.x[i]=sgetnum(d.field[i],&dummyflag);
      for (i=0;i<nnewvar;i++)
        { *dnew.x=evalnode(expr[i],x,NODATA);
          status=writedata(outfile,ttbout,&dnew,1,NULL,NULL,NULL,DEFAULT,
                           DEFAULT,i==nnewvar-1?1:0);
          if (status!=OK) exiterror(status,argv[0],outfname,0);
        }

      if (noinfile) status=EOF;
    }

  if (status!=EOF) exiterror(status,argv[0],argv[INFILE],lineno);

  if (infile) fclose(infile);
  fclose(outfile);
  return EXIT_SUCCESS;
}


int parse (char **s, char *dlim, int nvar, char *varname[], struct node
**head)

{ struct node *newnode,*curr;
  int status;

  *dlim='\0';

  if (**s=='\0')
    { *head=NULL;
      return OK;
    }

  status=getandexpr(s,dlim,nvar,varname,&newnode);
  if (status!=OK) return status;
  if (*dlim!='|')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  (*head)->op=OR;
  (*head)->arg=newnode;
  curr=newnode;
  do
    { status=getandexpr(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      curr->next=newnode;
      curr=newnode;
    }
  while (*dlim=='|');

  return OK;
}


int getandexpr (char **s, char *dlim, int nvar, char *varname[],
                struct node **head)

{ struct node *curr,*newnode;
  int status;

  if (!(isvnamech(**s) || isdigit(**s) || **s=='.' || **s=='(' || **s=='+' ||
        **s=='-'))
    return BADEXPR_AND;

  *dlim='\0';

  status=getcond(s,dlim,nvar,varname,&newnode);
  if (status!=OK) return status;
  if (*dlim!='&')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  (*head)->op=AND;
  (*head)->arg=newnode;
  curr=newnode;
  do
    { status=getcond(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      curr->next=newnode;
      curr=newnode;
    }
  while (*dlim=='&');

  return OK;
}


int getcond (char **s, char *dlim, int nvar, char *varname[],
             struct node **head)

{ struct node *newnode;
  int status;

  if (!(isvnamech(**s) || isdigit(**s) || **s=='.' || **s=='(' || **s=='+' ||
        **s=='-'))
    return BADEXPR_COND1;

  *dlim='\0';

  status=getseries(s,dlim,nvar,varname,&newnode);
  if (status!=OK) return status;
  if (*dlim!='=' && *dlim!='!' && *dlim!='<' && *dlim!='>')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  if (*dlim=='=') (*head)->op=EQ;
  else if (*dlim=='!')
    { if (**s=='=')
        { (*head)->op=NE;
          (*s)++;
        }
      else return BADEXPR_COND2;
    }
  else if (*dlim=='<')
    { if (**s=='=')
        { (*head)->op=LE;
          (*s)++;
        }
      else (*head)->op=LT;
    }
  else if (*dlim=='>')
    { if (**s=='=')
        { (*head)->op=GE;
          (*s)++;
        }
      else (*head)->op=GT;
    }
  (*head)->arg=newnode;

  if (((*head)->op==EQ || (*head)->op==NE) &&
      (**s==')' || **s==',' || **s=='\0'))
    { newnode=malloc(sizeof(struct node));
      if (newnode==NULL) return MEMERR;
      newnode->op=CONST;
      newnode->value=NODATA;
      *dlim=**s;
      (*s)++;
    }
  else
    { status=getseries(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
    }
  (*head)->arg2=newnode;

  return OK;
}


int getseries (char **s, char *dlim, int nvar, char *varname[],
               struct node **head)

{ struct node *curr,*newnode;
  int status;

  if (!(isvnamech(**s) || isdigit(**s) || **s=='.' || **s=='(' || **s=='+' ||
        **s=='-'))
    return BADEXPR_SER;

  *dlim='\0';

  status=getterm(s,dlim,nvar,varname,&newnode);
  if (status!=OK) return status;
  if (*dlim!='+' && *dlim!='-')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  (*head)->op=SUM;
  (*head)->arg=newnode;
  curr=newnode;
  do
    { status=getterm(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      curr->next=newnode;
      curr=newnode;
    }
  while (*dlim=='+' || *dlim=='-');

  return OK;
}


int getterm (char **s, char *dlim, int nvar, char *varname[],
             struct node **head)

{ struct node *curr,*newnode;
  int negate=0,status;

  if (*dlim=='\0')
    { if (**s=='-')
        { negate=1;
          (*s)++;
        }
      else if (**s=='+') (*s)++;
    }
  else if (*dlim=='-')
    { negate=1;
      *dlim='\0';
    }

  if (!(isvnamech(**s) || isdigit(**s) || **s=='.' || **s=='('))
    return BADEXPR_TERM;

  if (negate)
    { *head=malloc(sizeof(struct node));
      if (*head==NULL) return MEMERR;
      (*head)->op=MINUS;
      status=getterm(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      (*head)->arg=newnode;
      return OK;
    }

  status=getfactor(s,dlim,nvar,varname,&newnode);
  if (status!=OK) return status;
  if (*dlim!='*' && *dlim!='/')
    { *head=newnode;
      return OK;
    }

  *head=malloc(sizeof(struct node));
  if (*head==NULL) return MEMERR;
  (*head)->op=PROD;
  (*head)->arg=newnode;
  curr=newnode;
  do
    { status=getfactor(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      curr->next=newnode;
      curr=newnode;
    }
  while (*dlim=='*' || *dlim=='/');

  return OK;
}


int getfactor (char **s, char *dlim, int nvar, char *varname[],
                struct node **f)

{ struct node *newnode;
  char *p;
  int status,i,found,numerr;

  if (!(isvnamech(**s) || isdigit(**s) || **s=='.' || **s=='(')) 
    return BADEXPR_FACT1;

  if (*dlim=='/')
    { *dlim='\0';
      *f=malloc(sizeof(struct node));
      if (*f==NULL) return MEMERR;
      (*f)->op=DIV;
      status=getfactor(s,dlim,nvar,varname,&newnode);
      if (status!=OK) return status;
      (*f)->arg=newnode;
      return OK;
    }

  p=strpbrk(*s,"+-*/()|&=!<>,");

  if (p && *p=='(')

    { if (p==*s)
        { *s=p+1;
          status=parse(s,dlim,nvar,varname,&newnode);
          if (status!=OK) return status;
          *f=newnode;
        }

      else if (isalpha(**s))

        { *p='\0';
          *f=malloc(sizeof(struct node));
          if (*f==NULL) return MEMERR;
          if (strcmp(*s,"ln")==0) (*f)->op=LN;
          else if (strcmp(*s,"log10")==0) (*f)->op=LOG10;
          else if (strcmp(*s,"exp")==0) (*f)->op=EXP;
          else if (strcmp(*s,"pow")==0) (*f)->op=POW;
          else if (strcmp(*s,"sqr")==0) (*f)->op=SQR;
          else if (strcmp(*s,"sqrt")==0) (*f)->op=SQRT;
          else if (strcmp(*s,"sin")==0) (*f)->op=SIN;
          else if (strcmp(*s,"cos")==0) (*f)->op=COS;
          else if (strcmp(*s,"tan")==0) (*f)->op=TAN;
          else if (strcmp(*s,"asin")==0) (*f)->op=ASIN;
          else if (strcmp(*s,"acos")==0) (*f)->op=ACOS;
          else if (strcmp(*s,"atan")==0) (*f)->op=ATAN;
          else if (strcmp(*s,"sinh")==0) (*f)->op=SINH;
          else if (strcmp(*s,"cosh")==0) (*f)->op=COSH;
          else if (strcmp(*s,"tanh")==0) (*f)->op=TANH;
          else if (strcmp(*s,"ceil")==0) (*f)->op=CEIL;
          else if (strcmp(*s,"floor")==0) (*f)->op=FLOOR;
          else if (strcmp(*s,"int")==0) (*f)->op=INT;
          else if (strcmp(*s,"frac")==0) (*f)->op=FRAC;
          else if (strcmp(*s,"abs")==0) (*f)->op=ABS;
          else if (strcmp(*s,"min")==0) (*f)->op=MIN;
          else if (strcmp(*s,"max")==0) (*f)->op=MAX;
          else if (strcmp(*s,"if")==0) (*f)->op=IF;
          else if (strcmp(*s,"not")==0) (*f)->op=NOT;
          else if (strcmp(*s,"dtr")==0) (*f)->op=DTR;
          else if (strcmp(*s,"rtd")==0) (*f)->op=RTD;
          else if (strcmp(*s,"norm")==0) (*f)->op=NORM;
          else if (strcmp(*s,"erf")==0) (*f)->op=ERF;
          else if (strcmp(*s,"tdifd")==0) (*f)->op=TDIFD;
          else if (strcmp(*s,"tyears")==0) (*f)->op=TYEARS;
          else if (strcmp(*s,"tofyd")==0) (*f)->op=TOFYD;
          else if (strcmp(*s,"mofy")==0) (*f)->op=MOFY;
          else if (strcmp(*s,"dofm")==0) (*f)->op=DOFM;
          else if (strcmp(*s,"dofy")==0) (*f)->op=DOFY;
          else if (strcmp(*s,"isleap")==0) (*f)->op=ISLEAP;
#if defined(EXTFUNC_RANDOM)
          else if (strcmp(*s,"ran1")==0)
            { (*f)->op=RAN1;
              raninit(&(*f)->ranwk);
            }
          else if (strcmp(*s,"ran1norm")==0)
            { (*f)->op=RAN1NORM;
              raninit(&(*f)->ranwk);
            }
#endif
#if defined(EXTFUNC_GAMMA)
          else if (strcmp(*s,"gammaln")==0) (*f)->op=GAMMALN;
#endif
#if defined(EXTFUNC_BETA)
          else if (strcmp(*s,"betainc")==0) (*f)->op=BETAINC;
#endif
          else return BADFUNC;

          *s=p+1;
          if ((*f)->op==POW || (*f)->op==MIN || (*f)->op==MAX ||
              (*f)->op==TYEARS || (*f)->op==TOFYD ||
              (*f)->op==MOFY || (*f)->op==DOFM)
            { status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              if (*dlim!=',') return BADEXPR_FACT2;
              (*f)->arg=newnode;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg2=newnode;
            }
          else if ((*f)->op==IF)
            { status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              if (*dlim!=',') return BADEXPR_FACT3;
              (*f)->arg=newnode;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg2=newnode;
              if (*dlim!=',') (*f)->arg3=NULL;
              else
                { status=parse(s,dlim,nvar,varname,&newnode);
                  if (status!=OK) return status;
                  (*f)->arg3=newnode;
                }
            }
          else if ((*f)->op==NORM || (*f)->op==BETAINC || (*f)->op==DOFY)
            { status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              if (*dlim!=',') return BADEXPR_FACT9;
              (*f)->arg=newnode;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg2=newnode;
              if (*dlim!=',') return BADEXPR_FACT10;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg3=newnode;
            }
          else if ((*f)->op==TDIFD)
            { status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              if (*dlim!=',') return BADEXPR_FACT11;
              (*f)->arg=newnode;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg2=newnode;
              if (*dlim!=',') return BADEXPR_FACT12;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg3=newnode;
              if (*dlim!=',') return BADEXPR_FACT13;
              status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg4=newnode;
            }
          else
            { status=parse(s,dlim,nvar,varname,&newnode);
              if (status!=OK) return status;
              (*f)->arg=newnode;
            }
        }

      else return BADEXPR_FACT4;

      if (*dlim!=')') return BADEXPR_FACT5;
      *dlim=*((*s)++);
      if (strchr("+-*/|&=!<>),",*dlim)==NULL) return BADEXPR_FACT6;
    }

  else

    { *f=malloc(sizeof(struct node));
      if (*f==NULL) return MEMERR;
      if (isvnamech(**s))
        { if (p)
            { *dlim=*p;
              *p='\0';
            }
          else *dlim='\0';
          (*f)->op=ID;
          i=0;
          found=0;
          while (i<nvar && !(found=(strcmp(varname[i],*s)==0))) i++;
          if (!found)
            { if (strcmp("pi",*s)==0)
                { (*f)->op=CONST;
                  (*f)->value=PI;
                }
              else return VARERR;
            }
          else (*f)->index=i;
        }
      else if (isdigit(**s) || **s=='.')
        { if (p && *p!='\0' && (*(p-1)=='e' || *(p-1)=='E'))
            p=strpbrk(p+1,"+-*/)|&=!<>,");
          if (p)
            { *dlim=*p;
              *p='\0';
            }
          else *dlim='\0';
          (*f)->op=CONST;
          (*f)->value=sgetnum(*s,&numerr);
          if (numerr) return BADEXPR_FACT7;
        }
      else return BADEXPR_FACT8;
      if (p) *s=p+1;
    }

  return OK;
}


double evalnode (struct node *p, double x[], double nodata)

{ double value,value2,value3,value4,v,v2;
  int error;

  if (p==NULL) return nodata;

  switch (p->op)
    { case CONST: return p->value;
      case ID: return x[p->index];
      case SUM: return evalsum(p->arg,x,nodata);
      case PROD: return evalprod(p->arg,x,nodata);
      case OR: return evalor(p->arg,x,nodata);
      case AND: return evaland(p->arg,x,nodata);
      default:
        { value=evalnode(p->arg,x,nodata);
          if (value==nodata &&
              p->op!=EQ && p->op!=NE &&
              p->op!=TYEARS && p->op!=TOFYD &&
              p->op!=MOFY && p->op!=DOFM && p->op!=DOFY) return nodata;
          if (p->op==POW || p->op==MIN || p->op==MAX ||
              p->op==TYEARS || p->op==TOFYD ||
              p->op==MOFY || p->op==DOFM ||
              p->op==EQ || p->op==NE ||
              p->op==LT || p->op==GT ||
              p->op==LE || p->op==GE)
            { value2=evalnode(p->arg2,x,nodata);
              if (value2==nodata && p->op!=EQ && p->op!=NE) return nodata;
            }
          else if (p->op==IF || p->op==NORM || p->op==BETAINC || p->op==DOFY)
            { value2=evalnode(p->arg2,x,nodata);
              value3=evalnode(p->arg3,x,nodata);
              if ((value2==nodata || value3==nodata) && p->op!=IF) 
                return nodata;
            }
          else if (p->op==TDIFD)
            { value2=evalnode(p->arg2,x,nodata);
              value3=evalnode(p->arg3,x,nodata);
              value4=evalnode(p->arg4,x,nodata);
              if (value2==nodata || value3==nodata || value4==nodata)
                return nodata; 
            }
          switch (p->op)
            { case MINUS: return -value;
              case DIV:
                if (value==0) return nodata;
                else return 1/value;
              case LN:
                if (value>0) return log(value);
                else return nodata;
              case LOG10:
                if (value>0) return log10(value);
                else return nodata;
              case EXP: return exp(value);
              case POW:
                if (value==0 && value2<=0 || value<0 && value2!=floor(value2))
                  return nodata;
                return pow(value,value2);
              case SQR: return pow(value,2);
              case SQRT:
                if (value>=0) return sqrt(value);
                else return nodata;
              case SIN: return sin(value);
              case COS: return cos(value);
              case TAN: return tan(value);
              case ASIN:
                if (fabs(value)>1) return nodata;
                return asin(value);
              case ACOS:
                if (fabs(value)>1) return nodata;
                return acos(value);
              case ATAN: return atan(value);
              case SINH: return sinh(value);
              case TANH: return tanh(value);
              case CEIL: return ceil(value);
              case FLOOR: return floor(value);
              case INT:
                modf(value,&value2);
                return value2;
              case FRAC:
                return modf(value,&value2);
              case ABS: return fabs(value);
              case MIN:
                if (value<=value2) return value;
                else return value2;
              case MAX:
                if (value>=value2) return value;
                else return value2;
              case IF:
                if (value) return value2;
                else return value3;
              case NOT:
                if (value) return 0;
                else return 1;
              case RTD:
                return value*(180/PI);
              case DTR:
                return value*(PI/180);
              case NORM:
                v=(value-value2)/value3;
                return (1/(value3*sqrt(2*PI)))*exp(-0.5*v*v); 
              case ERF:
                v=fabs(value);
                v2=1-(1/pow(
                  1+
                  0.0705230784*v+
                  0.0422820123*v*v+
                  0.0092705272*v*v*v+ 
                  0.0001520143*v*v*v*v+ 
                  0.0002765672*v*v*v*v*v+ 
                  0.0000430638*v*v*v*v*v*v,16));
                if (value<0) return -v2;
                else return v2;
              case TDIFD:
                v=tdifd(value,value2,value3,value4);
                return v;
              case TYEARS:
                return tyears(value,value2,nodata,&value3);
              case TOFYD:
                tyears(value,value2,nodata,&value3);
                return value3;
              case MOFY:
                return mofy(value,value2,nodata);
              case DOFM:
                return dofm(value,value2,nodata);
              case DOFY:
                if (value2<1 || value3<1 || value2>=13 || value3>=32) 
                  return nodata;
                return dofy(value,(int)value2,(int)value3,nodata);
              case ISLEAP:
                v=(double)isleap(value);
                if (v<0) return nodata;
                else return v;
              case EQ:
                return value==value2;
              case NE:
                return value!=value2;
              case LT:
                return value<value2;
              case GT:
                return value>value2;
              case LE:
                return value<=value2;
              case GE:
                return value>=value2;
#if defined(EXTFUNC_RANDOM)
              case RAN1:
                v=ran1(value,&p->ranwk,&error);
                if (error) return nodata;
                return v;
              case RAN1NORM:
                v=ran1norm(value,&p->ranwk,&error);
                if (error) return nodata;
                return v;
#endif
#if defined(EXTFUNC_GAMMA)
              case GAMMALN:
                return gammln(value,nodata);
#endif
#if defined(EXTFUNC_BETA)
              case BETAINC:
                v=betai(value2,value3,value,nodata);
                return v;
#endif
              default: return nodata;
            }
        }
    }
}


double evalsum (struct node *curr, double x[], double nodata)

{ double sum=0,value;

  while (curr!=NULL)
    { value=evalnode(curr,x,nodata);
      if (value==nodata) return nodata;
      sum+=value;
      curr=curr->next;
    }
  return sum;
}


double evalprod (struct node *curr, double x[], double nodata)

{ double prod=1,value;

  while (curr!=NULL)
    { value=evalnode(curr,x,nodata);
      if (value==nodata) return nodata;
      prod*=value;
      curr=curr->next;
    }
  return prod;
}


double evalor (struct node *curr, double x[], double nodata)

{ double value;
  int missing=0;

  while (curr!=NULL)
    { value=evalnode(curr,x,nodata);
      if (value==nodata) missing=1;
      else if (value) return 1;
      curr=curr->next;
    }
  if (missing) return nodata;
  return 0;
}


double evaland (struct node *curr, double x[], double nodata)

{ double value;

  while (curr!=NULL)
    { value=evalnode(curr,x,nodata);
      if (value==nodata) return nodata;
      if (!value) return 0;
      curr=curr->next;
    }
  return 1;
}


double tdifd (double dyear1, double tdays1, double dyear2, double tdays2) 

/* Year is astronomical year so year sequence is continuous at origin */ 

{ static double year1,year2,y,n4,n100,n400,t1,t2;

  modf(dyear1,&year1); 
  modf(dyear2,&year2); 

  if (year1==year2) return tdays2-tdays1;

  if (year1>0) 
    { y=year1-1; 
      modf(y/4,&n4);
      modf(y/100,&n100);
      modf(y/400,&n400);
      t1=year1*365+1+n4-n100+n400+tdays1;
    }
  else
    { y=-year1;
      modf(y/4,&n4);
      modf(y/100,&n100);
      modf(y/400,&n400);
      t1=year1*365-(n4-n100+n400)+tdays1;
    }

  if (year2>0)
    { y=year2-1;
      modf(y/4,&n4);
      modf(y/100,&n100);
      modf(y/400,&n400);
      t2=year2*365+1+n4-n100+n400+tdays2;
    }
  else
    { y=-year2;
      modf(y/4,&n4);
      modf(y/100,&n100);
      modf(y/400,&n400);
      t2=year2*365-(n4-n100+n400)+tdays2;
    }

  return t2-t1;
}


double dofy (double dyear, int month, int day, double nodata)

{ static int leap;

  if (dyear==nodata) leap=-1;
  else leap=isleap(dyear);
 
  switch (month) 
    { case 1:
        if (day>31) return nodata; 
        return day; 
      case 2: 
        if (day>29 || (day==29 && leap==0))
          return nodata;
        return 31+day;
      case 3:
        if (day>31 || leap<0) return nodata;
        return 59+leap+day;
      case 4:
        if (day>30 || leap<0) return nodata;
        return 90+leap+day;
      case 5:
        if (day>31 || leap<0) return nodata;
        return 120+leap+day;
      case 6:
        if (day>30 || leap<0) return nodata;
        return 151+leap+day;
      case 7:
        if (day>31 || leap<0) return nodata;
        return 181+leap+day;
      case 8:
        if (day>31 || leap<0) return nodata;
        return 212+leap+day;
      case 9:
        if (day>30 || leap<0)  return nodata;
        return 243+leap+day;
      case 10:
        if (day>31 || leap<0) return nodata;
        return 273+leap+day;
      case 11:
        if (day>30 || leap<0) return nodata;
        return 304+leap+day;
      case 12:
        if (day>31 || leap<0) return nodata;
        return 334+leap+day;
      default: 
        return nodata;
    }
}


double mofy (double dyear, double tdays, double nodata)

{ static int leap;
  static double t,ty,year;

  ty=tyears(dyear,tdays,nodata,&t);
  if (t==nodata) return nodata;

  if (t<31) return 1;
  if (t<59) return 2;

  if (dyear==nodata || ty==nodata) return nodata;

  modf(dyear,&year);
  leap=isleap(year+floor(ty));
  if (leap<0) return nodata;

  if (t<59+leap) return 2;
  if (t<90+leap) return 3;
  if (t<120+leap) return 4;
  if (t<151+leap) return 5;
  if (t<181+leap) return 6;
  if (t<212+leap) return 7;
  if (t<243+leap) return 8;
  if (t<273+leap) return 9;
  if (t<304+leap) return 10;
  if (t<334+leap) return 11;
  return 12 ;
}


double dofm (double dyear, double tdays, double nodata)

{ static int leap,m;
  static double t,ty,year;

  ty=tyears(dyear,tdays,nodata,&t);
  if (t==nodata) return nodata;

  if (t<31) return (int)(t+1);
  if (t<59) return (int)(t-30);

  if (dyear==nodata || ty==nodata) return nodata;

  modf(dyear,&year);
  leap=isleap(year+floor(ty));
  if (leap<0) return nodata;

  if (t<59+leap) return (int)(t-30);
  if (t<90+leap) return (int)(t-58-leap);
  if (t<120+leap) return (int)(t-89-leap);
  if (t<151+leap) return (int)(t-119-leap);
  if (t<181+leap) return (int)(t-150-leap);
  if (t<212+leap) return (int)(t-180-leap);
  if (t<243+leap) return (int)(t-211-leap);
  if (t<273+leap) return (int)(t-242-leap);
  if (t<304+leap) return (int)(t-272-leap);
  if (t<334+leap) return (int)(t-303-leap);
  return (int)(t-333-leap);
}


double tyears (double dyear, double tdays, double nodata, double *trem)

/* Convert tdays to time in years (including fractional part)
   tdays is taken to be time relative to start of year 
   Time of year in days passed back in trem  */

{ static double year,refyear,n400,n100,n4,n,year2,y,t;
  static int leap;

  if (tdays==0) 
    { *trem=0;
      return 0;
    }

  if (dyear==nodata) leap=-1;
  else leap=isleap(dyear);

  /* year out of integer range - don't know if it's a leap year */

  if (leap<0) 
    { if (tdays>=0 && tdays<365) *trem=tdays;
      else *trem=nodata;
      return nodata; 
    }

  /* Check for simple case: +ve tdays less than 1 year */ 

  if (tdays>=0 && (tdays<365 || (leap==1 && tdays<366)))
    { *trem=tdays;
      return tdays/(365+leap);
    }

  /* Find time {year2,t} relative to a reference year at the start of a 400 
     year leap cycle (t is time of year in days)
     For refyear use year 0001 +/- multiple of 400 as origin so leap year is 
     always the last in 4 year period
     Choose year before the period between {year,0} and {year,tdays} to 
     ensure time is positive (y is set to year less than or equal to minimum 
     year in period and can be +ve or -ve) */

  modf(dyear,&year);

  if (tdays>0) y=year;
  else y=year+floor(tdays/365);
  modf(y/400,&n);
  refyear=400*n+1;
  while (refyear>y) refyear-=400;

  leap=isleap(refyear);
  if (leap<0) 
    { *trem=nodata;
      return nodata;
    }

  t=(tdifd(refyear,0,year,tdays));
  if (t<LONG_MIN || t>LONG_MAX) 
    { *trem=nodata;
      return nodata;
    }

  /* Find number of 400, 100, 4 and 1 year periods (+ve or -ve) 
     and add them up*/

  modf(t/146097,&n400);
  t-=n400*146097;

  modf(t/36524,&n100);
  if (n100>3) n100--;
  t-=n100*36524;

  modf(t/1461,&n4);
  t-=n4*1461;

  modf(t/365,&n);
  if (n>3) n--;
  t-=n*365;

  *trem=t;

  y=n400*400+n100*100+n4*4+n;
  year2=refyear+y;

  /* Add fractional part and subtract input year to return time in years */

  leap=isleap(1+y);
  if (leap<0) return nodata; 
  return year2+t/(365+leap)-year;
}


int isleap (double dyear)

{ static long year;

  if (dyear<LONG_MIN || dyear>LONG_MAX) return -1;

  year=(long)dyear;
  return (4*(long)(year/4)==year && 
          (100*(long)(year/100)!=year || 400*(long)(year/400)==year));
}
