/***********************************************************************

SOFTWARE RELEASE: MarMOT 1.1  DATE: 21 November 2013

Copyright Natural Environment Research Council, UK

Author: John Hemmings, National Oceanography Centre 
e-mail: j.hemmings@noc.ac.uk (until 30 Nov 2013) 
Alternative e-mail for author: marmot@wessexenv.mail1.co.uk 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms.

***********************************************************************/
/* marmot.c - John Hemmings */

#include "gfan.h"          /* GFAn header includes marmot.h (alias objfun.h) */
#include "fortname.h"      /* Compiler-dependent Fortran names */ 

#define MARMOT_VERSION "1.1 (14 Nov 2013) - John Hemmings, National Oceanography Centre (NERC)"
#define ATOM_C 12.011      /* Atomic mass of carbon */
#define VERTEXT_MAX 70     /* Max length for code version notes 
                              (Also used for model names) */
#define MODTEXT_MAX 160    /* Max length for model messages */ 
#define NP_MAX 500         /* Max number of model parameters */
#define PNAME_MAX 20       /* Max length for parameter names */ 
#define PDESC_MAX 100      /* Max length for parameter descriptions */
#define NK_MAX 500         /* Max number of model levels */
#define NAGG_MAX 6         /* Max number of primary tracers contributing to 
                              an aggregate tracer or diagnostic */
#define MAXDEPTH_DEF 50    /* Default maximum depth */
#define RELAXULIM 1000.0   /* Relaxation rate upper limit */
#define RELAXOVERRIDE 9999.0
                           /* Relaxation rate indcating override request */
#define N_TRACER_PRI 13    /* Number of primary tracers (initializable) */
#define N_RATIO 12         /* Number of composition ratios */
#define N_TRACER_DER 17    /* Number of derived tracers
                              (diagnosed from primary tracers + ratios) */
#define N_TRACER (N_TRACER_PRI+N_TRACER_DER)  
                           /* Number of tracers */
#define N_DIAG_PROF_BASIC 48 
#define N_DIAG_PROF (N_DIAG_PROF_BASIC+3*N_TRACER) 
                           /* Number of depth profile diagnostics */
#define N_DIAG_AGG 1       /* Number of aggregate diagnostics */
#define N_DIAG_SCAL 9      /* Number of scalar diagnostics */
#define N_IZ_DIAG 14       /* Number of depth integrals of prof. diagnostics */ 
#define N_ENVIRON 2        /* Number of environmental constants */
#define N_FSCALAR 11       /* Number of scalar forcing variables */  
#define N_FSCALAR_DA 4     /* Number of scalar forcing variables specific to  
                              data assimilation analysis */  
#define N_FSCALAR_NONDA (N_FSCALAR-N_FSCALAR_DA) 
                           /* Number of scalar forcing variables excuding DA
                              variables */
#define N_FPROF (3+4*(N_TRACER_PRI)) 
                           /* Number of depth profile forcing vars */
#define N_AUXVAR 5         /* Number of unprocessed auxiliary variables 
                              associated with depth prof. forcing variables */ 
#define N_OPTION 24+N_FSCALAR+N_FPROF+3*(N_TRACER_PRI)+N_RATIO
                           /* Number of option parameters */
#define N_PROFVAR (N_FPROF+N_TRACER+N_RATIO+N_DIAG_PROF)
                           /* Number of profile variables */
#define N_OUTVAR_PROF (14+N_PROFVAR) 
                           /* Number of depth profile output table variables */
#define N_OUTVAR_SCAL (11+N_FSCALAR+N_PROFVAR+N_DIAG_SCAL+N_TRACER+N_IZ_DIAG) 
                           /* Number of scalar output table variables */
#define N_OBSTYPE (N_FSCALAR+N_PROFVAR+N_DIAG_SCAL+N_TRACER+N_IZ_DIAG)
                           /* Number of observation types */ 
#define N_OUTVAR_MFIT (19+5*(N_OBSTYPE))
                           /* Number of misfit output table variables */ 
#define N_ASSIMPROFVAR 30 
                           /* Number of assimilation profile output variables */
#define N_ASSIMPROFTOPVAR 21 
                           /* Number of assimilation profile output variables
                              applicable to top level */
#define N_OUTVAR_ASSIMPROF (13+N_ASSIMPROFVAR) 
                           /* Number of assim. profile output table variables */
#define N_ASSIMSCALVAR 30  /* Number of assimilation scalar output variables 
                              (excl. 18 variables common to profile output) */
#define N_OUTVAR_ASSIMSCAL (10+N_ASSIMPROFTOPVAR+N_ASSIMSCALVAR) 
                           /* Number of assim. scalar output table variables */

/* Model indices */

#define NULLMOD 0
enum photmodeltype {NULLPHOTMOD,EP85,SMITH,PLATT90};
enum kdmodeltype {NULLKDMOD,LINEARPIG,A93KD,A93KD_ALT};
enum achlmodeltype {NULLACHLMOD,A93ACHL};

/* Dimension indices */

enum {TIME,YEAR,TIMEOFYEAR,LEVEL,DEPTH,ILOC,JLOC}; 

/* Item indices:
     6+N_MODEL+N_ASSIM parameter sets,
     2+N_FSCALITEM+N_FPROFITEM gridded domain items, 
     1 non-gridded domain item.
   N.B. model parameter set item numbers (index+1) match model numbers given in 
   description so that optimizable parameter set reported by GFAn matches
   model number
*/

enum {OPTION=N_MODEL+N_ASSIM,TIMEAXIS,TIMEPERIOD,ENVIRON,PERTSEED,MFITCTRL,
      ZLEVEL,INIT,FSCALAR};
enum {FPROF=FSCALAR+N_FSCALITEM};
enum {OBS=FPROF+N_FPROFITEM};

/* Item variable indices */

enum {GAIN,BALDINEXT,BALDINMIN,RLOSS,BETAG,BETAL,BETAM,SIGMAG,SIGMAL,SIGMAM,
      ZFRACB0,ZFRACB1,QRFMAX,QAFMAX,ZRFMAX,ZAFMAX,PRFMAX,INCPHYMIN,INTEGNSTEP,
      PTHRESHOLD,DINBALUPD,GROWLOSS,SUBSURFUPD,DEEPNEGUPD,DEEPPOSUPD,
      DINPROFUPD,RCCHLUPD};

enum {MODEL,PHOTMODEL,KDMODEL,KDPATHOPT,ATTENWATER,ATTENPIG,ACHLMODEL,
      ADVECTSCHEME,DIFFSCHEME,MIXOPT,PERTTR};
enum {PERTAUTO=PERTTR+N_TRACER_PRI,PERTMIXOPT,RLXCTRLFACT,RLXMASKFLAG,
      RLXMASKOPT,RLXMIXOPT,NSTEPDAY,NSTEPBIO,LEAPFROG,TFPARM,
      ASSIMSCHEME,NSTEPASSIM,STARTSTEPASSIM,FMODE,FMODES};
enum {RLXRATES=FMODES+N_FSCALAR+N_FPROF};
enum {DEFTRACERS=RLXRATES+N_TRACER_PRI};
enum {DEFRATIOS=DEFTRACERS+N_TRACER_PRI};

enum {BASEYEAR,PERIODIC,YEARLEN};
enum {MAXDEP,LATITUDE};
enum {STARTYEAR,STARTDAY,FINYEAR,FINDAY,STARTSTEP};
enum {ZBOT};
enum {SOL,SOLAV,MLD,SSS,PCO2ATM,DUST,FEDUST,CCHLDA,CHLDA,PGROWDA,PLOSSDA};  
enum {VDC,W,TEMP,PERT};
enum {PERTSD=PERT+N_TRACER_PRI}; 
enum {RLXRS=PERTSD+N_TRACER_PRI};
enum {REFS=RLXRS+N_TRACER_PRI};  /* N.B. Reference profiles are assumed to be 
                                    the last of the N_FPROF forcing vars */ 
enum {AUXVARS=REFS+N_TRACER_PRI};

/* Output table indices */

enum {OUTOPTION,OUTSCALAR,OUTPROFILE,OUTSCALAR_DAV,OUTPROFILE_DAV,
      OUTSCALAR_MAV,OUTPROFILE_MAV,OUTSCALAR_YAV,OUTPROFILE_YAV,OUTASSIMSCAL,
      OUTASSIMPROF,OUTMISFIT,OUTCOST};

/* Tracer indices (30 total, 13 primary) */

enum {DIN,NH4,PHY,PNDIA,PDIAT,ZOO,ZMI,ZME,DET,DIC,ALK,SIL,DFE,CHL,CHLNDIA,
      CHLDIAT,PHYC,PNDIAC,PDIATC,ZOOC,ZMIC,ZMEC,DETC,PDIATSI,NITROGEN,CARBON,
      SILICON,IRON,PON,POC};

/* Ratio indices */

enum {CCHL,CCHLNDIA,CCHLDIAT,CNPHY,CNPNDIA,CNPDIAT,CNZOO,CNZMI,CNZME,CNDET,
      SINPDIAT,FEN};

/* Profile diagnostic indices (Note separate indexing for 'id_agg' array) */

enum {FMIXED,PPROD,FESCAV,SINKFLXN,SINKFLXC,SINKFLXSI,SINKFLXCA,SINKFLXFE,
      KDPAR,PARFRACZTOP,PARZTOP,PARFRACZ,PARZ,
      PHOTMAX,LIMDIN,PMAX,ALPHA,PZ,PK,PHOT,
      PHOTMAXNDIA,LIMDINNDIA,LIMFENDIA,
      PMAXNDIA,ALPHANDIA,PZNDIA,PKNDIA,PHOTNDIA,
      PHOTMAXDIAT,LIMDINDIAT,LIMSIDIAT,LIMFEDIAT,
      PMAXDIAT,ALPHADIAT,PZDIAT,PKDIAT,PHOTDIAT,
      WDET,
      INSINKN,INSINKC,INSINKSI,INSINKCA,INSINKFE,
      DDTSINKN,DDTSINKC,DDTSINKSI,DDTSINKCA,DDTSINKFE,
      DDTH_TRACERS};
enum {DDTP_TRACERS=DDTH_TRACERS+N_TRACER};
enum {DDTR_TRACERS=DDTP_TRACERS+N_TRACER};

enum {PPROD_AGG};   /* Indexing for aggregate diagnostic array 'id_agg' */

/* Scalar diagnostic indices */

enum {DAYLEN,ZEUPH,PCO2,CO2FLUX,CO2IN,CO2OUT};

/* Depth integrated profile diagnostic indices */ 

enum {IZPPROD,IZFESCAV,
      IZDDTHN,IZDDTHC,IZDDTHSI,IZDDTHFE,
      IZDDTPN,IZDDTPC,IZDDTPSI,IZDDTPFE,
      IZDDTRN,IZDDTRC,IZDDTRSI,IZDDTRFE};

/* Forcing data interpolation options */

enum fmodetype {NOTAPPLIED,INTERPOLATED,NEAREST,PERSISTENT};

/* Simulation output variable categories & interpolation modes */

enum simcattype {FOUT_SCAL,FOUT_PROF,TRACER,RATIO,DIAG_PROF,DIAG_SCAL,
                 IZ_TRACER,IZ_DIAG};
enum zinterpmode {MIDDLE,BOTTOM,TOP};

/* Transform types */

enum transtype {NULLTRANS,LOGTRANS,SQRTTRANS};

/* Variable association maps for linking ratios, secondary tracers and 
   diagnostics to their respective primary tracers. 
   One map for each variable group (chlorophyll, carbon etc.).
   Each map has N_TRACER_PRI entries indexed by primary tracers:
   DIN,NH4,PHY,PNDIA,PDIAT,ZOO,ZMI,ZME,DET,DIC,ALK,SIL,DFE */

struct vamrec {
  int rcc[N_TRACER_PRI];
  int rc[N_TRACER_PRI];
  int rsi[N_TRACER_PRI];
  int rfe[N_TRACER_PRI];
  int chl[N_TRACER_PRI];
  int c[N_TRACER_PRI];
  int si[N_TRACER_PRI];
  int phot[N_TRACER_PRI];
  int w[N_TRACER_PRI];
};

static struct vamrec vamap = {
  -1,-1,CCHL,CCHLNDIA,CCHLDIAT,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,CNPHY,CNPNDIA,CNPDIAT,CNZOO,CNZMI,CNZME,CNDET,-1,-1,-1,-1,
  -1,-1,-1,-1,SINPDIAT,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,FEN,FEN,FEN,FEN,FEN,FEN,FEN,-1,-1,-1,-1,
  -1,-1,CHL,CHLNDIA,CHLDIAT,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,PHYC,PNDIAC,PDIATC,ZOOC,ZMIC,ZMEC,DETC,-1,-1,-1,-1,
  -1,-1,-1,-1,PDIATSI,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,PHOT,PHOTNDIA,PHOTDIAT,-1,-1,-1,-1,-1,-1,-1,-1,
  -1,-1,-1,-1,-1,-1,-1,-1,WDET,-1,-1,-1,-1
};

/* Observation record structure */

struct obsrec { 
  enum transtype transform[N_OBSTYPE];
  OBJF_FLOAT_TYPE *time;
  OBJF_FLOAT_TYPE *depth;
  OBJF_FLOAT_TYPE *value[N_OBSTYPE];
  OBJF_FLOAT_TYPE *weight[N_OBSTYPE];
  int i;
  double t;
  OBJF_FLOAT_TYPE z; 
  OBJF_FLOAT_TYPE k; 
  OBJF_FLOAT_TYPE kzbot; 
  OBJF_FLOAT_TYPE kztop; 
  int misfitflag[N_OBSTYPE]; 
  double modval[N_OBSTYPE];
  double moddiff[N_OBSTYPE];
  double misfit[N_OBSTYPE];
  double w[N_OBSTYPE];
};

/* Simulation output record structure */

struct modoutrec {
  double fscal[N_FSCALAR];
  double fprof[N_FPROF][NK_MAX];
  double tracer[N_TRACER][NK_MAX];
  double ratio[N_RATIO][NK_MAX];
  double diagscal[N_DIAG_SCAL];
  double diagprof[N_DIAG_PROF][NK_MAX];
  double iztracer[N_TRACER];
  double izdiag[N_IZ_DIAG];
};
     
/* Cost function record structure for accounting variables */
 
struct costrec {
  int mfitcount[N_OBSTYPE];
  double mfitsum[N_OBSTYPE];
};

/* Solver configuration record structure for diagnostics calculations */

struct solverrec {
  int leapfrog;
  int nstepbio;
  int advection;
  int dsink;
};
 

/* External function prototypes */

extern void FORTRAN_modelcodever
  (int *, char *, char *, char *, char *, char *, char *);
extern void FORTRAN_modeldefparm
  (int *, int *, char [][PNAME_MAX+1], char [][PDESC_MAX+1],
   double [], double []);
extern void FORTRAN_assimdefparm
  (int *, int *, char [][PNAME_MAX+1], char [][PDESC_MAX+1],
   double [], double []);
extern void FORTRAN_modelgetinfo
  (int *, int *, double [], int *, int *, int *, char *);
extern void FORTRAN_modelsetup
  (int *, int *, double [], enum photmodeltype *, enum kdmodeltype *, int *, 
   double *, double *, enum achlmodeltype *, int *, int *, int *, int [], 
   int [], enum transtype [], double [], int [], int *, int [], double [], 
   int *, int *, int *, double *, int *, double [], double *, int *, int *, 
   double *, int *, int [], int [][NAGG_MAX], int [][NAGG_MAX], int *, int *, 
   int *, int *, int *, int *, int *, int *, int *, int *, int *, int *);
extern void FORTRAN_assimsetup
  (int *, int *, double [], int *, int *, int *, int *, int *, int *,
   int *, int *, int *);
extern void FORTRAN_modelinit
  (double [][NK_MAX], double [][NK_MAX], double [][NK_MAX], double [][NK_MAX]);
extern void FORTRAN_modelstep
  (double *, int *, double *, double *, double *, double *, double *, double *, 
   double *, double [], double [], double [], double [][NK_MAX], 
   double [][NK_MAX], double [][NK_MAX], double [][NK_MAX], double [][NK_MAX], 
   double *, double *, double *, double *, double [][NK_MAX], double [][NK_MAX],
   int *, int *); 
extern void FORTRAN_getassimreport
  (double [][NK_MAX], double []);
extern void FORTRAN_getdiagnostics
  (double [][NK_MAX], double []);

/* Local function prototypes */ 

static int validitem_option 
  (int, struct itemrec *, int *, enum photmodeltype *, int *, 
   int *, int *, int *, char *);
static int validitem_taxis 
  (int, struct itemrec *, int *, int *, int *, char *);
static int validitem_timeperiod 
  (int, struct itemrec *, int, int *, int *, int *, int *, int *, char *);
static int validitem_environ (struct itemrec *, OBJF_FLOAT_TYPE *, OBJF_FLOAT_TYPE *);
static int validitem_zlevel (int, struct itemrec *, int *, OBJF_FLOAT_TYPE **, char *);
static int validitem_init
  (int, struct itemrec *, int *, OBJF_FLOAT_TYPE *, int *, OBJF_FLOAT_TYPE *, char *);
static int validitem_ft (int, struct itemrec *, char *, char *);
static int validitem_fkt (int, struct itemrec *, char *, int *, char *);
static int validitem_obset (int, struct itemrec *, char *);
static int validftime 
  (int, int, int, int, int, struct itemrec *, char *, char *);

static int daysinyear (int, int);
static int monthofyear (int, int, int);
static int dayofmonth (int, int, int);
static void getforcing_scal 
  (double, int, int, struct itemrec *[], int, int, 
   unsigned char [][N_FSCALAR+N_FPROF], enum fmodetype [], double []);
static int daysinmonth (int, int);
static void getforcing_fulldep 
  (double, int, int, int, struct itemrec *[], int, int, 
   unsigned char [][N_FSCALAR+N_FPROF], enum fmodetype [],
   unsigned char [][N_FPROF], enum fmodetype [], unsigned char [],
   unsigned char [], int, double [][NK_MAX]);
static void getforcing_prof_fromscal
  (double, int, int, struct itemrec *[], int, int, 
   unsigned char [][N_FSCALAR+N_FPROF], enum fmodetype [], double [][NK_MAX]);
static void getforcing_prof 
  (double, int, int, int, struct itemrec *[], int, int, 
   unsigned char [][N_FPROF], enum fmodetype [], int, double [][NK_MAX]);
static int findex (double, struct gridspecrec *, double *);
static int findex_periodic (double, int, struct gridspecrec *, double *, int *);
static void setdiagflags 
  (int [][NAGG_MAX], struct vamrec *, unsigned char [], unsigned char [], 
   unsigned char [], unsigned char [], unsigned char []);
static void setdiagflags_runtime
  (double [][NK_MAX], int [][NAGG_MAX], unsigned char [],
   unsigned char [], struct vamrec *, unsigned char []);
static void setaggregateflag (int, int [][NAGG_MAX], unsigned char [], int [], 
   int [], unsigned char [], unsigned char []);
static void setaggregateflag_pprod (int [][NAGG_MAX], unsigned char [], 
   unsigned char [], int [], double [][NK_MAX], int [], unsigned char []);
static void calcdiagnostics 
  (int, unsigned char [], unsigned char [], unsigned char [], 
   struct solverrec *, int, OBJF_FLOAT_TYPE [], double [][NK_MAX], 
   double [][NK_MAX], double [][NK_MAX], double [][NK_MAX], 
   double [][NK_MAX], int [][NAGG_MAX], int [][NAGG_MAX], struct vamrec *); 
static void calcfuntracers 
  (unsigned char [], int, double [][NK_MAX], double [][NK_MAX], 
   int [][NAGG_MAX], struct vamrec *);
static void calcratios 
  (unsigned char [], int, double [][NK_MAX], double [][NK_MAX], 
   struct vamrec *);
static void calcderived (int, int, int, int, double [], double [][NK_MAX]);
static void calcderived_extra (int, double [], double [], double []);
static void calcaggregate 
  (int, int, int, int [][NAGG_MAX], int [], double [][NK_MAX]);
static void calcaggregate_byr 
  (int, int, int, int [][NAGG_MAX], int [], double [][NK_MAX], 
   double [][NK_MAX]);
static void calcaggregate_pprod 
  (int, int [][NAGG_MAX], int [], double [][NK_MAX], int [], double [][NK_MAX]);
static void calcaggregate_pprod_intrp 
  (int, int [][NAGG_MAX], int [], double, double [][NK_MAX], double [][NK_MAX],
   int [], double [][NK_MAX]);
static void calciz 
  (unsigned char [], unsigned char [], int, OBJF_FLOAT_TYPE [], 
   double [][NK_MAX], double [][NK_MAX], double [], double []); 
static double zinteg (int, OBJF_FLOAT_TYPE [], double []);
static void resetaccumulator (struct modoutrec *, int, int, double);
static void updaccumulator 
  (struct modoutrec *acc, struct modoutrec *, int, int);
static void updatesums_scal (int, double [], double, double [], double);
static void updatesums_prof 
  (int, int, double [][NK_MAX], double, double [][NK_MAX], double);
static int writemod_scal 
  (int, FILE *, int, struct caseinputrec *, double, double, int, int, int, 
   OBJF_FLOAT_TYPE, int, int, OBJF_FLOAT_TYPE, OBJF_FLOAT_TYPE, OBJF_FLOAT_TYPE,
   double [], double [][NK_MAX], double [][NK_MAX], double [][NK_MAX], 
   double [][NK_MAX], double [], double [], double [], char **);
static int writemodav_scal 
  (int, FILE *, int, struct caseinputrec *, double, int, int, OBJF_FLOAT_TYPE, 
   int, int, OBJF_FLOAT_TYPE, OBJF_FLOAT_TYPE, OBJF_FLOAT_TYPE, 
   struct modoutrec *, int, char **);
static int writemod_prof 
  (int, FILE *, int, struct caseinputrec *, double, double, int, int, int, 
   OBJF_FLOAT_TYPE, int, int, OBJF_FLOAT_TYPE, int, OBJF_FLOAT_TYPE [], 
   OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE [], 
   double [][NK_MAX], double [][NK_MAX], double [][NK_MAX], double [][NK_MAX], 
   char **);
static int writemodav_prof 
  (int, FILE *, int, struct caseinputrec *, double, int, int, OBJF_FLOAT_TYPE, 
   int, int, OBJF_FLOAT_TYPE, int, OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE [], 
   OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE [], struct modoutrec *, int, char **);
static int writeassim_scal 
  (int, FILE *, int, struct caseinputrec *, double, int, int, int, 
   OBJF_FLOAT_TYPE, int, int, OBJF_FLOAT_TYPE, OBJF_FLOAT_TYPE, OBJF_FLOAT_TYPE,
   double [][NK_MAX], double [], char **);
static int writeassim_prof 
  (int, FILE *, int, struct caseinputrec *, double, int, int, int, 
   OBJF_FLOAT_TYPE, int, int, OBJF_FLOAT_TYPE, int, OBJF_FLOAT_TYPE [], 
   OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE [], 
   double [][NK_MAX], char **);
static void calcmisfit 
  (struct obsrec *, int, OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE [], 
   OBJF_FLOAT_TYPE, double [], double [][NK_MAX], double [][NK_MAX], 
   double [][NK_MAX], double [][NK_MAX], double [], double [], double []);
static int nextobs 
  (struct itemrec *, int, double, double, OBJF_FLOAT_TYPE, struct obsrec *);
static enum simcattype simcatofob (int, int *, enum zinterpmode *);
static OBJF_FLOAT_TYPE zinterp (int, OBJF_FLOAT_TYPE [], OBJF_FLOAT_TYPE, 
   int *, double *);
static double profvalue (enum zinterpmode, int, double, int, double, double []);
static double fltdiff (double, double);
static int writemisfit 
  (int, FILE *, int, struct caseinputrec *, double, double, int, int, int, 
   OBJF_FLOAT_TYPE, int, int, OBJF_FLOAT_TYPE, int, int, struct obsrec *, 
   char **);
static int writecost
  (int, FILE *, int, struct caseinputrec *, struct costrec *);
static double costfunc (struct costrec *, int *);


/* Function definitions */
 
int objfun_setnames (struct objfnamerec *name, char **errtxt)

{ extern double FORTRAN_modelobjfnodata;
  extern double FORTRAN_modelatomc;
  extern int FORTRAN_modelvertextlen;
  extern int FORTRAN_modelmsgtextlen;
  extern int FORTRAN_modelnp;
  extern int FORTRAN_modelpnamelen;
  extern int FORTRAN_modelpdesclen;
  extern double FORTRAN_modelrelaxulim;
  extern int FORTRAN_modelnk;
  extern int FORTRAN_modelnt;
  extern int FORTRAN_modelnr;
  extern int FORTRAN_modelndp;
  extern int FORTRAN_modelnds;
  extern int FORTRAN_modelntpri;
  extern int FORTRAN_modelntder;
  extern int FORTRAN_modelndpbasic;
  extern int FORTRAN_modelnagg;
  extern int FORTRAN_modelntagg;
  extern int FORTRAN_modelndagg;
  extern int FORTRAN_modelndpocda;
  extern int FORTRAN_modelndsocda;
  extern int FORTRAN_nmodel; 
  extern char FORTRAN_modname[N_MODEL][VERTEXT_MAX+1];
  extern int FORTRAN_modassimflag[N_MODEL]; 
  char *modname[N_MODEL+1]; 
  char *modver[N_MODEL+1]; 
  char *modver1[N_MODEL+1]; 
  char *modver2[N_MODEL+1]; 
  char *moddesc[N_MODEL+1];
  char *moddesc1[N_MODEL+1];
  char *moddesc2[N_MODEL+1];
  static char modcodever[N_MODEL][VERTEXT_MAX+1];
  static char modcodever1[N_MODEL][VERTEXT_MAX+1];
  static char modcodever2[N_MODEL][VERTEXT_MAX+1];
  static char modcodedesc[N_MODEL][VERTEXT_MAX+1];
  static char modcodedesc1[N_MODEL][VERTEXT_MAX+1];
  static char modcodedesc2[N_MODEL][VERTEXT_MAX+1];
  static char pname[N_MODEL+N_ASSIM][NP_MAX][PNAME_MAX+1];
  static char pdesc[N_MODEL+N_ASSIM][NP_MAX][PDESC_MAX+1];
  double pmin[N_MODEL+N_ASSIM][NP_MAX];
  double pmax[N_MODEL+N_ASSIM][NP_MAX];
  int model,i,j,k,jf,len,ioutprof_fprof,nassim;
  int itmid,istep,iday,idate,ithours,imth,ioutprof_dertracer,
      ioutscal_obstype,ioutassimprof_state,iobs_weights;
  char *s,*slimit;
  static char errline[200];

  *errtxt=errline;

  if (N_OUTVAR_MFIT>NFIELD_MAX)
    { sprintf(errline,
  "MarMOT error - N_OUTVAR_MFIT=%d, max no. of variables NFIELD_MAX=%d",
              N_OUTVAR_MFIT,NFIELD_MAX);
      return SPECIAL;
    }

  if (FORTRAN_modelobjfnodata!=OBJF_NODATA)
    { sprintf(errline,
  "bad MarMOT model configuration - expected objf_nodata=%g, found objf_nodata=%g",
              OBJF_NODATA,FORTRAN_modelobjfnodata);
      return SPECIAL;
    }

  if (fltdiff(FORTRAN_modelatomc,ATOM_C)!=0)
    { sprintf(errline,
  "bad MarMOT model configuration - expected atomc=%g, found atomc=%g",
              ATOM_C,FORTRAN_modelatomc);
      return SPECIAL;
    }

  if (FORTRAN_modelvertextlen > VERTEXT_MAX)
    { sprintf(errline,
 "bad MarMOT model configuration - expected vertextlen=%d, found vertextlen=%d",
              VERTEXT_MAX,FORTRAN_modelvertextlen);
      return SPECIAL;
    }

  if (FORTRAN_modelmsgtextlen > MODTEXT_MAX)
    { sprintf(errline,
 "bad MarMOT model configuration - expected msgtextlen=%d, found msgtextlen=%d",
              MODTEXT_MAX,FORTRAN_modelmsgtextlen);
      return SPECIAL;
    }

  if (FORTRAN_modelnp > NP_MAX)
    { sprintf(errline,
              "bad MarMOT model configuration - expected np=%d, found np=%d",
              NP_MAX,FORTRAN_modelnp);
      return SPECIAL;
    }

  if (FORTRAN_modelpnamelen != PNAME_MAX)
    { sprintf(errline,
  "bad MarMOT model configuration - expected pnamelen=%d, found pnamelen=%d",
              PNAME_MAX,FORTRAN_modelpnamelen);
      return SPECIAL;
    }

  if (FORTRAN_modelpdesclen != PDESC_MAX)
    { sprintf(errline,
  "bad MarMOT model configuration - expected pdesclen=%d, found pdesclen=%d",
              PDESC_MAX,FORTRAN_modelpdesclen);
      return SPECIAL;
    }

  if (FORTRAN_modelrelaxulim != RELAXULIM)
    { sprintf(errline,
  "bad MarMOT model configuration - expected relaxulim=%g, found relaxulim=%g",
              RELAXULIM,FORTRAN_modelrelaxulim);
      return SPECIAL;
    }

  if (FORTRAN_modelnk != NK_MAX)
    { sprintf(errline,
              "bad MarMOT model configuration - expected nk=%d, found nk=%d",
              NK_MAX,FORTRAN_modelnk);
      return SPECIAL;
    }

  if (FORTRAN_modelnt > N_TRACER)
    { sprintf(errline,
              "bad MarMOT model configuration - expected nt<=%d, found nt=%d",
              N_TRACER,FORTRAN_modelnt);
      return SPECIAL;
    }

  if (FORTRAN_modelnr > N_RATIO)
    { sprintf(errline,
              "bad MarMOT model configuration - expected nr<=%d, found nr=%d",
              N_RATIO,FORTRAN_modelnr);
      return SPECIAL;
    }

  if (FORTRAN_modelndp > N_DIAG_PROF)
    { sprintf(errline,
              "bad MarMOT model configuration - expected ndp<=%d, found ndp=%d",
              N_DIAG_PROF,FORTRAN_modelndp);
      return SPECIAL;
    }

  if (FORTRAN_modelnds > N_DIAG_SCAL)
    { sprintf(errline,
              "bad MarMOT model configuration - expected nds<=%d, found nds=%d",
              N_DIAG_SCAL,FORTRAN_modelnds);
      return SPECIAL;
    }

  if (FORTRAN_modelntagg != N_TRACER)
    { sprintf(errline,
      "bad MarMOT model configuration - expected ntagg=%d, found ntagg=%d",
              N_TRACER,FORTRAN_modelntagg);
      return SPECIAL;
    }

  if (FORTRAN_modelndagg != N_DIAG_AGG)
    { sprintf(errline,
      "bad MarMOT model configuration - expected ndagg=%d, found ndagg=%d",
              N_DIAG_AGG,FORTRAN_modelndagg);
      return SPECIAL;
    }

  if (FORTRAN_modelnagg != NAGG_MAX)
    { sprintf(errline,
        "bad MarMOT model configuration - expected nagg=%d, found nagg=%d",
              NAGG_MAX,FORTRAN_modelnagg);
      return SPECIAL;
    }

  if (FORTRAN_modelntpri != N_TRACER_PRI)
    { sprintf(errline,
        "bad MarMOT model configuration - expected ntpri=%d, found ntpri=%d",
              N_TRACER_PRI,FORTRAN_modelntpri);
      return SPECIAL;
    }

  if (FORTRAN_modelntder != N_TRACER_DER)
    { sprintf(errline,
        "bad MarMOT model configuration - expected ntder=%d, found ntder=%d",
              N_TRACER_DER,FORTRAN_modelntder);
      return SPECIAL;
    }

  if (FORTRAN_modelndpbasic != N_DIAG_PROF_BASIC)
    { sprintf(errline,
     "bad MarMOT model configuration - expected ndpbasic=%d, found ndpbasic=%d",
              N_DIAG_PROF_BASIC,FORTRAN_modelndpbasic);
      return SPECIAL;
    }

  if (FORTRAN_modelndpocda > N_ASSIMPROFVAR)
    { sprintf(errline,
  "bad MarMOT model configuration - expected ndp_ocda<=%d, found ndp_ocda=%d",
              N_ASSIMPROFVAR,FORTRAN_modelndpocda);
      return SPECIAL;
    }

  if (FORTRAN_modelndsocda > N_ASSIMSCALVAR)
    { sprintf(errline,
  "bad MarMOT model configuration - expected nds_ocda<=%d, found nds_ocda=%d",
              N_ASSIMSCALVAR,FORTRAN_modelndsocda);
      return SPECIAL;
    }

  if (FORTRAN_nmodel<N_MODEL) 
    { sprintf(errline,
        "bad MarMOT model configuration - expected nmodel>=%d, found nmodel=%d",
        N_MODEL,FORTRAN_nmodel);
      return SPECIAL;
    }

  nassim=0;
  for (i=1;i<=N_MODEL;i++) nassim+=FORTRAN_modassimflag[i-1];
  if (nassim<N_ASSIM) 
    { sprintf(errline,
        "bad MarMOT model configuration - expected nassim>=%d, found nassim=%d",
        N_ASSIM,nassim);
      return SPECIAL;
    }

  modname[NULLMOD]="no ecosystem model";
  modver[NULLMOD]="";
  modver1[NULLMOD]="";
  modver2[NULLMOD]="";
  moddesc[NULLMOD]="";
  moddesc1[NULLMOD]="";
  moddesc2[NULLMOD]="";

  for (i=1;i<=N_MODEL;i++) 

    { modname[i]=FORTRAN_modname[i-1];
      j=VERTEXT_MAX;
      modname[i][j--]='\0';
      while (j>0 && modname[i][j]==' ') modname[i][j--]='\0';

      modver[i]="unspecified"; 
      modver1[i]="";
      modver2[i]="";
      moddesc[i]="";
      moddesc1[i]="";
      moddesc2[i]="";
      *modcodever[i]='\0';
      *modcodever1[i]='\0';
      *modcodever2[i]='\0';
      *modcodedesc[i]='\0';
      *modcodedesc1[i]='\0';
      *modcodedesc2[i]='\0';

      FORTRAN_modelcodever(&i,modcodever[i],modcodever1[i],modcodever2[i],
                           modcodedesc[i],modcodedesc1[i],modcodedesc2[i]);

      if (*modcodever[i]) 
        { j=VERTEXT_MAX;
          modcodever[i][j--]='\0'; 
          while (j>0 && modcodever[i][j]==' ') modcodever[i][j--]='\0';
          modver[i]=modcodever[i];
        }
      if (*modcodever1[i]) 
        { j=VERTEXT_MAX;
          modcodever1[i][j--]='\0'; 
          while (j>0 && modcodever1[i][j]==' ') modcodever1[i][j--]='\0';
          modver1[i]=modcodever1[i];
        }
      if (*modcodever2[i]) 
        { j=VERTEXT_MAX;
          modcodever2[i][j--]='\0'; 
          while (j>0 && modcodever2[i][j]==' ') modcodever2[i][j--]='\0';
          modver2[i]=modcodever2[i];
        }
      if (*modcodedesc[i]) 
        { j=VERTEXT_MAX;
          modcodedesc[i][j--]='\0'; 
          while (j>0 && modcodedesc[i][j]==' ') modcodedesc[i][j--]='\0';
          moddesc[i]=modcodedesc[i];
        }
      if (*modcodedesc1[i]) 
        { j=VERTEXT_MAX;
          modcodedesc1[i][j--]='\0'; 
          while (j>0 && modcodedesc1[i][j]==' ') modcodedesc1[i][j--]='\0';
          moddesc1[i]=modcodedesc1[i];
        }
      if (*modcodedesc2[i]) 
        { j=VERTEXT_MAX;
          modcodedesc2[i][j--]='\0'; 
          while (j>0 && modcodedesc2[i][j]==' ') modcodedesc2[i][j--]='\0';
          moddesc2[i]=modcodedesc2[i];
        }
    }

  len=2500;
  for (i=0;i<=N_MODEL;i++) 
    len+=8+strlen(modname[i])+
         22+strlen(modver[i])+
         11+strlen(modver1[i])+
         11+strlen(modver2[i])+
         21+strlen(moddesc[i])+
         11+strlen(moddesc1[i])+
         11+strlen(moddesc2[i]);
  s=malloc(len+1);
  if (!s) return MEMERR; 
  slimit=s+len; 

  name->function="MarMOT Model Evaluator (MME)"; 
  name->version=MARMOT_VERSION;
  name->description=s;

  strcpy(s,
"Model evaluation engine for the Marine Model Optimization Testbed.\nPerforms a water-column simulation of a plankton ecosystem for each prescribed\ncase associated with a site and/or ensemble member.\nReturns a cost function value to GFAn summarizing the misfit of the simulations\nto reference data over all cases.\n");
                                             /* approx. 300 characters */
  s+=strlen(s);
  strcpy(s,"\nCopyright Natural Envrionment Research Council\nAdvection and diffusion code derived from NEMO 3.4 (copyright NEMO Consortium)\nLicence: CeCILL\nDifferent conditions may apply to individual ecosystem models\n(refer to source code for more information).\n");
                                            /* approx. 270 characters */
  s+=strlen(s);
  strcpy(s,"\nEcosystem model options\n");  /* 25 characters */
  s+=strlen(s);
  for (i=0;i<=N_MODEL;i++)
    { sprintf(s,"\n  %2d : ",i);
      s+=strlen(s);
      strcpy(s,modname[i]);
      s+=strlen(s);
      if (*modver[i])
        { strcpy(s,"\n       code version: ");
          s+=strlen(s);
          strcpy(s,modver[i]);
          s+=strlen(s);
        }
      if (*modver1[i])
        { strcpy(s,"\n        - ");
          s+=strlen(s);
          strcpy(s,modver1[i]);
          s+=strlen(s);
        }
      if (*modver2[i])
        { strcpy(s,"\n        - ");
          s+=strlen(s);
          strcpy(s,modver2[i]);
          s+=strlen(s);
        }
      if (*moddesc[i]) 
        { strcpy(s,"\n       description: ");
          s+=strlen(s);
          strcpy(s,moddesc[i]);
          s+=strlen(s);
        }
      if (*moddesc1[i])
        { strcpy(s,"\n        - ");
          s+=strlen(s);
          strcpy(s,moddesc1[i]);
          s+=strlen(s);
        }
      if (*moddesc2[i])
        { strcpy(s,"\n        - ");
          s+=strlen(s);
          strcpy(s,moddesc2[i]);
          s+=strlen(s);
        }
    }

  if (s > slimit) return MEMERR;

  len=5000;
  for (i=0;i<=N_MODEL;i++) len+=8+strlen(modname[i]);
  s=malloc(len+1);
  if (!s) return MEMERR; 
  slimit=s+len; 
  name->notes=s;

  strcpy(s,"====================================================\n"); 
  s+=strlen(s);
  strcpy(s," Codes for variables in parameter set item 'option' \n"); 
  s+=strlen(s);
  strcpy(s,"====================================================\n");
                                                     /* 3 * 54 characters */ 

  s+=strlen(s);
  strcpy(s,"\nmodel - ecosystem model number\n");  /* 32 characters */
  s+=strlen(s);
  for (i=0;i<=N_MODEL;i++)
    { sprintf(s,"\n  %2d : ",i);
      s+=strlen(s);
      strcpy(s,modname[i]);
      s+=strlen(s);
    }
 
  strcpy(s,"\n\nphotmodel - photosynthesis light limitation model number\n");   
                                                     /* 59 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : no photosynthesis (default)\n   1 : Evans and Parslow 1985 daily mean photosynthesis\n   2 : point-in-time version of Evans and Parslow 1985\n   3 : Platt et al. 1990 daily mean photosynthesis");
                                                     /* 198 characters */

  s+=strlen(s);
  strcpy(s,"\n\nkdmodel - light attenuation model number\n");
                                                     /* 43 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : constant attenuation (default)\n   1 : depth-independent attenuation\n   2 : Anderson 1993 attenuation model - original layers (0-5m,5-23m,>23m)\n   3 : Anderson 1993 attenuation model - alternative layers (0-10m,10-20m,>20m)");
                                                     /* 230 characters */

  s+=strlen(s);
  strcpy(s,"\n\nkdpathopt - light attenuation model option for path length adjustment\n");
                                                     /* 72 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : no adjustment (default)\n   1 : Oschlies and Garcon 1999 noon path length adjustment");
                                                     /* 91 characters */

  s+=strlen(s);
  strcpy(s,"\n\nachlmodel - light absorption model number\n");   
                                                     /* 44 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : depth independent absorption (default)\n   1 : Anderson 1993 spectrally-averaged chlorophyll absorption");
                                                     /* 110 characters */

  s+=strlen(s);
  strcpy(s,"\n\nadvection - advection scheme number\n");   
                                                     /* 38 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : no advection (default); * no active tracer vel. unless handled by model *\n   1 : Basic upstream scheme\n   2 : Monotonic Upstream Scheme for Conservative Laws (MUSCL) * Recommended *");
                                                     /* 183 characters */

  s+=strlen(s);
  strcpy(s,"\n\ndiffusion - diffusion scheme\n");    /* 31 characters */
  s+=strlen(s);
  strcpy(s,"\n   0 : no diffusion (default)\n   1 : backward (implicit) diffusion scheme");
                                                     /* 74 characters */

  s+=strlen(s);
  strcpy(s,"\n\nmixopt/pertmixopt/rlxmixopt - boundary layer mixing options\n");   
                                                     /* 62 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : none (default)\n   1 : above MLD only \n   2 : partial mix at MLD");
                                                     /* 70 characters */

  s+=strlen(s);
  strcpy(s,"\n\nfmode - forcing data interpolation modes\n");   
                                                     /* 43 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : not applied\n   1 : linear (default)\n   2 : nearest\n   3 : persistent");
                                                     /* 76 characters */

  s+=strlen(s);
  strcpy(s,
"\n\nperttr<varname> - transformation types for tracer perturbation\n");
                                                     /* 66 characters */
  s+=strlen(s);
  strcpy(s,"\n   0 : no transformation (default)\n   1 : log10\n   2 : sqrt");
                                                     /* 60 characters */

  s+=strlen(s);
  strcpy(s,"\n\nrlxmask - tracer relaxation masking flag\n");  
                                                     /* 43 characters */
  s+=strlen(s);
  strcpy(s,
"\n   0 : no mask (default)\n  -1 : mask at z and above\n   1 : mask at z and below");
                                                     /* 79 characters */

  s+=strlen(s);
  strcpy(s,"\n\nrlxmaskopt - tracer relaxation masking reference depth\n"); 
                                                     /* 57 characters */
  s+=strlen(s);
  strcpy(s,
"\n   1 : z=mld\n   2 : z=zeuphotic\n   3 : z=max(mld,zeuphotic)");
                                                     /* 60 characters */

  s+=strlen(s);
  strcpy(s,"\n\n");                                  /* 2 characters */
  s+=strlen(s);
  strcpy(s,"======================================================\n");  
  s+=strlen(s);  
  strcpy(s," Codes for variables in parameter set item 'mfitctrl' \n");
  s+=strlen(s); 
  strcpy(s,"======================================================\n");
                                                     /* 3 * 56 characters */

  s+=strlen(s);
  strcpy(s,"\ntr.<varname> - transformation types for model-data differencing\n");
                                                     /* 65 characters */
  s+=strlen(s);
  strcpy(s,"\n   0 : no transformation (default)\n   1 : log10\n   2 : sqrt");
                                                     /* 60 characters */

  if (s+strlen(s) > slimit) return MEMERR;

  name->dimension[TIME]="t";
  name->dimension[YEAR]="year";
  name->dimension[TIMEOFYEAR]="tofyr";
  name->dimension[LEVEL]="k";
  name->dimension[DEPTH]="z";
  name->dimension[ILOC]="i";
  name->dimension[JLOC]="j";
  name->casevar[0]="site";
  name->casevar[1]="member";

  for (i=0;i<N_ITEM;i++) 
    { name->item[i].dimenabled[TIME]=0;
      name->item[i].dimenabled[YEAR]=0;
      name->item[i].dimenabled[TIMEOFYEAR]=0;
      name->item[i].dimenabled[LEVEL]=0;
      name->item[i].dimenabled[DEPTH]=0;
      name->item[i].dimenabled[ILOC]=0;
      name->item[i].dimenabled[JLOC]=0;
    }

  /* Model & assimilation parameter set items */

  len=0;
  for (i=0;i<N_MODEL+N_ASSIM;i++) len+=9;
  s=malloc(len);
  if (!s) return MEMERR;

  for (i=0;i<N_MODEL;i++)
    { name->item[i].name=s;
      strcpy(s,"model");
      s+=strlen(s);
      sprintf(s,"%d",i+1);
      s+=strlen(s)+1;
    }

  for (j=0; j<N_MODEL && i<(N_MODEL+N_ASSIM); j++) if (FORTRAN_modassimflag[j])
    { name->item[i++].name=s;
      strcpy(s,"assim");
      s+=strlen(s);
      sprintf(s,"%d",j+1);
      s+=strlen(s)+1;
    }

  for (i=0;i<N_MODEL+N_ASSIM;i++)
    { name->item[i].optional=1;
      name->item[i].nvar=0;
      for (j=0;j<NP_MAX;j++)
        { pname[i][j][0]='\0';
          pdesc[i][j][0]='\0';
          pmin[i][j]=OBJF_NODATA;
          pmax[i][j]=OBJF_NODATA;
        }
    }

  for (i=0;i<N_MODEL;i++)
    { model=i+1;
      FORTRAN_modeldefparm(&model,&name->item[i].nvar,pname[i],pdesc[i],
                           pmin[i],pmax[i]);
    }

  for (j=0; j<N_MODEL && i<(N_MODEL+N_ASSIM); j++) if (FORTRAN_modassimflag[j])
      { model=j+1;
        FORTRAN_assimdefparm(&model,&name->item[i].nvar,pname[i],pdesc[i],
                             pmin[i],pmax[i]);
        i++;
      } 

  for (i=0;i<N_MODEL+N_ASSIM;i++)
    { for (j=0;j<name->item[i].nvar;j++) 
        { k=PNAME_MAX;
          pname[i][j][k--]='\0';
          while (k>0 && pname[i][j][k]==' ') pname[i][j][k--]='\0';
          name->item[i].varname[j]=pname[i][j];
          k=PDESC_MAX;
          pdesc[i][j][k--]='\0';
          while (k>0 && pdesc[i][j][k]==' ') pdesc[i][j][k--]='\0';
          name->item[i].vartext[j]=pdesc[i][j];
          name->item[i].varmin[j]=pmin[i][j];   
          name->item[i].varmax[j]=pmax[i][j];
        }
    }

  /* Gridded domain items */

  name->item[ZLEVEL].name="zlevel";
  name->item[ZLEVEL].optional=1;
  name->item[ZLEVEL].dimenabled[LEVEL]=1;
  name->item[ZLEVEL].dimenabled[ILOC]=1;
  name->item[ZLEVEL].dimenabled[JLOC]=1;
  name->item[ZLEVEL].dimoptional[ILOC]=1;
  name->item[ZLEVEL].dimoptional[JLOC]=1;
  name->item[ZLEVEL].nvar=1;
  name->item[ZLEVEL].vartext[ZBOT]="depth at bottom of level (m)";
  name->item[ZLEVEL].varname[ZBOT]="zbot";

  name->item[INIT].name="init";
  name->item[INIT].optional=1;
  name->item[INIT].dimenabled[TIME]=1;
  name->item[INIT].dimenabled[YEAR]=1;
  name->item[INIT].dimenabled[TIMEOFYEAR]=1;
  name->item[INIT].dimenabled[LEVEL]=1;
  name->item[INIT].dimenabled[ILOC]=1;
  name->item[INIT].dimenabled[JLOC]=1;
  name->item[INIT].dimoptional[TIME]=1;
  name->item[INIT].dimoptional[YEAR]=1;
  name->item[INIT].dimoptional[TIMEOFYEAR]=1;
  name->item[INIT].dimoptional[ILOC]=1;
  name->item[INIT].dimoptional[JLOC]=1;
  name->item[INIT].nvar=N_TRACER_PRI+N_RATIO;          
  name->item[INIT].vartext[DIN]="dissolved inorganic nitrogen (mmolN/m^3)";
  name->item[INIT].varname[DIN]="din";
  name->item[INIT].vartext[NH4]="ammonium (mmolN/m^3)";
  name->item[INIT].varname[NH4]="nh4";
  name->item[INIT].vartext[PHY]="phytoplankton nitrogen (mmolN/m^3)";
  name->item[INIT].varname[PHY]="phy";
  name->item[INIT].vartext[PNDIA]=
    "non-diatom phytoplankton nitrogen (mmolN/m^3)";
  name->item[INIT].varname[PNDIA]="pndia";
  name->item[INIT].vartext[PDIAT]="diatom nitrogen (mmolN/m^3)";
  name->item[INIT].varname[PDIAT]="pdiat";
  name->item[INIT].vartext[ZOO]="zooplankton nitrogen (mmolN/m^3)";
  name->item[INIT].varname[ZOO]="zoo";
  name->item[INIT].vartext[ZMI]="microzooplankton nitrogen (mmolN/m^3)";
  name->item[INIT].varname[ZMI]="zmi";
  name->item[INIT].vartext[ZME]="mesozooplankton nitrogen (mmolN/m^3)";
  name->item[INIT].varname[ZME]="zme";
  name->item[INIT].vartext[DET]="detrital nitrogen (mmolN/m^3)";
  name->item[INIT].varname[DET]="det";
  name->item[INIT].vartext[DIC]="dissolved inorganic carbon (mmolC/m^3)";
  name->item[INIT].varname[DIC]="dic";
  name->item[INIT].vartext[ALK]="alkalinity (meq/m^3)";
  name->item[INIT].varname[ALK]="alk";
  name->item[INIT].vartext[SIL]="dissolved silicon (silicic acid) (mmolSi/m^3)";
  name->item[INIT].varname[SIL]="dsi";
  name->item[INIT].vartext[DFE]="dissolved iron (mmolFe/m^3)";
  name->item[INIT].varname[DFE]="dfe";
  name->item[INIT].vartext[CCHL+N_TRACER_PRI]="C:Chl ratio (gC/gChl)";
  name->item[INIT].varname[CCHL+N_TRACER_PRI]="rcchl";
  name->item[INIT].vartext[CCHLNDIA+N_TRACER_PRI]=
    "non-diatom C:Chl ratio (gC/gChl)";
  name->item[INIT].varname[CCHLNDIA+N_TRACER_PRI]="rcchlndia";
  name->item[INIT].vartext[CCHLDIAT+N_TRACER_PRI]=
    "diatom C:Chl ratio (gC/gChl)";
  name->item[INIT].varname[CCHLDIAT+N_TRACER_PRI]="rcchldiat";
  name->item[INIT].vartext[CNPHY+N_TRACER_PRI]="phytoplankton C:N ratio";
  name->item[INIT].varname[CNPHY+N_TRACER_PRI]="rcnphy";
  name->item[INIT].vartext[CNPNDIA+N_TRACER_PRI]=
    "non-diatom phytoplankton C:N ratio";
  name->item[INIT].varname[CNPNDIA+N_TRACER_PRI]="rcnpndia";
  name->item[INIT].vartext[CNPDIAT+N_TRACER_PRI]="diatom C:N ratio";
  name->item[INIT].varname[CNPDIAT+N_TRACER_PRI]="rcnpdiat";
  name->item[INIT].vartext[CNZOO+N_TRACER_PRI]="zooplankton C:N ratio";
  name->item[INIT].varname[CNZOO+N_TRACER_PRI]="rcnzoo";
  name->item[INIT].vartext[CNZMI+N_TRACER_PRI]="microzooplankton C:N ratio";
  name->item[INIT].varname[CNZMI+N_TRACER_PRI]="rcnzmi";
  name->item[INIT].vartext[CNZME+N_TRACER_PRI]="mesozooplankton C:N ratio";
  name->item[INIT].varname[CNZME+N_TRACER_PRI]="rcnzme";
  name->item[INIT].vartext[CNDET+N_TRACER_PRI]="detritus C:N ratio";
  name->item[INIT].varname[CNDET+N_TRACER_PRI]="rcndet";
  name->item[INIT].vartext[SINPDIAT+N_TRACER_PRI]="diatom Si:N ratio";
  name->item[INIT].varname[SINPDIAT+N_TRACER_PRI]="rsinpdiat";
  name->item[INIT].vartext[FEN+N_TRACER_PRI]="particulate organic Fe:N ratio";
  name->item[INIT].varname[FEN+N_TRACER_PRI]="rfen";
  for (i=0;i<name->item[INIT].nvar;i++) name->item[INIT].varoptional[i]=1;

  if (N_AUXVAR>99)
    { sprintf(errline,
  "MarMOT error - N_AUXVAR=%d, max no. of auxiliary variables allowed is 99",
              N_AUXVAR);
      return SPECIAL;
    }

  len=0;
  for (i=0;i<N_TRACER_PRI;i++) len+=4*(strlen(name->item[INIT].varname[i])+7);
  for (i=0;i<N_AUXVAR;i++) len+=9;
  s=malloc(len);
  if (!s) return MEMERR; 

  name->item[FPROF].name="fkt";
  name->item[FPROF].optional=1;
  name->item[FPROF].dimenabled[TIME]=1;
  name->item[FPROF].dimenabled[LEVEL]=1;
  name->item[FPROF].dimenabled[ILOC]=1;
  name->item[FPROF].dimenabled[JLOC]=1;
  name->item[FPROF].dimoptional[TIME]=1;
  name->item[FPROF].dimoptional[ILOC]=1;
  name->item[FPROF].dimoptional[JLOC]=1;
  name->item[FPROF].nvar=N_FPROF+N_AUXVAR;          
  name->item[FPROF].vartext[VDC]=
    "vertical diffusion coefficient forcing (m^2/d, depth zbot)";
  name->item[FPROF].varname[VDC]="vdc";
  name->item[FPROF].vartext[W]=
    "vertical velocity forcing (m/d, upward, depth zbot)";
  name->item[FPROF].varname[W]="w";
  name->item[FPROF].vartext[TEMP]="temperature forcing (degC)";
  name->item[FPROF].varname[TEMP]="temp";
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[FPROF].vartext[i+PERT]=
        "perturbation rate for tracer (transformed tracer units /d)";
      name->item[FPROF].varname[i+PERT]=s;
      strcpy(s,"pert");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[i]);
      s+=strlen(s)+1;
    }
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[FPROF].vartext[i+PERTSD]=
        "perturbation rate s.d. for tracer (transformed tracer units /d)";
      name->item[FPROF].varname[i+PERTSD]=s;
      strcpy(s,"pertsd");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[i]);
      s+=strlen(s)+1;
    }
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[FPROF].vartext[i+RLXRS]="relaxation rate for tracer (/d)";
      name->item[FPROF].varname[i+RLXRS]=s;
      strcpy(s,"rlx");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[i]);
      s+=strlen(s)+1;
    }
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[FPROF].vartext[i+REFS]=name->item[INIT].vartext[i]; 
      name->item[FPROF].varname[i+REFS]=s; 
      strcpy(s,"ref");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[i]);
      s+=strlen(s)+1;
    }
  for (i=0;i<N_AUXVAR;i++)
    { name->item[FPROF].vartext[i+AUXVARS]="unprocessed auxiliary variable";
      name->item[FPROF].varname[i+AUXVARS]=s;
      strcpy(s,"auxvar");
      s+=strlen(s);
      sprintf(s,"%d",i+1);
      s+=strlen(s)+1;
    }
  for (i=0;i<name->item[FPROF].nvar;i++) name->item[FPROF].varoptional[i]=1;

  if (N_FPROFITEM>99)
    { sprintf(errline,
        "MarMOT error - N_FPROFITEM=%d, max no. of 'fkt' items allowed is 99", 
              N_FPROFITEM);
      return SPECIAL;
    }

  len=0;
  for (j=1;j<N_FPROFITEM;j++) len+=strlen(name->item[FPROF].name)+3;
  s=malloc(len);
  if (!s) return MEMERR;

  for (j=1;j<N_FPROFITEM;j++)

    { jf=FPROF+j;
      name->item[jf].name=s;
      sprintf(s,"%s%d",name->item[FPROF].name,j+1);
      s+=strlen(s)+1;
      name->item[jf].optional=1;
      name->item[jf].dimenabled[TIME]=1;
      name->item[jf].dimenabled[LEVEL]=1;
      name->item[jf].dimenabled[ILOC]=1;
      name->item[jf].dimenabled[JLOC]=1;
      name->item[jf].dimoptional[TIME]=1;
      name->item[jf].dimoptional[ILOC]=1;
      name->item[jf].dimoptional[JLOC]=1;
      name->item[jf].nvar=name->item[FPROF].nvar;          
      for (i=0;i<name->item[FPROF].nvar;i++)
        { name->item[jf].vartext[i]=name->item[FPROF].vartext[i];
          name->item[jf].varname[i]=name->item[FPROF].varname[i];
        }
      for (i=0;i<name->item[FPROF].nvar;i++) name->item[jf].varoptional[i]=1;
    }

  name->item[FSCALAR].name="ft";
  name->item[FSCALAR].optional=1;
  name->item[FSCALAR].dimenabled[TIME]=1;
  name->item[FSCALAR].dimenabled[LEVEL]=1;
  name->item[FSCALAR].dimenabled[DEPTH]=1;
  name->item[FSCALAR].dimenabled[ILOC]=1;
  name->item[FSCALAR].dimenabled[JLOC]=1;
  name->item[FSCALAR].dimoptional[LEVEL]=1;
  name->item[FSCALAR].dimoptional[DEPTH]=1;
  name->item[FSCALAR].dimoptional[ILOC]=1;
  name->item[FSCALAR].dimoptional[JLOC]=1;
  name->item[FSCALAR].nvar=N_FSCALAR+name->item[FPROF].nvar;
  name->item[FSCALAR].vartext[SOL]="surface solar irradiance forcing (W/m^2)";
  name->item[FSCALAR].varname[SOL]="sol";
  name->item[FSCALAR].vartext[SOLAV]=
    "daily mean surface solar irradiance forcing (W/m^2)";
  name->item[FSCALAR].varname[SOLAV]="solav";
  name->item[FSCALAR].vartext[MLD]="mixed layer depth forcing (m)";
  name->item[FSCALAR].varname[MLD]="mld"; 
  name->item[FSCALAR].vartext[SSS]="sea surface salinity forcing";
  name->item[FSCALAR].varname[SSS]="sss";
  name->item[FSCALAR].vartext[PCO2ATM]="atmospheric pCO2 forcing (uatm)";
  name->item[FSCALAR].varname[PCO2ATM]="pco2atm";
  name->item[FSCALAR].vartext[DUST]=
    "aeolian dust deposition (mg/m^2/d)";
  name->item[FSCALAR].varname[DUST]="dust";
  name->item[FSCALAR].vartext[FEDUST]=
    "aeolian input of soluble iron (mmolFe/m^2/d)";
  name->item[FSCALAR].varname[FEDUST]="fedust";
  name->item[FSCALAR].vartext[CCHLDA]=
    "surface C:Chl forcing for data assimilation scheme (gC/gChl)";
  name->item[FSCALAR].varname[CCHLDA]="rcchlda";
  name->item[FSCALAR].vartext[CHLDA]=
    "surface chlorophyll forcing for data assimilation scheme (mg/m^3)";
  name->item[FSCALAR].varname[CHLDA]="chlda";
  name->item[FSCALAR].vartext[PGROWDA]=
"phytoplankton specific growth rate forcing for data assimilation scheme (/d)";
  name->item[FSCALAR].varname[PGROWDA]="pgrowda";
  name->item[FSCALAR].vartext[PLOSSDA]=
"phytoplankton specific loss rate forcing for data assimilation scheme (/d)";
  name->item[FSCALAR].varname[PLOSSDA]="plossda";
  for (i=0;i<name->item[FPROF].nvar;i++)
    { name->item[FSCALAR].vartext[i+N_FSCALAR]=name->item[FPROF].vartext[i];
      name->item[FSCALAR].varname[i+N_FSCALAR]=name->item[FPROF].varname[i];
    }
  for (i=0;i<name->item[FSCALAR].nvar;i++) name->item[FSCALAR].varoptional[i]=1;

  if (N_FSCALITEM>99)
    { sprintf(errline,
        "MarMOT error - N_FSCALITEM=%d, max no. of 'ft' items allowed is 99",
              N_FSCALITEM);
      return SPECIAL;
    }

  len=0;
  for (j=1;j<N_FSCALITEM;j++) len+=strlen(name->item[FSCALAR].name)+3;
  s=malloc(len);
  if (!s) return MEMERR;

  for (j=1;j<N_FSCALITEM;j++)

    { jf=FSCALAR+j;
      name->item[jf].name=s;
      sprintf(s,"%s%d",name->item[FSCALAR].name,j+1);
      s+=strlen(s)+1;
      name->item[jf].optional=1;
      name->item[jf].dimenabled[TIME]=1;
      name->item[jf].dimenabled[LEVEL]=1;
      name->item[jf].dimenabled[DEPTH]=1;
      name->item[jf].dimenabled[ILOC]=1;
      name->item[jf].dimenabled[JLOC]=1;
      name->item[jf].dimoptional[LEVEL]=1;
      name->item[jf].dimoptional[DEPTH]=1;
      name->item[jf].dimoptional[ILOC]=1;
      name->item[jf].dimoptional[JLOC]=1;
      name->item[jf].nvar=name->item[FSCALAR].nvar;
      for (i=0;i<name->item[FSCALAR].nvar;i++)
        { name->item[jf].vartext[i]=name->item[FSCALAR].vartext[i];
          name->item[jf].varname[i]=name->item[FSCALAR].varname[i];
        }
      for (i=0;i<name->item[FSCALAR].nvar;i++) name->item[jf].varoptional[i]=1;
    }


  /* MarMOT parameter set items */

  len=0;
  for (i=0;i<N_FSCALAR;i++) len+=strlen(name->item[FSCALAR].varname[i])+6;
  for (i=0;i<N_FPROF;i++) len+=strlen(name->item[FPROF].varname[i])+6;
  for (i=0;i<N_TRACER_PRI;i++) len+=2*(strlen(name->item[INIT].varname[i])+7);
  s=malloc(len);
  if (!s) return MEMERR;

  name->item[OPTION].name="option";
  name->item[OPTION].optional=1;
  name->item[OPTION].nvar=N_OPTION;
  name->item[OPTION].vartext[MODEL]="ecosystem model number";
  name->item[OPTION].varname[MODEL]="model";
  name->item[OPTION].varmin[MODEL]=0;
  name->item[OPTION].varmax[MODEL]=N_MODEL;
  name->item[OPTION].vartext[PHOTMODEL]=
    "photosynthesis light limitation model number (0-3)";
  name->item[OPTION].varname[PHOTMODEL]="photmodel";
  name->item[OPTION].varmin[PHOTMODEL]=0;
  name->item[OPTION].varmax[PHOTMODEL]=3;
  name->item[OPTION].vartext[KDMODEL]="light attenuation model number (0-3)";
  name->item[OPTION].varname[KDMODEL]="kdmodel";
  name->item[OPTION].varmin[KDMODEL]=0;
  name->item[OPTION].varmax[KDMODEL]=3;
  name->item[OPTION].vartext[KDPATHOPT]="light attenuation model option for path length adjustment (0-1)";
  name->item[OPTION].varname[KDPATHOPT]="kdpathopt";
  name->item[OPTION].varmin[KDPATHOPT]=0;
  name->item[OPTION].varmax[KDPATHOPT]=1;
  name->item[OPTION].vartext[ATTENWATER]=
    "downwelling PAR attenuation due to water (/m)";
  name->item[OPTION].varname[ATTENWATER]="attenwater";
  name->item[OPTION].varmin[ATTENWATER]=0;
  name->item[OPTION].vartext[ATTENPIG]=
    "downwelling PAR attenuation due to pigment (/m/(mg/m^3))";
  name->item[OPTION].varname[ATTENPIG]="attenpig";
  name->item[OPTION].varmin[ATTENPIG]=0;
  name->item[OPTION].vartext[ACHLMODEL]="light absorption model number (0-1)";
  name->item[OPTION].varname[ACHLMODEL]="achlmodel";
  name->item[OPTION].varmin[ACHLMODEL]=0;
  name->item[OPTION].varmax[ACHLMODEL]=1;
  name->item[OPTION].vartext[ADVECTSCHEME]="advection scheme (0-2)";
  name->item[OPTION].varname[ADVECTSCHEME]="advection";
  name->item[OPTION].varmin[ADVECTSCHEME]=0;
  name->item[OPTION].varmax[ADVECTSCHEME]=2;
  name->item[OPTION].vartext[DIFFSCHEME]=
    "diffusion scheme (0-1)";
  name->item[OPTION].varname[DIFFSCHEME]="diffusion";
  name->item[OPTION].varmin[DIFFSCHEME]=0;
  name->item[OPTION].varmax[DIFFSCHEME]=1;
  name->item[OPTION].vartext[MIXOPT]="b.l. mixing option (0-2)";
  name->item[OPTION].varname[MIXOPT]="mixopt";
  name->item[OPTION].varmin[MIXOPT]=0;
  name->item[OPTION].varmax[MIXOPT]=2;
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[OPTION].vartext[i+PERTTR]=
        "transformation type for tracer perturbation (0-2)";
      name->item[OPTION].varname[i+PERTTR]=s;
      strcpy(s,"perttr");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[i]);
      s+=strlen(s)+1;
      name->item[OPTION].varmin[i+PERTTR]=0;
      name->item[OPTION].varmax[i+PERTTR]=2;
    }
  name->item[OPTION].vartext[PERTAUTO]=
    "autocorrelation of random tracer perturbation rates after 24 h";
  name->item[OPTION].varname[PERTAUTO]="pertautocorr";
  name->item[OPTION].varmin[PERTAUTO]=-1;
  name->item[OPTION].varmax[PERTAUTO]=1;
  name->item[OPTION].vartext[PERTMIXOPT]=
    "perturbation b.l. mixing option (0-2)";
  name->item[OPTION].varname[PERTMIXOPT]="pertmixopt";
  name->item[OPTION].varmin[PERTMIXOPT]=0;
  name->item[OPTION].varmax[PERTMIXOPT]=2;
  name->item[OPTION].vartext[RLXCTRLFACT]=
    "multiple of perturbation s.d. for limiting relaxation rate";
  name->item[OPTION].varname[RLXCTRLFACT]="rlxctrlfact";
  name->item[OPTION].varmin[RLXCTRLFACT]=0;
  name->item[OPTION].vartext[RLXMASKFLAG]="relaxation mask flag (0,-1,1)";
  name->item[OPTION].varname[RLXMASKFLAG]="rlxmask";
  name->item[OPTION].varmin[RLXMASKFLAG]=-1;
  name->item[OPTION].varmax[RLXMASKFLAG]=1;
  name->item[OPTION].vartext[RLXMASKOPT]="relaxation mask depth option (1-3)";
  name->item[OPTION].varname[RLXMASKOPT]="rlxmaskopt";
  name->item[OPTION].varmin[RLXMASKOPT]=1;
  name->item[OPTION].varmax[RLXMASKOPT]=3;
  name->item[OPTION].vartext[RLXMIXOPT]=
    "relaxation reference profile b.l. mixing option (0-2)";
  name->item[OPTION].varname[RLXMIXOPT]="rlxmixopt";
  name->item[OPTION].varmin[RLXMIXOPT]=0;
  name->item[OPTION].varmax[RLXMIXOPT]=2;
  name->item[OPTION].vartext[NSTEPDAY]="number of model time steps per day";
  name->item[OPTION].varname[NSTEPDAY]="nstepday";
  name->item[OPTION].varmin[NSTEPDAY]=1;
  name->item[OPTION].varmax[NSTEPDAY]=1440;
  name->item[OPTION].vartext[NSTEPBIO]=
    "number of biology sub-steps per model time step";
  name->item[OPTION].varname[NSTEPBIO]="nstepbio";
  name->item[OPTION].varmin[NSTEPBIO]=1;
  name->item[OPTION].varmax[NSTEPBIO]=1440;
  name->item[OPTION].vartext[LEAPFROG]=
    "switch for leapfrog time-stepping (0-1)";
  name->item[OPTION].varname[LEAPFROG]="leapfrog";
  name->item[OPTION].varmin[LEAPFROG]=0;
  name->item[OPTION].varmax[LEAPFROG]=1;
  name->item[OPTION].vartext[TFPARM]=
    "Robert-Asselin time filter parameter for leapfrog scheme";
  name->item[OPTION].varname[TFPARM]="tfparm";
  name->item[OPTION].varmin[TFPARM]=0;
  name->item[OPTION].varmax[TFPARM]=0.25;
  name->item[OPTION].vartext[ASSIMSCHEME]=
    "ocean colour assimilation scheme (0-1)";
  name->item[OPTION].varname[ASSIMSCHEME]="assimilation";
  name->item[OPTION].varmin[ASSIMSCHEME]=0;
  name->item[OPTION].varmax[ASSIMSCHEME]=1;
  name->item[OPTION].vartext[NSTEPASSIM]=
    "number of model time steps per assimilation time step";
  name->item[OPTION].varname[NSTEPASSIM]="nstepassim";
  name->item[OPTION].varmin[NSTEPASSIM]=1;
  name->item[OPTION].vartext[STARTSTEPASSIM]=
    "model step number at start of first assimilation time step";
  name->item[OPTION].varname[STARTSTEPASSIM]="startstepassim";
  name->item[OPTION].varmin[STARTSTEPASSIM]=1;
  name->item[OPTION].vartext[FMODE]="default forcing interpolation mode (0-3)";
  name->item[OPTION].varname[FMODE]="fmode";
  name->item[OPTION].varmin[FMODE]=0;
  name->item[OPTION].varmax[FMODE]=3;
  for (i=0;i<N_FSCALAR;i++)
    { name->item[OPTION].vartext[i+FMODES]=
        "interpolation mode for forcing variable";
      name->item[OPTION].varname[i+FMODES]=s;
      strcpy(s,"fmode");
      s+=strlen(s);
      strcpy(s,name->item[FSCALAR].varname[i]);
      s+=strlen(s)+1;
      name->item[OPTION].varmin[i+FMODES]=0;
      name->item[OPTION].varmax[i+FMODES]=3;
    }
  for (i=0;i<N_FPROF;i++)
    { name->item[OPTION].vartext[i+FMODES+N_FSCALAR]=
        "interpolation mode for forcing variable";
      name->item[OPTION].varname[i+FMODES+N_FSCALAR]=s;
      strcpy(s,"fmode");
      s+=strlen(s);
      strcpy(s,name->item[FPROF].varname[i]);
      s+=strlen(s)+1;
      name->item[OPTION].varmin[i+FMODES+N_FSCALAR]=0;
      name->item[OPTION].varmax[i+FMODES+N_FSCALAR]=3;
    }
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[OPTION].vartext[i+RLXRATES]=
        "relaxation rate for tracer (/d)";
      name->item[OPTION].varname[i+RLXRATES]=s;
      strcpy(s,"rlx");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[i]);
      s+=strlen(s)+1;
    }
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[OPTION].vartext[i+DEFTRACERS]=name->item[INIT].vartext[i];
      name->item[OPTION].varname[i+DEFTRACERS]=name->item[INIT].varname[i];
    }
  for (i=0;i<N_RATIO;i++) 
    { name->item[OPTION].vartext[i+DEFRATIOS]=
        name->item[INIT].vartext[i+N_TRACER_PRI];
      name->item[OPTION].varname[i+DEFRATIOS]=
        name->item[INIT].varname[i+N_TRACER_PRI];
    }
  for (i=0;i<name->item[OPTION].nvar;i++) name->item[OPTION].varoptional[i]=1;

  name->item[TIMEAXIS].name="taxis";
  name->item[TIMEAXIS].optional=1;
  name->item[TIMEAXIS].nvar=3;
  name->item[TIMEAXIS].vartext[BASEYEAR]="first calendar year of time axis";
  name->item[TIMEAXIS].varname[BASEYEAR]="baseyear";
  name->item[TIMEAXIS].vartext[PERIODIC]=
"switch for periodic annual forcing with fixed year length (0: off, 1: on)";
  name->item[TIMEAXIS].varname[PERIODIC]="periodic";
  name->item[TIMEAXIS].varmin[PERIODIC]=0;
  name->item[TIMEAXIS].varmax[PERIODIC]=1;
  name->item[TIMEAXIS].vartext[YEARLEN]="year length if fixed (days)";
  name->item[TIMEAXIS].varname[YEARLEN]="yearlen";
  name->item[TIMEAXIS].varmin[YEARLEN]=0;
  for (i=0;i<name->item[TIMEAXIS].nvar;i++) 
    name->item[TIMEAXIS].varoptional[i]=1;

  name->item[TIMEPERIOD].name="timeperiod";
  name->item[TIMEPERIOD].optional=1;
  name->item[TIMEPERIOD].nvar=5;
  name->item[TIMEPERIOD].vartext[STARTYEAR]="first calendar year of model run";
  name->item[TIMEPERIOD].varname[STARTYEAR]="startyear";
  name->item[TIMEPERIOD].vartext[STARTDAY]="first day of model run in year";
  name->item[TIMEPERIOD].varname[STARTDAY]="startday";
  name->item[TIMEPERIOD].varmin[STARTDAY]=1; 
  name->item[TIMEPERIOD].vartext[FINYEAR]="last calendar year of model run";
  name->item[TIMEPERIOD].varname[FINYEAR]="finyear";
  name->item[TIMEPERIOD].vartext[FINDAY]="last day of model run in year";
  name->item[TIMEPERIOD].varname[FINDAY]="finday";
  name->item[TIMEPERIOD].varmin[FINDAY]=1; 
  name->item[TIMEPERIOD].vartext[STARTSTEP]="first step of model run in day";
  name->item[TIMEPERIOD].varname[STARTSTEP]="startstep";
  name->item[TIMEPERIOD].varmin[STARTSTEP]=1; 
  for (i=0;i<name->item[TIMEPERIOD].nvar;i++) 
    name->item[TIMEPERIOD].varoptional[i]=1;

  name->item[ENVIRON].name="environ";
  name->item[ENVIRON].optional=1;
  name->item[ENVIRON].nvar=N_ENVIRON+name->item[FSCALAR].nvar-N_AUXVAR;
  name->item[ENVIRON].vartext[MAXDEP]="max. depth of water column (m)";
  name->item[ENVIRON].varname[MAXDEP]="maxdepth";
  name->item[ENVIRON].varmin[MAXDEP]=0; 
  name->item[ENVIRON].vartext[LATITUDE]="latitude (deg N)";
  name->item[ENVIRON].varname[LATITUDE]="lat";
  name->item[ENVIRON].varmin[LATITUDE]=-90; 
  name->item[ENVIRON].varmax[LATITUDE]=90; 
  for (i=0;i<name->item[FSCALAR].nvar-N_AUXVAR;i++)
    { name->item[ENVIRON].vartext[i+N_ENVIRON]=name->item[FSCALAR].vartext[i];
      name->item[ENVIRON].varname[i+N_ENVIRON]=name->item[FSCALAR].varname[i];
    }
  for (i=0;i<name->item[ENVIRON].nvar;i++) name->item[ENVIRON].varoptional[i]=1;
 
  len=0;
  for (i=0;i<N_TRACER_PRI;i++) len+=strlen(name->item[INIT].varname[i])+5;
  s=malloc(len);
  if (!s) return MEMERR;

  name->item[PERTSEED].name="pertseed";
  name->item[PERTSEED].optional=1;
  name->item[PERTSEED].nvar=N_TRACER_PRI;
  for (i=0;i<N_TRACER_PRI;i++)
    { name->item[PERTSEED].vartext[i]=
        "seed for random perturbation of tracer (integer)";
      name->item[PERTSEED].varname[i]=s;
      strcpy(s,"seed");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[i]);
      s+=strlen(s)+1;
    }
  for (i=0;i<name->item[PERTSEED].nvar;i++) 
    name->item[PERTSEED].varoptional[i]=1;

  /* Parameter output table */

  name->outtab[OUTOPTION].name="outoption";
  name->outtab[OUTOPTION].nvar=N_OPTION;
  for (i=0;i<N_OPTION;i++) 
    { name->outtab[OUTOPTION].vartext[i]=name->item[OPTION].vartext[i];
      name->outtab[OUTOPTION].varname[i]=name->item[OPTION].varname[i];
    }

  /* End-of-time-step model output tables */

  name->outtab[OUTPROFILE].name="outkt";
  name->outtab[OUTPROFILE].nvar=N_OUTVAR_PROF;
  i=0;
  name->outtab[OUTPROFILE].vartext[i]="time from t axis origin (days)";
  name->outtab[OUTPROFILE].varname[i++]="t";
  itmid=i;
  name->outtab[OUTPROFILE].vartext[i]=
    "time from t axis origin at time step mid-point (days)";
  name->outtab[OUTPROFILE].varname[i++]="tmid";
  name->outtab[OUTPROFILE].vartext[i]="calendar year";
  name->outtab[OUTPROFILE].varname[i++]="year";
  iday=i;
  name->outtab[OUTPROFILE].vartext[i]="day of year (1-366)";
  name->outtab[OUTPROFILE].varname[i++]="day";
  istep=i;
  name->outtab[OUTPROFILE].vartext[i]="model step number within day";
  name->outtab[OUTPROFILE].varname[i++]="step";
  name->outtab[OUTPROFILE].vartext[i]=
    "time since start of calendar year (days)";
  name->outtab[OUTPROFILE].varname[i++]="tofyr";
  imth=i;
  name->outtab[OUTPROFILE].vartext[i]="month of year";
  name->outtab[OUTPROFILE].varname[i++]="mth";
  idate=i;
  name->outtab[OUTPROFILE].vartext[i]="day of month";
  name->outtab[OUTPROFILE].varname[i++]="date";
  ithours=i;
  name->outtab[OUTPROFILE].vartext[i]="time of day (h)";
  name->outtab[OUTPROFILE].varname[i++]="tofd";
  name->outtab[OUTPROFILE].vartext[i]="level number";
  name->outtab[OUTPROFILE].varname[i++]="k";
  name->outtab[OUTPROFILE].vartext[i]="depth at level mid-point (m)";
  name->outtab[OUTPROFILE].varname[i++]="z";
  name->outtab[OUTPROFILE].vartext[i]="depth at top of level (m)";
  name->outtab[OUTPROFILE].varname[i++]="ztop";
  name->outtab[OUTPROFILE].vartext[i]="depth at bottom of level (m)";
  name->outtab[OUTPROFILE].varname[i++]="zbot";
  name->outtab[OUTPROFILE].vartext[i]="level height (m)";
  name->outtab[OUTPROFILE].varname[i++]="dz";

  ioutprof_fprof=i;

  for (j=0;j<N_FPROF;j++)
    { name->outtab[OUTPROFILE].vartext[i]=name->item[FPROF].vartext[j];
      name->outtab[OUTPROFILE].varname[i++]=name->item[FPROF].varname[j];
    }

  for (j=0;j<N_TRACER_PRI;j++)
    { name->outtab[OUTPROFILE].vartext[i]=name->item[INIT].vartext[j];
      name->outtab[OUTPROFILE].varname[i++]=name->item[INIT].varname[j];
    }

  for (j=0;j<N_RATIO;j++)
    { name->outtab[OUTPROFILE].vartext[i]=
        name->item[INIT].vartext[j+N_TRACER_PRI];
      name->outtab[OUTPROFILE].varname[i++]=
        name->item[INIT].varname[j+N_TRACER_PRI];
    }

  ioutprof_dertracer=i;

  name->outtab[OUTPROFILE].vartext[i]="chlorophyll (mg/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="chl";
  name->outtab[OUTPROFILE].vartext[i]="non-diatom chlorophyll (mg/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="chlndia";
  name->outtab[OUTPROFILE].vartext[i]="diatom chlorophyll (mg/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="chldiat";
  name->outtab[OUTPROFILE].vartext[i]="phytoplankton carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="phyc";
  name->outtab[OUTPROFILE].vartext[i]=
    "non-diatom phytoplankton carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="pndiac";
  name->outtab[OUTPROFILE].vartext[i]="diatom carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="pdiatc";
  name->outtab[OUTPROFILE].vartext[i]="zooplankton carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="zooc";
  name->outtab[OUTPROFILE].vartext[i]="mesozooplankton carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="zmec";
  name->outtab[OUTPROFILE].vartext[i]="microzooplankton carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="zmic";
  name->outtab[OUTPROFILE].vartext[i]="detrital carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="detc";
  name->outtab[OUTPROFILE].vartext[i]="diatom silicon (mmolSi/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="pdiatsi";
  name->outtab[OUTPROFILE].vartext[i]="total system nitrogen (mmolN/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="totaln";
  name->outtab[OUTPROFILE].vartext[i]="total system carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="totalc";
  name->outtab[OUTPROFILE].vartext[i]="total system silicon (mmolSi/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="totalsi";
  name->outtab[OUTPROFILE].vartext[i]="total system iron (mmolFe/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="totalfe";
  name->outtab[OUTPROFILE].vartext[i]=
    "particulate organic nitrogen (mmolN/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="pon";
  name->outtab[OUTPROFILE].vartext[i]="particulate organic carbon (mmolC/m^3)";
  name->outtab[OUTPROFILE].varname[i++]="poc";

  name->outtab[OUTPROFILE].vartext[i]="fraction of level mixed";
  name->outtab[OUTPROFILE].varname[i++]="fmixed"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "level mean primary production (mmolC/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="pprod"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "iron scavenging (mmolFe/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="fescav"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "sinking particle flux of nitrogen (mmolN/m^2/d, downward, depth zbot)";
  name->outtab[OUTPROFILE].varname[i++]="sinkflxn"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "sinking particle flux of carbon (mmolC/m^2/d, downward, depth zbot)";
  name->outtab[OUTPROFILE].varname[i++]="sinkflxc"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "sinking particle flux of silicon (mmolSi/m^2/d, downward, depth zbot)";
  name->outtab[OUTPROFILE].varname[i++]="sinkflxsi"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "sinking particle flux of calcium (mmolCa/m^2/d, downward, depth zbot)";
  name->outtab[OUTPROFILE].varname[i++]="sinkflxca"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "sinking particle flux of iron (mmolFe/m^2/d, downward, depth zbot)";
  name->outtab[OUTPROFILE].varname[i++]="sinkflxfe"; 
  name->outtab[OUTPROFILE].vartext[i]=
    "diffuse attenuation coefficient for downwelling PAR (/m)";
  name->outtab[OUTPROFILE].varname[i++]="kdpar";
  name->outtab[OUTPROFILE].vartext[i]=
    "fraction of surface downwelling PAR at top of level ztop";
  name->outtab[OUTPROFILE].varname[i++]="parfracztop";
  name->outtab[OUTPROFILE].vartext[i]=
    "downwelling PAR at top of level ztop (E/m^2/d)";
  name->outtab[OUTPROFILE].varname[i++]="parztop";
  name->outtab[OUTPROFILE].vartext[i]=
    "fraction of surface downwelling PAR at level mid-point z";
  name->outtab[OUTPROFILE].varname[i++]="parfracz";
  name->outtab[OUTPROFILE].vartext[i]=
    "downwelling PAR at level mid-point z (E/m^2/d)";
  name->outtab[OUTPROFILE].varname[i++]="parz";
  name->outtab[OUTPROFILE].vartext[i]=
    "max. potential photosynthetic carbon fixation rate (/d)";
  name->outtab[OUTPROFILE].varname[i++]="photmax";
  name->outtab[OUTPROFILE].vartext[i]="DIN uptake rate limitation factor";
  name->outtab[OUTPROFILE].varname[i++]="limdin";
  name->outtab[OUTPROFILE].vartext[i]=
    "light saturated carbon fixation rate for photosynthesis-PAR curve (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pmax";
  name->outtab[OUTPROFILE].vartext[i]=
    "initial slope of photosynthesis-PAR curve for downwelling PAR (/(E/m^2))";
  name->outtab[OUTPROFILE].varname[i++]="alphad";
  name->outtab[OUTPROFILE].vartext[i]=
    "potential carbon fixation at z, fn. of 'pmax' and light limitation (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pz";
  name->outtab[OUTPROFILE].vartext[i]=
"potential level mean carbon fixation, fn. of 'pmax' and light limitation (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pk";
  name->outtab[OUTPROFILE].vartext[i]=
    "realized level mean carbon fixation rate (/d)";
  name->outtab[OUTPROFILE].varname[i++]="phot";
  name->outtab[OUTPROFILE].vartext[i]=
    "non-diatom max. potential photosynthetic carbon fixation rate (/d)";
  name->outtab[OUTPROFILE].varname[i++]="photmaxndia";
  name->outtab[OUTPROFILE].vartext[i]=    "non-diatom DIN uptake rate limitation factor";
  name->outtab[OUTPROFILE].varname[i++]="limdinndia";
  name->outtab[OUTPROFILE].vartext[i]=
    "non-diatom iron uptake rate limitation factor";  name->outtab[OUTPROFILE].varname[i++]="limfendia";
  name->outtab[OUTPROFILE].vartext[i]=
"non-diatom light saturated carbon fixation rate for photosynthesis-PAR curve (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pmaxndia";
  name->outtab[OUTPROFILE].vartext[i]=
"non-diatom initial slope of photosynthesis-PAR curve for downwelling PAR (/(E/m^2))";
  name->outtab[OUTPROFILE].varname[i++]="alphadndia";
  name->outtab[OUTPROFILE].vartext[i]=
"non-diatom potential carbon fixation at z, fn. of 'pmax' and light limitation (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pzndia";
  name->outtab[OUTPROFILE].vartext[i]=
"non-diatom potential level mean carbon fixation, fn. of 'pmax' and light limitation (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pkndia";
  name->outtab[OUTPROFILE].vartext[i]=
    "non-diatom realized level mean carbon fixation rate (/d)";
  name->outtab[OUTPROFILE].varname[i++]="photndia";
  name->outtab[OUTPROFILE].vartext[i]=
    "diatom max. potential photosynthetic carbon fixation rate (/d)";
  name->outtab[OUTPROFILE].varname[i++]="photmaxdiat";
  name->outtab[OUTPROFILE].vartext[i]=
    "diatom DIN uptake rate limitation factor";
  name->outtab[OUTPROFILE].varname[i++]="limdindiat";
  name->outtab[OUTPROFILE].vartext[i]=
    "diatom silicon uptake rate limitation factor";
  name->outtab[OUTPROFILE].varname[i++]="limsidiat";
  name->outtab[OUTPROFILE].vartext[i]=
    "diatom iron uptake rate limitation factor";
  name->outtab[OUTPROFILE].varname[i++]="limfediat";
  name->outtab[OUTPROFILE].vartext[i]=
"diatom light saturated carbon fixation rate for photosynthesis-PAR curve (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pmaxdiat";
  name->outtab[OUTPROFILE].vartext[i]=
"diatom initial slope of photosynthesis-PAR curve for downwelling PAR (/(E/m^2))";
  name->outtab[OUTPROFILE].varname[i++]="alphaddiat";
  name->outtab[OUTPROFILE].vartext[i]=
"diatom potential carbon fixation at z, fn. of 'pmax' and light limitation (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pzdiat";
  name->outtab[OUTPROFILE].vartext[i]=
"diatom potential level mean carbon fixation, fn. of 'pmax' and light limitation (/d)";
  name->outtab[OUTPROFILE].varname[i++]="pkdiat";
  name->outtab[OUTPROFILE].vartext[i]=
    "diatom realized level mean carbon fixation rate (/d)";
  name->outtab[OUTPROFILE].varname[i++]="photdiat";
  name->outtab[OUTPROFILE].vartext[i]=
    "detrital sinking rate (m/d, upward, depth zbot)";
  name->outtab[OUTPROFILE].varname[i++]="wdet";
  name->outtab[OUTPROFILE].vartext[i]=
    "nitrogen influx from fast sinking particles (mmolN/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="insinkn";  
  name->outtab[OUTPROFILE].vartext[i]=
    "carbon influx from fast sinking particles (mmolC/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="insinkc";  
  name->outtab[OUTPROFILE].vartext[i]=
    "silicon influx from fast sinking particles (mmolSi/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="insinksi";  
  name->outtab[OUTPROFILE].vartext[i]=
    "calcium influx from fast sinking particles (mmolCa/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="insinkca";  
  name->outtab[OUTPROFILE].vartext[i]=
    "iron influx from fast sinking particles (mmolFe/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="insinkfe";  
  name->outtab[OUTPROFILE].vartext[i]=
    "nitrogen change due to sinking particle flux (mmolN/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="ddtsinkn";
  name->outtab[OUTPROFILE].vartext[i]=
    "carbon change due to sinking particle flux (mmolC/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="ddtsinkc";
  name->outtab[OUTPROFILE].vartext[i]=
    "silicon change due to sinking particle flux (mmolSi/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="ddtsinksi";
  name->outtab[OUTPROFILE].vartext[i]=
    "calcium change due to sinking particle flux (mmolCa/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="ddtsinkca";
  name->outtab[OUTPROFILE].vartext[i]=
    "iron change due to sinking particle flux (mmolFe/m^3/d)";
  name->outtab[OUTPROFILE].varname[i++]="ddtsinkfe";

  len=0;
  for (j=0;j<N_TRACER_PRI;j++) len+=3*strlen(name->item[INIT].varname[j])+15;
  for (j=0;j<N_TRACER_DER;j++) 
    len+=3*strlen(name->outtab[OUTPROFILE].varname[ioutprof_dertracer+j])+15;
  s=malloc(len);
  if (!s) return MEMERR;

  for (j=0;j<N_TRACER_PRI;j++)
    { name->outtab[OUTPROFILE].vartext[i]=
        "tracer change due to implicit horizontal flow (tracer units /d)";
      name->outtab[OUTPROFILE].varname[i++]=s;
      strcpy(s,"ddth");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<N_TRACER_DER;j++)
    { name->outtab[OUTPROFILE].vartext[i]=
        "tracer change due to implicit horizontal flow (tracer units /d)";
      name->outtab[OUTPROFILE].varname[i++]=s;
      strcpy(s,"ddth");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTPROFILE].varname[ioutprof_dertracer+j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<N_TRACER_PRI;j++)
    { name->outtab[OUTPROFILE].vartext[i]=
        "tracer change due to perturbation (tracer units /d)";
      name->outtab[OUTPROFILE].varname[i++]=s;
      strcpy(s,"ddtp");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<N_TRACER_DER;j++)
    { name->outtab[OUTPROFILE].vartext[i]=
        "tracer change due to perturbation (tracer units /d)";
      name->outtab[OUTPROFILE].varname[i++]=s;
      strcpy(s,"ddtp");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTPROFILE].varname[ioutprof_dertracer+j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<N_TRACER_PRI;j++)
    { name->outtab[OUTPROFILE].vartext[i]=
        "tracer change due to relaxation (tracer units /d)";
      name->outtab[OUTPROFILE].varname[i++]=s;
      strcpy(s,"ddtr");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<N_TRACER_DER;j++)
    { name->outtab[OUTPROFILE].vartext[i]=
        "tracer change due to relaxation (tracer units /d)";
      name->outtab[OUTPROFILE].varname[i++]=s;
      strcpy(s,"ddtr");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTPROFILE].varname[ioutprof_dertracer+j]);
      s+=strlen(s)+1;
    }

  len=0;
  for (i=0;i<N_TRACER_PRI;i++) len+=strlen(name->item[INIT].varname[i])+4;
  for (i=0;i<N_TRACER_DER;i++) 
    len+=strlen(name->outtab[OUTPROFILE].varname[ioutprof_dertracer+i])+4;
  s=malloc(len);
  if (!s) return MEMERR;

  name->outtab[OUTSCALAR].name="outt";
  name->outtab[OUTSCALAR].nvar=N_OUTVAR_SCAL;

  for (i=0;i<9;i++)
    { name->outtab[OUTSCALAR].vartext[i]=name->outtab[OUTPROFILE].vartext[i];
      name->outtab[OUTSCALAR].varname[i]=name->outtab[OUTPROFILE].varname[i];
    }

  name->outtab[OUTSCALAR].vartext[i]="depth at level mid-point z (m)";
  name->outtab[OUTSCALAR].varname[i++]="z";
  name->outtab[OUTSCALAR].vartext[i]="depth at bottom of level (m)";
  name->outtab[OUTSCALAR].varname[i++]="zbot";

  ioutscal_obstype=i;

  for (j=0;j<N_FSCALAR;j++)
    { name->outtab[OUTSCALAR].vartext[i]=name->item[FSCALAR].vartext[j];
      name->outtab[OUTSCALAR].varname[i++]=name->item[FSCALAR].varname[j];
    }

  for (j=0;j<N_PROFVAR;j++)
    { name->outtab[OUTSCALAR].vartext[i]=
        name->outtab[OUTPROFILE].vartext[ioutprof_fprof+j];
      name->outtab[OUTSCALAR].varname[i++]=
        name->outtab[OUTPROFILE].varname[ioutprof_fprof+j];
    }
 
  name->outtab[OUTSCALAR].vartext[i]=
    "relaxation rate scaling factor due to flux limitation scheme";
  name->outtab[OUTSCALAR].varname[i++]="rlxfrac";
  name->outtab[OUTSCALAR].vartext[i]="day length (d)";
  name->outtab[OUTSCALAR].varname[i++]="daylen";
  name->outtab[OUTSCALAR].vartext[i]="euphotic zone depth, 1% PAR (m)";
  name->outtab[OUTSCALAR].varname[i++]="zeuphotic";
  name->outtab[OUTSCALAR].vartext[i]="sea surface pCO2 (uatm)";
  name->outtab[OUTSCALAR].varname[i++]="pco2";
  name->outtab[OUTSCALAR].vartext[i]="air-sea CO2 flux (mmolC/m^2/d, downward)";
  name->outtab[OUTSCALAR].varname[i++]="co2flux";
  name->outtab[OUTSCALAR].vartext[i]="CO2 invasion (mmolC/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="co2in";
  name->outtab[OUTSCALAR].vartext[i]="CO2 evasion (mmolC/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="co2out";
  name->outtab[OUTSCALAR].vartext[i]="iron influx at surface (mmolFe/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="feinsurf";
  name->outtab[OUTSCALAR].vartext[i]="iron influx from sea bed (mmolFe/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="feinbed";

  for (j=0;j<N_TRACER_PRI;j++)
    { name->outtab[OUTSCALAR].vartext[i]=
        "depth integrated tracer (tracer units * m)";
      name->outtab[OUTSCALAR].varname[i++]=s;
      strcpy(s,"iz.");
      s+=strlen(s);
      strcpy(s,name->item[INIT].varname[j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<N_TRACER_DER;j++)
    { name->outtab[OUTSCALAR].vartext[i]=
        "depth integrated tracer (tracer units * m)";
      name->outtab[OUTSCALAR].varname[i++]=s;
      strcpy(s,"iz.");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTPROFILE].varname[ioutprof_dertracer+j]);
      s+=strlen(s)+1;
    }

  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated primary production (mmolC/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.pprod";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated iron scavenging (mmolFe/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.fescav";
  name->outtab[OUTSCALAR].vartext[i]=
"depth integrated nitrogen change due to implicit horizontal flow (mmolN/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddthn";
  name->outtab[OUTSCALAR].vartext[i]=
"depth integrated carbon change due to implicit horizontal flow (mmolC/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddthc";
  name->outtab[OUTSCALAR].vartext[i]=
"depth integrated silicon change due to implicit horizontal flow (mmolSi/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddthsi";
  name->outtab[OUTSCALAR].vartext[i]=
"depth integrated iron change due to implicit horizontal flow (mmolFe/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddthfe";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated nitrogen change due to tracer perturbation (mmolN/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtpn";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated carbon change due to tracer perturbation (mmolC/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtpc";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated silicon change due to tracer perturbation (mmolSi/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtpsi";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated iron change due to tracer perturbation (mmolFe/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtpfe";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated nitrogen change due to tracer relaxation (mmolN/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtrn";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated carbon change due to tracer relaxation (mmolC/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtrc";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated silicon change due to tracer relaxation (mmolSi/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtrsi";
  name->outtab[OUTSCALAR].vartext[i]=
    "depth integrated iron change due to tracer relaxation (mmolFe/m^2/d)";
  name->outtab[OUTSCALAR].varname[i++]="iz.ddtrfe";

  /* Daily averaged model output tables 
     (Don't copy 'tmid' or 'step' from end-of-time-step model output) */

  name->outtab[OUTPROFILE_DAV].name="outktday";
  name->outtab[OUTPROFILE_DAV].nvar=N_OUTVAR_PROF-2;
  i=0;
  for (j=0;j<N_OUTVAR_PROF;j++) if (j!=itmid && j!=istep)
    { name->outtab[OUTPROFILE_DAV].vartext[i]=
        name->outtab[OUTPROFILE].vartext[j];
      name->outtab[OUTPROFILE_DAV].varname[i++]=
        name->outtab[OUTPROFILE].varname[j];
    }

  name->outtab[OUTSCALAR_DAV].name="outtday";
  name->outtab[OUTSCALAR_DAV].nvar=N_OUTVAR_SCAL-2;
  i=0;
  for (j=0;j<N_OUTVAR_SCAL;j++) if (j!=itmid && j!=istep)
    { name->outtab[OUTSCALAR_DAV].vartext[i]=
        name->outtab[OUTSCALAR].vartext[j];
      name->outtab[OUTSCALAR_DAV].varname[i++]=
        name->outtab[OUTSCALAR].varname[j];
    }

  /* Monthly averaged model output tables
     (Don't copy 'tmid','step','day','date','tofd') */

  name->outtab[OUTPROFILE_MAV].name="outktmth";
  name->outtab[OUTPROFILE_MAV].nvar=N_OUTVAR_PROF-5;
  i=0;
  for (j=0;j<N_OUTVAR_PROF;j++) 
    if (j!=itmid && j!=istep && j!=iday && j!=idate && j!=ithours)
      { name->outtab[OUTPROFILE_MAV].vartext[i]=
          name->outtab[OUTPROFILE].vartext[j];
        name->outtab[OUTPROFILE_MAV].varname[i++]=
          name->outtab[OUTPROFILE].varname[j];
      }

  name->outtab[OUTSCALAR_MAV].name="outtmth";
  name->outtab[OUTSCALAR_MAV].nvar=N_OUTVAR_SCAL-5;
  i=0;
  for (j=0;j<N_OUTVAR_SCAL;j++) 
    if (j!=itmid && j!=istep && j!=iday && j!=idate && j!=ithours)
      { name->outtab[OUTSCALAR_MAV].vartext[i]=
          name->outtab[OUTSCALAR].vartext[j];
        name->outtab[OUTSCALAR_MAV].varname[i++]=
          name->outtab[OUTSCALAR].varname[j];
      }

  /* Yearly averaged model output tables
     (Don't copy 'tmid','step','day','date','tofd','mth') */

  name->outtab[OUTPROFILE_YAV].name="outktyr";
  name->outtab[OUTPROFILE_YAV].nvar=N_OUTVAR_PROF-6;
  i=0;
  for (j=0;j<N_OUTVAR_PROF;j++) 
    if (j!=itmid && j!=istep && j!=iday && j!=idate && j!=ithours && j!=imth)
      { name->outtab[OUTPROFILE_YAV].vartext[i]=
          name->outtab[OUTPROFILE].vartext[j];
        name->outtab[OUTPROFILE_YAV].varname[i++]=
          name->outtab[OUTPROFILE].varname[j];
      }
  
  name->outtab[OUTSCALAR_YAV].name="outtyr";
  name->outtab[OUTSCALAR_YAV].nvar=N_OUTVAR_SCAL-6; 
  i=0;
  for (j=0;j<N_OUTVAR_SCAL;j++) 
    if (j!=itmid && j!=istep && j!=iday && j!=idate && j!=ithours && j!=imth)
      { name->outtab[OUTSCALAR_YAV].vartext[i]=
          name->outtab[OUTSCALAR].vartext[j];
        name->outtab[OUTSCALAR_YAV].varname[i++]=
          name->outtab[OUTSCALAR].varname[j];
      }
    
  /* Assimilation scheme output tables
     (Leading variables match those in end-of-time-step model output
      omitting 'tmid') */ 

  name->outtab[OUTASSIMPROF].name="outktassim";
  name->outtab[OUTASSIMPROF].nvar=N_OUTVAR_ASSIMPROF;
  i=0;
  for (j=0;j<ioutprof_fprof;j++) if (j!=itmid)
    { name->outtab[OUTASSIMPROF].vartext[i]=
        name->outtab[OUTPROFILE].vartext[j];
      name->outtab[OUTASSIMPROF].varname[i++]=
        name->outtab[OUTPROFILE].varname[j];
    }

  ioutassimprof_state=i;

  name->outtab[OUTASSIMPROF].vartext[i]="background DIN (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="bdin";
  name->outtab[OUTASSIMPROF].vartext[i]="background phytoplankton (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="bphy";
  name->outtab[OUTASSIMPROF].vartext[i]="background zooplankton (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="bzoo";
  name->outtab[OUTASSIMPROF].vartext[i]="background detritus (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="bdet";
  name->outtab[OUTASSIMPROF].vartext[i]="background DIC (mmolC/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="bdic";
  name->outtab[OUTASSIMPROF].vartext[i]="background alkalinity (meq/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="balk";
  name->outtab[OUTASSIMPROF].vartext[i]="analysis DIN (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="adin";
  name->outtab[OUTASSIMPROF].vartext[i]="analysis phytoplankton (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="aphy";
  name->outtab[OUTASSIMPROF].vartext[i]="analysis zooplankton (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="azoo";
  name->outtab[OUTASSIMPROF].vartext[i]="analysis detritus (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="adet";
  name->outtab[OUTASSIMPROF].vartext[i]="analysis DIC (mmolC/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="adic";
  name->outtab[OUTASSIMPROF].vartext[i]="analysis alkalinity (meq/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="aalk";
  name->outtab[OUTASSIMPROF].vartext[i]="DIN increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incdin";
  name->outtab[OUTASSIMPROF].vartext[i]="phytoplankton increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incphy";
  name->outtab[OUTASSIMPROF].vartext[i]="zooplankton increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="inczoo";
  name->outtab[OUTASSIMPROF].vartext[i]="detritus increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incdet";
  name->outtab[OUTASSIMPROF].vartext[i]="DIC increment (mmolC/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incdic";
  name->outtab[OUTASSIMPROF].vartext[i]="alkalinity increment (meq/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incalk";
  name->outtab[OUTASSIMPROF].vartext[i]="background nitrogen (mmol/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="bnitrogen";
  name->outtab[OUTASSIMPROF].vartext[i]="analysis nitrogen (mmol/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="anitrogen";
  name->outtab[OUTASSIMPROF].vartext[i]="total nitrogen increment (mmol/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incnitrogen";
  name->outtab[OUTASSIMPROF].vartext[i]="primary DIN increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incdin1";
  name->outtab[OUTASSIMPROF].vartext[i]=
    "primary phytoplankton increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incphy1";
  name->outtab[OUTASSIMPROF].vartext[i]=
    "primary zooplankton increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="inczoo1";
  name->outtab[OUTASSIMPROF].vartext[i]=
    "primary detritus increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incdet1";
  name->outtab[OUTASSIMPROF].vartext[i]="secondary DIN increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incdin2";
  name->outtab[OUTASSIMPROF].vartext[i]=
    "secondary phytoplankton increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incphy2";
  name->outtab[OUTASSIMPROF].vartext[i]=
    "secondary zooplankton increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="inczoo2";
  name->outtab[OUTASSIMPROF].vartext[i]=
    "secondary detritus increment (mmolN/m^3)";
  name->outtab[OUTASSIMPROF].varname[i++]="incdet2";
  name->outtab[OUTASSIMPROF].vartext[i]=
    "unadjusted phytoplankton balancing factor";
  name->outtab[OUTASSIMPROF].varname[i++]="ubalphy";

  name->outtab[OUTASSIMSCAL].name="outtassim";
  name->outtab[OUTASSIMSCAL].nvar=N_OUTVAR_ASSIMSCAL;
  i=0;
  for (j=0;j<ioutscal_obstype;j++) if (j!=itmid)
    { name->outtab[OUTASSIMSCAL].vartext[i]=
        name->outtab[OUTSCALAR].vartext[j];
      name->outtab[OUTASSIMSCAL].varname[i++]=
        name->outtab[OUTSCALAR].varname[j];
    }
  for (j=0;j<N_ASSIMPROFTOPVAR;j++) 
    { name->outtab[OUTASSIMSCAL].vartext[i]=
        name->outtab[OUTASSIMPROF].vartext[ioutassimprof_state+j];
      name->outtab[OUTASSIMSCAL].varname[i++]=
        name->outtab[OUTASSIMPROF].varname[ioutassimprof_state+j];
    }
  name->outtab[OUTASSIMSCAL].vartext[i]="observed C:Chl ratio (gC/gChl)";
  name->outtab[OUTASSIMSCAL].varname[i++]="obsrcchl";
  name->outtab[OUTASSIMSCAL].vartext[i]="background C:Chl ratio (gC/gChl)";
  name->outtab[OUTASSIMSCAL].varname[i++]="brcchl";
  name->outtab[OUTASSIMSCAL].vartext[i]="analysis C:Chl ratio (gC/gChl)";
  name->outtab[OUTASSIMSCAL].varname[i++]="arcchl";
  name->outtab[OUTASSIMSCAL].vartext[i]="C:Chl increment (gC/gChl)";
  name->outtab[OUTASSIMSCAL].varname[i++]="incrcchl";
  name->outtab[OUTASSIMSCAL].vartext[i]="observed chlorophyll (mg/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="obschl";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "background chlorophyll, after any C:Chl update (mg/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="bchl";
  name->outtab[OUTASSIMSCAL].vartext[i]="analysis chlorophyll (mg/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="achl";
  name->outtab[OUTASSIMSCAL].vartext[i]="chlorophyll increment (mg/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="incchl";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "mixed layer phytoplankton for assimilation time step (mmolN/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="phy0";
  name->outtab[OUTASSIMSCAL].vartext[i]=
"mixed layer phytoplankton-specfic growth rate for assimilation time step (/d)";
  name->outtab[OUTASSIMSCAL].varname[i++]="pgrow0";
  name->outtab[OUTASSIMSCAL].vartext[i]=
"mixed layer phytoplankton-specific loss rate for assimilation time step (/d)";
  name->outtab[OUTASSIMSCAL].varname[i++]="ploss0";
  name->outtab[OUTASSIMSCAL].vartext[i]="mixed layer depth (m)";
  name->outtab[OUTASSIMSCAL].varname[i++]="mld";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "mixed layer depth maximum for assimilation time step (m)";
  name->outtab[OUTASSIMSCAL].varname[i++]="mldmax";
  name->outtab[OUTASSIMSCAL].vartext[i]="unadjusted din balancing factor";
  name->outtab[OUTASSIMSCAL].varname[i++]="ubaldin";
  name->outtab[OUTASSIMSCAL].vartext[i]="top level with low DIN constraint";
  name->outtab[OUTASSIMSCAL].varname[i++]="kdinlo";
  name->outtab[OUTASSIMSCAL].vartext[i]="top level with high DIN constraint";
  name->outtab[OUTASSIMSCAL].varname[i++]="kdinhi";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "top level with low phytoplankton constraint";
  name->outtab[OUTASSIMSCAL].varname[i++]="kphylo";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "top level with low zooplankton constraint";
  name->outtab[OUTASSIMSCAL].varname[i++]="kzoolo";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "top level with high zooplankton constraint";
  name->outtab[OUTASSIMSCAL].varname[i++]="kzoohi";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "top level with low detritus constraint";
  name->outtab[OUTASSIMSCAL].varname[i++]="kdetlo";
  name->outtab[OUTASSIMSCAL].vartext[i]="deepest level for surface increments";
  name->outtab[OUTASSIMSCAL].varname[i++]="kmld";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "deepest level partly or fully in mixed layer during assimilation time step";
  name->outtab[OUTASSIMSCAL].varname[i++]="kmldmax";
  name->outtab[OUTASSIMSCAL].vartext[i]="deepest level of analysis updates";
  name->outtab[OUTASSIMSCAL].varname[i++]="kmax";
  name->outtab[OUTASSIMSCAL].vartext[i]=
"lower bound for phytoplankton error component p.d.f. integration (mmolN/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="xmin";
  name->outtab[OUTASSIMSCAL].vartext[i]=
"upper bound for phytoplankton error component p.d.f. integration (mmolN/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="xmax";
  name->outtab[OUTASSIMSCAL].vartext[i]=
"p.d.f. integral on line where total error = - phy. increment (/(mmolN/m^3))";
  name->outtab[OUTASSIMSCAL].varname[i++]="pinteg";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "conditional expected phytoplankton error due to growth errors (mmolN/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="exx";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "conditional expected phytoplankton error due to loss errors (mmolN/m^3)";
  name->outtab[OUTASSIMSCAL].varname[i++]="exy";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "conditional expected growth rate error fraction";
  name->outtab[OUTASSIMSCAL].varname[i++]="exug";
  name->outtab[OUTASSIMSCAL].vartext[i]=
    "conditional expected phytoplankton-specific turnover rate (/d)";
  name->outtab[OUTASSIMSCAL].varname[i++]="extovr";

  /* Allocate space for derived obs-related variables */

  len=0;
  for (j=0;j<N_OBSTYPE;j++)
    len+=5*(strlen(name->outtab[OUTSCALAR].varname[ioutscal_obstype+j])+4);
  s=malloc(len+1);
  if (!s) return MEMERR; 

  /* Non-gridded domain item 'obset' */
    
  name->item[OBS].name="obset";
  name->item[OBS].optional=1;
  name->item[OBS].dimenabled[TIME]=1;
  name->item[OBS].dimenabled[DEPTH]=1;
  name->item[OBS].dimoptional[DEPTH]=1;
  name->item[OBS].nvar=2*(N_OBSTYPE);

  i=0;

  for (j=0;j<N_OBSTYPE;j++)
    { name->item[OBS].vartext[i]=
        name->outtab[OUTSCALAR].vartext[ioutscal_obstype+j];
      name->item[OBS].varname[i]=
        name->outtab[OUTSCALAR].varname[ioutscal_obstype+j];
      name->item[OBS].varoptional[i++]=1;
    }

  iobs_weights=i;

  for (j=0;j<N_OBSTYPE;j++)
    { name->item[OBS].vartext[i]="misfit weighting factor"; 
      name->item[OBS].varname[i]=s; 
      strcpy(s,"w.");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTSCALAR].varname[ioutscal_obstype+j]);
      s+=strlen(s)+1;
      name->item[OBS].varoptional[i++]=1;
    }

  /* Parameter set item 'mfitctrl' */

  name->item[MFITCTRL].name="mfitctrl";
  name->item[MFITCTRL].optional=1;
  name->item[MFITCTRL].nvar=N_OBSTYPE;

  i=0;
  for (j=0;j<N_OBSTYPE;j++)
    { name->item[MFITCTRL].vartext[i]=
        "transformation type for model-data differencing (0-2)";
      name->item[MFITCTRL].varname[i]=s;
      strcpy(s,"tr.");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTSCALAR].varname[ioutscal_obstype+j]);
      s+=strlen(s)+1;
      name->item[MFITCTRL].varmin[i]=0.;
      name->item[MFITCTRL].varmax[i]=2.;
      name->item[MFITCTRL].varoptional[i++]=1;
    }

  /* Misfit output table */

  name->outtab[OUTMISFIT].name="outmfit";
  name->outtab[OUTMISFIT].nvar=N_OUTVAR_MFIT;

  i=0;
  name->outtab[OUTMISFIT].vartext[i]="observation number";
  name->outtab[OUTMISFIT].varname[i++]="num";
  name->outtab[OUTMISFIT].vartext[i]=
    "observation time relative to origin (days)";
  name->outtab[OUTMISFIT].varname[i++]="ob.t";
  name->outtab[OUTMISFIT].vartext[i]="observation calendar year";
  name->outtab[OUTMISFIT].varname[i++]="ob.year";
  name->outtab[OUTMISFIT].vartext[i]="observation day of year (1-366)";
  name->outtab[OUTMISFIT].varname[i++]="ob.day";
  name->outtab[OUTMISFIT].vartext[i]=
    "observation time relative to start of calendar year (days)";
  name->outtab[OUTMISFIT].varname[i++]="ob.tofyr";
  name->outtab[OUTMISFIT].vartext[i]="depth (m)";
  name->outtab[OUTMISFIT].varname[i++]="z";
  name->outtab[OUTMISFIT].vartext[i]=
    "time difference from observation time (days)";
  name->outtab[OUTMISFIT].varname[i++]="d.t";

  for (j=0;j<N_OBSTYPE;j++)
    { name->outtab[OUTMISFIT].vartext[i]=
        "model-data difference (transformed variable units)"; 
      name->outtab[OUTMISFIT].varname[i++]=s; 
      strcpy(s,"d.");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTSCALAR].varname[ioutscal_obstype+j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<N_OBSTYPE;j++)
    { name->outtab[OUTMISFIT].vartext[i]="weighted misfit"; 
      name->outtab[OUTMISFIT].varname[i++]=s; 
      strcpy(s,"mf.");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTSCALAR].varname[ioutscal_obstype+j]);
      s+=strlen(s)+1;
    }

  for (j=0;j<9;j++)
    { name->outtab[OUTMISFIT].vartext[i]=name->outtab[OUTPROFILE].vartext[j];
      name->outtab[OUTMISFIT].varname[i++]=name->outtab[OUTPROFILE].varname[j];
    }

  name->outtab[OUTMISFIT].vartext[i]=
    "level number for variables defined at level mid-point";
  name->outtab[OUTMISFIT].varname[i++]="k";
  name->outtab[OUTMISFIT].vartext[i]=
    "level number for variables defined at level top";
  name->outtab[OUTMISFIT].varname[i++]="kztop";
  name->outtab[OUTMISFIT].vartext[i]=
    "level number for variables defined at level bottom";
  name->outtab[OUTMISFIT].varname[i++]="kzbot";

  for (j=0;j<N_OBSTYPE;j++)
    { name->outtab[OUTMISFIT].vartext[i]=
        name->outtab[OUTSCALAR].vartext[ioutscal_obstype+j];
      name->outtab[OUTMISFIT].varname[i++]=
        name->outtab[OUTSCALAR].varname[ioutscal_obstype+j];
    }
 
  for (j=0;j<N_OBSTYPE;j++)
    { name->outtab[OUTMISFIT].vartext[i]="observation value";
      name->outtab[OUTMISFIT].varname[i++]=s;
      strcpy(s,"ob.");
      s+=strlen(s);
      strcpy(s,name->outtab[OUTSCALAR].varname[ioutscal_obstype+j]);
      s+=strlen(s)+1;
    } 

  for (j=0;j<N_OBSTYPE;j++)
    { name->outtab[OUTMISFIT].vartext[i]=
        name->item[OBS].vartext[iobs_weights+j]; 
      name->outtab[OUTMISFIT].varname[i++]=
        name->item[OBS].varname[iobs_weights+j]; 
    }

  /* Cost output table */

  name->outtab[OUTCOST].name="outcost";
  name->outtab[OUTCOST].nvar=2; 

  i=0;
  name->outtab[OUTCOST].vartext[i]="misfit cost: mean weighted misfit";
  name->outtab[OUTCOST].varname[i++]="cost";
  name->outtab[OUTCOST].vartext[i]="number of observations";
  name->outtab[OUTCOST].varname[i++]="nobs";

  return OK;
}


int objfun_inputcheck (struct casetabrec *in, char **errtxt)

{ struct caseinputrec *tabentry;
  int model;           
  enum photmodeltype photmodel;
  int advection;
  int kdpathopt;
  int nstepday,nstepbio;
  int baseyear,periodicflag,fixedyearlen;
  OBJF_FLOAT_TYPE maxdepth,latitude;
  int startyear,startday,finyear,finday,startstep;
  OBJF_FLOAT_TYPE *zlevbot;
  int nk,nkwet,nkinit,nkfkt[N_FPROFITEM];
  OBJF_FLOAT_TYPE inittime,inittdays;
  int inityear;
  int tstart,tfin;
  int found[N_FSCALAR+N_FPROF];
  OBJF_FLOAT_TYPE *xvector;
  double modelparm[NP_MAX];
  double *pvector;
  int photflag,latflag,advectflag;
  int i,j,jf,k,n,m,y;
  static char errline[200],*s;

  *errtxt=errline;

  for (i=0;i<in->ndata;i++)

    { tabentry=&in->table[i];

      model=NULLMOD;
      photmodel=NULLPHOTMOD;
      kdpathopt=0;
      advection=0;
      nstepday=24;
      nstepbio=1;
      if (tabentry->item[OPTION] &&
          !validitem_option(i+1,tabentry->item[OPTION],&model,&photmodel,
                            &kdpathopt,&advection,&nstepday,&nstepbio,errline)) 

        return SPECIAL; 

      baseyear=1;
      periodicflag=1;
      fixedyearlen=365;     
      if (tabentry->item[TIMEAXIS] && 
          !validitem_taxis(i+1,tabentry->item[TIMEAXIS],&baseyear,
                           &periodicflag,&fixedyearlen,errline)) return SPECIAL;

      if (tabentry->item[TIMEPERIOD])
        { startyear=baseyear;
          startday=1;
          finyear=startyear;
          finday=daysinyear(fixedyearlen,finyear);
          startstep=1;
          if (tabentry->item[TIMEPERIOD] &&
              !validitem_timeperiod(i+1,tabentry->item[TIMEPERIOD],fixedyearlen,
                                    &startyear,&startday,&finyear,&finday,
                                    &startstep,errline)) return SPECIAL;
        }

      maxdepth=MAXDEPTH_DEF;
      latitude=OBJF_NODATA;
      if (tabentry->item[ENVIRON])
        validitem_environ(tabentry->item[ENVIRON],&maxdepth,&latitude);

      nk=1;
      if (tabentry->item[ZLEVEL] &&
          !validitem_zlevel(i+1,tabentry->item[ZLEVEL],&nk,&zlevbot,errline))
            return SPECIAL;

      if (tabentry->item[INIT] && 
          !validitem_init(i+1,tabentry->item[INIT],&nkinit,&inittime,
                          &inityear,&inittdays,errline)) return SPECIAL;

      for (j=0;j<N_FSCALITEM;j++)
        if (tabentry->item[FSCALAR+j] &&
            !validitem_ft(i+1,tabentry->item[FSCALAR+j],
                          in->objfname->item[FSCALAR+j].name,errline)) 
          return SPECIAL;

      for (j=0;j<N_FPROFITEM;j++)
        if (tabentry->item[FPROF+j] && 
            !validitem_fkt(i+1,tabentry->item[FPROF+j],
                           in->objfname->item[FPROF+j].name,&nkfkt[j],errline)) 
          return SPECIAL;

      if (tabentry->item[OBS] && 
          !validitem_obset(i+1,tabentry->item[OBS],errline)) return SPECIAL;

      /* Check 'option' against optimizable parameter set */

      if (in->optpsetnum > 0 && in->optpsetnum <= N_MODEL && 
          model != in->optpsetnum)
        { sprintf(errline,
"model=%d in 'option' item (or default) is inconsistent with optimizable parameter set no. (case %d)",
                  model,i+1);
          return SPECIAL;
        }

      /* Check 'timeperiod' against 'option' */

      if (tabentry->item[TIMEPERIOD] && (startstep>nstepday || startstep<1))
        { sprintf(errline,
      "startstep=%d out of range in 'timeperiod' item (or default) (case %d)",
                  startstep,i+1);
          return SPECIAL;
        }

      /* Check 'timeperiod' against 'taxis' */

      if (tabentry->item[TIMEPERIOD])

        { if (startyear<baseyear)
            { sprintf(errline,
"startyear=%d in 'timeperiod' item (or default) is before baseyear=%d (case %d)",
                      startyear,baseyear,i+1);
              return SPECIAL; 
            }     

          if (startday>daysinyear(fixedyearlen,startyear))
            { sprintf(errline,
                "startday=%d in 'timeperiod' item is after year end (case %d)",
                      startday,i+1);
              return SPECIAL; 
            }
     
          if (finday>daysinyear(fixedyearlen,finyear))
            { sprintf(errline,
                "finday=%d in 'timeperiod' item is after year end (case %d)",
                      finday,i+1);
              return SPECIAL; 
            }

          tstart=startday-1;
          for (y=baseyear;y<startyear;y++) tstart+=daysinyear(fixedyearlen,y);
          tfin=finday;
          for (y=baseyear;y<finyear;y++) tfin+=daysinyear(fixedyearlen,y);
        }      

      /* Check 'zlevel' against 'environ' */

      if (tabentry->item[ZLEVEL] && maxdepth<0.5*zlevbot[0]) 
        { sprintf(errline,
"maxdepth=%g in 'environ' item (or default) is too shallow for grid (case %d)",
                  maxdepth,i+1);
          return SPECIAL;
        }

      nkwet=1;
      for (k=1;k<nk;k++) if (0.5*(zlevbot[k-1]+zlevbot[k])<=maxdepth) nkwet++;

      if (tabentry->item[INIT])

        { /* Check 'init' against 'taxis' */

          if (inittdays!=OBJF_NODATA && 
              inittdays>daysinyear(fixedyearlen,inityear))
            { sprintf(errline,
                      "inittdays=%g in item 'init' is after year end (case %d)",
                      inittdays,i+1);
              return SPECIAL;
            }

          /* Check 'init' against 'environ' */

          if (nkinit<nkwet)
            { sprintf(errline,
          "missing grid points above bottom boundary in item 'init' (case %d)",
                      i+1);
              return SPECIAL;
            }

          /* Check 'init' against 'timeperiod' 
              - if specified, time must match start day or day before */
       
          if (tabentry->item[TIMEPERIOD])
 
            { if ((inittime!=OBJF_NODATA && 
                    (inittime<tstart-1 || inittime>tstart+1)) ||
                  (inittdays!=OBJF_NODATA &&
                    (inityear<startyear-1 || inityear>startyear ||
                      (inityear==startyear-1 && 
                        (startday>1 || 
                         inittdays<daysinyear(fixedyearlen,inityear)-1)) ||
                      (inityear==startyear && 
                        (inittdays<startday-2 || inittdays>startday)))))
                { sprintf(errline,
                          "item 'init' has incorrect time (case %d)",i+1);
                  return SPECIAL;
                }
            }
        }

      /* Check forcing data against 'environ' */

      for (j=0;j<N_FPROFITEM;j++)
        if (tabentry->item[FPROF+j] && nkfkt[j]<nkwet)
          { sprintf(errline,
            "missing grid points above bottom boundary in item '%s' (case %d)",
                    in->objfname->item[FPROF+j].name,i+1);
            return SPECIAL;
          }

      /* Check forcing data against 'taxis' (if periodic) or 'timeperiod' */ 

      if (tabentry->item[TIMEPERIOD])

        { for (j=0;j<N_FSCALITEM;j++)
            if (tabentry->item[FSCALAR+j] &&
                !validftime(i+1,periodicflag,fixedyearlen,tstart,tfin,
                            tabentry->item[FSCALAR+j],
                            in->objfname->item[FSCALAR+j].name,errline))
              return SPECIAL;

          for (j=0;j<N_FPROFITEM;j++)
            if (tabentry->item[FPROF+j] &&
                !validftime(i+1,periodicflag,fixedyearlen,tstart,tfin,
                            tabentry->item[FPROF+j],
                            in->objfname->item[FPROF+j].name,errline))
              return SPECIAL;
        }

      /* Check for duplicate forcing variables */

      for (m=0;m<N_FSCALAR+N_FPROF;m++) found[m]=0;

      for (j=0;j<N_FSCALITEM;j++) if (tabentry->item[FSCALAR+j])
        for (m=0;m<N_FSCALAR+N_FPROF;m++)
          { xvector=tabentry->item[FSCALAR+j]->value[m];
            if (xvector)
              { if (found[m])
                  { if (xvector[0]!=OBJF_NODATA)
                      { sprintf(errline,
              "duplicate forcing variable '%s' found in '%s' item (case %d)",
                                in->objfname->item[FSCALAR+j].varname[m],
                                in->objfname->item[FSCALAR+j].name,i+1);
                          return SPECIAL;
                      }
                  }
                else found[m]=(xvector[0]!=OBJF_NODATA);
              }
          }

      for (m=0;m<N_FPROF;m++) found[m]=0;

      for (j=0;j<N_FPROFITEM;j++) if (tabentry->item[FPROF+j])
        for (m=0;m<N_FPROF;m++) 
          { xvector=tabentry->item[FPROF+j]->value[m];
            if (xvector)
              { if (found[m])
                  { if (xvector[0]!=OBJF_NODATA)
                      { sprintf(errline,
              "duplicate forcing variable '%s' found in '%s' item (case %d)",
                                in->objfname->item[FPROF+j].varname[m],
                                in->objfname->item[FPROF+j].name,i+1);
                          return SPECIAL;
                      }
                  }
                else found[m]=(xvector[0]!=OBJF_NODATA);
              }
          }

      /* Check 'environ' against 'option' */

      if ((photmodel==EP85 || photmodel==PLATT90 || kdpathopt>0) 
          && latitude==OBJF_NODATA)
        { sprintf(errline,
"photmodel=%d and/or kdpathopt=%d in 'option' item needs 'lat' in 'environ' item (case %d)",
                  photmodel,kdpathopt,i+1);
            return SPECIAL;
        }

      /* Check model requirements */

      if (model!=NULLMOD) 

        { n=0;
          pvector=NULL;
          if (tabentry->item[model-1])
            { n=in->objfname->item[model-1].nvar;
              xvector=tabentry->item[model-1]->parm;
              for (j=0;j<n;j++) modelparm[j]=xvector[j];
              pvector=modelparm;
            }
          errline[0]='\0';
          FORTRAN_modelgetinfo(&model,&n,pvector,
                               &photflag,&latflag,&advectflag,errline);

          if (*errline)
            { j=MODTEXT_MAX;
              errline[j--]='\0';
              while (j>0 && errline[j]==' ') errline[j--]='\0';
              sprintf(&errline[j+1]," (model %d, case %d)",model,i+1);
              return SPECIAL;
            }

          if (photflag>0 && photmodel==0) 
            { sprintf(errline,
          "model %d is needing non-zero 'photmodel' in 'option' item (case %d)",
                      model,i+1);
              return SPECIAL;
            }

          if (photflag<0 && photmodel!=0)
            { sprintf(errline,
                "model %d is needing photmodel=0 in 'option' item (case %d)",
                      model,i+1);
              return SPECIAL;
            }

          if (photflag>1 && photmodel!=SMITH)
            { sprintf(errline,
                "model %d is needing photmodel=%d in 'option' item (case %d)",
                      model,SMITH,i+1);
              return SPECIAL;
            }

          if (latflag>0 && latitude==OBJF_NODATA) 
            { sprintf(errline,
                "model %d is needing 'lat' in 'environ' item (case %d)",
                      model,i+1);
              return SPECIAL;
            }

          if (advectflag==1 && advection==0)
            { sprintf(errline,
                "model %d is needing 'advection' in 'option' item (case %d)",
                      model,i+1);
              return SPECIAL;
            }
        }
    }

  return OK;
}


static int validitem_option (int casenum, struct itemrec *option, int *model, 
                             enum photmodeltype *photmodel, int *kdpathopt, 
                             int *advection, int *nstepday, 
                             int *nstepbio, char *errline)

{ int kdmodel,leapfrogflag,mixopt;
  OBJF_FLOAT_TYPE x;
  int i;

  x=option->parm[MODEL];
  if (x!=OBJF_NODATA) *model=floor(x+0.5);

  if (*model!=NULLMOD)

    { x=option->parm[PHOTMODEL];
      if (x!=OBJF_NODATA) *photmodel=(enum photmodeltype)floor(x+0.5);

      if (*photmodel!=NULLPHOTMOD)

        { x=option->parm[KDMODEL];
          if (x!=OBJF_NODATA) kdmodel=floor(x+0.5);
          else kdmodel=NULLKDMOD;

          x=option->parm[KDPATHOPT];
          if (x!=OBJF_NODATA) *kdpathopt=floor(x+0.5);

          if (kdmodel==NULLKDMOD || kdmodel==LINEARPIG)
            { x=option->parm[ATTENWATER];
              if (x==OBJF_NODATA)
                { sprintf(errline,
                    "kdmodel=%d needs 'attenwater' in 'option' item (case %d)",
                          kdmodel,casenum);
                  return 0;
                }
            }

          if (kdmodel==LINEARPIG)
            { x=option->parm[ATTENPIG]; 
              if (x==OBJF_NODATA)
                { sprintf(errline,
                    "kdmodel=%d needs 'attenpig' in 'option' item (case %d)",
                          kdmodel,casenum);
                  return 0;
                }
            }
        }
    }

  x=option->parm[ADVECTSCHEME];
  if (x!=OBJF_NODATA) *advection=floor(x+0.5);

  x=option->parm[MIXOPT];
  if (x!=OBJF_NODATA) mixopt=floor(x+0.5);
  else mixopt=0;

  x=option->parm[PERTMIXOPT];
  if (x!=OBJF_NODATA)
    { i=floor(x+0.5);
      if (i>mixopt)
        { sprintf(errline,
      "pertmixopt=%d incompatible with mixopt=%d in 'option' item (case %d)",
                  i,mixopt,casenum);
          return 0;
        }
    }

  x=option->parm[RLXMASKFLAG];
  if (x!=OBJF_NODATA)
    { i=floor(x+0.5);
      if (i!=0)
        { x=option->parm[RLXMASKOPT];
          if (x==OBJF_NODATA)
            { sprintf(errline,"missing 'rlxmaskopt' in 'option' item (case %d)",
                      casenum);
              return 0;
            }
        }
    }

  x=option->parm[RLXMIXOPT];
  if (x!=OBJF_NODATA)
    { i=floor(x+0.5);
      if (i>mixopt) 
        { sprintf(errline,
          "rlxmixopt=%d incompatible with mixopt=%d in 'option' item (case %d)",
                  i,mixopt,casenum);
          return 0;
        }
    }

  x=option->parm[NSTEPDAY];
  if (x!=OBJF_NODATA) *nstepday=floor(x+0.5);

  if (*model!=NULLMOD)
    { x=option->parm[NSTEPBIO];
      if (x!=OBJF_NODATA) *nstepbio=floor(x+0.5);
    }

  x=option->parm[LEAPFROG];
  if (x!=OBJF_NODATA) leapfrogflag=floor(x+0.5);
  else leapfrogflag=0;

  if (leapfrogflag)
    { x=option->parm[TFPARM];
      if (x==OBJF_NODATA)
        { sprintf(errline,"Missing 'tfparm' in 'option' item (case %d)",
                  casenum);
          return 0;
        }
      if (*nstepbio>1 && 2*(*nstepbio/2)!=*nstepbio)
        { sprintf(errline,
    "nstepbio=%d incompatible with leapfrog scheme in 'option' item (case %d)",
                  *nstepbio,casenum);
          return 0;
        }
    }

  for (i=0;i<N_TRACER_PRI;i++)
    { x=option->parm[i+RLXRATES];
      if (x!=OBJF_NODATA && x<0.0)   
        { sprintf(errline,
            "relaxation rate = %g out of range in 'option' item (case %d)",
                  x,casenum);
          return 0;
        }
    }

  return 1;
}
 

static int validitem_taxis (int casenum, struct itemrec *taxis,
                            int *baseyear, int *periodic, int *yearlen,
                            char *errline)

{ OBJF_FLOAT_TYPE x;

  x=taxis->parm[BASEYEAR];
  if (x!=OBJF_NODATA) 
    { *baseyear=floor(x+0.5);
      *periodic=0;
      *yearlen=0;
    }

  x=taxis->parm[PERIODIC];
  if (x!=OBJF_NODATA) 
    { *periodic=floor(x+0.5);
      *yearlen=365;
    }

  x=taxis->parm[YEARLEN];
  if (x!=OBJF_NODATA) 
    { *yearlen=floor(x+0.5);
      if (*yearlen<1)
        { if (*periodic) sprintf(errline,
          "yearlen=%d in 'taxis' item is invalid, omit if not fixed (case %d)",
                                 *yearlen,casenum);
          else sprintf(errline,
                       "yearlen=%d in 'taxis' item is invalid (case %d)",
                       *yearlen,casenum);
          return 0;
        }
    }

  return 1;
}


static int validitem_timeperiod (int casenum, struct itemrec *timeperiod,
                                 int yearlen, int *startyear, int *startday,
                                 int *finyear, int *finday, int *startstep,
                                 char *errline)

{ OBJF_FLOAT_TYPE x;
 
  x=timeperiod->parm[STARTYEAR];
  if (x!=OBJF_NODATA) 
    { *startyear=floor(x+0.5);
      *finyear=*startyear;
      *finday=daysinyear(yearlen,*finyear);
    }

  x=timeperiod->parm[FINYEAR];
  if (x!=OBJF_NODATA) 
    { *finyear=floor(x+0.5);
      *finday=daysinyear(yearlen,*finyear);
    }

  x=timeperiod->parm[STARTDAY];
  if (x!=OBJF_NODATA) *startday=floor(x+0.5);

  x=timeperiod->parm[FINDAY];
  if (x!=OBJF_NODATA) *finday=floor(x+0.5);

  x=timeperiod->parm[STARTSTEP];
  if (x!=OBJF_NODATA) *startstep=floor(x+0.5);

  if (*finyear < *startyear || 
       (*finyear == *startyear && *finday < *startday))
    { sprintf(errline,
"finish (yr %d day %d) before start (yr %d day %d) - bad 'timeperiod' item (case %d)",
              *finyear,*finday,*startyear,*startday,casenum);
      return 0;
    }

  return 1;
}


static int validitem_environ (struct itemrec *environ,
                              OBJF_FLOAT_TYPE *maxdepth, 
                              OBJF_FLOAT_TYPE *latitude)

{ OBJF_FLOAT_TYPE x;

  x=environ->parm[MAXDEP];
  if (x!=OBJF_NODATA) *maxdepth=x;

  x=environ->parm[LATITUDE];
  if (x!=OBJF_NODATA) *latitude=x;

  return 1;
}


static int validitem_zlevel (int casenum, struct itemrec *zlevel, int *n,
                             OBJF_FLOAT_TYPE **zlevbot, char *errline)

{ int i;

  *n=zlevel->gspec[LEVEL].n;
  *zlevbot=zlevel->value[ZBOT];

  if (zlevel->objfunflag) return 1;
  zlevel->objfunflag=1; 
  
  if (*n<1 || *n>NK_MAX-1)
    { sprintf(errline,"nk=%d in 'zlevel' item is out of range (case %d)",
              *n,casenum);
      return 0;
    } 
      
  if (zlevel->gspec[LEVEL].first!=1) 
    { sprintf(errline,"bad 'zlevel' item - expecting k1=1 (case %d)",casenum);
      return 0;
    }

  if (*n>1 && zlevel->gspec[LEVEL].step!=1)
    { sprintf(errline,"bad 'zlevel' item - expecting kstep=1 (case %d)",
              casenum);
      return 0;
    }

  if (zlevel->dimflag[ILOC] && zlevel->gspec[ILOC].n!=1)
    { sprintf(errline,
  "bad 'zlevel' item spec - expecting single point in 'i' dimension (case %d)",
              casenum);
      return 0;
    }

  if (zlevel->dimflag[JLOC] && zlevel->gspec[JLOC].n!=1)
    { sprintf(errline,
  "bad 'zlevel' item spec - expecting single point in 'j' dimension (case %d)",
              casenum);
      return 0;
    }

  for (i=0;i<zlevel->gspec[LEVEL].n;i++) 
    if (zlevel->value[ZBOT][i]==OBJF_NODATA)
      { sprintf(errline,
    "one or more missing data values in 'zlevel' item record (case %d, k=%d)",
                casenum,i+1);
        return 0;
      }

  return 1;
}


static int validitem_init (int casenum, struct itemrec *init, int *n, 
                           OBJF_FLOAT_TYPE *t, int *year, 
                           OBJF_FLOAT_TYPE *tdays, char *errline)

{ *n=init->gspec[LEVEL].n;

  if (*n<1)
    { sprintf(errline,"nk=%d in 'init' item is invalid (case %d)",*n,casenum);
      return 0;
    }

  if (init->gspec[LEVEL].first!=1) 
    { sprintf(errline,"bad 'init' item - expecting k1=1 (case %d)",casenum);
      return 0;
    }

  if (*n>1 && init->gspec[LEVEL].step!=1)
    { sprintf(errline,"bad 'init' item - expecting kstep=1 (case %d)",
              casenum);
      return 0;
    }

  if (init->dimflag[TIME] && (init->dimflag[YEAR] || init->dimflag[TIMEOFYEAR]))
    { sprintf(errline,
"expecting spec for 't1' OR ('year1','tofyr1') in item 'init' - found both (case %d)",casenum);
      return 0;
    }

  if (init->dimflag[YEAR] && !init->dimflag[TIMEOFYEAR])
    { sprintf(errline,"missing spec for 'tofyr1' in item 'init' (case %d)",
              casenum);
      return 0;
    }

  if (init->dimflag[TIMEOFYEAR] && !init->dimflag[YEAR])
    { sprintf(errline,"missing spec for 'year1' in item 'init' (case %d)",
              casenum);
      return 0;
    }

  if ((init->dimflag[TIME] && init->gspec[TIME].n!=1) || 
      (init->dimflag[YEAR] && 
        (init->gspec[YEAR].n!=1 || init->gspec[TIMEOFYEAR].n!=1)))
    { sprintf(errline,
        "bad 'init' item spec - expecting single time point (case %d)",
              casenum);
      return 0;
    }

  *t=OBJF_NODATA;
  *year=0;
  *tdays=OBJF_NODATA;
  if (init->dimflag[TIME]) *t=init->gspec[TIME].first;
  else if (init->dimflag[YEAR])
    { *year=floor(init->gspec[YEAR].first+0.5);
      *tdays=init->gspec[TIMEOFYEAR].first;
      if (*tdays<0) 
        { sprintf(errline,"tofyr1=%g in 'init' item is invalid (case %d)",
                  *tdays,casenum);
          return 0;
        }
    }

  if (init->dimflag[ILOC] && init->gspec[ILOC].n!=1)
    { sprintf(errline,
  "bad 'init' item spec - expecting single point in 'i' dimension (case %d)",
              casenum);
      return 0;
    }

  if (init->dimflag[JLOC] && init->gspec[JLOC].n!=1)
    { sprintf(errline,
  "bad 'init' item spec - expecting single point in 'j' dimension (case %d)",
              casenum);
      return 0;
    }

  return 1;
}


static int validitem_ft (int casenum, struct itemrec *ft, char *itemname, 
                         char *errline)

{ int i,j;

  if (ft->objfunflag) return 1;
  ft->objfunflag=1;

  if (ft->gspec[TIME].step<0)
    { sprintf(errline,"negative time interval 'tstep' in '%s' item (case %d)",
              itemname,casenum);
      return 0;
    }

  if (ft->dimflag[LEVEL] && ft->gspec[LEVEL].n!=1)
    { sprintf(errline,
        "bad '%s' item spec - expecting single point in 'k' dimension (case %d)",
              itemname,casenum);
      return 0;
    }

  if (ft->dimflag[DEPTH] && ft->gspec[DEPTH].n!=1)
    { sprintf(errline,
        "bad '%s' item spec - expecting single point in 'z' dimension (case %d)",
              itemname,casenum);
      return 0;
    }

  if (ft->dimflag[ILOC] && ft->gspec[ILOC].n!=1)
    { sprintf(errline,
        "bad '%s' item spec - expecting single point in 'i' dimension (case %d)",
              itemname,casenum);
      return 0;
    }

  if (ft->dimflag[JLOC] && ft->gspec[JLOC].n!=1)
    { sprintf(errline,
        "bad '%s' item spec - expecting single point in 'j' dimension (case %d)",
              itemname,casenum);
      return 0;
    }

  for (j=0;j<N_FSCALAR+N_FPROF;j++) if (ft->value[j])
    for (i=0;i<ft->gspec[TIME].n;i++) if (ft->value[j][i]==OBJF_NODATA)
      { sprintf(errline,
          "one or more missing data values in '%s' item record (case %d, t=%g)",
                itemname,casenum,ft->gspec[TIME].first+ft->gspec[TIME].step*i);
        return 0; 
      }
 
  return 1;
}


static int validitem_fkt (int casenum, struct itemrec *fkt, char *itemname, 
                          int *n, char *errline)

{ int i,j,k,m;

  *n=fkt->gspec[LEVEL].n;

  if (fkt->objfunflag) return 1;
  fkt->objfunflag=1;

  if (fkt->dimflag[TIME] && fkt->gspec[TIME].step<0)
    { sprintf(errline,"negative time interval 'tstep' in '%s' item (case %d)",
              itemname,casenum);
      return 0;
    }
 
  if (*n<1)
    { sprintf(errline,"nk=%d in '%s' item is invalid (case %d)",
              *n,itemname,casenum);
      return 0;
    } 
      
  if (fkt->gspec[LEVEL].first!=1) 
    { sprintf(errline,"bad '%s' item - expecting k1=1 (case %d)",
              itemname,casenum);
      return 0;
    }

  if (*n>1 && fkt->gspec[LEVEL].step!=1)
    { sprintf(errline,"bad '%s' item - expecting kstep=1 (case %d)",
              itemname,casenum);
      return 0;
    }

  if (fkt->dimflag[ILOC] && fkt->gspec[ILOC].n!=1)
    { sprintf(errline,
        "bad '%s' item spec - expecting single point in 'i' dimension (case %d)",
              itemname,casenum);
      return 0;
    }

  if (fkt->dimflag[JLOC] && fkt->gspec[JLOC].n!=1)
    { sprintf(errline,
        "bad '%s' item spec - expecting single point in 'j' dimension (case %d)",
              itemname,casenum);
      return 0;
    }

  for (m=0;m<N_FPROF;m++) if (fkt->value[m])
    { k=0;
      for (i=0;i<fkt->gspec[TIME].n;i++)
        for (j=0;j<fkt->gspec[LEVEL].n;j++)
          { if (fkt->value[m][k]==OBJF_NODATA)
              { sprintf(errline,
  "one or more missing data values in '%s' item record (case %d, t=%g, k=%d)",
                        itemname,casenum,
                        fkt->gspec[TIME].first+fkt->gspec[TIME].step*i,
                        j+1);
                return 0;
              }
            k++;
          }
    }

  return 1;
}


static int validitem_obset (int casenum, struct itemrec *obset, char *errline)

{ OBJF_FLOAT_TYPE *obst;
  enum simcattype simcat;
  enum zinterpmode dummy;
  int i,j,simindex;

  if (obset->objfunflag) return 1;
  obset->objfunflag=1;

  obst=obset->dim[TIME];
  for (i=1; i<obset->npoint; i++)
    { if (obst[i]<obst[i-1])
        { sprintf(errline,
            "'obset' item data not ordered by time (case %d, observation %d)",
                  casenum,i+1);
          return 0;
        }
    }

  for (i=0;i<obset->npoint;i++)
    { if (obset->dim[DEPTH] && obset->dim[DEPTH][i]!=OBJF_NODATA)
        { for (j=0;j<N_OBSTYPE;j++) 
            if (obset->value[j] && obset->value[j][i]!=OBJF_NODATA)
              { simcat=simcatofob(j,&simindex,&dummy);
                if (simcat==FOUT_SCAL || simcat==DIAG_SCAL ||
                    simcat==IZ_TRACER || simcat==IZ_DIAG)
                  { sprintf(errline,
"scalar variable value found in depth-specific 'obset' item record (case %d, observation %d)",
                            casenum,i+1);
                    return 0;
                  }
              }
        }
      else
        { for (j=0;j<N_OBSTYPE;j++) 
            if (obset->value[j] && obset->value[j][i]!=OBJF_NODATA)
              { simcat=simcatofob(j,&simindex,&dummy);
                if (simcat!=FOUT_SCAL && simcat!=DIAG_SCAL &&
                    simcat!=IZ_TRACER && simcat !=IZ_DIAG)
                  { sprintf(errline,
"depth-dependent variable value found in depth-independent 'obset' item record (case %d, observation %d)",
                            casenum,i+1);
                    return 0;
                  }
              }
        }
    }
 
  return 1;
}


static int validftime (int casenum, int periodic, int periodlen, int tstart,
                       int tfin, struct itemrec *item, char *itemname, 
                       char *errline)

{ if (!item->dimflag[TIME]) return 1; 

  if (periodic)
 
    { if (item->gspec[TIME].min>periodlen)
        { sprintf(errline,
 "'%s' item starts after year end: can't set up periodic forcing (case %d)",
                  itemname,casenum);
          return 0;
        }

      if (item->gspec[TIME].max<0)
        { sprintf(errline,
 "'%s' item ends before t=0: can't set up periodic forcing (case %d)",
                  itemname,casenum);
          return 0;
        }

      if (findex_periodic(OBJF_NODATA,periodlen,&item->gspec[TIME],
                          NULL,NULL)<0)
        { sprintf(errline,
 "'%s' item has no data within year: can't set up periodic forcing (case %d)",
                  itemname,casenum);
          return 0;
        }

    }

  else

    { if (item->gspec[TIME].min>tstart)
        { sprintf(errline,
          "forcing in '%s' item starts after start time t=%d (case %d)",
                  itemname,tstart,casenum);
          return 0;
        }

      if (item->gspec[TIME].max<tfin)
        { sprintf(errline,
          "forcing in '%s' item ends before fin time t=%d (case %d)",
                  itemname,tfin,casenum);
          return 0;
        }
    }

  return 1;
}


double objfun (struct casetabrec *in, FILE *outfile[],int *status, 
               char **errtxt)

{ extern int FORTRAN_modassimflag[N_MODEL];
  int nt_nullmod,it_nullmod[N_TRACER_PRI];
  int it_agg[N_TRACER][NAGG_MAX]; 
  int id_agg[N_DIAG_AGG][NAGG_MAX]; 
  OBJF_FLOAT_TYPE optionparm[N_OPTION];
  int model;           
  enum photmodeltype photmodel; 
  enum kdmodeltype kdmodel; 
  enum achlmodeltype achlmodel; 
  int advection;
  double attenwater,attenpig;
  int kdpathopt,diffusion,mixopt,pertmixopt,relaxmixopt;
  int leapfrogflag,prevdiagflag;
  double tfparm;           
  int nstep_day,nstepbio;
  OBJF_FLOAT_TYPE maxdepth;
  double latitude;
  int baseyear,periodicflag,fixedyearlen;
  int year,day,finyear,finday;
  int tstart,tfin,yearlen,mthlen,yearlen_steps,mthlen_steps;
  int startstep,hfirst;
  double t,tf,frac,tmid,tmidnext,tfinobs;
  int mth,date;
  OBJF_FLOAT_TYPE tdays,thours;
  double timestep,tofyr;
  int nk,nkwet,nkavout;
  OBJF_FLOAT_TYPE *zlevbot;
  OBJF_FLOAT_TYPE zlevbot_def[1];
  OBJF_FLOAT_TYPE zlevdz[NK_MAX];
  OBJF_FLOAT_TYPE zlevmid[NK_MAX];
  OBJF_FLOAT_TYPE zlevtop[NK_MAX];
  double zlevdata[NK_MAX];
  struct obsrec obs;
  struct costrec cost,casecost;
  double fscal[N_FSCALAR];
  double fscalconst[N_FSCALAR];
  double fprof[N_FPROF][NK_MAX];
  double fprof0[N_FPROF][NK_MAX];
  double (*fprofpert)[NK_MAX],(*fprofpertsd)[NK_MAX];
  double (*fprofref0)[NK_MAX],(*fprofref)[NK_MAX],(*fprofrlx)[NK_MAX];
  double tracer[N_TRACER][NK_MAX];
  double tracer0[N_TRACER][NK_MAX];
  double tracer_ic[N_TRACER][NK_MAX];
  double ratio[N_RATIO][NK_MAX];
  double ratio0[N_RATIO][NK_MAX];
  double ratio_ic[N_RATIO][NK_MAX];
  double diagscal[N_DIAG_SCAL];
  double diagprof[N_DIAG_PROF][NK_MAX];
  double iztracer[N_TRACER];
  double izdiag[N_IZ_DIAG];
  enum fmodetype fmode_default;
  enum fmodetype fscalmode[N_FSCALAR];
  enum fmodetype fscalreq[N_FSCALAR+N_FPROF];
  enum fmodetype fprofmode[N_FPROF];
  enum fmodetype fprofreq[N_FPROF];
  unsigned char fproftop[N_FPROF];
  unsigned char fproffromtop[N_FPROF];
  unsigned char fconstreq[N_FSCALAR+N_FPROF];
  int fscalflag[N_FSCALAR];
  int fprofflag[N_FPROF-2*(N_TRACER_PRI)];
  int fflag_sst;
  int dsinkflag;
  enum transtype perttransform[N_TRACER_PRI];
  double pertautocorr,relaxctrlfact;
  int relaxmaskflag,relaxmaskopt;
  OBJF_FLOAT_TYPE relaxrate[N_TRACER_PRI]; 
  int initfound[N_TRACER_PRI];
  int ratiofound[N_TRACER_PRI];
  int pertfound[N_TRACER_PRI];
  int pertsdfound[N_TRACER_PRI];
  int pertseed[N_TRACER_PRI];
  int ranpertflag[N_TRACER_PRI];
  int pertflag[N_TRACER_PRI];
  int frelaxfound[N_TRACER_PRI];
  int relaxflag[N_TRACER_PRI];
  unsigned char tracerflag[N_TRACER];
  unsigned char ratioflag[N_TRACER];
  unsigned char diagtracerflag[N_TRACER];
  unsigned char diagratioflag[N_RATIO];
  unsigned char diagprofflag[N_DIAG_PROF];
  unsigned char iztracerflag[N_TRACER];
  unsigned char izdiagflag[N_IZ_DIAG];
  unsigned char fscalfound[N_FSCALITEM][N_FSCALAR+N_FPROF];
  unsigned char fproffound[N_FPROFITEM][N_FPROF];
  int firststep,firststep_lf,rdfset;
  int obsfound,misfitreq,valid,flag;
  int modoutputreq,dayoutputreq,mthoutputreq,yearoutputreq,timeavoutputreq;
  int assimoutputreq,misfitoutputreq;
  int fullmth,fullyear;
  int newmth;
  struct modoutrec dsum,msum,ysum;
  int assimpset[N_MODEL+1];
  int assimilation,nstepassim,startstepassim;
  int assimnext,assimupd,assimscalreq,assimprofreq;
  double assimscal[N_ASSIMSCALVAR];
  double assimprof[N_ASSIMPROFVAR][NK_MAX];
  int flushflag=1;
  struct caseinputrec *tabentry;
  OBJF_FLOAT_TYPE *xvector,x;
  double modelparm[NP_MAX];
  double *pvector;
  struct solverrec solver;
  double w;
  int i,icase,j,k,m,mm,y,h,n;

  /* --- 1. CASE-INDEPENDENT INITIALIZATION --- */

  *status=OK;
  *errtxt=NULL;

  dayoutputreq=(outfile[OUTSCALAR_DAV] || outfile[OUTPROFILE_DAV]);
  mthoutputreq=(outfile[OUTSCALAR_MAV] || outfile[OUTPROFILE_MAV]);
  yearoutputreq=(outfile[OUTSCALAR_YAV] || outfile[OUTPROFILE_YAV]);
  timeavoutputreq=(dayoutputreq ||  mthoutputreq || yearoutputreq); 
  modoutputreq=(outfile[OUTSCALAR] || outfile[OUTPROFILE] || timeavoutputreq);
  assimoutputreq=(outfile[OUTASSIMSCAL] || outfile[OUTASSIMPROF]);
  misfitoutputreq=(outfile[OUTMISFIT]!=NULL);

  /* Assign assimilation scheme parameter set items to models for cross-ref */

  j=N_MODEL;
  for (i=0;i<N_MODEL;i++) 
    { if (FORTRAN_modassimflag[i]) assimpset[i+1]=j++;
      else assimpset[i+1]=0;
    }

  /* Initialize cost function data */

  for (m=0;m<N_OBSTYPE;m++)
    { cost.mfitcount[m]=0;
      cost.mfitsum[m]=0.0;
    } 

  /* --- 2. PROCESS CASE TABLE RECORDS --- */

  for (icase=0;icase<in->ndata;icase++)

    { tabentry=&in->table[icase];

      for (m=0;m<N_OBSTYPE;m++)
        { casecost.mfitcount[m]=0;
          casecost.mfitsum[m]=0.0;
        } 

      /* 2.1. --- GET TIME-INVARIANT INPUT DATA --- */

      /* Set transformation options for misfit calculations */

      for (m=0;m<N_OBSTYPE;m++) obs.transform[m]=NULLTRANS;
      if (tabentry->item[MFITCTRL])
        for (m=0;m<N_OBSTYPE;m++)
          { x=tabentry->item[MFITCTRL]->parm[m];
            if (x>0 && x<2.5) obs.transform[m]=(enum transtype)(x+0.5);
          }

      /* Set up vertical grid */

      maxdepth=MAXDEPTH_DEF;
      if (tabentry->item[ENVIRON]) 
        { x=tabentry->item[ENVIRON]->parm[MAXDEP];
          if (x!=OBJF_NODATA) maxdepth=x;
        }

      nk=1;
      zlevbot_def[0]=maxdepth;
      zlevbot=zlevbot_def;
      if (tabentry->item[ZLEVEL])
        { nk=tabentry->item[ZLEVEL]->gspec[LEVEL].n;
          zlevbot=tabentry->item[ZLEVEL]->value[ZBOT];
        }

      zlevdz[0]=zlevbot[0];
      zlevmid[0]=0.5*zlevbot[0];
      zlevtop[0]=0;
      for (k=1;k<nk;k++) 
        { zlevdz[k]=zlevbot[k]-zlevbot[k-1];
          zlevmid[k]=0.5*(zlevbot[k-1]+zlevbot[k]);
          zlevtop[k]=zlevbot[k-1];
        }

      nkwet=1;
      for (k=1;k<nk;k++) if (zlevmid[k]<=maxdepth) nkwet++;

      /* Set up time axis */

      baseyear=1;
      periodicflag=1;
      fixedyearlen=365;
      if (tabentry->item[TIMEAXIS])
        { x=tabentry->item[TIMEAXIS]->parm[BASEYEAR];
          if (x!=OBJF_NODATA) 
            { baseyear=floor(x+0.5);
              periodicflag=0;
              fixedyearlen=0;
            }
          x=tabentry->item[TIMEAXIS]->parm[PERIODIC];
          if (x!=OBJF_NODATA) 
            { periodicflag=floor(x+0.5); 
              fixedyearlen=365;
            }
          x=tabentry->item[TIMEAXIS]->parm[YEARLEN];
          if (x!=OBJF_NODATA) fixedyearlen=floor(x+0.5); 
        }

      /* Set start and end times */

      year=baseyear;
      day=1;
      finyear=year;
      finday=daysinyear(fixedyearlen,finyear);
      startstep=1;
      if (tabentry->item[TIMEPERIOD])
        { x=tabentry->item[TIMEPERIOD]->parm[STARTYEAR];
          if (x!=OBJF_NODATA) 
            { year=floor(x+0.5);
              finyear=year;
              finday=daysinyear(fixedyearlen,finyear);
            }
          x=tabentry->item[TIMEPERIOD]->parm[STARTDAY];
          if (x!=OBJF_NODATA) day=floor(x+0.5);
          x=tabentry->item[TIMEPERIOD]->parm[FINYEAR];
          if (x!=OBJF_NODATA) 
            { finyear=floor(x+0.5);
              finday=daysinyear(fixedyearlen,finyear);
            }
          x=tabentry->item[TIMEPERIOD]->parm[FINDAY];
          if (x!=OBJF_NODATA) finday=floor(x+0.5);
          x=tabentry->item[TIMEPERIOD]->parm[STARTSTEP];
          if (x!=OBJF_NODATA) startstep=floor(x+0.5);
        }

      tstart=day-1;
      for (y=baseyear;y<year;y++) tstart+=daysinyear(fixedyearlen,y);

      if (tabentry->item[TIMEPERIOD])
        { tfin=finday;
          for (y=baseyear;y<finyear;y++) tfin+=daysinyear(fixedyearlen,y); 
        }
      else tfin=tstart;

      /* Initialize time variables */

      if (startstep==0) hfirst=0;
      else hfirst=startstep-1;

      yearlen=daysinyear(fixedyearlen,year);

      if (modoutputreq || assimoutputreq || misfitoutputreq)
        { mth=monthofyear(fixedyearlen,yearlen,day);
          date=dayofmonth(fixedyearlen,yearlen,day);
          fullmth=(date==1 && hfirst==0);
          fullyear=(day==1 && hfirst==0);
        }

      /* Set run options */

      model=NULLMOD;
      photmodel=NULLPHOTMOD;
      kdmodel=NULLKDMOD;
      kdpathopt=0;
      attenwater=OBJF_NODATA;
      attenpig=OBJF_NODATA;
      achlmodel=NULLACHLMOD;
      advection=0;
      diffusion=0;
      mixopt=0;
      for (m=0;m<N_TRACER_PRI;m++) perttransform[m]=NULLTRANS;
      pertautocorr=0;
      pertmixopt=0;
      relaxctrlfact=OBJF_NODATA;
      relaxmaskflag=0;
      relaxmaskopt=0;
      relaxmixopt=0;
      nstep_day=24;
      nstepbio=1;
      leapfrogflag=0;
      tfparm=OBJF_NODATA;
      assimilation=0;
      nstepassim=nstep_day;
      startstepassim=nstep_day/2+1;
      fmode_default=INTERPOLATED; 
      for (m=0;m<N_FSCALAR;m++) fscalmode[m]=fmode_default;
      for (m=0;m<N_FPROF;m++) fprofmode[m]=fmode_default;
      for (m=0;m<N_TRACER_PRI;m++) relaxrate[m]=0; 

      for (m=0;m<N_OPTION;m++) optionparm[m]=OBJF_NODATA;

      if (tabentry->item[OPTION])

        { x=tabentry->item[OPTION]->parm[MODEL];
          if (x!=OBJF_NODATA) model=floor(x+0.5);
          optionparm[MODEL]=model;

          if (model!=NULLMOD)

            { x=tabentry->item[OPTION]->parm[PHOTMODEL];
              if (x!=OBJF_NODATA) photmodel=(enum photmodeltype)(x+0.5);
              optionparm[PHOTMODEL]=photmodel;

              if (photmodel!=NULLPHOTMOD)

                { x=tabentry->item[OPTION]->parm[KDMODEL];
                  if (x!=OBJF_NODATA) kdmodel=(enum kdmodeltype)(x+0.5);
                  optionparm[KDMODEL]=kdmodel;

                  x=tabentry->item[OPTION]->parm[KDPATHOPT];
                  if (x!=OBJF_NODATA) kdpathopt=(int)(x+0.5);
                  optionparm[KDPATHOPT]=kdpathopt;

                  x=tabentry->item[OPTION]->parm[ACHLMODEL];
                  if (x!=OBJF_NODATA) achlmodel=(enum achlmodeltype)(x+0.5);
                  optionparm[ACHLMODEL]=achlmodel;

                  if (kdmodel==NULLKDMOD || kdmodel==LINEARPIG)
                    { attenwater=tabentry->item[OPTION]->parm[ATTENWATER];
                      optionparm[ATTENWATER]=attenwater;
                    }
                  if (kdmodel==LINEARPIG)
                    { attenpig=tabentry->item[OPTION]->parm[ATTENPIG];
                      optionparm[ATTENPIG]=attenpig;
                    }
                }
            }

          x=tabentry->item[OPTION]->parm[ADVECTSCHEME];
          if (x!=OBJF_NODATA) advection=(int)(x+0.5);
          optionparm[ADVECTSCHEME]=advection;

          x=tabentry->item[OPTION]->parm[DIFFSCHEME];
          if (x!=OBJF_NODATA) diffusion=(int)(x+0.5);
          optionparm[DIFFSCHEME]=diffusion;

          x=tabentry->item[OPTION]->parm[MIXOPT];
          if (x!=OBJF_NODATA) mixopt=(int)(x+0.5);
          optionparm[MIXOPT]=mixopt;

          for (m=0;m<N_TRACER_PRI;m++)
            { x=tabentry->item[OPTION]->parm[m+PERTTR];
              if (x!=OBJF_NODATA) perttransform[m]=(enum transtype)(x+0.5);
              optionparm[m+PERTTR]=perttransform[m];
            }

          x=tabentry->item[OPTION]->parm[PERTAUTO];
          if (x!=OBJF_NODATA) pertautocorr=x;
          optionparm[PERTAUTO]=pertautocorr;

          x=tabentry->item[OPTION]->parm[PERTMIXOPT];
          if (x!=OBJF_NODATA) pertmixopt=(int)(x+0.5);
          optionparm[PERTMIXOPT]=pertmixopt;

          x=tabentry->item[OPTION]->parm[RLXCTRLFACT];
          if (x!=OBJF_NODATA) relaxctrlfact=x;
          optionparm[RLXCTRLFACT]=relaxctrlfact;

          x=tabentry->item[OPTION]->parm[RLXMASKFLAG];
          if (x!=OBJF_NODATA) relaxmaskflag=floor(x+0.5);
          optionparm[RLXMASKFLAG]=relaxmaskflag;

          x=tabentry->item[OPTION]->parm[RLXMASKOPT];
          if (x!=OBJF_NODATA) relaxmaskopt=(int)(x+0.5);
          optionparm[RLXMASKOPT]=relaxmaskopt;

          x=tabentry->item[OPTION]->parm[RLXMIXOPT];
          if (x!=OBJF_NODATA) relaxmixopt=(int)(x+0.5);
          optionparm[RLXMIXOPT]=relaxmixopt;

          x=tabentry->item[OPTION]->parm[NSTEPDAY];
          if (x!=OBJF_NODATA) nstep_day=(int)(x+0.5);
          optionparm[NSTEPDAY]=nstep_day;

          if (model!=NULLMOD)
            { x=tabentry->item[OPTION]->parm[NSTEPBIO];
              if (x!=OBJF_NODATA) nstepbio=(int)(x+0.5);
              optionparm[NSTEPBIO]=nstepbio;
            }

          x=tabentry->item[OPTION]->parm[LEAPFROG];
          if (x!=OBJF_NODATA) leapfrogflag=(int)(x+0.5);
          optionparm[LEAPFROG]=leapfrogflag;

          if (leapfrogflag) 
            { tfparm=tabentry->item[OPTION]->parm[TFPARM];
              optionparm[TFPARM]=tfparm;
            }

          if (model!=NULLMOD)
            { x=tabentry->item[OPTION]->parm[ASSIMSCHEME];
              if (x!=OBJF_NODATA) assimilation=(int)(x+0.5);
              optionparm[ASSIMSCHEME]=assimilation;

              x=tabentry->item[OPTION]->parm[NSTEPASSIM];
              if (x!=OBJF_NODATA) nstepassim=(int)(x+0.5);
              optionparm[NSTEPASSIM]=nstepassim;

              x=tabentry->item[OPTION]->parm[STARTSTEPASSIM];
              if (x!=OBJF_NODATA) startstepassim=(int)(x+0.5);
              optionparm[STARTSTEPASSIM]=startstepassim;
            }

          x=tabentry->item[OPTION]->parm[FMODE];
          if (x!=OBJF_NODATA) fmode_default=(enum fmodetype)(x+0.5);
          optionparm[FMODE]=fmode_default;

          for (m=0;m<N_FSCALAR;m++) 
            { x=tabentry->item[OPTION]->parm[m+FMODES];
              if (x!=OBJF_NODATA) fscalmode[m]=(enum fmodetype)(x+0.5);
              else fscalmode[m]=fmode_default;
              optionparm[m+FMODES]=fscalmode[m];
            }

          for (m=0;m<N_FPROF;m++) 
            { x=tabentry->item[OPTION]->parm[m+FMODES+N_FSCALAR];
              if (x!=OBJF_NODATA) fprofmode[m]=(enum fmodetype)(x+0.5);
              else fprofmode[m]=fmode_default;
              optionparm[m+FMODES+N_FSCALAR]=fprofmode[m];
            }

          for (m=0;m<N_TRACER_PRI;m++) 
            { x=tabentry->item[OPTION]->parm[m+RLXRATES];
              if (x!=OBJF_NODATA) relaxrate[m]=x;
              if (relaxrate[m]>RELAXULIM) relaxrate[m]=RELAXOVERRIDE;
              optionparm[m+RLXRATES]=relaxrate[m];
            }

          for (m=0;m<N_TRACER_PRI;m++) 
            optionparm[m+DEFTRACERS]=
              tabentry->item[OPTION]->parm[m+DEFTRACERS];

          for (m=0;m<N_RATIO;m++)
            optionparm[m+DEFRATIOS]=
              tabentry->item[OPTION]->parm[m+DEFRATIOS];
        }

      /* Set seeds for random tracer perturbation */

      for (m=0;m<N_TRACER_PRI;m++) pertseed[m]=0;

      if (tabentry->item[PERTSEED])
        
        { for (m=0;m<N_TRACER_PRI;m++) 
            { x=tabentry->item[PERTSEED]->parm[m];
              if (x!=OBJF_NODATA) pertseed[m]=(int)(x>0?x+0.5:x-0.5);
            }
        }

      /* 2.2. --- SETUP ACCESS TO FORCING DATA --- */
 
      /* Locate profile forcing variables and set forcing data modes 
         to mode reqested in run options for all variables present 
          - For variables not found the request is set to 'NOTAPPLIED' 
         Two flags are also initialized here according to the 
         available data:
          - fproffromtop=1 indicates that a full profile should be
            generated from a scalar forcing variable 
            (set to 0 if profile found)
          - fproftop=1 indicates that only the top value from a profile 
            forcing variable is to be used 
         These flags may be updated later according to specific model 
         requirements */

      for (m=0;m<N_FPROF;m++) 
        { fprofreq[m]=NOTAPPLIED; 
          fproffromtop[m]=1;
          fproftop[m]=0;
        }

      for (j=0;j<N_FPROFITEM;j++) if (tabentry->item[FPROF+j])
        for (m=0;m<N_FPROF;m++) 
          { xvector=tabentry->item[FPROF+j]->value[m];
            if (xvector)
              { fproffound[j][m]=(xvector[0]!=OBJF_NODATA);
                fprofreq[m]=fprofmode[m];
                fproffromtop[m]=0;
              }
            else fproffound[j][m]=0;
          }

      /* Locate scalar forcing variables and set forcing data modes for all 
         variables present
          - Include scalar versions of profile variables not found in profile
            forcing data 
          - For variables not found the request is set to 'NOTAPPLIED' */

      for (m=0;m<N_FSCALAR+N_FPROF;m++) fscalreq[m]=NOTAPPLIED; 

      for (j=0;j<N_FSCALITEM;j++) if (tabentry->item[FSCALAR+j])
        { for (m=0;m<N_FSCALAR;m++) 
            { xvector=tabentry->item[FSCALAR+j]->value[m]; 
              if (xvector) 
                { fscalfound[j][m]=(xvector[0]!=OBJF_NODATA);
                  fscalreq[m]=fscalmode[m];
                }
              else fscalfound[j][m]=0;
            }
          for (m=0;m<N_FPROF;m++) if (fprofreq[m]==NOTAPPLIED) 
            { xvector=tabentry->item[FSCALAR+j]->value[m+N_FSCALAR]; 
              if (xvector) 
                { fscalfound[j][m+N_FSCALAR]=(xvector[0]!=OBJF_NODATA); 
                  fscalreq[m+N_FSCALAR]=fprofmode[m];
                }
              else fscalfound[j][m+N_FSCALAR]=0;
            }
        }

      /* Set request flag to get forcing variables from 'environ' parameter set
         if not found elsewhere (At this point it is assumed that all 
         forcing variables are used) */
 
      for (m=0;m<N_FSCALAR;m++) 
        { if (fscalreq[m]==NOTAPPLIED) fconstreq[m]=1;
          else fconstreq[m]=0;
        }
      for (m=0;m<N_FPROF;m++) 
        { if (fprofreq[m]==NOTAPPLIED && fscalreq[m+N_FSCALAR]==NOTAPPLIED) 
            fconstreq[m+N_FSCALAR]=1;
          else fconstreq[m+N_FSCALAR]=0;
        }

      /* Set tracer perturbation flags and suppress unnecessary forcing 
         profiles
          - Perturbation for a tracer is activated if perturbation rate forcing
            data are found or if stochastic forcing is specified by presence 
            of a seed and a perturbation std. dev.
          - If no constant data are requested then we know forcing data has
            been found, otherwise we need to check 'environ' parameter set
            (unless forcing data mode suppresses forcing data anyway)
          - If a perturbation rate constant or std. dev. constant is zero 
            it is treated as not found 
          - N.B. perturbation flags set here are ignored for unused tracers */

      for (m=0;m<N_TRACER_PRI;m++)
 
        { if (fprofmode[m+PERT]==NOTAPPLIED) pertfound[m]=0;
          else if (!fconstreq[m+PERT+N_FSCALAR]) pertfound[m]=1;
          else if (!tabentry->item[ENVIRON] ||
                   tabentry->item[ENVIRON]->parm[m+N_ENVIRON+N_FSCALAR+PERT]
                     ==OBJF_NODATA) pertfound[m]=0;
          else
            pertfound[m]=
              (tabentry->item[ENVIRON]->parm[m+N_ENVIRON+N_FSCALAR+PERT]!=0);

          if (fprofmode[m+PERTSD]==NOTAPPLIED) pertsdfound[m]=0;
          else if (!fconstreq[m+PERT+N_FSCALAR]) pertsdfound[m]=1; 
          else if (!tabentry->item[ENVIRON] ||
                     tabentry->item[ENVIRON]->parm[m+N_ENVIRON+N_FSCALAR+PERTSD]
                       ==OBJF_NODATA) pertsdfound[m]=0;
          else
            pertsdfound[m]=
              (tabentry->item[ENVIRON]->parm[m+N_ENVIRON+N_FSCALAR+PERTSD]>0);

          ranpertflag[m]=(pertseed[m]!=0 && pertsdfound[m]);
          pertflag[m]=(pertfound[m] || ranpertflag[m]); 

          if (!pertfound[m])
            { fprofreq[m+PERT]=NOTAPPLIED;
              fscalreq[m+PERT+N_FSCALAR]=NOTAPPLIED;
              fconstreq[m+PERT+N_FSCALAR]=NOTAPPLIED;
            }

          if (!ranpertflag[m] && relaxctrlfact<0) 
            { fprofreq[m+PERTSD]=NOTAPPLIED;
              fscalreq[m+PERTSD+N_FSCALAR]=NOTAPPLIED;
              fconstreq[m+PERTSD+N_FSCALAR]=NOTAPPLIED;
            }
        }

      /* Set tracer relaxation flags and suppress unnecessary forcing profiles 
         All relaxation is suppressed if a relaxation control factor of zero
         is specified. Otherwise:
          - Relaxation for a tracer is activated if relaxation rate forcing 
            data are found or if non-zero default rates have been set in 
            'option' parameter set 
          - If no constant data are requested then we know forcing data has 
            been found, otherwise we need to check 'environ' parameter set 
            (unless forcing data mode suppresses forcing data anyway)
          - If a relaxation rate constant is zero (or less) then relaxation 
            is not activated
          N.B. relaxation flags set here are ignored for unused tracers */ 

      for (m=0;m<N_TRACER_PRI;m++)

        { if (relaxctrlfact==0) relaxflag[m]=0; 
          else
            { if (fprofmode[m+RLXRS]==NOTAPPLIED) 
                { frelaxfound[m]=0;
                  relaxflag[m]=(relaxrate[m]>0);
                }
              else if (!fconstreq[m+N_FSCALAR+RLXRS]) 
                { frelaxfound[m]=1;
                  relaxflag[m]=1;
                }
              else if (!tabentry->item[ENVIRON] ||
                tabentry->item[ENVIRON]->parm[m+N_ENVIRON+N_FSCALAR+RLXRS]==
                       OBJF_NODATA) 
                { frelaxfound[m]=0;
                  relaxflag[m]=(relaxrate[m]>0);
                }
              else 
                { frelaxfound[m]=1;
                  relaxflag[m]=
                 (tabentry->item[ENVIRON]->parm[m+N_ENVIRON+N_FSCALAR+RLXRS]>0);
                }
             }

          if (!relaxflag[m]) 
            { fprofreq[m+RLXRS]=NOTAPPLIED;
              fscalreq[m+RLXRS+N_FSCALAR]=NOTAPPLIED;
              fconstreq[m+RLXRS+N_FSCALAR]=NOTAPPLIED; 
              fprofreq[m+REFS]=NOTAPPLIED;
              fscalreq[m+REFS+N_FSCALAR]=NOTAPPLIED;
              fconstreq[m+REFS+N_FSCALAR]=NOTAPPLIED; 
              if (!ranpertflag[m])
                { fprofreq[m+PERTSD]=NOTAPPLIED;
                  fscalreq[m+PERTSD+N_FSCALAR]=NOTAPPLIED;
                  fconstreq[m+PERTSD+N_FSCALAR]=NOTAPPLIED;
                }
            }
        }

      /* 2.3. --- DO MODEL-INDEPENDENT INITIALIZATION --- */

      /* Set initial conditions (model independent)
         All tracers and ratios are initialized here 
         or set to OBJF_NODATA if unavailable - the current ecosystem model 
         then selects its own subset
         Also set up indexing for null model tracers - any tracers initialized
         perturbed or relaxed are activated if there is no ecosystem model */ 

      nt_nullmod=0;
      for (m=0;m<N_TRACER_PRI;m++)

        { xvector=NULL;
          x=OBJF_NODATA;
          if (tabentry->item[INIT]) xvector=tabentry->item[INIT]->value[m];
          if (!xvector) 
            { x=optionparm[m+DEFTRACERS]; 
              initfound[m]=0;
            }
          else initfound[m]=1;
          for (k=0;k<nkwet;k++)
            { if (xvector) tracer_ic[m][k]=xvector[k];
              else tracer_ic[m][k]=x;
            }
          for (k=nkwet;k<nk;k++) tracer_ic[m][k]=OBJF_NODATA;

          if (xvector || x!=OBJF_NODATA || pertflag[m] || relaxflag[m]) 
            { it_nullmod[nt_nullmod]=m+1;
              nt_nullmod++;
            }
        }

      for (m=0;m<N_RATIO;m++)
        { xvector=NULL;
          x=OBJF_NODATA;
          if (tabentry->item[INIT]) 
            xvector=tabentry->item[INIT]->value[m+N_TRACER_PRI];
          if (!xvector) 
            { x=optionparm[m+DEFRATIOS];
              ratiofound[m]=0;
            }
          else ratiofound[m]=1;
          for (k=0;k<nkwet;k++)
            { if (xvector) ratio_ic[m][k]=xvector[k];
              else ratio_ic[m][k]=x;
            }
          for (k=nkwet;k<nk;k++) ratio_ic[m][k]=OBJF_NODATA;
        }

      /* 2.4. --- SET UP MODEL --- */

      n=0;
      pvector=NULL;
      if (model!=NULLMOD && tabentry->item[model-1])
        { n=in->objfname->item[model-1].nvar;
          xvector=tabentry->item[model-1]->parm;
          for (j=0;j<n;j++) modelparm[j]=xvector[j];
          pvector=modelparm;
        }

      latitude=OBJF_NODATA;
      if (tabentry->item[ENVIRON])
        latitude=tabentry->item[ENVIRON]->parm[LATITUDE];

      for (j=0;j<nk;j++) zlevdata[j]=zlevbot[j];

      timestep=1./nstep_day;

      FORTRAN_modelsetup(&model,&n,pvector,&photmodel,&kdmodel,&kdpathopt,
        &attenwater,&attenpig,&achlmodel,&advection,&diffusion,&mixopt,
        pertfound,ranpertflag,perttransform,&pertautocorr,pertseed,
        &pertmixopt,  
        relaxflag,&relaxctrlfact,&relaxmaskflag,&relaxmaskopt,&relaxmixopt,
        &latitude,&nkwet,zlevdata,&timestep,&nstepbio,&leapfrogflag,
        &tfparm,&nt_nullmod,it_nullmod,it_agg,id_agg,
        &fscalflag[SOL],&fscalflag[SOLAV],&fscalflag[MLD],&fflag_sst,
        &fscalflag[SSS],&fscalflag[PCO2ATM],&fscalflag[DUST],&fscalflag[FEDUST],
        &fprofflag[TEMP],&fprofflag[W],&fprofflag[VDC],&dsinkflag); 

      /* Set solver information for diagnostics calculations */

      solver.leapfrog=leapfrogflag;
      solver.nstepbio=nstepbio;
      solver.advection=advection;
      solver.dsink=dsinkflag;

      /* Do assimilation scheme setup */

      n=0;
      pvector=NULL;
      j=assimpset[model];
      if (j && assimilation>0 && tabentry->item[j])
        { n=in->objfname->item[j].nvar; 
          xvector=tabentry->item[j]->parm; 
          for (j=0;j<n;j++) modelparm[j]=xvector[j];
          pvector=modelparm;
        }

      assimscalreq=(outfile[OUTASSIMSCAL]!=NULL);
      assimprofreq=(outfile[OUTASSIMPROF]!=NULL);

      FORTRAN_assimsetup(&assimilation,&n,pvector,&nstepassim,
        &startstepassim,&assimscalreq,&assimprofreq,&fscalflag[MLD],
        &fscalflag[CCHLDA],&fscalflag[CHLDA],
        &fscalflag[PGROWDA],&fscalflag[PLOSSDA]); 

      /* Override forcing data modes for any redundant data
         to save processing and suppress in output tables and adjust 
         forcing control flags for temperature if only the surface value 
         is needed */

      for (m=0;m<N_FSCALAR;m++) if (!fscalflag[m])
        { fscalreq[m]=NOTAPPLIED;
          fconstreq[m]=NOTAPPLIED;
        }

      for (m=0;m<PERT;m++) if (!fprofflag[m])
        { if (m==TEMP && fflag_sst)
            { if (fprofreq[TEMP]) fproftop[TEMP]=1; 
              else fproffromtop[TEMP]=0;
            }
          else 
            { fprofreq[m]=NOTAPPLIED;
              fscalreq[m+N_FSCALAR]=NOTAPPLIED;
              fconstreq[m+N_FSCALAR]=NOTAPPLIED;
            }
        }

      /* --- 2.5 INITIALIZE STATE, FORCING & DIAGNOSTICS --- */

      /* Clear workspace for model output (for output tables and/or misfit 
         calculation) */ 

      if (modoutputreq || tabentry->item[OBS])

        { for (m=0;m<N_FSCALAR;m++) fscal[m]=OBJF_NODATA;
          for (m=0;m<N_DIAG_SCAL;m++) diagscal[m]=OBJF_NODATA;
          for (m=0;m<N_TRACER;m++) iztracer[m]=OBJF_NODATA;
          for (m=0;m<N_IZ_DIAG;m++) izdiag[m]=OBJF_NODATA;
          for (k=0;k<nk;k++)
            { for (m=0;m<N_FPROF;m++) fprof[m][k]=OBJF_NODATA;
              for (m=0;m<N_TRACER;m++) tracer[m][k]=OBJF_NODATA;
              for (m=0;m<N_RATIO;m++) ratio[m][k]=OBJF_NODATA;
              for (m=0;m<N_DIAG_PROF;m++) diagprof[m][k]=OBJF_NODATA;
            }
        }

      if (assimoutputreq)
        { for (m=0;m<N_ASSIMSCALVAR;m++) assimscal[m]=OBJF_NODATA;
          for (k=0;k<nk;k++)
            for (m=0;m<N_ASSIMPROFVAR;m++) assimprof[m][k]=OBJF_NODATA;
        }

      /* Initialize model state (subsection of workspace: tracers & ratios) */

      FORTRAN_modelinit(tracer_ic,ratio_ic,tracer,ratio);

      /* Initialize flags to show which tracers and ratios are available */

      for (m=0;m<N_TRACER;m++) tracerflag[m]=(tracer[m][0]!=OBJF_NODATA);
      for (m=0;m<N_RATIO;m++) ratioflag[m]=(ratio[m][0]!=OBJF_NODATA);

      /* Suppress perturbation and relaxation forcing for 
         absent primary tracers */ 

      for (m=0;m<N_TRACER_PRI;m++) if (!tracerflag[m])
        { fprofreq[m+PERT]=NOTAPPLIED;
          fscalreq[m+PERT+N_FSCALAR]=NOTAPPLIED;
          fconstreq[m+PERT+N_FSCALAR]=NOTAPPLIED;
          fprofreq[m+PERTSD]=NOTAPPLIED;
          fscalreq[m+PERTSD+N_FSCALAR]=NOTAPPLIED;
          fconstreq[m+PERTSD+N_FSCALAR]=NOTAPPLIED;
          fprofreq[m+RLXRS]=NOTAPPLIED;
          fscalreq[m+RLXRS+N_FSCALAR]=NOTAPPLIED;
          fconstreq[m+RLXRS+N_FSCALAR]=NOTAPPLIED;
          fprofreq[m+REFS]=NOTAPPLIED;
          fscalreq[m+REFS+N_FSCALAR]=NOTAPPLIED;
          fconstreq[m+REFS+N_FSCALAR]=NOTAPPLIED;
        }

      /* Set reference profile forcing from environmental constants if required.
         (default=0 if not present or input data has been deactivated by
         setting the forcing data mode to 'NOTAPPLIED') */ 
    
      for (m=REFS;m<REFS+N_TRACER_PRI;m++) if (fconstreq[m+N_FSCALAR])
          
        { if (fprofmode[m]==NOTAPPLIED ||
              !tabentry->item[ENVIRON] ||
              (fprof[m][0]=tabentry->item[ENVIRON]->parm[N_ENVIRON+N_FSCALAR+m])
               ==OBJF_NODATA) fprof[m][0]=0;

          if (fproffromtop[m])
            for (k=1;k<nkwet;k++) fprof[m][k]=fprof[m][0];
        }

      /* Set time at forcing time point = start of time step. */
          
      frac=(double)hfirst/nstep_day;
      t=tstart+frac; 
      tdays=day-1+frac;
          
      /* Interpolate reference profile forcing data to forcing time point 
         (If forcing is not year specific then use time of year) */
          
      if (periodicflag) tf=tdays;
      else tf=t;

      firststep=1;
      getforcing_fulldep(tf,periodicflag,yearlen,firststep,
                         tabentry->item,REFS,N_TRACER_PRI,fscalfound,
                         fscalreq,fproffound,fprofreq,fproftop,
                         fproffromtop,nkwet,fprof);

      /* --- 2.6. WRITE OPTION OUTPUT TABLE --- */

      /* Write parameter values to be used in the run 
         (Suppress any irrelevant values first) */

      if (outfile[OUTOPTION])

        { /* Tracer perturbation parameters */

          flag=0;
          for (m=0;m<N_TRACER_PRI;m++) 
            { if (!tracerflag[m] || !pertflag[m]) 
                optionparm[m+PERTTR]=OBJF_NODATA;
              else flag=1;
            }

          if (flag) 
            { flag=0;
              for (m=0;m<N_TRACER_PRI;m++) 
                if (tracerflag[m] && ranpertflag[m]) flag=1;
              if (!flag) optionparm[PERTAUTO]=OBJF_NODATA;
            }
          else
            { optionparm[PERTMIXOPT]=OBJF_NODATA;
              optionparm[PERTAUTO]=OBJF_NODATA;
            }

          /* Relaxation parameters */

          if (relaxctrlfact==0) 
            { optionparm[RLXMASKFLAG]=OBJF_NODATA;
              optionparm[RLXMASKOPT]=OBJF_NODATA;
            }
          else
            { flag=0; 
              for (m=0;m<N_TRACER_PRI;m++)
                if (tracerflag[m] && relaxflag[m]!=0) flag=1;
              if (!flag) 
                { optionparm[RLXCTRLFACT]=OBJF_NODATA; 
                  optionparm[RLXMASKFLAG]=OBJF_NODATA;
                  optionparm[RLXMASKOPT]=OBJF_NODATA;
                }
              else if (relaxmaskflag==0)
                optionparm[RLXMASKOPT]=OBJF_NODATA;
            }

          /* Assimilation parameters */

          if (assimilation==0)
            { optionparm[NSTEPASSIM]=OBJF_NODATA;
              optionparm[STARTSTEPASSIM]=OBJF_NODATA;
            }

          /* Forcing data modes 
             N.B. data mode is suppressed if there are no time-varying data
             to interpolate (except where mode is 'NOT APPLIED' and this
             could be the cause!) */

          flag=0;
          for (m=0;m<N_FSCALAR;m++)
            { if (!fscalflag[m] || fconstreq[m]) 
                optionparm[m+FMODES]=OBJF_NODATA;
              else flag=1;
            }
          for (m=0;m<PERT;m++)
            { if ((!fprofflag[m] && !(m==TEMP && fflag_sst)) || 
                  fconstreq[m+N_FSCALAR])
                optionparm[m+FMODES+N_FSCALAR]=OBJF_NODATA;
              else flag=1;
            }
          for (m=0;m<N_TRACER_PRI;m++)
            { if (!tracerflag[m] || 
                  (!pertfound[m] && fprofmode[m+PERT]!=NOTAPPLIED) ||
                  fconstreq[m+PERT+N_FSCALAR])  
                optionparm[m+FMODES+N_FSCALAR+PERT]=OBJF_NODATA;
              else flag=1;
              if (!tracerflag[m] || 
                  (pertseed[m]==0 && relaxctrlfact<=0) ||
                  fconstreq[m+PERTSD+N_FSCALAR])
                optionparm[m+FMODES+N_FSCALAR+PERTSD]=OBJF_NODATA;
              else flag=1;
              if (!tracerflag[m] || 
                  (!relaxflag[m] && fprofmode[m+RLXRS]!=NOTAPPLIED) ||
                  relaxctrlfact==0 ||
                  fconstreq[m+RLXRS+N_FSCALAR])
                optionparm[m+FMODES+N_FSCALAR+RLXRS]=OBJF_NODATA;
              else flag=1;
              if (!tracerflag[m] ||
                  !relaxflag[m] || 
                  fconstreq[m+REFS+N_FSCALAR])
                optionparm[m+FMODES+N_FSCALAR+REFS]=OBJF_NODATA;
              else flag=1;
            }
          if (!flag) optionparm[FMODE]=OBJF_NODATA;

          /* Default relaxation rates */

          for (m=0;m<N_TRACER_PRI;m++)
            if (!tracerflag[m] || relaxctrlfact==0 || frelaxfound[m]) 
              optionparm[m+RLXRATES]=OBJF_NODATA;

          /* Default initial state */ 

          for (m=0;m<N_TRACER_PRI;m++)
            if (!tracerflag[m] || initfound[m]) 
              optionparm[m+DEFTRACERS]=OBJF_NODATA;

          for (m=0;m<N_RATIO;m++)
            if (!ratioflag[m] || ratiofound[m]) 
              optionparm[m+DEFRATIOS]=OBJF_NODATA;

          *status=writeouttabdata_array(OUTOPTION,outfile[OUTOPTION],flushflag,
                                        1,tabentry,N_OPTION,optionparm);
          if (*status!=OK) return -1;
        }

      /* --- 2.7. DO SETUP FOR OUTPUT DATA PRODUCTION --- */

      /* Do setup required specifically for producing model output 
         and diagnostics and write initial output records */
            
      if (modoutputreq || tabentry->item[OBS])

        { /* Initialize diagnostics flags used to show which 
             tracers/ratios/diagnostics/integrals need updating by MarMOT */

          for (m=0;m<N_TRACER;m++) diagtracerflag[m]=0;
          for (m=0;m<N_RATIO;m++) diagratioflag[m]=0;
          for (m=0;m<N_DIAG_PROF;m++) diagprofflag[m]=0;
          for (m=0;m<N_IZ_DIAG;m++) izdiagflag[m]=0;

          if (model!=NULLMOD)

            { /* Initialize any relevant ratios that are not already 
                 initialized by model using ratio values taken from the 
                 'init' item.
                 Attempt to initialize all ratios associated with primary
                 tracers for which model output is available. 
                 If the required ratios are not present then try alternatives
                 associated with appropriate aggregate tracers. */

              for (m=0;m<N_RATIO;m++) 
                { if (!ratioflag[m])
                    { switch (m)
                        { case CCHL: valid=tracerflag[PHY]; break; 
                          case CCHLNDIA: valid=tracerflag[PNDIA]; break; 
                          case CCHLDIAT: valid=tracerflag[PDIAT]; break; 
                          case CNPHY: valid=tracerflag[PHY]; break; 
                          case CNPNDIA: valid=tracerflag[PNDIA]; break; 
                          case CNPDIAT: valid=tracerflag[PDIAT]; break; 
                          case CNZOO: valid=tracerflag[ZOO]; break; 
                          case CNZMI: valid=tracerflag[ZMI]; break; 
                          case CNZME: valid=tracerflag[ZME]; break; 
                          case CNDET: valid=tracerflag[DET]; break;
                          case SINPDIAT: valid=tracerflag[PDIAT]; break;
                          case FEN: valid=
                                      tracerflag[PHY] ||
                                      tracerflag[PNDIA] ||
                                      tracerflag[PDIAT] ||
                                      tracerflag[ZOO] ||
                                      tracerflag[ZMI] ||
                                      tracerflag[ZME] ||
                                      tracerflag[DET]; break;
                          default: valid=0; 
                        }
                      if (valid)
                        { ratioflag[m]=(ratio_ic[m][0]!=OBJF_NODATA); 
                          if (ratioflag[m]) 
                            for (k=0;k<nkwet;k++) ratio[m][k]=ratio_ic[m][k];
                          else
                            { mm=-1;
                              switch(m)
                                { case CCHLNDIA: mm=CCHL; break; 
                                  case CCHLDIAT: mm=CCHL; break; 
                                  case CNPNDIA: mm=CNPHY; break; 
                                  case CNPDIAT: mm=CNPHY; break; 
                                  case CNZMI: mm=CNZOO; break; 
                                  case CNZME: mm=CNZOO; break; 
                                }
                              if (mm>=0) 
                                { ratioflag[m]=(ratio_ic[mm][0]!=OBJF_NODATA); 
                                  if (ratioflag[m]) 
                                    for (k=0;k<nkwet;k++) 
                                      ratio[m][k]=ratio_ic[mm][k];
                                }
                            }
                        }
                    }
                }

              /* Set diagnostic flags & associated depth integral flags
                 (excluding those needing runtime input) */

              setdiagflags(it_agg,&vamap,tracerflag,ratioflag,
                           diagtracerflag,diagratioflag,diagprofflag);

              for (m=IZDDTHN;m<IZDDTPN;m++)  
                izdiagflag[m]=diagprofflag[m-IZDDTHN+DDTH_TRACERS+NITROGEN];
              for (m=IZDDTPN;m<IZDDTRN;m++)  
                izdiagflag[m]=diagprofflag[m-IZDDTPN+DDTH_TRACERS+NITROGEN];
              for (m=IZDDTRN;m<N_IZ_DIAG;m++)  
                izdiagflag[m]=diagprofflag[m-IZDDTRN+DDTH_TRACERS+NITROGEN];
            }
          
          /* Set depth integral flags according to tracers available */

          for (m=0;m<N_TRACER;m++) iztracerflag[m]=tracerflag[m];

          /* Diagnose additional tracers & ratios */

          calcdiagnostics(0,diagratioflag,diagtracerflag,NULL,NULL,nkwet,zlevdz,
                          tracer,ratio,NULL,NULL,NULL,it_agg,NULL,&vamap);

          calciz(iztracerflag,NULL,nkwet,zlevdz,tracer,NULL,iztracer,NULL);

          prevdiagflag=1;
   
          /* Save beginning-of-time-step tracers and ratios if potentially
             needed for calculating diagnostics at first time step. */

          if (!leapfrogflag)
            { for (k=0;k<nk;k++)
                { for (m=0;m<N_TRACER;m++) tracer0[m][k]=tracer[m][k];
                  for (m=0;m<N_RATIO;m++) ratio0[m][k]=ratio[m][k];
                }
            }

          /* Write initial state */

          frac=(double)hfirst/nstep_day; 
          t=tstart+frac;
          tdays=day-1+frac;
          thours=frac*24;
          tmid=OBJF_NODATA;
 
          if (outfile[OUTSCALAR])
            { *status=writemod_scal(OUTSCALAR,outfile[OUTSCALAR],flushflag,
                        tabentry,t,tmid,year,day,hfirst,tdays,mth,date,
                        thours,zlevmid[0],zlevbot[0],fscal,fprof,tracer,ratio,
                        diagprof,diagscal,iztracer,izdiag,errtxt);
              if (*status!=OK) return -1;
            }  
                
          if (outfile[OUTPROFILE])
            { *status=writemod_prof(OUTPROFILE,outfile[OUTPROFILE],flushflag,
                        tabentry,t,tmid,year,day,hfirst,tdays,mth,date,
                        thours,nk,zlevmid,zlevtop,zlevbot,zlevdz,fprof,tracer,
                        ratio,diagprof,errtxt);
              if (*status!=OK) return -1;
            }

          /* Reset sums for averaging in accumulators */

          if (timeavoutputreq)
            { resetaccumulator(&dsum,0,nkwet,0.0);
              resetaccumulator(&dsum,nkwet,nk-nkwet,OBJF_NODATA);
            }
          if (mthoutputreq || yearoutputreq)
            { resetaccumulator(&msum,0,nkwet,0.0);
              resetaccumulator(&msum,nkwet,nk-nkwet,OBJF_NODATA);
            }
          if (yearoutputreq)
            { resetaccumulator(&ysum,0,nkwet,0.0);
              resetaccumulator(&ysum,nkwet,nk-nkwet,OBJF_NODATA);
            }

          /* Set deepest level for which accumulator updates are needed */

          if (outfile[OUTPROFILE_DAV] || outfile[OUTPROFILE_MAV] || 
              outfile[OUTPROFILE_YAV]) nkavout=nkwet;
          else nkavout=1;
        }

      /* --- 2.8. GET CONSTANT FORCING DATA & PREPARE FOR MODEL RUN --- */ 

      if (tfin>tstart) 

        { /* Initialize forcing data from environment constants if required
             (default=0 if not present or input data has been deactivated by 
             setting the forcing data mode to 'NOTAPPLIED' or for relaxation 
             rate default is from 'option' parameters).

             Some forcing variables require special treatment: 
              - For data assimilation forcing, a copy is kept to refresh the
                variables at assimilation times since they are blanked out at 
                time steps where not used. 
              - Modify relaxation rate if -ve or standardize rate if above
                the upper limit. The latter is taken to indicate complete 
                override by external data. 
              - Set vertical velocity and diffusion coeff. to zero at max. 
                depth - transport schemes do not allow for bottom slopes */
        
          /* Scalar forcing */
 
          for (m=0;m<N_FSCALAR;m++) if (fconstreq[m])

            { if (fscalmode[m]==NOTAPPLIED ||
                  !tabentry->item[ENVIRON]) fscal[m]=0;
              else
                { fscal[m]=tabentry->item[ENVIRON]->parm[N_ENVIRON+m];
                  if (fscal[m]==OBJF_NODATA) fscal[m]=0;
                }
            }

          for (m=N_FSCALAR_NONDA;m<N_FSCALAR;m++) fscalconst[m]=fscal[m];

          /* Profile forcing (excluding reference profiles; already done) */

          for (m=0;m<REFS;m++) if (fconstreq[m+N_FSCALAR]) 

            { if (fprofmode[m]==NOTAPPLIED || 
                  !tabentry->item[ENVIRON] ||
                  (fprof[m][0]=
                   tabentry->item[ENVIRON]->parm[N_ENVIRON+N_FSCALAR+m])
                    ==OBJF_NODATA) 

                { if (m>=RLXRS && m-RLXRS<N_TRACER_PRI) 
                    fprof[m][0]=relaxrate[m-RLXRS]; 
                  else fprof[m][0]=0;
                }

              else if (m>=RLXRS && m-RLXRS<N_TRACER_PRI)
                { if (fprof[m][0]<0) fprof[m][0]=0;
                  else if (fprof[m][0]>RELAXULIM) fprof[m][0]=RELAXOVERRIDE;
                }

              if (fproffromtop[m])
                for (k=1;k<nkwet;k++) fprof[m][k]=fprof[m][0];
            }

          if (fconstreq[VDC+N_FSCALAR]) fprof[VDC][nkwet-1]=0;
          if (fconstreq[W+N_FSCALAR]) fprof[W][nkwet-1]=0;

          /* Set time at middle of first time step and
             find first observation (within specified time period) */

          t=tstart+(hfirst+0.5)/nstep_day;
          tfinobs=tfin+0.5/nstep_day;
          obsfound=nextobs(tabentry->item[OBS],1,t,tfinobs,maxdepth,&obs);

          /* Initialize data assimilation flags */ 

          assimnext=0;
          assimupd=0;
        }

      /* --- 2.9. RUN MODEL FOR SPECIFIED TIME PERIOD --- */

      firststep=1;
      firststep_lf=leapfrogflag;
      rdfset=0;

      /* If using leapfrog time-stepping scheme a single forward Euler 
         step is needed to get started */

      if (leapfrogflag)

        { /* Set time at forcing time point = middle of time step. */

          frac=(hfirst+0.5)/nstep_day;
          t=tstart+frac; 
          tdays=day-1+frac;

          /* Interpolate forcing data to forcing time point for all
             forcing variables not needing special treatment
             (If forcing is not year specific then use time of year) */

          if (periodicflag) tf=tdays;
          else tf=t;

          getforcing_scal(tf,periodicflag,yearlen,tabentry->item,
                          0,N_FSCALAR_NONDA,fscalfound,fscalreq,fscal);

          getforcing_fulldep(tf,periodicflag,yearlen,firststep,
                             tabentry->item,0,REFS,fscalfound,fscalreq,
                             fproffound,fprofreq,fproftop,
                             fproffromtop,nkwet,fprof);

          /* Reference tracer profiles used as forcing variables need
             special treatment since they must be consistent with the model
             state at the time level used in the tendency calculation.
             Normally this is the start of the time step (for stability)
             but end-of-time-step reference profiles are needed where the
             relaxation rate indicates override.
             Forcing time point is end of time step; for normal relaxation use
             values from previous time step. */

          for (m=REFS;m<REFS+N_TRACER_PRI;m++)
            for (k=0;k<nkwet;k++)
              fprof0[m][k]=fprof[m][k]; 

          tf+=0.5/nstep_day;
          getforcing_fulldep(tf,periodicflag,yearlen,firststep,
                             tabentry->item,REFS,N_TRACER_PRI,fscalfound,
                             fscalreq,fproffound,fprofreq,fproftop,
                             fproffromtop,nkwet,fprof);

          /* Run model for one time step with no output */

          tofyr=tdays;
          fprofpert=&fprof[PERT];
          fprofpertsd=&fprof[PERTSD];
          fprofrlx=&fprof[RLXRS];
          fprofref0=&fprof0[REFS];
          fprofref=&fprof[REFS];

          FORTRAN_modelstep(&tofyr,&yearlen,
            &fscal[SOL],&fscal[SOLAV],&fscal[MLD],&fscal[SSS],
            &fscal[PCO2ATM],&fscal[DUST],&fscal[FEDUST],fprof[TEMP],
            fprof[VDC],fprof[W],fprofpert,fprofpertsd,fprofrlx,
            fprofref0,fprofref,NULL,NULL,NULL,NULL,tracer,ratio,NULL,NULL);

          firststep=0;
        }

      for (j=0;j<tfin-tstart;j++)

        { /* Initialize/reset daily sums for averaging purposes.
             Daily means are calculated from end-of-time-step output
             taking into account time of validity and implicit variation
             between valid times.
             Time of validity for all tracer variables and composition ratios,
             including reference tracer values used as forcing data, is end 
             of time step.
             Concentrations are assumed to vary linearly during the
             time step. 
             Time of validity for all forcing variables specific to a 
             data assimilation analysis is also end of time step. (These 
             variables are only defined at the end of the assimilation time 
             step.)
             Other forcing variables and non-tracer diagnostics are
             constant for the time step and time of validity is mid-step.
             If leapfrog time-stepping scheme is used, the time of validity
             for non-tracer forcing and diagnostics becomes the middle of
             a double time step which is the end of a standard time step
             and these values are treated as varying linearly between valid 
             times for averaging purposes. */

          if (timeavoutputreq) 

            { updatesums_scal(N_FSCALAR_DA,&dsum.fscal[CCHLDA],0,
                              &fscal[CCHLDA],0.5); 
              updatesums_prof(N_TRACER_PRI,nkavout,&dsum.fprof[REFS],0,
                              &fprof[REFS],0.5);
              updatesums_prof(N_TRACER,nkavout,dsum.tracer,0,tracer,0.5);
              updatesums_prof(N_RATIO,nkavout,dsum.ratio,0,ratio,0.5);
              updatesums_scal(N_TRACER,dsum.iztracer,0,iztracer,0.5);

              if (leapfrogflag)
                { updatesums_scal(N_FSCALAR-N_FSCALAR_DA,dsum.fscal,0,
                                  fscal,0.5);
                  updatesums_prof(REFS,nkavout,dsum.fprof,0,fprof,0.5);
                  updatesums_scal(N_DIAG_SCAL,dsum.diagscal,0,diagscal,0.5);
                  updatesums_prof(N_DIAG_PROF,nkavout,dsum.diagprof,0,diagprof,
                                  0.5);
                  updatesums_scal(N_IZ_DIAG,dsum.izdiag,0,izdiag,0.5);
                }
            }

          /* Run model for one day */ 

          for (h=hfirst;h<nstep_day;h++) 

            { /* Set time at forcing time point, which is middle of time step. 
                 If using leapfrog scheme then this is the middle of a 
                 double time step that spans the current output time 
                 (i.e. forcing time point = current output time = 
                 time at end of standard time step). */

              if (leapfrogflag)
                { frac=(double)(h+1)/nstep_day;
                  t=tstart+j+frac;
                  tdays=day-1+frac;
                }
              else
                { frac=(h+0.5)/nstep_day; 
                  t=tstart+j+frac;
                  tdays=day-1+frac;
                }

              /* Interpolate forcing data to forcing time point for all
                 forcing variables not needing special treatment 
                 (If forcing is not year specific then use time of year) */

              if (periodicflag) tf=tdays;
              else tf=t;

              getforcing_scal(tf,periodicflag,yearlen,tabentry->item,
                              0,N_FSCALAR_NONDA,fscalfound,fscalreq,fscal);

              getforcing_fulldep(tf,periodicflag,yearlen,firststep,
                                 tabentry->item,0,REFS,fscalfound,fscalreq,
                                 fproffound,fprofreq,fproftop,
                                 fproffromtop,nkwet,fprof);

              /* Reference tracer profiles used as forcing variables need
                 special treatment since they must be consistent with the model
                 state at the time level used in the tendency calculation.
                 Normally this is the start of the time step (for stability)
                 but end-of-time-step reference profiles are needed where the 
                 relaxation rate indicates override.  
                 Forcing time point is end of time step (i.e. the output time);
                 for normal relaxation use values from previous time step. 

                 Forcing time point:
                  - in forward Euler scheme, half a step on from the middle of
                    the current time step. 
                  - In leapfrog scheme, middle of double time step (as for 
                    other forcing data). 
                 In the leapfrog scheme, we already have these profiles for 
                 the first double time step so we can skip this initially. */

              if (!firststep_lf) 

                { for (m=REFS;m<REFS+N_TRACER_PRI;m++)
                    for (k=0;k<nkwet;k++)
                      fprof0[m][k]=fprof[m][k];

                  if (!leapfrogflag) tf+=0.5/nstep_day;

                  getforcing_fulldep(tf,periodicflag,yearlen,firststep,
                                     tabentry->item,REFS,N_TRACER_PRI,
                                     fscalfound,fscalreq,fproffound,fprofreq,
                                     fproftop,fproffromtop,nkwet,fprof);
                }

              /* Data assimilation forcing variables need special treatment:
                 They are interpolated to the output time rather than 
                 the middle of the current time step. This is the analysis time
                 for steps in which assimilation occurs (i.e. the last model 
                 time step of each assimilation time step).
                 In the leapfrog scheme the forcing is still interpolated to 
                 the analysis time but the assimilation increments are applied 
                 to both the output tracer values and the tracer values at the 
                 end of the current double time step */
              
              if (assimnext)

                { for (m=N_FSCALAR_NONDA;m<N_FSCALAR;m++) 
                    fscal[m]=fscalconst[m]; 

                  if (leapfrogflag) frac=0;
                  else frac=0.5/nstep_day;
                  if (periodicflag) tf=tdays+frac;
                  else tf=t+frac;

                  getforcing_scal(tf,periodicflag,yearlen,tabentry->item,
                    N_FSCALAR_NONDA,N_FSCALAR_DA,fscalfound,fscalreq,fscal);
                }

              else 
                for (m=N_FSCALAR_NONDA;m<N_FSCALAR;m++) fscal[m]=OBJF_NODATA;

              if (modoutputreq || tabentry->item[OBS])

                { /* Note: the following forcing data modifications reflect
                     MarMOT's interpretation of the unmodified forcing data 
                     i.e. the simulation should be the same whether they are 
                     applied or not.
                     They are included here to ensure that the forcing data 
                     in the output is consistent with the dynamics. */ 
 
                  /* Set vertical velocity and diffusion to zero at max. depth. 
                     Transport routines do not allow for a bottom slope 
                     so the input values are not used. */

                  if (fprofreq[VDC] || 
                      (fscalreq[VDC+N_FSCALAR] && fproffromtop[VDC]))
                    fprof[VDC][nkwet-1]=0;

                  if (fprofreq[W] || (fscalreq[W+N_FSCALAR] && fproffromtop[W]))
                    fprof[W][nkwet-1]=0;

                  /* Modify relaxation rate to remove -ve rates and standardize 
                     rates above the upper limit. The latter are taken to 
                     indicate complete override by external data. */

                  for (m=RLXRS;m<RLXRS+N_TRACER_PRI;m++)
                  if (fprofreq[m] || (fscalreq[m+N_FSCALAR] && fproffromtop[m]))
                    for (k=0;k<nkwet;k++) 
                      { if (fprof[m][k]>RELAXULIM) fprof[m][k]=RELAXOVERRIDE;
                        else if (fprof[m][k]<0) fprof[m][k]=0;
                      }
                }

              /* Run model for one time step and get diagnostics if required 
                 for output table or misfit calculation
                 (Observation search window is from middle of time step to
                 middle of next time step) */

              tofyr=tdays;
              fprofpert=&fprof[PERT]; 
              fprofpertsd=&fprof[PERTSD]; 
              fprofrlx=&fprof[RLXRS]; 
              fprofref0=&fprof0[REFS];
              fprofref=&fprof[REFS];

              FORTRAN_modelstep(&tofyr,&yearlen,
                &fscal[SOL],&fscal[SOLAV],&fscal[MLD],&fscal[SSS],
                &fscal[PCO2ATM],&fscal[DUST],&fscal[FEDUST],fprof[TEMP],
                fprof[VDC],fprof[W],fprofpert,fprofpertsd,fprofrlx,
                fprofref0,fprofref,
                &fscal[CCHLDA],&fscal[CHLDA],&fscal[PGROWDA],&fscal[PLOSSDA],
                tracer,ratio,&assimnext,&assimupd);

              firststep=0;
              firststep_lf=0;

              /* Set time at end of time step */

              frac=(double)(h+1)/nstep_day;
              t=tstart+j+frac;
              tdays=day-1+frac;
              thours=frac*24;

              /* Write details of any data assimilation updates */

              if (assimoutputreq && assimupd)

                { FORTRAN_getassimreport(assimprof,assimscal);

                  if (outfile[OUTASSIMSCAL])
                    *status=writeassim_scal(OUTASSIMSCAL,
                      outfile[OUTASSIMSCAL],flushflag,tabentry,
                      t,year,day,h+1,tdays,mth,date,thours,
                      zlevmid[0],zlevbot[0],assimprof,assimscal,errtxt);
                  if (*status!=OK) return -1;

                  if (outfile[OUTASSIMPROF])
                    *status=writeassim_prof(OUTASSIMPROF,
                      outfile[OUTASSIMPROF],flushflag,tabentry,
                      t,year,day,h+1,tdays,mth,date,thours,nk,
                      zlevmid,zlevtop,zlevbot,zlevdz,assimprof,errtxt);
                  if (*status!=OK) return -1;
                }

              /* Get diagnostics if needed for output or misfit calcs. 
                 Tracer and ratio values from the previous time step may be
                 needed but some of these are diagnosed values and may not be 
                 available. 'prevdiagflag' indicates availability; if not 
                 available the diagnosed tracer and ratio values must be 
                 calculated prior to the other diagnostics, in addition to 
                 those for the current time step. */

              tmid=tstart+j+(h+0.5)/nstep_day;
              tmidnext=tmid+1./nstep_day;
              misfitreq=(obsfound && obs.t<tmidnext); 

              if (modoutputreq || misfitreq)

                { FORTRAN_getdiagnostics(diagprof,diagscal);

                  if (model!=NULLMOD)

                    { if (!rdfset) 
                        { setdiagflags_runtime(diagprof,id_agg,tracerflag,
                                               ratioflag,&vamap,diagprofflag);
                          rdfset=1;
                        }

                      calcdiagnostics(prevdiagflag,diagratioflag,
                                      diagtracerflag,diagprofflag,&solver,
                                      nkwet,zlevdz,tracer,ratio,tracer0,ratio0,
                                      diagprof,it_agg,id_agg,&vamap);

                      izdiagflag[IZPPROD]=diagprofflag[PPROD];
                      izdiagflag[IZFESCAV]=diagprof[FESCAV][0]!=OBJF_NODATA;
                    }

                  calciz(iztracerflag,izdiagflag,nkwet,zlevdz,tracer,diagprof,
                         iztracer,izdiag);

                  prevdiagflag=1;
                }

              else prevdiagflag=0;
 
              /* Save beginning-of-time-step tracers and ratios if potentially
                 needed for calculating diagnostics at next time step. */

              if (!leapfrogflag && (modoutputreq || tabentry->item[OBS]))
                { for (k=0;k<nk;k++)
                    { for (m=0;m<N_TRACER;m++) tracer0[m][k]=tracer[m][k];
                      for (m=0;m<N_RATIO;m++) ratio0[m][k]=ratio[m][k];
                    }
                }

              /* Determine misfits:
                 process any observation records for current time step */

              while (misfitreq)

                { calcmisfit(&obs,nkwet,zlevmid,zlevbot,maxdepth,fscal,fprof,
                             tracer,ratio,diagprof,diagscal,iztracer,izdiag);
                  valid=0; 
                  for (m=0;m<N_OBSTYPE;m++) if (obs.misfitflag[m])
                    { casecost.mfitcount[m]++;
                      casecost.mfitsum[m]+=obs.misfit[m];
                      valid=1;
                    }

                  if (valid && outfile[OUTMISFIT])
                    { *status=
                        writemisfit(OUTMISFIT,outfile[OUTMISFIT],flushflag,
                          tabentry,t,tmid,year,day,h+1,tdays,mth,date,thours,
                          yearlen,nstep_day,&obs,errtxt);
                      if (*status!=OK) return -1;
                    }

                  obsfound=nextobs(tabentry->item[OBS],0,tmid,tfinobs,maxdepth,
                                   &obs);
                  misfitreq=(obsfound && obs.t<tmidnext); 
                }

              /* Write end of time step data */

              if (outfile[OUTSCALAR])
                { *status=
                     writemod_scal(OUTSCALAR,outfile[OUTSCALAR],flushflag,
                       tabentry,t,tmid,year,day,h+1,tdays,mth,date,
                       thours,zlevmid[0],zlevbot[0],fscal,fprof,tracer,ratio,
                       diagprof,diagscal,iztracer,izdiag,errtxt);
                  if (*status!=OK) return -1;
                }  

              if (outfile[OUTPROFILE])
                { *status=
                     writemod_prof(OUTPROFILE,outfile[OUTPROFILE],flushflag,
                       tabentry,t,tmid,year,day,h+1,tdays,mth,date,
                       thours,nk,zlevmid,zlevtop,zlevbot,zlevdz,fprof,tracer,
                       ratio,diagprof,errtxt);
                  if (*status!=OK) return -1;
                }  

              /* Accumulate daily sums for averaging */

              if (timeavoutputreq) 

                { if (h==nstep_day-1) w=0.5;
                  else w=1;

                  updatesums_scal(N_FSCALAR_DA,&dsum.fscal[CCHLDA],1,
                                  &fscal[CCHLDA],w);
                  updatesums_prof(N_TRACER_PRI,nkavout,&dsum.fprof[REFS],1,
                                  &fprof[REFS],w);
                  updatesums_prof(N_TRACER,nkavout,dsum.tracer,1,tracer,w);
                  updatesums_prof(N_RATIO,nkavout,dsum.ratio,1,ratio,w);
                  updatesums_scal(N_TRACER,dsum.iztracer,1,iztracer,w);

                  if (leapfrogflag) 
                    { updatesums_scal(N_FSCALAR-N_FSCALAR_DA,dsum.fscal,1,
                                      fscal,w);
                      updatesums_prof(REFS,nkavout,dsum.fprof,1,fprof,w);
                      updatesums_scal(N_DIAG_SCAL,dsum.diagscal,1,diagscal,w);
                      updatesums_prof(N_DIAG_PROF,nkavout,dsum.diagprof,1,
                                      diagprof,w);
                      updatesums_scal(N_IZ_DIAG,dsum.izdiag,1,izdiag,w);
                    }
                  else
                    { updatesums_scal(N_FSCALAR-N_FSCALAR_DA,dsum.fscal,1,
                                      fscal,1);
                      updatesums_prof(REFS,nkavout,dsum.fprof,1,fprof,1);
                      updatesums_scal(N_DIAG_SCAL,dsum.diagscal,1,diagscal,1);
                      updatesums_prof(N_DIAG_PROF,nkavout,dsum.diagprof,1,
                                      diagprof,1);
                      updatesums_scal(N_IZ_DIAG,dsum.izdiag,1,izdiag,1);
                    }
                }
            }

          /* Write daily means and accumulate for monthlies
             (skip if not a full day) */

          if (hfirst==0)

            { t=tstart+j+0.5;
              tdays=day-0.5;
              thours=12;

              if (outfile[OUTSCALAR_DAV])
                { *status=
                     writemodav_scal(OUTSCALAR_DAV,outfile[OUTSCALAR_DAV],
                       flushflag,tabentry,t,year,day,tdays,mth,date,
                       thours,zlevmid[0],zlevbot[0],&dsum,nstep_day,errtxt);
                  if (*status!=OK) return -1;
                } 

              if (outfile[OUTPROFILE_DAV])
                { *status=
                     writemodav_prof(OUTPROFILE_DAV,outfile[OUTPROFILE_DAV],
                       flushflag,tabentry,t,year,day,tdays,mth,date,thours,nk,
                       zlevmid,zlevtop,zlevbot,zlevdz,&dsum,nstep_day,errtxt);
                  if (*status!=OK) return -1;
                }

              if (mthoutputreq)
                updaccumulator(&msum,&dsum,nkavout,
                               outfile[OUTSCALAR_MAV]!=NULL);

              if (yearoutputreq)
                updaccumulator(&ysum,&dsum,nkavout,
                               outfile[OUTSCALAR_YAV]!=NULL);
            }

          /* Check what month it is tomorrow */

          if (modoutputreq || assimoutputreq || misfitoutputreq)
            { if (day==yearlen) newmth=monthofyear(fixedyearlen,yearlen,1);
              else newmth=monthofyear(fixedyearlen,yearlen,day+1); 
            }

          /* If end of month, write monthly means */

          if (mthoutputreq && newmth!=mth)

            { mthlen=daysinmonth(yearlen,mth);   
              t=tstart+j+1-0.5*mthlen;
              tdays=day-0.5*mthlen;
              mthlen_steps=mthlen*nstep_day;

              if (outfile[OUTSCALAR_MAV] && fullmth && mthlen>0)
                { *status=writemodav_scal(OUTSCALAR_MAV,outfile[OUTSCALAR_MAV],
                            flushflag,tabentry,t,year,0,tdays,mth,0,0,
                            zlevmid[0],zlevbot[0],&msum,mthlen_steps,errtxt);
                  if (*status!=OK) return -1;
                }

              if (outfile[OUTPROFILE_MAV] && fullmth && mthlen>0)
                { *status=writemodav_prof(OUTPROFILE_MAV,
                            outfile[OUTPROFILE_MAV],flushflag,tabentry,t,
                            year,0,tdays,mth,0,0,nk,zlevmid,zlevtop,
                            zlevbot,zlevdz,&msum,mthlen_steps,errtxt);
                  if (*status!=OK) return -1;
                }

              resetaccumulator(&msum,0,nkavout,0.0);
              fullmth=1;
            }

          /* if end of year, write annual means */

          if (yearoutputreq && day==yearlen)

            { t=tstart+j+1-0.5*yearlen;
              tdays=day-0.5*yearlen;
              yearlen_steps=yearlen*nstep_day; 

              if (outfile[OUTSCALAR_YAV] && fullyear)
                { *status=writemodav_scal(OUTSCALAR_YAV,
                            outfile[OUTSCALAR_YAV],flushflag,tabentry,t,
                            year,0,tdays,0,0,0,zlevmid[0],zlevbot[0],
                            &ysum,yearlen_steps,errtxt);
                  if (*status!=OK) return -1;
                }

              if (outfile[OUTPROFILE_YAV] && fullyear)
                { *status=writemodav_prof(OUTPROFILE_YAV,
                            outfile[OUTPROFILE_YAV],flushflag,tabentry,t,
                            year,0,tdays,0,0,0,nk,zlevmid,zlevtop,
                            zlevbot,zlevdz,&ysum,yearlen_steps,errtxt);
                  if (*status!=OK) return -1;
                }

              resetaccumulator(&ysum,0,nkavout,0.0);
              fullyear=1;
            }

          /* Setup for next day */

          day++;
          hfirst=0;

          if (day>yearlen)
            { year++;
              day=1;
              yearlen=daysinyear(fixedyearlen,year);
            }

          if (modoutputreq || assimoutputreq || misfitoutputreq)
            { mth=newmth;
              date=dayofmonth(fixedyearlen,yearlen,day);
            }

          resetaccumulator(&dsum,0,nkavout,0.0);
        }

      /* --- 2.10 END SIMULATION CASE --- */

      /* Write case summary statistics */ 

      if (outfile[OUTCOST]) 
        { *status=writecost(OUTCOST,outfile[OUTCOST],flushflag,tabentry,
                            &casecost);
          if (*status!=OK) return -1;
        }

      /* Update cost function */

      for (m=0;m<N_OBSTYPE;m++)
        { cost.mfitsum[m]+=casecost.mfitsum[m];
          cost.mfitcount[m]+=casecost.mfitcount[m];
        }
    }

  /* --- 3. EVALUATE COST FUNCTION --- */

  return costfunc(&cost,NULL);
}


static int daysinyear (int fixedyearlen, int y)

{ if (fixedyearlen>0) return fixedyearlen; 

  if ((double)y/4==y/4 && 
      ((double)y/100!=y/100 || 
       (double)y/400==y/400)) return 366;
  else return 365;
}


static int monthofyear (int fixedyearlen, int yearlen, int day)

{ int leap;

  if (fixedyearlen && (yearlen<365 || yearlen>366))
    { if (yearlen==360) return floor((day-0.5)/30)+1;
      else return 0; 
    }

  if (day<=31) return 1;
  if (day<=59) return 2;

  leap=(yearlen==366);

  if (day<=59+leap) return 2;
  if (day<=90+leap) return 3;
  if (day<=120+leap) return 4;
  if (day<=151+leap) return 5;
  if (day<=181+leap) return 6;
  if (day<=212+leap) return 7;
  if (day<=243+leap) return 8;
  if (day<=273+leap) return 9;
  if (day<=304+leap) return 10;
  if (day<=334+leap) return 11;
  return 12 ;
}


static int dayofmonth (int fixedyearlen, int yearlen, int day)

{ int leap;

  if (fixedyearlen && (yearlen<365 || yearlen>366))
    { if (yearlen==360) return day-30*floor((day-0.5)/30);
      else return 0; 
    }

  if (day<=31) return (int)day;
  if (day<=59) return (int)(day-31);

  leap=(yearlen==366);

  if (day<=59+leap) return (int)(day-31);
  if (day<=90+leap) return (int)(day-59-leap);
  if (day<=120+leap) return (int)(day-90-leap);
  if (day<=151+leap) return (int)(day-120-leap);
  if (day<=181+leap) return (int)(day-151-leap);
  if (day<=212+leap) return (int)(day-181-leap);
  if (day<=243+leap) return (int)(day-212-leap);
  if (day<=273+leap) return (int)(day-243-leap);
  if (day<=304+leap) return (int)(day-273-leap);
  if (day<=334+leap) return (int)(day-304-leap);
  return (int)(day-334-leap);
}


static int daysinmonth (int yearlen, int mth)

{ /* For valid 'mth' assume 'yearlen' is 360, 365 or 366 */

  if (mth<1) return 0;
  else if (yearlen==360) return 30;
  else if (mth==2)
    { if (yearlen==366) return 29;
      else return 28;
    }
  else if (mth==4 || mth==6 || mth==9 || mth==11) return 30;
  else return 31;
}


static void getforcing_scal (double t, int periodic, int periodlen, 
                             struct itemrec *item[], int ivar, int nvar, 
                             unsigned char varfound[][N_FSCALAR+N_FPROF], 
                             enum fmodetype freq[], double fval[])

/* Gets scalar forcing data values for time 't' from depth-invariant 
   input items */

{ int i,inext,j,jf,m;
  double w1;

  for (jf=0;jf<N_FSCALITEM;jf++) if (item[FSCALAR+jf])

    { j=FSCALAR+jf;
  
      if (periodic) 
        i=findex_periodic(t,periodlen,&item[j]->gspec[TIME],&w1,&inext);
      else
        { i=findex(t,&item[j]->gspec[TIME],&w1);
          inext=i+1;
        }

      for (m=ivar;m<ivar+nvar;m++) if (freq[m] && varfound[jf][m])
        { if (i==-1) fval[m]=item[j]->value[m][0];
          else if (w1==0) 
            fval[m]=item[j]->value[m][i];
          else if (freq[m]==INTERPOLATED)
            fval[m]=(1-w1)*item[j]->value[m][i]+w1*item[j]->value[m][inext];
          else if (freq[m]==NEAREST)
            { if (w1<=0.5) fval[m]=item[j]->value[m][i];
              else fval[m]=item[j]->value[m][inext];
            }
          else if (freq[m]==PERSISTENT) fval[m]=item[j]->value[m][i];
        }
    }
}


static void getforcing_fulldep 
              (double t, int periodic, int periodlen, int first, 
               struct itemrec *item[], int ivar, int nvar,
               unsigned char fscalfound[][N_FSCALAR+N_FPROF],
               enum fmodetype fscalreq[], unsigned char fproffound[][N_FPROF],
               enum fmodetype fprofreq[], unsigned char fproftop[],
               unsigned char fproffromtop[], int nk, double fprof[][NK_MAX])

/* Gets profile forcing data values from available input items */

{ int k,m;

  getforcing_prof_fromscal(t,periodic,periodlen,item,ivar,nvar,
                           fscalfound,fscalreq,fprof);

  getforcing_prof(t,periodic,periodlen,first,item,ivar,nvar,
                  fproffound,fprofreq,nk,fprof);

  for (m=ivar;m<ivar+nvar;m++)
    { if (fprofreq[m] && fproftop[m])
        for (k=1;k<nk;k++) fprof[m][k]=OBJF_NODATA;
      if (fscalreq[m+N_FSCALAR] && fproffromtop[m])
        for (k=1;k<nk;k++) fprof[m][k]=fprof[m][0];
    }
}


static void getforcing_prof_fromscal 
              (double t, int periodic, int periodlen,
               struct itemrec *item[], int ivar, int nvar,
               unsigned char varfound[][N_FSCALAR+N_FPROF],
               enum fmodetype freq[], double fprofval[][NK_MAX])

/* Gets profile forcing data values for time 't' from depth-invariant 
   input items - only the top level is filled */

{ int i,inext,j,jf,m,ms;
  double w1;

  for (jf=0;jf<N_FSCALITEM;jf++) if (item[FSCALAR+jf])

    { j=FSCALAR+jf;
  
      if (periodic) 
        i=findex_periodic(t,periodlen,&item[j]->gspec[TIME],&w1,&inext);
      else
        { i=findex(t,&item[j]->gspec[TIME],&w1);
          inext=i+1;
        }

      for (m=ivar;m<nvar;m++) 
        if (freq[N_FSCALAR+m] && varfound[jf][N_FSCALAR+m])

          { ms=N_FSCALAR+m;
            if (i==-1) fprofval[m][0]=item[j]->value[ms][0];
            else if (w1==0) 
              fprofval[m][0]=item[j]->value[ms][i];
            else if (freq[ms]==INTERPOLATED)
              fprofval[m][0]=
                (1-w1)*item[j]->value[ms][i]+w1*item[j]->value[ms][inext];
            else if (freq[ms]==NEAREST)
              { if (w1<=0.5) fprofval[m][0]=item[j]->value[ms][i];
                else fprofval[m][0]=item[j]->value[ms][inext];
              }
            else if (freq[ms]==PERSISTENT) fprofval[m][0]=item[j]->value[ms][i];
          }
    }
}


static void getforcing_prof (double t, int periodic, int periodlen, 
                             int first, struct itemrec *item[], 
                             int ivar, int nvar,
                             unsigned char varfound[][N_FPROF], 
                             enum fmodetype freq[], int nk, 
                             double fval[][NK_MAX])
    
/* Gets profile forcing data values for time 't' from depth-dependent
   input items */

{ double w1;
  int base0,base1;
  int nzlevel,i,inext,j,jf,k,m;

  for (jf=0;jf<N_FPROFITEM;jf++) if (item[FPROF+jf])

    { j=FPROF+jf;
 
      if (!item[j]->dimflag[TIME] || item[j]->gspec[TIME].n==1) 

        { /* Assign time-invariant profile */

          if (first)  
            { for (m=ivar;m<ivar+nvar;m++) if (freq[m] && varfound[jf][m])
                for (k=0;k<nk;k++) fval[m][k]=item[j]->value[m][k];
            }
        }

      else

        { /* Locate profile(s) and interpolate or assign as required */

          nzlevel=item[j]->gspec[LEVEL].n;
          if (periodic) 
            i=findex_periodic(t,periodlen,&item[j]->gspec[TIME],&w1,&inext);
          else
            { i=findex(t,&item[j]->gspec[TIME],&w1);
              inext=i+1;
            }

          base0=i*nzlevel;
          base1=(inext)*nzlevel;
          for (m=ivar;m<ivar+nvar;m++) if (freq[m] && varfound[jf][m])
            { if (i==-1) 
                for (k=0;k<nk;k++) fval[m][k]=item[j]->value[m][k]; 
              else if (w1==0) 
                for (k=0;k<nk;k++) fval[m][k]=item[j]->value[m][base0+k];
              else if (freq[m]==INTERPOLATED)
                for (k=0;k<nk;k++) 
                  fval[m][k]=(1-w1)*item[j]->value[m][base0+k]+
                             w1*item[j]->value[m][base1+k];
              else if (freq[m]==NEAREST)
                { if (w1<=0.5) 
                    for (k=0;k<nk;k++) fval[m][k]=item[j]->value[m][base0+k];
                  else 
                    for (k=0;k<nk;k++) fval[m][k]=item[j]->value[m][base1+k];
                }
              else if (freq[m]==PERSISTENT) 
                for (k=0;k<nk;k++) fval[m][k]=item[j]->value[m][base0+k];
            }
        }
    }
}


static int findex (double t, struct gridspecrec *g, double *frac)

/* Returns index of data point at or before 't' or -1 if none.
   'frac' contains proximity (as fraction of data interval) to use for 
   interpolation if index is in the range 0..n-2, otherwise 't' is outside 
   data range (or at upper limit) and 'frac' is set to zero. */ 
    
{ double conv,offset;
  int i;

  if (g->n==1) 
    { *frac=0;
      if (t>=g->first) return 0;
      else return -1;
    } 

  if (g->iregflag) conv=g->n/(g->max-g->min);
  else conv=(g->n-1)/(g->max-g->min);

  offset=(t-g->first)*conv;
  i=floor(offset);
  *frac=offset-i;
  if (i<0) 
    { *frac=0;
      i=-1;
    }
  else if (i > g->n-2) 
    { *frac=0;
      i=g->n-1;
    }

  return i;
}


static int findex_periodic (double t, int periodlen, struct gridspecrec *g, 
                            double *frac, int *inext)

/* Returns index of data point at or conceptually before 't' or first data 
   point in period if 't' is absent. If 't' is specified:
   'frac' contains proximity (as fraction of data interval).
   'inext' contains index of data point conceptually after t. 
   All data points outside period are ignored.  Returns -1 if no data points
   in period and 'frac' and 'inext' are undefined. */

{ double conv,offset,offsetpstart,offsetpfin,frachead,fractail,fracwrap;
  int i,ipfirst,iplast;

  /* If only 1 data point available then just check if it is in period or not 
     (The return value is independent of 't'; set 'frac and 'inext' by 
     request instead of by presence of 't')  */

  if (g->n==1)
    { if (frac) *frac=0;
      if (inext) *inext=0; 
      if (g->first<0 || g->first>periodlen) return -1;
      else return 0;
    }

  /* Determine fractional index values at period start and end */ 

  if (g->iregflag) conv=g->n/(g->max-g->min);
  else conv=(g->n-1)/(g->max-g->min);

  offsetpstart=-g->first*conv;
  offsetpfin=(periodlen-g->first)*conv;

  /* Start of period (t=0) */

  ipfirst=ceil(offsetpstart);
  if (ipfirst<0) ipfirst=0;
  if (ipfirst>offsetpfin) return -1;
  frachead=ipfirst-offsetpstart;

  /* End of period (t=periodlen) */

  iplast=floor(offsetpfin);
  if (iplast>g->n-1) iplast=g->n-1;
  if (iplast<offsetpstart) return -1;
  fractail=offsetpfin-iplast;

  /* Missing t - just return 'ipfirst' to show that data is present */

  if (t<0) return ipfirst;

  /* Determine size of wrap step in fractions of a grid step 
     If first and last data points in period are within 1/100th of a grid step
     then they are taken to be co-located and the last data point is ignored */

  fracwrap=frachead+fractail;  
  if (iplast>ipfirst && fracwrap<0.01) 
    { iplast--;
      fracwrap+=1;
    }

  /* If only one data value in period then use that for whole period */ 

  if (iplast==ipfirst)
    { *frac=0;
      *inext=ipfirst;
      return ipfirst;
    }

  /* Current time t (assumed positive) - adjust to ensure it is in period */ 

  while (t>periodlen) t-=periodlen;
  offset=(t-g->first)*conv;   
  i=floor(offset);
 
  /* If nearest grid point before t (index i) is outside period then use last 
     grid point in period instead and weight according to fraction of wrap step 
     (ipfirst-offset is number of grid steps before first grid point in period).
     If nearest grid point after t (index i+1) is outside period then use first
     grid point in period instead and weight according to fraction of wrap step
     (offset-iplast is number of grid steps after last grid point in period). */ 

  if (i<ipfirst) 
    { i=iplast;
      *frac=(fracwrap-(ipfirst-offset))/fracwrap;
      *inext=ipfirst;
    }
  else if (i>=iplast)
    { i=iplast;
      *frac=(offset-iplast)/fracwrap; 
      *inext=ipfirst;
    }
  else 
    { *frac=offset-i;
      *inext=i+1;
    }

  return i;
}


static void setdiagflags (int it_agg[][NAGG_MAX], struct vamrec *iv,
                          unsigned char tracerflag[], 
                          unsigned char ratioflag[], 
                          unsigned char diagtracerflag[], 
                          unsigned char diagratioflag[],
                          unsigned char diagprofflag[])

{ int i,j,m,offset;

  /* Model output can include primary tracers, secondary tracers and ratios.
     Categories of diagnosable tracers not output directly by the model:
      - aggregate primary tracers
      - secondary tracers: ratio-derived tracers (i.e. derived from primary 
          tracers using ratios), aggregate or non-aggregate 
      - tertiary tracers: additional tracers derived by aggregation
     Whether tracers are aggregate or non-aggregate is model dependent.
     Pre-requisites for aggregate tracers are determined from a model-dependent
     list of contributing primary tracers. 
     Primary and secondary tracers are collectively referred to as fundamental
     tracers (i.e. tracers that can potentially be aggregated). 
     Diagnosis flags are set to 1 or 2 if diagnosis is required; 2 is used for
     aggregate tracers or ratios thereof or aggregate tracer tendencies.  */

  /* Set flags to diagnose ratios of fundamental tracers present in the 
     model output and update ratio availability flags.
     This is done first to ensure ratio availability flags are finalized for 
     all non-aggregate tracers. They can then be used in assessing 
     diagnosability of aggregate secondary tracers. 
     Diagnosability of C:Chl ratios is checked last to allow for dependency 
     on C:N ratios. */

  for (i=0;i<N_TRACER_PRI;i++)
    { j=iv->c[i];
      m=iv->rc[i];
      if (m>=0 && j>=0 && tracerflag[i] && tracerflag[j])
        { diagratioflag[m]=1;
          ratioflag[m]=1;
        }
    }
  if (tracerflag[PDIAT] && tracerflag[PDIATSI])
    { diagratioflag[SINPDIAT]=1;
      ratioflag[SINPDIAT]=1;
    }
  if (tracerflag[CHL] &&
    (tracerflag[PHYC] || (tracerflag[PHY] && ratioflag[CNPHY])))
    { diagratioflag[CCHL]=1;
      ratioflag[CCHL]=1;
    }
  if (tracerflag[CHLNDIA] &&
    (tracerflag[PNDIAC] || (tracerflag[PNDIA] && ratioflag[CNPNDIA])))
    { diagratioflag[CCHLNDIA]=1;
      ratioflag[CCHLNDIA]=1;
    }
  if (tracerflag[CHLDIAT] &&
    (tracerflag[PDIATC] || (tracerflag[PDIAT] && ratioflag[CNPDIAT])))
    { diagratioflag[CCHLDIAT]=1;
      ratioflag[CCHLDIAT]=1;
    }

  /* Set flags to diagnose aggregate primary tracers from model tracers. 
     These may be needed for diagnosing aggregate tracer ratios. 
     Tracer availability flags are updated. */

  setaggregateflag(PHY,it_agg,NULL,NULL,NULL,tracerflag,diagtracerflag);
  setaggregateflag(ZOO,it_agg,NULL,NULL,NULL,tracerflag,diagtracerflag);

  /* Set flags to diagnose aggregate secondary tracers.
     These may be needed for diagnosing aggregate tracer ratios. 
     They will be diagnosed from secondary tracers. However, diagnosability 
     is determined from availability of primary tracers and ratios since 
     secondary tracer availability flags are not yet finalized. 
     Tracer availability flags are updated here for the aggregate secondary 
     tracers. */ 

  setaggregateflag(PHYC,it_agg,ratioflag,iv->rc,NULL,tracerflag,diagtracerflag);
  setaggregateflag(ZOOC,it_agg,ratioflag,iv->rc,NULL,tracerflag,diagtracerflag);
  setaggregateflag(CHL,it_agg,ratioflag,iv->rc,iv->rcc,tracerflag,
                   diagtracerflag);

  /* Set flags to diagnose aggregate tracer ratios and update ratio 
     availability flags. 
     Diagnosability of C:Chl ratios is checked last to allow for dependency 
     on C:N ratios. */

  if (tracerflag[PHY] && tracerflag[PHYC] &&
       (diagtracerflag[PHY] || diagtracerflag[PHYC])) 
    { diagratioflag[CNPHY]=2;
      ratioflag[CNPHY]=1;
    }

  if (tracerflag[ZOO] && tracerflag[ZOOC] &&
       (diagtracerflag[ZOO] || diagtracerflag[ZOOC]))
    { diagratioflag[CNZOO]=2;
      ratioflag[CNZOO]=1;
    }

  if (tracerflag[CHL] && (
       (tracerflag[PHYC] && (diagtracerflag[CHL] || diagtracerflag[PHYC])) ||
       (tracerflag[PHY] && ratioflag[CNPHY] && 
        (diagtracerflag[CHL] || diagtracerflag[PHY]) )))
    { diagratioflag[CCHL]=2;
      ratioflag[CCHL]=1;
    }

  /* Set flags to diagnose secondary tracers (aggregate or non-aggregate) 
     that are not otherwise available.
     These are only set after assessing diagnosability of ratios; they cannot
     generally be used to derive ratios because the ratios are used in their 
     derivation. (An exception is that carbon tracers can be used to derive 
     chl.)
     Diagnosability of chl. tracers is checked last to allow for dependency 
     on carbon tracers. 
     Tracer availability flags are updated. */

  for (i=0;i<N_TRACER_PRI;i++)
    { j=iv->c[i];
      m=iv->rc[i];
      if (m>=0 && j>=0 && !tracerflag[j])
        { diagtracerflag[j]=tracerflag[i]&&ratioflag[m];
          tracerflag[j]=diagtracerflag[j];
        }
    }
  if (!tracerflag[PDIATSI])
    { diagtracerflag[PDIATSI]=tracerflag[PDIAT] && ratioflag[SINPDIAT];
      tracerflag[PDIATSI]=diagtracerflag[PDIATSI];
    }
  if (!tracerflag[CHL])
    { diagtracerflag[CHL]=tracerflag[PHYC] && ratioflag[CCHL];
      tracerflag[CHL]=diagtracerflag[CHL];
    }
  if (!tracerflag[CHLNDIA])
    { diagtracerflag[CHLNDIA]=tracerflag[PNDIAC] && ratioflag[CCHLNDIA];
      tracerflag[CHLNDIA]=diagtracerflag[CHLNDIA];
    }
  if (!tracerflag[CHLDIAT])
    { diagtracerflag[CHLDIAT]=tracerflag[PDIATC] && ratioflag[CCHLDIAT];
      tracerflag[CHLDIAT]=diagtracerflag[CHLDIAT];
    }

  /* Set flags to diagnose tertiary tracers and update tracer availability 
     flags */ 

  setaggregateflag(NITROGEN,it_agg,NULL,NULL,NULL,tracerflag,diagtracerflag);
  setaggregateflag(CARBON,it_agg,ratioflag,iv->rc,NULL,tracerflag,
                   diagtracerflag);
  setaggregateflag(SILICON,it_agg,ratioflag,iv->rsi,NULL,tracerflag,
                   diagtracerflag);
  setaggregateflag(IRON,it_agg,ratioflag,iv->rfe,NULL,tracerflag,
                   diagtracerflag);
  setaggregateflag(PON,it_agg,NULL,NULL,NULL,tracerflag,diagtracerflag);
  setaggregateflag(POC,it_agg,ratioflag,iv->rc,NULL,tracerflag,diagtracerflag);
  
  /* Set diagnostics flags for tracer tendencies */ 

  for (i=0;i<3;i++)
    { offset=DDTH_TRACERS+i*N_TRACER;
      for (j=0;j<N_TRACER;j++)
        diagprofflag[j+offset]=diagtracerflag[j];
    }
}


static void setdiagflags_runtime(double diagprof[][NK_MAX], 
                                 int id_agg[][NAGG_MAX], 
                                 unsigned char tracerflag[],
                                 unsigned char ratioflag[], 
                                 struct vamrec *iv,
                                 unsigned char diagprofflag[])

{ int prim,i,flag;

  /* Set flag to calculate primary production as an aggregate diagnostic.
     Pre-requisites are determined from  model-dependent list of 1 or more
     contributing primary tracers. */

  setaggregateflag_pprod(id_agg,tracerflag,ratioflag,iv->rc,diagprof,iv->phot,
                         diagprofflag);

  /* Set flags to calculate sinking particle fluxes.
     Slow sinking and fast sinking particles are supported.
     Slow sinking effects are represented by explicit tracer velocities. 
     Fast sinking effects are represented by vertical redistribution of 
     material via outflux and influx rates.
     Slow sinking is not currently supported for silicon and calcium fluxes.
     For other elements, sinking particle flux output requires both the  
     velocity and sinking particle influxes (either of which which may be zero).
     Only one type of slow sinking particle (detritus) is currently 
     supported. Addition of further types will require the sinking particle
     fluxes to become aggregate diagnostics. */
 
  diagprofflag[SINKFLXN]=tracerflag[DET] &&
                         diagprof[iv->w[DET]][0]!=OBJF_NODATA &&
                         diagprof[INSINKN][0]!=OBJF_NODATA;

  diagprofflag[SINKFLXC]=tracerflag[DET] && ratioflag[CNDET] &&
                         diagprof[iv->w[DET]][0]!=OBJF_NODATA &&
                         diagprof[INSINKC][0]!=OBJF_NODATA;

  diagprofflag[SINKFLXSI]=diagprof[INSINKSI][0]!=OBJF_NODATA;

  diagprofflag[SINKFLXFE]=tracerflag[DET] && ratioflag[FEN] &&
                          diagprof[iv->w[DET]][0]!=OBJF_NODATA &&
                          diagprof[INSINKFE][0]!=OBJF_NODATA;

  diagprofflag[SINKFLXCA]=diagprof[INSINKCA][0]!=OBJF_NODATA;

  diagprofflag[DDTSINKN]=diagprofflag[SINKFLXN];
  diagprofflag[DDTSINKC]=diagprofflag[SINKFLXC];
  diagprofflag[DDTSINKSI]=diagprofflag[SINKFLXSI];
  diagprofflag[DDTSINKFE]=diagprofflag[SINKFLXFE];
  diagprofflag[DDTSINKCA]=diagprofflag[SINKFLXCA];
}


static void setaggregateflag (int it, int iagg[][NAGG_MAX], 
                              unsigned char ratioflag[], int ir_assoc1[],
                              int ir_assoc2[], unsigned char tracerflag[],
                              unsigned char diagtracerflag[])

{ int i,prim;

  prim=iagg[it][0]-1;
  diagtracerflag[it]=(prim>=0);
  i=0;
  while (prim>=0)
    { diagtracerflag[it]=diagtracerflag[it]&&tracerflag[prim];
      if (ir_assoc1 && ir_assoc1[prim]>=0)
        diagtracerflag[it]=diagtracerflag[it]&&ratioflag[ir_assoc1[prim]];
      if (ir_assoc2 && ir_assoc2[prim]>=0)
        diagtracerflag[it]=diagtracerflag[it]&&ratioflag[ir_assoc2[prim]];
      if (++i>=NAGG_MAX) prim=-1;
      else prim=iagg[it][i]-1;
    }
  if (diagtracerflag[it])
    { diagtracerflag[it]=2;
      tracerflag[it]=1;
    }
}


static void setaggregateflag_pprod 
              (int iagg[][NAGG_MAX], unsigned char tracerflag[],
               unsigned char ratioflag[], int irc_assoc[],
               double diagprof[][NK_MAX], int iphot_assoc[], 
               unsigned char diagprofflag[])

{ int i,prim;

  prim=iagg[PPROD_AGG][0]-1;
  diagprofflag[PPROD]=(prim>=0);
  i=0;
  while (prim>=0)
    { diagprofflag[PPROD]=
        diagprofflag[PPROD]&&tracerflag[prim]&&ratioflag[irc_assoc[prim]]&&
        diagprof[iphot_assoc[prim]][0]!=OBJF_NODATA;
      if (++i>=NAGG_MAX) prim=-1;
      else prim=iagg[PPROD_AGG][i]-1;
    }
  if (diagprofflag[PPROD]) diagprofflag[PPROD]=2;
}


static void calcdiagnostics (int prevdiag, 
                             unsigned char drflag[], unsigned char dtflag[], 
                             unsigned char dpflag[], struct solverrec *solver, 
                             int nk, OBJF_FLOAT_TYPE dz[], 
                             double tracer[][NK_MAX], double ratio[][NK_MAX], 
                             double tracer0[][NK_MAX], double ratio0[][NK_MAX], 
                             double diagprof[][NK_MAX], int it_agg[][NAGG_MAX],
                             int id_agg[][NAGG_MAX], struct vamrec *iv) 

{ double tracer_tdiag[N_TRACER][NK_MAX];
  double detfe[NK_MAX];
  double detfe0[NK_MAX];
  double detfe_tdiag[NK_MAX];
  double fastflxn[nk];  
  double fastflxc[nk]; 
  double fastflxsi[nk]; 
  double fastflxfe[nk]; 
  double fastflxca[nk]; 
  double r[nk];
  double r0[nk];
  int itbase,i,j,k,m,offset;
  double x,x0,x1,wsink,wsinktop,slowflxn,slowflxc,slowflxfe;
  double w0bio,w0slowsink,w1;
  int beforeval,advect_ctflag;

  /* Secondary tracers and aggregate primary tracers.
     Tracers are calculated before ratios since carbon tracers may be needed
     for diagnosing C:Chl ratios. */

  if (dtflag) 
    { calcfuntracers(dtflag,nk,tracer,ratio,it_agg,iv);
      if (dpflag && !prevdiag) 
        calcfuntracers(dtflag,nk,tracer0,ratio0,it_agg,iv);
    }

  /* Tracer ratios */

  if (drflag) 
    { calcratios(drflag,nk,tracer,ratio,iv);
      if (dpflag && !prevdiag) calcratios(dtflag,nk,tracer0,ratio0,iv);
    }
     
  /* Tertiary tracers */

  if (dtflag)

    { if (dtflag[NITROGEN]) calcaggregate(nk,NITROGEN,0,it_agg,NULL,tracer);
      if (dtflag[CARBON]) calcaggregate(nk,CARBON,0,it_agg,iv->c,tracer);
      if (dtflag[SILICON]) calcaggregate(nk,SILICON,0,it_agg,iv->si,tracer);
      if (dtflag[IRON])
        calcaggregate_byr(nk,IRON,0,it_agg,iv->rfe,ratio,tracer);
      if (dtflag[PON]) calcaggregate(nk,PON,0,it_agg,NULL,tracer);
      if (dtflag[POC]) calcaggregate(nk,POC,0,it_agg,iv->c,tracer);
    }

  /* Diagnostic variables

     Variables represent time-step mean values. Time of validity is thus middle
     of time step for forward Euler or end of time step for leapfrog. 
     Tracer/ratio values used in calculating diagnostics are for the same time 
     level as the tracer values used for calculating the associated tendencies 
     in the model (or the average time level). Details depend on solver
     numerics.

     NOTES RE. APPROXIMATION OF MODEL NUMERICS
     1) In leapfrog scheme, the 'now' tracer values used here are the final 
     output values NOT the 'after' values from the previous time step that
     are used in the model. 
     2) Rate of change due to slow sinking particles is calculated here
     using a simple upstream scheme. This may be different from the scheme 
     used for the simulation, in which case the sinking flux diagnostics will 
     only approximate the actual fluxes. 
     Where sinking is handled by a different MarMOT advection scheme, the
     vertical divergence here should ideally be compatible with that scheme 
     but this is not currently implemented.
     For models handling sinking internally, provision of sinking 
     velocity data by the model would be taken to imply that use of the upsteam 
     scheme is acceptable. 
  */

  if (dpflag)

    { if (solver->nstepbio>1) 

        { /* 'before' tracer values are used at each sub-step 
             so set weight to estimate average of these by 
             interpolating to the appropriate time */

          m=solver->nstepbio;
          w1=0.;
          for (i=1;i<m;i++) w1+=i;
          w1=w1/(m*m);
          if (solver->leapfrog) w0bio=1.-2.*w1;
          else w0bio=1.-w1; 
        }
 
      else

        { if (solver->leapfrog) w0bio=0.;  /* 'now' tracer values are used */
          else w0bio=1.;                   /* 'before' tracer values are used */
        }
       
      if (solver->dsink)      /* Sinking handled by advection scheme */
        { if (solver->leapfrog && 
              (solver->advection==1 || solver->advection==3)) 
            w0slowsink=0.;
          else w0slowsink=1.;
        }
      else w0slowsink=w0bio;  /* Sinking handled internally by the model */

      /* Diagnostics: primary production */

      if (dpflag[PPROD]==2)
        { if (w0bio==0.)
            calcaggregate_pprod(nk,id_agg,iv->c,tracer,iv->phot,diagprof);
          else if (w0bio==1.)
            calcaggregate_pprod(nk,id_agg,iv->c,tracer0,iv->phot,diagprof);
          else
            calcaggregate_pprod_intrp(nk,id_agg,iv->c,w0bio,tracer0,tracer,
                                      iv->phot,diagprof);
        }

      /* Diagnostics: sinking particle fluxes */

      if (dpflag[SINKFLXN] || dpflag[DDTSINKN])
        { if (w0slowsink==0.)
            for (k=0;k<nk;k++) tracer_tdiag[DET][k]=tracer[DET][k];
          else if (w0slowsink==1.)
            for (k=0;k<nk;k++) tracer_tdiag[DET][k]=tracer0[DET][k];
          else
            { w1=1.-w0slowsink;
              for (k=0;k<nk;k++) 
                { x0=tracer0[DET][k];
                  x1=tracer[DET][k];
                  if (x0==OBJF_NODATA || x1==OBJF_NODATA)
                    tracer_tdiag[DET][k]=OBJF_NODATA; 
                  else tracer_tdiag[DET][k]=w0slowsink*x0+w1*x1;
                }
            }
        }

      if (dpflag[SINKFLXC] || dpflag[DDTSINKC])
        { if (w0slowsink==0.)
            for (k=0;k<nk;k++) tracer_tdiag[DETC][k]=tracer[DETC][k];
          else if (w0slowsink==1.)
            for (k=0;k<nk;k++) tracer_tdiag[DETC][k]=tracer0[DETC][k];
          else
            { w1=1.-w0slowsink;
              for (k=0;k<nk;k++) 
                { x0=tracer0[DETC][k];
                  x1=tracer[DETC][k];
                  if (x0==OBJF_NODATA || x1==OBJF_NODATA)
                    tracer_tdiag[DETC][k]=OBJF_NODATA; 
                  else tracer_tdiag[DETC][k]=w0slowsink*x0+w1*x1;
                }
            }
        }

      if (dpflag[SINKFLXFE] || dpflag[DDTSINKFE])
        { if (w0slowsink==0.)
            calcderived_extra(nk,tracer[DET],ratio[FEN],detfe_tdiag);
          else if (w0slowsink==1.)
            calcderived_extra(nk,tracer0[DET],ratio0[FEN],detfe_tdiag);
          else
            { calcderived_extra(nk,tracer[DET],ratio[FEN],detfe);
              calcderived_extra(nk,tracer0[DET],ratio0[FEN],detfe0);
              w1=1.-w0slowsink;
              for (k=0;k<nk;k++) 
                { x0=detfe0[k]; 
                  x1=detfe[k]; 
                  if (x0==OBJF_NODATA || x1==OBJF_NODATA)
                    detfe_tdiag[k]=OBJF_NODATA; 
                  else detfe_tdiag[k]=w0slowsink*x0+w1*x1;
                }
            }
        }

      for (k=0;k<nk;k++)

        { wsink=-diagprof[iv->w[DET]][k];
          if (wsink==-(OBJF_NODATA)) 
            { slowflxn=OBJF_NODATA;
              slowflxc=OBJF_NODATA;
              slowflxfe=OBJF_NODATA;
            }
          else
            { if (dpflag[SINKFLXN] || dpflag[DDTSINKN]) 
                { x=tracer_tdiag[DET][k];
                  if (x==OBJF_NODATA) slowflxn=OBJF_NODATA;
                  else slowflxn=wsink*x;
                }
              if (dpflag[SINKFLXC] || dpflag[DDTSINKC]) 
                { x=tracer_tdiag[DETC][k];
                  if (x==OBJF_NODATA) slowflxc=OBJF_NODATA;
                  else slowflxc=wsink*x;
                }
              if (dpflag[SINKFLXFE] || dpflag[DDTSINKFE]) 
                { x=detfe_tdiag[k];
                  if (x==OBJF_NODATA) slowflxfe=OBJF_NODATA;
                  else slowflxfe=wsink*x;
                }
            }

          if (dpflag[SINKFLXN] && slowflxn!=OBJF_NODATA)
            { if (k==0) fastflxn[k]=-diagprof[INSINKN][k]*dz[k];
              else fastflxn[k]=fastflxn[k-1]-diagprof[INSINKN][k]*dz[k];
              diagprof[SINKFLXN][k]=fastflxn[k]+slowflxn;
            }

          if (dpflag[SINKFLXC] && slowflxc!=OBJF_NODATA)
            { if (k==0) fastflxc[k]=-diagprof[INSINKC][k]*dz[k];
              else fastflxc[k]=fastflxc[k-1]-diagprof[INSINKC][k]*dz[k];
              diagprof[SINKFLXC][k]=fastflxc[k]+slowflxc; 
            }

          if (dpflag[SINKFLXSI])
            { if (k==0) fastflxsi[k]=-diagprof[INSINKSI][k]*dz[k];
              else fastflxsi[k]=fastflxsi[k-1]-diagprof[INSINKSI][k]*dz[k];
              diagprof[SINKFLXSI][k]=fastflxsi[k];
            }

          if (dpflag[SINKFLXCA])
            { if (k==0) fastflxca[k]=-diagprof[INSINKCA][k]*dz[k];
              else fastflxca[k]=fastflxca[k-1]-diagprof[INSINKCA][k]*dz[k];
              diagprof[SINKFLXCA][k]=fastflxca[k];
            }

          if (dpflag[SINKFLXFE] && slowflxfe!=OBJF_NODATA)
            { if (k==0) fastflxfe[k]=-diagprof[INSINKFE][k]*dz[k];
              else fastflxfe[k]=fastflxfe[k-1]-diagprof[INSINKFE][k]*dz[k];
              diagprof[SINKFLXFE][k]=fastflxfe[k]+slowflxfe;
            }

          wsinktop=-diagprof[iv->w[DET]][k-1];

          if (dpflag[DDTSINKN] && slowflxn!=OBJF_NODATA)
            { if (k==0) 
                diagprof[DDTSINKN][k]=-slowflxn/dz[0]+diagprof[INSINKN][k]; 
              else 
                { x=tracer_tdiag[DET][k-1];
                  if (x==OBJF_NODATA || wsinktop==-(OBJF_NODATA)) 
                    diagprof[DDTSINKN][k]=OBJF_NODATA;
                  else 
                    diagprof[DDTSINKN][k]=
                      (wsinktop*x-slowflxn)/dz[k]+diagprof[INSINKN][k];
                }
            }

          if (dpflag[DDTSINKC] && slowflxc!=OBJF_NODATA)
            { if (k==0)
                diagprof[DDTSINKC][k]=-slowflxc/dz[0]+diagprof[INSINKC][k];
              else
                { x=tracer_tdiag[DETC][k-1];
                  if (x==OBJF_NODATA || wsinktop==-(OBJF_NODATA))
                    diagprof[DDTSINKC][k]=OBJF_NODATA; 
                  else 
                    diagprof[DDTSINKC][k]= 
                      (wsinktop*x-slowflxc)/dz[k]+diagprof[INSINKC][k];
                }
            }

          if (dpflag[DDTSINKSI]) diagprof[DDTSINKSI][k]=diagprof[INSINKSI][k];

          if (dpflag[DDTSINKCA]) diagprof[DDTSINKCA][k]=diagprof[INSINKCA][k];

          if (dpflag[DDTSINKFE] && slowflxfe!=OBJF_NODATA)
            { if (k==0)
                diagprof[DDTSINKFE][k]=-slowflxfe/dz[0]+diagprof[INSINKFE][k];
              else
                { x=detfe_tdiag[k-1];
                  if (x==OBJF_NODATA || wsinktop==-(OBJF_NODATA))
                    diagprof[DDTSINKFE][k]=OBJF_NODATA;
                  else 
                    diagprof[DDTSINKFE][k]=
                      (wsinktop*x-slowflxfe)/dz[k]+diagprof[INSINKFE][k];
                }
            }
        }

      /* Diagnostics: tracer tendencies
          
         Compatible with use of 'before' tracer values in forward Euler
         and 'before' or 'now' tracer values in leapfrog as noted below. 

         i=0: ddth (leapfrog uses 'before' tracer values with non-centred 
                    advection scheme &'now' values with centred scheme) 
         i=1: ddtp (leapfrog uses 'now' tracer values)
         i=2: ddtr (leapfrog uses 'before' tracer values)
      */

      advect_ctflag=0; /* No centred schemes in this code version */

      for (i=0;i<3;i++)

        { offset=DDTH_TRACERS+i*N_TRACER;

          beforeval=(!solver->leapfrog || (i==0 && advect_ctflag==0) || i==2); 

          /* Tendencies for ratio-derived tracers */

          for (itbase=0;itbase<N_TRACER_PRI;itbase++)
            { j=iv->c[itbase];
              m=iv->rc[itbase];
              if (m>=0 && dpflag[j+offset]==1)
                { if (beforeval) 
                    calcderived(nk,j,offset,itbase,ratio0[m],diagprof);
                  else
                    calcderived(nk,j,offset,itbase,ratio[m],diagprof);
                }
            }

          if (dpflag[PDIATSI+offset]==1)
            { if (beforeval)
                calcderived(nk,PDIATSI,offset,PDIAT,ratio0[SINPDIAT],diagprof); 
              else
                calcderived(nk,PDIATSI,offset,PDIAT,ratio[SINPDIAT],diagprof); 
            } 
     
          if (dpflag[CHL+offset]==1)
            { for (k=0;k<nk;k++) 
                { if (beforeval) r[k]=ATOM_C/ratio0[CCHL][k];
                  else r[k]=ATOM_C/ratio[CCHL][k];
                }
              calcderived(nk,CHL,offset,PHYC,r,diagprof);
            }
        
          if (dpflag[CHLNDIA+offset]==1)
            { for (k=0;k<nk;k++) 
                { if (beforeval) r[k]=ATOM_C/ratio0[CCHLNDIA][k];
                  else r[k]=ATOM_C/ratio[CCHLNDIA][k];
                }
              calcderived(nk,CHLNDIA,offset,PNDIAC,r,diagprof);
            }
        
          if (dpflag[CHLDIAT+offset]==1)
            { for (k=0;k<nk;k++) 
                { if (beforeval) r[k]=ATOM_C/ratio0[CCHLDIAT][k];
                  else r[k]=ATOM_C/ratio[CCHLDIAT][k];
                }
              calcderived(nk,CHLDIAT,offset,PDIATC,r,diagprof);
            }

          /* Tendencies for aggregate tracers */

          if (dpflag[PHY+offset]==2)
            calcaggregate(nk,PHY,offset,it_agg,NULL,diagprof);

          if (dpflag[ZOO+offset]==2)
            calcaggregate(nk,ZOO,offset,it_agg,NULL,diagprof);

          if (dpflag[PHYC+offset]==2)
            calcaggregate(nk,PHYC,offset,it_agg,iv->c,diagprof);

          if (dpflag[ZOOC+offset]==2)
            calcaggregate(nk,ZOOC,offset,it_agg,iv->c,diagprof);

          if (dpflag[CHL+offset]==2)
            calcaggregate(nk,CHL,offset,it_agg,iv->chl,diagprof);
    
          if (dpflag[NITROGEN+offset])
            calcaggregate(nk,NITROGEN,offset,it_agg,NULL,diagprof);

          if (dpflag[CARBON+offset])
            calcaggregate(nk,CARBON,offset,it_agg,iv->c,diagprof);

          if (dpflag[SILICON+offset])
            calcaggregate(nk,SILICON,offset,it_agg,iv->si,diagprof);

          if (dpflag[IRON+offset])
            { if (beforeval) 
                calcaggregate_byr(nk,IRON,offset,it_agg,iv->rfe,ratio0,
                                  diagprof);
              else
                calcaggregate_byr(nk,IRON,offset,it_agg,iv->rfe,ratio,
                                  diagprof);
            }

          if (dpflag[PON+offset])
            calcaggregate(nk,PON,offset,it_agg,NULL,diagprof);

          if (dpflag[POC+offset])
            calcaggregate(nk,POC,offset,it_agg,iv->c,diagprof);
        }
    }
}


static void calcfuntracers (unsigned char dtflag[], int nk, 
                            double tracer[][NK_MAX], double ratio[][NK_MAX],
                            int it_agg[][NAGG_MAX], struct vamrec *iv)

{ int itbase,j,m,k;
  double r[nk];
 
  /* Non-aggregate secondary tracers */

  for (itbase=0;itbase<N_TRACER_PRI;itbase++)
    { j=iv->c[itbase];
      m=iv->rc[itbase];
      if (m>=0 && dtflag[j]==1) calcderived(nk,j,0,itbase,ratio[m],tracer);
    }

  if (dtflag[PDIATSI]==1)
    calcderived(nk,PDIATSI,0,PDIAT,ratio[SINPDIAT],tracer);

  for (k=0;k<nk;k++) r[k]=ATOM_C/ratio[CCHL][k];
  if (dtflag[CHL]==1) calcderived(nk,CHL,0,PHYC,r,tracer);

  for (k=0;k<nk;k++) r[k]=ATOM_C/ratio[CCHLNDIA][k];
  if (dtflag[CHLNDIA]==1) calcderived(nk,CHLNDIA,0,PNDIAC,r,tracer);

  for (k=0;k<nk;k++) r[k]=ATOM_C/ratio[CCHLDIAT][k];
  if (dtflag[CHLDIAT]==1) calcderived(nk,CHLDIAT,0,PDIATC,r,tracer);

  /* Aggregate primary & secondary tracers */

  if (dtflag[PHY]==2) calcaggregate(nk,PHY,0,it_agg,NULL,tracer);
  if (dtflag[ZOO]==2) calcaggregate(nk,ZOO,0,it_agg,NULL,tracer);
  if (dtflag[PHYC]==2) calcaggregate(nk,PHYC,0,it_agg,iv->c,tracer);
  if (dtflag[ZOOC]==2) calcaggregate(nk,ZOOC,0,it_agg,iv->c,tracer);
  if (dtflag[CHL]==2) calcaggregate(nk,CHL,0,it_agg,iv->chl,tracer);
}


static void calcratios (unsigned char drflag[], int nk, 
                        double tracer[][NK_MAX], double ratio[][NK_MAX],
                        struct vamrec *iv)

{ int i,j,k,m;

 /* Ratios of non-aggregate tracers */

  for (k=0;k<nk;k++)

    { for (i=0;i<N_TRACER_PRI;i++)
        { j=iv->c[i];
          m=iv->rc[i];
          if (j>=0 && drflag[m]==1)
            { if (tracer[j][k]<0.||tracer[i][k]<=0.) ratio[m][k]=OBJF_NODATA;
              else ratio[m][k]=tracer[j][k]/tracer[i][k];
            }
        }
      if (drflag[SINPDIAT]==1)
        { if (tracer[PDIATSI][k]<0.||tracer[PDIAT][k]<=0.)
            ratio[SINPDIAT][k]=OBJF_NODATA;
          else ratio[SINPDIAT][k]=tracer[PDIATSI][k]/tracer[PDIAT][k];
        }
      if (drflag[CCHL]==1)
        { if (tracer[PHYC][k]<0.||tracer[CHL][k]<=0.)
            ratio[CCHL][k]=OBJF_NODATA;
          else ratio[CCHL][k]=tracer[PHYC][k]*ATOM_C/tracer[CHL][k];
        }
      if (drflag[CCHLNDIA]==1)
        { if (tracer[PNDIAC][k]<0.||tracer[CHLNDIA][k]<=0.)
            ratio[CCHLNDIA][k]=OBJF_NODATA;
          else ratio[CCHLNDIA][k]=tracer[PNDIAC][k]*ATOM_C/tracer[CHLNDIA][k];
        }
      if (drflag[CCHLDIAT]==1)
        { if (tracer[PDIATC][k]<0.||tracer[CHLDIAT][k]<=0.)
            ratio[CCHLDIAT][k]=OBJF_NODATA;
          else ratio[CCHLDIAT][k]=tracer[PDIATC][k]*ATOM_C/tracer[CHLDIAT][k];
        }
    }

  /* Ratios of aggregate tracers */

  for (k=0;k<nk;k++)

    { if (drflag[CNPHY]==2)
        { if (tracer[PHYC][k]<0.||tracer[PHY][k]<=0.)
            ratio[CNPHY][k]=OBJF_NODATA;
          else ratio[CNPHY][k]=tracer[PHYC][k]/tracer[PHY][k];
        }
      if (drflag[CNZOO]==2)
        { if (tracer[ZOOC][k]<0.||tracer[ZOO][k]<=0.)
            ratio[CNZOO][k]=OBJF_NODATA;
          else ratio[CNZOO][k]=tracer[ZOOC][k]/tracer[ZOO][k];
        }
      if (drflag[CCHL]==2)
        { if (tracer[PHYC][k]<0.||tracer[CHL][k]<=0.)
            ratio[CCHL][k]=OBJF_NODATA;
          else ratio[CCHL][k]=tracer[PHYC][k]*ATOM_C/tracer[CHL][k];
        }
    }
}


static void calcderived (int nk, int it, int offset, int itbase, double r[],
                         double a[][NK_MAX])

{ int i,j,k;
  double x;

  i=itbase+offset;
  j=it+offset;
  for (k=0;k<nk;k++)
    { x=a[i][k];
      if (x==0.) a[j][k]=0.;
      else if (r[k]<=0. || x==OBJF_NODATA) a[j][k]=OBJF_NODATA;
      else a[j][k]=x*r[k];
    }
}


static void calcderived_extra (int nk, double trbase[], double r[], 
                               double trnew[])
                  
{ int k;
  double x;
 
  for (k=0;k<nk;k++)
    { x=trbase[k];
      if (x==0.) trnew[k]=0.;
      else if (r[k]<=0. || x==OBJF_NODATA) trnew[k]=OBJF_NODATA;
      else trnew[k]=x*r[k];
    }
}      


static void calcaggregate (int nk, int it, int offset, int iagg[][NAGG_MAX],
                           int it_assoc[], double a[][NK_MAX])

{ int i,j,k,prim;
  double x;

  for (k=0;k<nk;k++)

    { j=it+offset;
      prim=iagg[it][0]-1;
      if (prim>=0) a[j][k]=0.;
      i=0;
      while (prim>=0 && a[j][k]!=OBJF_NODATA)
        { if (it_assoc)
            { if (it_assoc[prim]<0) x=a[prim+offset][k];
              else x=a[it_assoc[prim]+offset][k];
              if (x==OBJF_NODATA) a[j][k]=OBJF_NODATA;
              else a[j][k]+=x;
            }
          else
            { x=a[prim+offset][k];
              if (x==OBJF_NODATA) a[j][k]=OBJF_NODATA;
              else a[j][k]+=x;
            }
          if (++i>=NAGG_MAX) prim=-1;
          else prim=iagg[it][i]-1;
        }
    }
}


static void calcaggregate_byr (int nk, int it, int offset, 
                               int iagg[][NAGG_MAX], int ir_assoc[], 
                               double ratio[][NK_MAX], double a[][NK_MAX])

{ int i,j,k,prim;
  double x,r;

  for (k=0;k<nk;k++)

    { j=it+offset;
      prim=iagg[it][0]-1;
      if (prim>=0) a[j][k]=0.;
      i=0;
      while (prim>=0 && a[j][k]!=OBJF_NODATA)
        { if (ir_assoc[prim]<0)
            { x=a[prim+offset][k];
              if (x==OBJF_NODATA) a[j][k]=OBJF_NODATA;
              else a[j][k]+=x;
            }
          else
            { r=ratio[ir_assoc[prim]][k];
              x=a[prim+offset][k];
              if ((r<=0. && x!=0.) || x==OBJF_NODATA) a[j][k]=OBJF_NODATA;
              else a[j][k]+=x*r;
            }
          if (++i>=NAGG_MAX) prim=-1;
          else prim=iagg[it][i]-1;
        }
    }
}


static void calcaggregate_pprod (int nk, int iagg[][NAGG_MAX], 
                                 int ic_assoc[], double tracer[][NK_MAX], 
                                 int iphot_assoc[], double diagprof[][NK_MAX])

{ int i,k,prim;
  double x,phot;

  for (k=0;k<nk;k++)
    { prim=iagg[PPROD_AGG][0]-1;
      if (prim>=0) diagprof[PPROD][k]=0.;
      i=0;
      while (prim>=0 && diagprof[PPROD][k]!=OBJF_NODATA)
        { phot=diagprof[iphot_assoc[prim]][k];
          x=tracer[ic_assoc[prim]][k];
          if (phot==OBJF_NODATA || x==OBJF_NODATA)
            diagprof[PPROD][k]=OBJF_NODATA;
          else if (x>0.)
            diagprof[PPROD][k]+=phot*x;
          if (++i>=NAGG_MAX) prim=-1;
          else prim=iagg[PPROD_AGG][i]-1;
        }
    }
}


static void calcaggregate_pprod_intrp (int nk, int iagg[][NAGG_MAX],
                                       int ic_assoc[], double w0, 
                                       double tracer0[][NK_MAX],
                                       double tracer1[][NK_MAX],
                                       int iphot_assoc[], 
                                       double diagprof[][NK_MAX])

{ int i,k,prim;
  double w1,x,x0,x1,phot;

  w1=1.-w0;
  if (w1<0.) w1=0.;

  for (k=0;k<nk;k++)
    { prim=iagg[PPROD_AGG][0]-1;
      if (prim>=0) diagprof[PPROD][k]=0.;
      i=0;
      while (prim>=0 && diagprof[PPROD][k]!=OBJF_NODATA)
        { phot=diagprof[iphot_assoc[prim]][k];
          x0=tracer0[ic_assoc[prim]][k]; 
          x1=tracer1[ic_assoc[prim]][k];
          if (phot==OBJF_NODATA || (w0>0. && x0==OBJF_NODATA) || 
              (w1>0. && x1==OBJF_NODATA))
            diagprof[PPROD][k]=OBJF_NODATA;
          else 
            { x=w0*x0+w1*x1;
              if (x>0.) diagprof[PPROD][k]+=phot*x;
            }
          if (++i>=NAGG_MAX) prim=-1;
          else prim=iagg[PPROD_AGG][i]-1;
        }
    }
}


static void calciz (unsigned char iztflag[], unsigned char izdflag[], 
                    int nk, OBJF_FLOAT_TYPE dz[], double tracer[][NK_MAX], 
                    double diagprof[][NK_MAX], double iztracer[], 
                    double izdiag[]) 

{ int i;

  if (iztflag)
    { for (i=0;i<N_TRACER;i++) 
        if (iztflag[i]) iztracer[i]=zinteg(nk,dz,tracer[i]);
    }

  if (izdflag)
    { if (izdflag[IZPPROD]) izdiag[IZPPROD]=zinteg(nk,dz,diagprof[PPROD]);
      if (izdflag[IZFESCAV]) izdiag[IZFESCAV]=zinteg(nk,dz,diagprof[FESCAV]);
      for (i=IZDDTHN;i<IZDDTPN;i++) if (izdflag[i]) 
        izdiag[i]=zinteg(nk,dz,diagprof[DDTH_TRACERS+NITROGEN+i-IZDDTHN]);
      for (i=IZDDTPN;i<IZDDTRN;i++) if (izdflag[i])
        izdiag[i]=zinteg(nk,dz,diagprof[DDTP_TRACERS+NITROGEN+i-IZDDTPN]);
      for (i=IZDDTRN;i<N_IZ_DIAG;i++) if (izdflag[i]) 
        izdiag[i]=zinteg(nk,dz,diagprof[DDTR_TRACERS+NITROGEN+i-IZDDTRN]);
    }
}


static double zinteg (int nk, OBJF_FLOAT_TYPE dz[], double x[])

{ int k;
  double sum; 

  sum=0;
  for (k=0;k<nk;k++) 
    { if (x[k]==OBJF_NODATA) return OBJF_NODATA;
      sum+=dz[k]*x[k];
    }

  return sum;
}


static void resetaccumulator (struct modoutrec *acc, int kmin, int nk, double x)

{ int m,k;

  if (kmin==0)
    { for (m=0;m<N_FSCALAR;m++) acc->fscal[m]=x;
      for (m=0;m<N_DIAG_SCAL;m++) acc->diagscal[m]=x;
      for (m=0;m<N_TRACER;m++) acc->iztracer[m]=x;
      for (m=0;m<N_IZ_DIAG;m++) acc->izdiag[m]=x;
    }

  for (k=kmin;k<kmin+nk;k++)
    { for (m=0;m<N_FPROF;m++) acc->fprof[m][k]=x;
      for (m=0;m<N_TRACER;m++) acc->tracer[m][k]=x;
      for (m=0;m<N_RATIO;m++) acc->ratio[m][k]=x;
      for (m=0;m<N_DIAG_PROF;m++) acc->diagprof[m][k]=x;
    }
}


static void updaccumulator (struct modoutrec *acc, struct modoutrec *new, 
                            int nk, int updscalars)

{ if (updscalars) 
    { updatesums_scal(N_FSCALAR,acc->fscal,1,new->fscal,1);
      updatesums_scal(N_DIAG_SCAL,acc->diagscal,1,new->diagscal,1);
      updatesums_scal(N_TRACER,acc->iztracer,1,new->iztracer,1);
      updatesums_scal(N_IZ_DIAG,acc->izdiag,1,new->izdiag,1);
    }
  updatesums_prof(N_FPROF,nk,acc->fprof,1,new->fprof,1);
  updatesums_prof(N_TRACER,nk,acc->tracer,1,new->tracer,1);
  updatesums_prof(N_RATIO,nk,acc->ratio,1,new->ratio,1);
  updatesums_prof(N_DIAG_PROF,nk,acc->diagprof,1,new->diagprof,1);
}


static void updatesums_scal (int n, double sum[], double w0, 
                             double newval[], double w1)

{ int i;

  if (w0==0 && w1==0) 
    for (i=0;i<n;i++) sum[i]=0;
  else if (w0==0) 
    for (i=0;i<n;i++) 
      { if (newval[i]==OBJF_NODATA) sum[i]=OBJF_NODATA;
        else sum[i]=w1*newval[i];
      }
  else if (w1==0) 
    for (i=0;i<n;i++) 
      { if (sum[i]!=OBJF_NODATA) sum[i]=w0*sum[i];
      }
  else 
    for (i=0;i<n;i++) 
      { if (newval[i]==OBJF_NODATA) sum[i]=OBJF_NODATA;
        else if (sum[i]!=OBJF_NODATA) sum[i]=w0*sum[i]+w1*newval[i];
      }
}


static void updatesums_prof (int n, int nk, double sum[][NK_MAX], double w0,
                             double newval[][NK_MAX], double w1)

{ int i,j;

  if (w0==0 && w1==0)
    for (i=0;i<n;i++) 
      for (j=0;j<nk;j++) sum[i][j]=0;
  else if (w0==0)
    for (i=0;i<n;i++)
      for (j=0;j<nk;j++)
        { if (newval[i][j]==OBJF_NODATA) sum[i][j]=OBJF_NODATA;
          else sum[i][j]=w1*newval[i][j];
        }
  else if (w1==0)
    for (i=0;i<n;i++)
      for (j=0;j<nk;j++)
        { if (sum[i][j]!=OBJF_NODATA) sum[i][j]=w0*sum[i][j];
        }
  else
    for (i=0;i<n;i++) 
      for (j=0;j<nk;j++) 
        { if (newval[i][j]==OBJF_NODATA) 
            sum[i][j]=OBJF_NODATA;
          else if (sum[i][j]!=OBJF_NODATA) 
            sum[i][j]=w0*sum[i][j]+w1*newval[i][j];
        }
}


static int writemod_scal (int outtab, FILE *outfile, int flushflag,
                          struct caseinputrec *c, double t, double tmid, 
                          int year, int day, int step, OBJF_FLOAT_TYPE tdays,
                          int mth, int date, OBJF_FLOAT_TYPE thours, 
                          OBJF_FLOAT_TYPE z, OBJF_FLOAT_TYPE zbot,
                          double fscal[], double fprof[][NK_MAX],
                          double tracer[][NK_MAX],
                          double ratio[][NK_MAX],
                          double diagprof[][NK_MAX],
                          double diagscal[], double iztracer[], double izdiag[],
                          char **errtxt)

{ OBJF_FLOAT_TYPE outdata[N_OUTVAR_SCAL];
  int nvar,m;

  nvar=0;
  outdata[nvar++]=t;
  outdata[nvar++]=tmid;
  outdata[nvar++]=year;
  outdata[nvar++]=day;
  outdata[nvar++]=step;
  outdata[nvar++]=tdays;
  outdata[nvar++]=(mth<1)?OBJF_NODATA:mth;
  outdata[nvar++]=(date<1)?OBJF_NODATA:date;
  outdata[nvar++]=thours;
  outdata[nvar++]=z;
  outdata[nvar++]=zbot;
  for (m=0;m<N_FSCALAR;m++) outdata[nvar++]=fscal[m];
  for (m=0;m<N_FPROF;m++) outdata[nvar++]=fprof[m][0];
  for (m=0;m<N_TRACER_PRI;m++) outdata[nvar++]=tracer[m][0];
  for (m=0;m<N_RATIO;m++) outdata[nvar++]=ratio[m][0];
  for (m=0;m<N_TRACER_DER;m++) outdata[nvar++]=tracer[m+N_TRACER_PRI][0];
  for (m=0;m<N_DIAG_PROF;m++) outdata[nvar++]=diagprof[m][0];
  for (m=0;m<N_DIAG_SCAL;m++) outdata[nvar++]=diagscal[m];
  for (m=0;m<N_TRACER;m++) outdata[nvar++]=iztracer[m];
  for (m=0;m<N_IZ_DIAG;m++) outdata[nvar++]=izdiag[m];

  if (nvar>N_OUTVAR_SCAL)
    { *errtxt="too many output variables - MarMOT error!";
      return SPECIAL;
    }

  return writeouttabdata_array(outtab,outfile,flushflag,1,c,nvar,outdata);
}


static int writemodav_scal (int outtab, FILE *outfile, int flushflag, 
                            struct caseinputrec *c, double t, int year, 
                            int day, OBJF_FLOAT_TYPE tdays, int mth,
                            int date, OBJF_FLOAT_TYPE thours, OBJF_FLOAT_TYPE z,
                            OBJF_FLOAT_TYPE zbot, struct modoutrec *sum,
                            int n, char **errtxt)

{ OBJF_FLOAT_TYPE outdata[N_OUTVAR_SCAL-2];
  int nvar,m;

  nvar=0;
  outdata[nvar++]=t;
  outdata[nvar++]=year;
  if (outtab==OUTSCALAR_DAV) outdata[nvar++]=day;
  outdata[nvar++]=tdays;
  if (outtab==OUTSCALAR_DAV || outtab==OUTSCALAR_MAV) 
    outdata[nvar++]=(mth<1)?OBJF_NODATA:mth;
  if (outtab==OUTSCALAR_DAV) outdata[nvar++]=(date<1)?OBJF_NODATA:date;
  if (outtab==OUTSCALAR_DAV) outdata[nvar++]=thours;
  outdata[nvar++]=z;
  outdata[nvar++]=zbot;
  for (m=0;m<N_FSCALAR;m++)
    { if (sum->fscal[m]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->fscal[m]/n;
    }
  for (m=0;m<N_FPROF;m++)
    { if (sum->fprof[m][0]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->fprof[m][0]/n;
    }
  for (m=0;m<N_TRACER_PRI;m++)
    { if (sum->tracer[m][0]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->tracer[m][0]/n;
    }
  for (m=0;m<N_RATIO;m++)
    { if (sum->ratio[m][0]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->ratio[m][0]/n;
    }
  for (m=0;m<N_TRACER_DER;m++)
    { if (sum->tracer[m+N_TRACER_PRI][0]==OBJF_NODATA) 
        outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->tracer[m+N_TRACER_PRI][0]/n;
    }
  for (m=0;m<N_DIAG_PROF;m++)
    { if (sum->diagprof[m][0]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->diagprof[m][0]/n;
    }
  for (m=0;m<N_DIAG_SCAL;m++)
    { if (sum->diagscal[m]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->diagscal[m]/n;
    }
  for (m=0;m<N_TRACER;m++)
    { if (sum->iztracer[m]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->iztracer[m]/n;
    }
  for (m=0;m<N_IZ_DIAG;m++)
    { if (sum->izdiag[m]==OBJF_NODATA) outdata[nvar++]=OBJF_NODATA;
      else outdata[nvar++]=sum->izdiag[m]/n;
    }

  if (nvar>N_OUTVAR_SCAL-2)
    { *errtxt="too many output variables - MarMOT error!";
      return SPECIAL;
    }

  return writeouttabdata_array(outtab,outfile,flushflag,1,c,nvar,outdata);
}


static int writemod_prof (int outtab, FILE *outfile, int flushflag,
                          struct caseinputrec *c,
                          double t, double tmid, int year, int day, int step, 
                          OBJF_FLOAT_TYPE tdays, int mth, int date, 
                          OBJF_FLOAT_TYPE thours, int nk,
                          OBJF_FLOAT_TYPE z[], OBJF_FLOAT_TYPE ztop[], 
                          OBJF_FLOAT_TYPE zbot[], OBJF_FLOAT_TYPE dz[],
                          double fprof[][NK_MAX],
                          double tracer[][NK_MAX],
                          double ratio[][NK_MAX],
                          double diag[][NK_MAX],
                          char **errtxt)

{ OBJF_FLOAT_TYPE outdata[NK_MAX][N_OUTVAR_PROF];
  int nvar,m,k;

  for (k=0;k<nk;k++)

    { nvar=0;
      outdata[k][nvar++]=t;
      outdata[k][nvar++]=tmid;
      outdata[k][nvar++]=year;
      outdata[k][nvar++]=day;
      outdata[k][nvar++]=step;
      outdata[k][nvar++]=tdays;
      outdata[k][nvar++]=(mth<1)?OBJF_NODATA:mth;
      outdata[k][nvar++]=(date<1)?OBJF_NODATA:date;
      outdata[k][nvar++]=thours;
      outdata[k][nvar++]=k+1;
      outdata[k][nvar++]=z[k];
      outdata[k][nvar++]=ztop[k];
      outdata[k][nvar++]=zbot[k];
      outdata[k][nvar++]=dz[k];
      for (m=0;m<N_FPROF;m++) outdata[k][nvar++]=fprof[m][k];
      for (m=0;m<N_TRACER_PRI;m++) outdata[k][nvar++]=tracer[m][k];
      for (m=0;m<N_RATIO;m++) outdata[k][nvar++]=ratio[m][k];
      for (m=0;m<N_TRACER_DER;m++) outdata[k][nvar++]=tracer[m+N_TRACER_PRI][k];
      for (m=0;m<N_DIAG_PROF;m++) outdata[k][nvar++]=diag[m][k];

      if (nvar!=N_OUTVAR_PROF)
        { *errtxt="unexpected number of output variables - MarMOT error!";
          return SPECIAL;
        }
    }

  return writeouttabdata_array(outtab,outfile,flushflag,nk,c,nvar,
                               &outdata[0][0]);
}


static int writemodav_prof (int outtab, FILE *outfile, int flushflag, 
                            struct caseinputrec *c, double t, int year, int day,
                            OBJF_FLOAT_TYPE tdays, int mth, int date,
                            OBJF_FLOAT_TYPE thours, int nk, OBJF_FLOAT_TYPE z[],
                            OBJF_FLOAT_TYPE ztop[], OBJF_FLOAT_TYPE zbot[], 
                            OBJF_FLOAT_TYPE dz[], struct modoutrec *sum,
                            int n, char **errtxt)

{ OBJF_FLOAT_TYPE outdata[NK_MAX][N_OUTVAR_PROF-2];
  int nvarout,nvar,m,k;

  if (outtab==OUTPROFILE_DAV) nvarout=N_OUTVAR_PROF-2;
  else if (outtab==OUTPROFILE_MAV) nvarout=N_OUTVAR_PROF-5;
  else nvarout=N_OUTVAR_PROF-6; 

  for (k=0;k<nk;k++)

    { nvar=0;
      outdata[k][nvar++]=t;
      outdata[k][nvar++]=year;
      if (outtab==OUTPROFILE_DAV) outdata[k][nvar++]=day;
      outdata[k][nvar++]=tdays;
      if (outtab==OUTPROFILE_DAV || outtab==OUTPROFILE_MAV) 
        outdata[k][nvar++]=(mth<1)?OBJF_NODATA:mth;
      if (outtab==OUTPROFILE_DAV) outdata[k][nvar++]=(date<1)?OBJF_NODATA:date;
      if (outtab==OUTPROFILE_DAV) outdata[k][nvar++]=thours;
      outdata[k][nvar++]=k+1;
      outdata[k][nvar++]=z[k];
      outdata[k][nvar++]=ztop[k];
      outdata[k][nvar++]=zbot[k];
      outdata[k][nvar++]=dz[k];
      for (m=0;m<N_FPROF;m++)
        { if (sum->fprof[m][k]==OBJF_NODATA) outdata[k][nvar++]=OBJF_NODATA;
          else outdata[k][nvar++]=sum->fprof[m][k]/n;
        }
      for (m=0;m<N_TRACER_PRI;m++)
        { if (sum->tracer[m][k]==OBJF_NODATA) outdata[k][nvar++]=OBJF_NODATA; 
          else outdata[k][nvar++]=sum->tracer[m][k]/n;
        }
      for (m=0;m<N_RATIO;m++)
        { if (sum->ratio[m][k]==OBJF_NODATA) outdata[k][nvar++]=OBJF_NODATA;
          else outdata[k][nvar++]=sum->ratio[m][k]/n;
        }
      for (m=0;m<N_TRACER_DER;m++)
        { if (sum->tracer[m+N_TRACER_PRI][k]==OBJF_NODATA) 
            outdata[k][nvar++]=OBJF_NODATA; 
          else outdata[k][nvar++]=sum->tracer[m+N_TRACER_PRI][k]/n;
        }
      for (m=0;m<N_DIAG_PROF;m++)
        { if (sum->diagprof[m][k]==OBJF_NODATA) outdata[k][nvar++]=OBJF_NODATA;
          else outdata[k][nvar++]=sum->diagprof[m][k]/n;
        }

      if (nvar!=nvarout)
        { *errtxt="unexpected number of output variables - MarMOT error!";
          return SPECIAL;
        }
    }

  return writeouttabdata_array(outtab,outfile,flushflag,nk,c,nvar,
                               &outdata[0][0]);
}


static int writeassim_scal (int outtab, FILE *outfile, int flushflag,
                            struct caseinputrec *c, double t, int year, int day,
                            int step, OBJF_FLOAT_TYPE tdays, int mth, int date,
                            OBJF_FLOAT_TYPE thours, OBJF_FLOAT_TYPE z, 
                            OBJF_FLOAT_TYPE zbot, double prof[][NK_MAX],
                            double scal[], char **errtxt)

{ OBJF_FLOAT_TYPE outdata[N_OUTVAR_ASSIMSCAL];
  int nvar,m;

  nvar=0;
  outdata[nvar++]=t;
  outdata[nvar++]=year;
  outdata[nvar++]=day;
  outdata[nvar++]=step;
  outdata[nvar++]=tdays;
  outdata[nvar++]=(mth<1)?OBJF_NODATA:mth;
  outdata[nvar++]=(date<1)?OBJF_NODATA:date;
  outdata[nvar++]=thours;
  outdata[nvar++]=z;
  outdata[nvar++]=zbot;
  for (m=0;m<N_ASSIMPROFTOPVAR;m++) outdata[nvar++]=prof[m][0];
  for (m=0;m<N_ASSIMSCALVAR;m++) outdata[nvar++]=scal[m];

  if (nvar>N_OUTVAR_ASSIMSCAL)
    { *errtxt="too many output variables - MarMOT error!";
      return SPECIAL;
    }

  return writeouttabdata_array(outtab,outfile,flushflag,1,c,nvar,outdata);
}


static int writeassim_prof (int outtab, FILE *outfile, int flushflag,
                            struct caseinputrec *c, double t, int year, int day,
                            int step, OBJF_FLOAT_TYPE tdays, int mth, int date,
                            OBJF_FLOAT_TYPE thours, int nk, OBJF_FLOAT_TYPE z[],
                            OBJF_FLOAT_TYPE ztop[], OBJF_FLOAT_TYPE zbot[], 
                            OBJF_FLOAT_TYPE dz[], double prof[][NK_MAX],
                            char **errtxt)

{ OBJF_FLOAT_TYPE outdata[NK_MAX][N_OUTVAR_ASSIMPROF];
  int nvar,m,k;

  for (k=0;k<nk;k++)

    { nvar=0;
      outdata[k][nvar++]=t;
      outdata[k][nvar++]=year;
      outdata[k][nvar++]=day;
      outdata[k][nvar++]=step;
      outdata[k][nvar++]=tdays;
      outdata[k][nvar++]=(mth<1)?OBJF_NODATA:mth;
      outdata[k][nvar++]=(date<1)?OBJF_NODATA:date;
      outdata[k][nvar++]=thours;
      outdata[k][nvar++]=k+1;
      outdata[k][nvar++]=z[k];
      outdata[k][nvar++]=ztop[k];
      outdata[k][nvar++]=zbot[k];
      outdata[k][nvar++]=dz[k];
      for (m=0;m<N_ASSIMPROFVAR;m++) outdata[k][nvar++]=prof[m][k];

      if (nvar!=N_OUTVAR_ASSIMPROF)
        { *errtxt="unexpected number of output variables - MarMOT error!";
          return SPECIAL;
        }
    }

  return writeouttabdata_array(outtab,outfile,flushflag,nk,c,nvar,
                               &outdata[0][0]);
}


static int nextobs (struct itemrec *item, int newcase, double t0, 
                    double t1, OBJF_FLOAT_TYPE zmax, struct obsrec *obs)

{ int i;

  if (!item) return 0;

  /* Update array index for obs record at least once
     (initialize array pointers if new case)  */
 
  if (newcase) 
    { obs->time=item->dim[TIME];
      obs->depth=item->dim[DEPTH];
      for (i=0;i<N_OBSTYPE;i++) 
        { obs->value[i]=item->value[i];
          obs->weight[i]=item->value[i+N_OBSTYPE];
        }
      obs->i=0;
    }
  else obs->i++;             

  /* Return if out of records */

  if (obs->i >= item->npoint) return 0;

  /* Skip obs that are too early or too deep and return 0 if none in range */

  while (obs->time[obs->i] < t0) 
    if (++obs->i >= item->npoint) return 0;

  if (obs->depth) 
    { while (obs->depth[obs->i] > zmax)
        if (++obs->i >= item->npoint) return 0;
      obs->z=obs->depth[obs->i];
      if (obs->z > zmax) return 0; 
    }
  else obs->z=OBJF_NODATA; 

  obs->t=obs->time[obs->i];
  if (obs->t > t1) return 0; 

  /* Observation found */

  return 1; 
} 
  

static void calcmisfit (struct obsrec *obs, int nk, OBJF_FLOAT_TYPE zlevmid[], 
                        OBJF_FLOAT_TYPE zlevbot[], OBJF_FLOAT_TYPE zmax, 
                        double fscal[], double fprof[][NK_MAX], 
                        double tracer[][NK_MAX], double ratio[][NK_MAX], 
                        double diagprof[][NK_MAX], double diagscal[], 
                        double iztracer[], double izdiag[])
  
{ double obsval,weight,diff;
  int simindex,i,k,kb;
  double kw,kbw;
  enum zinterpmode interpopt;

  if (obs->z==OBJF_NODATA)

    { obs->k=OBJF_NODATA;
      obs->kztop=OBJF_NODATA;
      obs->kzbot=OBJF_NODATA;
    }

  else

    { if (obs->z>zmax || obs->z<0) 
        { obs->k=OBJF_NODATA; 
          for (i=0;i<N_OBSTYPE;i++) obs->misfitflag[i]=0;
          return;
        }

      /* Try to get an exact depth match to a level mid point
         (if depth is out of bounds then use nearest mid-point).
         Otherwise set weight 'kw' to interpolate between mid points. */

      kw=1;
      for (i=0; i<nk && obs->z > zlevmid[i]; i++);
  
      k=-1;
      if (i==0) k=0;
      else if (i==nk) k=nk-1;
      else if (obs->z-zlevmid[i-1] > zlevmid[i]-obs->z)
        { diff=fltdiff(obs->z,zlevmid[i]);
          if (diff==0) k=i; 
        }
      else
        { diff=fltdiff(obs->z,zlevmid[i-1]);
          if (diff==0) k=i-1;
        }

      if (k<0) obs->k=zinterp(nk,zlevmid,obs->z,&k,&kw)+1;
      else obs->k=k+1;

      /* Try to get an exact depth match to a level boundary
         (if depth is below bottom bound then use bottom bound). 
         Otherwise set weight 'kbw' to interpolate between level boundaries. */

      kbw=1;
      for (i=0; i<nk && obs->z > zlevbot[i]; i++);

      kb=-2;
      if (i==nk) kb=nk-1; 
      else if (obs->z < zlevbot[i]) 
        { if (i==0) 
            { diff=fltdiff(obs->z,0.);
              if (diff==0)
                { kb=-1;
                  kbw=1; 
                }
            }
          else 
            { diff=fltdiff(obs->z,zlevbot[i-1]);
              if (diff==0)
                { kb=i-1;
                  kbw=1;
                }
            }
        }
      else
        { diff=fltdiff(obs->z,zlevbot[i]); 
          if (diff==0)
            { kb=i;
              kbw=1;
            }
        }

      if (kb<-1) obs->kzbot=zinterp(nk,zlevbot,obs->z,&kb,&kbw)+1;
      else obs->kzbot=kb+1;

      obs->kztop=obs->kzbot+1;
    }

  /* Get interpolated value and calculate misfit for each observation type */
 
  for (i=0;i<N_OBSTYPE;i++)

    { if (obs->value[i]) obsval=obs->value[i][obs->i];
      else obsval=OBJF_NODATA;

      if (obsval==OBJF_NODATA) obs->misfitflag[i]=0;
      else
        { obs->modval[i]=OBJF_NODATA;
          switch (simcatofob(i,&simindex,&interpopt))
            { case FOUT_SCAL:
                obs->modval[i]=fscal[simindex];
                break; 
              case FOUT_PROF:
                if (obs->z==OBJF_NODATA) break;
                obs->modval[i]=profvalue(interpopt,k,kw,kb,kbw,fprof[simindex]);
                break; 
              case TRACER:
                if (obs->z==OBJF_NODATA) break;
                obs->modval[i]=
                  profvalue(interpopt,k,kw,kb,kbw,tracer[simindex]);
                break;
              case RATIO:
                if (obs->z==OBJF_NODATA) break;
                obs->modval[i]=profvalue(interpopt,k,kw,kb,kbw,ratio[simindex]);
                break; 
              case DIAG_PROF:
                if (obs->z==OBJF_NODATA) break;
                obs->modval[i]=
                  profvalue(interpopt,k,kw,kb,kbw,diagprof[simindex]);
                break;
              case DIAG_SCAL: 
                obs->modval[i]=diagscal[simindex]; 
                break;
              case IZ_TRACER: 
                obs->modval[i]=iztracer[simindex]; 
                break;
              case IZ_DIAG: 
                obs->modval[i]=izdiag[simindex]; 
                break;
            }
          if (obs->modval[i]==OBJF_NODATA) obs->misfitflag[i]=0;
          else 
            { obs->misfitflag[i]=1;
              switch (obs->transform[i])
                { case NULLTRANS: 
                    diff=fltdiff(obs->modval[i],obsval); 
                    break;
                  case LOGTRANS:
                    if (obsval<=0 || obs->modval[i]<=0) obs->misfitflag[i]=0; 
                    else diff=fltdiff(log10(obs->modval[i]),log10(obsval)); 
                    break;
                  case SQRTTRANS:
                    if (obsval<0 || obs->modval[i]<0) obs->misfitflag[i]=0; 
                    else diff=fltdiff(sqrt(obs->modval[i]),sqrt(obsval)); 
                    break;
                }
              if (obs->misfitflag[i])
                { if (obs->weight[i]) 
                    { weight=obs->weight[i][obs->i]; 
                      if (weight==OBJF_NODATA) weight=1; 
                    }
                  else weight=1;
                  obs->moddiff[i]=diff;
                  obs->w[i]=weight;
                  obs->misfit[i]=diff*diff*weight;
                } 
            }
        }
    }   
}       
     

static enum simcattype simcatofob (int i, int *simindex, 
                                   enum zinterpmode *interpopt)

{ static int isimcat_fprof=N_FSCALAR;
  static int isimcat_tracer_pri=N_FSCALAR+N_FPROF;
  static int isimcat_ratio=N_FSCALAR+N_FPROF+N_TRACER_PRI;
  static int isimcat_tracer_der=N_FSCALAR+N_FPROF+N_TRACER_PRI+N_RATIO;
  static int isimcat_diagprof=N_FSCALAR+N_FPROF+N_TRACER+N_RATIO;
  static int isimcat_diags=N_FSCALAR+N_FPROF+N_TRACER+N_RATIO+N_DIAG_PROF;
  static int isimcat_iztracer=
    N_FSCALAR+N_FPROF+N_TRACER+N_RATIO+N_DIAG_PROF+N_DIAG_SCAL;
  static int isimcat_izdiag=
    N_FSCALAR+N_FPROF+N_TRACER+N_RATIO+N_DIAG_PROF+N_DIAG_SCAL+N_TRACER;

  *interpopt=MIDDLE; 
 
  if (i<isimcat_fprof)
    { *simindex=i;
      return FOUT_SCAL;
    }

  if (i<isimcat_tracer_pri)
    { *simindex=i-isimcat_fprof;
      if (*simindex==VDC || *simindex==W) *interpopt=BOTTOM;
      return FOUT_PROF;
    }

  if (i<isimcat_ratio)
    { *simindex=i-isimcat_tracer_pri;
      return TRACER;
    }

  if (i<isimcat_tracer_der)
    { *simindex=i-isimcat_ratio;
      return RATIO;
    }

  if (i<isimcat_diagprof)
    { *simindex=i-isimcat_tracer_der+N_TRACER_PRI;
      return TRACER;
    }

  if (i<isimcat_diags)
    { *simindex=i-isimcat_diagprof;
      if (*simindex>=SINKFLXN && *simindex<=SINKFLXCA) *interpopt=BOTTOM;
      else if (*simindex==WDET) *interpopt=BOTTOM;
      else if (*simindex==PARZTOP || PARFRACZTOP) *interpopt=TOP; 
      return DIAG_PROF;
    }

  if (i<isimcat_iztracer)
    { *simindex=i-isimcat_diags;
      return DIAG_SCAL;
    }

  if (i<isimcat_izdiag)
    { *simindex=i-isimcat_iztracer;
      return IZ_TRACER;
    }

  *simindex=i-isimcat_izdiag;
  return IZ_DIAG;
}


static OBJF_FLOAT_TYPE zinterp (int nk, OBJF_FLOAT_TYPE zlev[], 
                                OBJF_FLOAT_TYPE z, int *k, double *kw) 

/* Returns equivalent to depth 'z' in index units based on indexing of depth 
   level array 'zlev'.
   Sets 'k' to index of level above required depth 'z' 
   and 'kw' to weight for value at 'k'. */

{ /* Above top of level depth array set k=-1. 
     'kw' is weight for a surface value (if it exists!).
     Return value assumes zlev[-1]=0 */

  if (z<=zlev[0]) 
    { *kw=(zlev[0]-z)/zlev[0];
      *k=-1;
      return -(*kw);
    }

  /* Look for level depth below 'z' */

  *k=1;
  while (*k<nk && z>zlev[*k]) (*k)++;

  /* If none found then return deepest point with undefined weight */

  if (*k==nk) 
    { *kw=OBJF_NODATA;
      *k=nk-1;
      return *k; 
    }

  /* Found level depth below 'z' so interpolate between levels */

  *kw=(zlev[*k]-z)/(zlev[*k]-zlev[(*k)-1]);
  (*k)--;
  return (*k)+1-(*kw);
}


static double profvalue (enum zinterpmode interpopt, int k, double kw, int kb, 
                         double kbw, double prof[])

/* 'k' is index at or above required depth for variables defined at level
    mid point (k=0 for depths throughout surface level: mid-point value is
    extended to zero depth by setting 'kw' to 1 above mid point).
    'kw' is weight for variable at index 'k', decreasing to zero at the
    mid point of the next level down (1 for depths out of range).
    'kb' is index at or above required depth for variables defined at level
    bottom (kb=-1 for depths in surface level).
    'kbw' is weight for variable at index 'kb', decreasing to zero at the
    bottom of the next level down (1 for depths below bottom bound).
    For variables defined at level top add 1 to 'kb'. Profile array(s) for
    such variables must include the level with top boundary at or below the
    maximum observation depth 'zmax'. */

{ switch (interpopt)
    { case MIDDLE:
        if (kw==1) return prof[k];
        else
          { if (prof[k]==OBJF_NODATA || prof[k+1]==OBJF_NODATA) 
              return OBJF_NODATA;
            else
              return kw*prof[k]+(1-kw)*prof[k+1];
          }
        break;
      case BOTTOM:
        if (kb==-1)
          { if (prof[0]==OBJF_NODATA) return OBJF_NODATA;
            else return (1-kbw)*prof[0];
          }
        else if (kbw==1) return prof[kb];
        else 
          { if (prof[kb]==OBJF_NODATA || prof[kb+1]==OBJF_NODATA)
              return OBJF_NODATA;
            else
              return kbw*prof[kb]+(1-kbw)*prof[kb+1];
          }
        break;
      case TOP:
        if (kbw==1) return prof[kb+1];
        else 
          { if (prof[kb+1]==OBJF_NODATA || prof[kb+2]==OBJF_NODATA)
              return OBJF_NODATA;
            else
              return kbw*prof[kb+1]+(1-kbw)*prof[kb+2];
          }
        break;
    }
  return OBJF_NODATA;
}


static double fltdiff (double a, double b)

{ double absa,absb,x,scale,dp,neg;
  int d;

  if (isnan(a) || isnan(b)) return sqrt(-1);
  if (a*b<=0) return a-b;

  /* Sign is the same so larger value determines level of accuracy 
     Round to this level before subtraction */

  absa=fabs(a);
  absb=fabs(b);
  x=(absa>=absb)?absa:absb;
  dp=FLT_DIG-ceil(log10(x));
  scale=pow(10,dp);
  if (a<0) neg=-1;
  else neg=1;
  d=(int)(scale*a+neg*0.5)-(int)(scale*b+neg*0.5); 
  return d/scale;
} 

        
static int writemisfit (int outtab, FILE *outfile, int flushflag, 
                        struct caseinputrec *c, 
                        double t, double tmid, int year, int day, int step, 
                        OBJF_FLOAT_TYPE tdays, int mth, int date, 
                        OBJF_FLOAT_TYPE thours, int yearlen, int nstep_day, 
                        struct obsrec *obs, char **errtxt)
    
{ OBJF_FLOAT_TYPE outdata[N_OUTVAR_MFIT];
  int nvar,i,newday,newyear;

  newday=(step==nstep_day && obs->t > t);
  newyear=(newday && day==yearlen);
  
  nvar=0;
  outdata[nvar++]=obs->i+1;
  outdata[nvar++]=obs->t;
  outdata[nvar++]=newyear ? year+1 : year;
  outdata[nvar++]=newyear ? 1 : (newday ? day+1 : day); 
  outdata[nvar++]=newyear ? fltdiff(obs->t,t) : tdays+fltdiff(obs->t,t);
  outdata[nvar++]=obs->z;
  outdata[nvar++]=fltdiff(t,obs->t);
  for (i=0;i<N_OBSTYPE;i++)
    { if (obs->misfitflag[i]) outdata[nvar++]=obs->moddiff[i];
      else outdata[nvar++]=OBJF_NODATA;
    }
  for (i=0;i<N_OBSTYPE;i++)
    { if (obs->misfitflag[i]) outdata[nvar++]=obs->misfit[i];
      else outdata[nvar++]=OBJF_NODATA;
    }
  outdata[nvar++]=t;
  outdata[nvar++]=tmid;
  outdata[nvar++]=year;
  outdata[nvar++]=day;
  outdata[nvar++]=step;
  outdata[nvar++]=tdays;
  outdata[nvar++]=(mth<1)?OBJF_NODATA:mth;
  outdata[nvar++]=(date<1)?OBJF_NODATA:date;
  outdata[nvar++]=thours;
  outdata[nvar++]=obs->k;
  outdata[nvar++]=obs->kztop;
  outdata[nvar++]=obs->kzbot;
  for (i=0;i<N_OBSTYPE;i++)
    { if (obs->misfitflag[i]) outdata[nvar++]=obs->modval[i];
      else outdata[nvar++]=OBJF_NODATA;
    }
  for (i=0;i<N_OBSTYPE;i++) 
    { if (obs->misfitflag[i]) outdata[nvar++]=obs->value[i][obs->i];
      else outdata[nvar++]=OBJF_NODATA;
    }
  for (i=0;i<N_OBSTYPE;i++)
    { if (obs->misfitflag[i]) outdata[nvar++]=obs->w[i];
      else outdata[nvar++]=OBJF_NODATA;
    }

  if (nvar>N_OUTVAR_MFIT)
    { *errtxt="too many output variables - MarMOT error!";
      return SPECIAL;
    }

  return writeouttabdata_array(outtab,outfile,flushflag,1,c,nvar,outdata);
}


static int writecost (int outtab, FILE *outfile, int flushflag, 
                      struct caseinputrec *c, struct costrec *cost)

{ OBJF_FLOAT_TYPE outdata[2];
  int nobs=0,nvar=2;

  outdata[0]=costfunc(cost,&nobs);
  outdata[1]=nobs;

  return writeouttabdata_array(outtab,outfile,flushflag,1,c,nvar,outdata);
}


static double costfunc (struct costrec *cost, int *nobs)

{ double total;
  int i,count;

  total=0.0;
  count=0;
 
  for (i=0;i<N_OBSTYPE;i++)
    { total+=cost->mfitsum[i];
      count+=cost->mfitcount[i];
    }

  if (nobs) *nobs=count;

  if (count==0) return 0;
  return total/count;
}
 
